# -*- coding: utf-8 -*-
# pylint: disable=line-too-long, wildcard-import, unused-wildcard-import
"""
Version 4 of *HydPy-L-Land* is still under development and likely to change in the
future.  Its goal is to extend |lland_v3| with processes dealing with the interception
of snow in canopies, similar to the `SCHNEEINTERZEPTION` option of `LARSIM`_.  The
publically available `LARSIM`_ documentation does not provide all information required
for implementing a perfect emulation, but detailed comparisons suggest that the
essential process equations are defined similarly.

.. _`LARSIM`: http://www.larsim.de/en/the-model/

The following list summarises the current status of |lland_v4| by discussing its
main differences to |lland_v3|:

 * Snow interception occurs only in tree canopies, which is why the results of
   |lland_v4| differ from those of |lland_v3| only for the land-use types |LAUBW|,
   |MISCHW|, and |NADELW|.
 * There is are two additional storages, |SInz| and |STInz|, which handle the total
   and the frozen amount of intercepted snowfall, being placed between the "normal"
   interception and the "normal" snow module.
 * If not stated otherwise, |SInz| and |STInz| correspond to |WAeS| and |WATS| and
   are modified by the same or functionally identical methods.
 * If the `LARSIM`_ documentation does not suggest otherwise, we reuse the names
   of the "normal" snow module and suffix "Inz" to it or replace its "Snow" suffix
   (for example, |WaDa| becomes |WaDaInz| and |NetRadiationSnow| becomes
   |NetRadiationInz|).
 * Only a fraction of the water amount leaving the "normal" interception module reaches
   the snow interception module (see |Calc_NBesInz_V1| and |Calc_SnowIntRate_V1|).
 * The surface temperature of the intercepted snow is always identical with its bulk
   temperature (|TempSInz|), except for the calculation of the latent heat flux
   (see |Return_WSurfInz_V1|).
 * The energy balance of the intercepted snow does not (need to) consider the soil
   heat flux.
 * The radiation exchange with the atmosphere depends on how much the intercepted
   snow shields the ground (see |Calc_NetShortwaveRadiationInz_V1| and
   |Return_NetLongwaveRadiationInz_V1|).
 * During periods with intercepted snow, water evaporates (or sublimates) from the snow
   interception module (|EvSInz|) but not from the "normal" interception module (|EvI|).

:cite:`ref-LARSIM` discusses many of these points in more detail.

In our experience, |lland_v4| tends to calculate thicker snowpacks on the ground
than |lland_v3| (the highest values of |WAeS| are usually larger when applying
|lland_v4|).  This counterintuitive behaviour seems to be related to disabling the
"normal" interception evaporation (|EvI|) during snow interception periods and is
still under discussion.

We will add further explanations and references as soon as we could consolidate
|lland_v4|.

The following integration tests are identical to the ones of |lland_v3|, except for
the additional snow interception processes.

Integration tests
=================

.. how_to_understand_integration_tests::

daily simulation
________________

>>> from hydpy import pub, Timegrid
>>> pub.timegrids = "1997-08-01", "1997-09-01", "1d"
>>> from hydpy.models.lland_v4 import *
>>> parameterstep("1h")
>>> from hydpy import Node, Element
>>> outlet = Node("outlet")
>>> inlet = Node("inlet")
>>> land = Element("land", inlets=inlet, outlets=outlet)
>>> land.model = model
>>> nhru(1)
>>> ft(1.0)
>>> fhru(1.0)
>>> from hydpy import IntegrationTest
>>> IntegrationTest.plotting_options.axis1 = inputs.nied
>>> IntegrationTest.plotting_options.axis2 = fluxes.qah
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

.. _lland_v4_acker_summer_daily:

acre (summer)
-------------

>>> lnk(ACKER)
>>> kg(0.94)
>>> kt(0.0)
>>> hinz(0.2)
>>> lai.acker_aug = 3.5
>>> tgr(0.0)
>>> trefn(0.0)
>>> tsp(4.0)
>>> pwmax(1.43)
>>> wmax(309.0)
>>> fk(199.0)
>>> pwp(119.4)
>>> kapgrenz(option="0_WMax/10")
>>> kapmax(0.08)
>>> beta(0.001/24)
>>> fbeta(1.0)
>>> rbeta(False)
>>> dmax(r_dmax=0.55)
>>> dmin(r_dmin=10.0)
>>> bsf(0.3)
>>> volbmax(inf)
>>> gsbmax(1.0)
>>> gsbgrad1(inf)
>>> gsbgrad2(inf)
>>> a1(0.0)
>>> a2(0.5)
>>> tind(0.191056)
>>> eqb(35000.0)
>>> eqi1(400.0)
>>> eqi2(400.0)
>>> eqd1(200.0)
>>> eqd2(50.0)
>>> negq(False)

>>> latitude(54.1)
>>> longitude(9.7)
>>> measuringheightwindspeed(10.0)
>>> angstromconstant(0.25)
>>> angstromfactor(0.5)
>>> angstromalternative(0.15)
>>> emissivity(0.95)
>>> py(119.4)
>>> cropheight.acker_aug = 0.4
>>> albedo.acker_aug = 0.24
>>> wg2z.aug = -0.25
>>> surfaceresistance.acker = 40.0

>>> p1strahl(0.5)
>>> p2strahl(1.0/35.0)
>>> p1wind(0.6)
>>> p2wind(1.0/70.0)
>>> turb0(0.0072)
>>> turb1(0.00576)
>>> albedo0snow(0.9)
>>> snowagingfactor(0.35)
>>> refreezeflag(True)
>>> ktschnee(inf)

>>> fvf(0.3)
>>> bsff(2.0)

>>> p1simax(8.0)
>>> p2simax(1.5)
>>> p1sirate(0.2)
>>> p2sirate(0.02)
>>> p3sirate(0.003)

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.stinz, 0.0),
...     (states.sinz, 0.0),
...     (states.esnowinz, 0.0),
...     (states.asinz, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 30.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     13.3, 0.1, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     2.7, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.3, 0.0, 0.1, 10.5, 6.5,
...     15.0, 9.2, 0.0, 0.0)
>>> inputs.teml.series = (
...     16.1, 17.3, 18.9, 17.8, 18.5, 19.9, 20.8, 20.9, 22.3, 23.0, 23.1,
...     23.2, 23.2, 21.7, 20.8, 19.5, 21.0, 21.6, 21.1, 22.4, 22.5, 19.6,
...     20.3, 24.1, 23.7, 21.1, 20.9, 20.2, 16.1, 17.0, 18.2)
>>> inputs.sunshineduration.series = (
...     6.3, 1.7, 4.5, 12.4, 13.9, 13.0, 13.8, 12.3, 13.1, 12.8, 13.1, 13.3,
...     12.7, 10.2, 9.4, 10.3, 11.1, 11.0, 8.5, 11.3, 12.4, 0.1, 6.7, 10.4,
...     6.5, 4.9, 6.6, 0.3, 0.1, 5.0, 3.8)
>>> inputs.windspeed.series = (
...     2.4, 1.6, 1.5, 2.8, 3.0, 3.5, 3.8, 3.3, 1.8, 1.9, 3.0, 2.5, 1.6, 2.7,
...     3.2, 1.3, 1.9, 2.0, 2.6, 3.6, 2.5, 2.1, 1.5, 2.2, 1.9, 2.3, 2.1, 2.6,
...     2.6, 2.1, 2.1)
>>> inputs.relativehumidity.series = (
...     86.2, 85.8, 82.4, 77.6, 74.9, 77.9, 73.5, 80.1, 72.9, 68.5, 68.6,
...     66.0, 69.0, 75.6, 81.4, 79.0, 75.9, 70.0, 70.3, 69.0, 62.1, 84.5,
...     83.6, 76.5, 76.6, 84.2, 85.8, 86.5, 89.6, 78.3, 78.5)
>>> inputs.atmosphericpressure.series = (
...     100.7, 101.3, 101.6, 101.7, 101.8, 101.8, 101.9, 101.9, 101.9, 102.0,
...     102.1, 102.2, 101.9, 101.5, 101.6, 101.9, 101.9, 101.8, 101.9, 101.9,
...     101.7, 101.5, 101.5, 101.1, 100.8, 100.6, 100.5, 100.0, 100.4, 101.1,
...     101.4)
>>> inlet.sequences.sim.series = 0.0


.. integration-test::

    >>> conditions_acker_summer = test(
    ...     "lland_v4_acker_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                      0.0 |                 12.492674 |                  12.492674 |                  1.877708 |                     0.0 |                      0.0 |    10.614966 |             0.0 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 0.196005 | 0.002269 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                      0.0 |                  8.382718 |                   8.382718 |                   0.96939 |                     0.0 |                      0.0 |     7.413327 |             0.0 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.263667 | 0.003052 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.003052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                      0.0 |                 10.795247 |                  10.795247 |                  1.386217 |                     0.0 |                      0.0 |      9.40903 |             0.0 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.140696 | 0.001628 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                      0.0 |                 17.664962 |                  17.664962 |                  3.108707 |                     0.0 |                      0.0 |    14.556255 |             0.0 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.075077 | 0.000869 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                      0.0 |                 18.901295 |                  18.901295 |                  3.414868 |                     0.0 |                      0.0 |    15.486427 |             0.0 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.040062 | 0.000464 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                      0.0 |                 18.036211 |                  18.036211 |                  2.845741 |                     0.0 |                      0.0 |     15.19047 |             0.0 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.021377 | 0.000247 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                      0.0 |                 18.651721 |                  18.651721 |                  3.051087 |                     0.0 |                      0.0 |    15.600634 |             0.0 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.011407 | 0.000132 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                      0.0 |                 17.267745 |                  17.267745 |                   2.45312 |                     0.0 |                      0.0 |    14.814625 |             0.0 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.006087 |  0.00007 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                      0.0 |                 17.877445 |                  17.877445 |                  2.688326 |                     0.0 |                      0.0 |    15.189119 |             0.0 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003248 | 0.000038 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                      0.0 |                  17.53355 |                   17.53355 |                   2.76572 |                     0.0 |                      0.0 |     14.76783 |             0.0 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001733 |  0.00002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                      0.0 |                 17.704758 |                  17.704758 |                  2.799361 |                     0.0 |                      0.0 |    14.905396 |             0.0 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000925 | 0.000011 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                      0.0 |                 17.786952 |                  17.786952 |                  2.986484 |                     0.0 |                      0.0 |    14.800468 |             0.0 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000494 | 0.000006 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                      0.0 |                 17.184083 |                  17.184083 |                  2.708378 |                     0.0 |                      0.0 |    14.475704 |             0.0 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000263 | 0.000003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                      0.0 |                 14.968904 |                  14.968904 |                  2.239353 |                     0.0 |                      0.0 |    12.729551 |             0.0 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000141 | 0.000002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                      0.0 |                 14.206821 |                  14.206821 |                  2.021579 |                     0.0 |                      0.0 |    12.185243 |             0.0 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 0.012919 |  0.00015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                      0.0 |                 14.881951 |                  14.881951 |                  2.487815 |                     0.0 |                      0.0 |    12.394137 |             0.0 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.017318 |   0.0002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.0002 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                      0.0 |                 15.465744 |                  15.465744 |                  2.520953 |                     0.0 |                      0.0 |     12.94479 |             0.0 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 0.009173 | 0.000106 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |   0.0 | 0.000106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                      0.0 |                 15.289825 |                  15.289825 |                  2.702703 |                     0.0 |                      0.0 |    12.587122 |             0.0 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 0.004366 | 0.000051 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |   0.0 | 0.000051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                      0.0 |                 13.113987 |                  13.113987 |                  2.307856 |                     0.0 |                      0.0 |    10.806131 |             0.0 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 0.000839 |  0.00001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |   0.0 |  0.00001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                      0.0 |                 15.350398 |                  15.350398 |                  2.688476 |                     0.0 |                      0.0 |    12.661922 |             0.0 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                      0.0 |                 16.161308 |                  16.161308 |                  3.308913 |                     0.0 |                      0.0 |    12.852395 |             0.0 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                      0.0 |                  5.948207 |                   5.948207 |                   0.60663 |                     0.0 |                      0.0 |     5.341577 |             0.0 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                      0.0 |                 11.290596 |                  11.290596 |                  1.641158 |                     0.0 |                      0.0 |     9.649438 |             0.0 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                      0.0 |                 14.219527 |                  14.219527 |                  1.868809 |                     0.0 |                      0.0 |    12.350718 |             0.0 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                      0.0 |                 10.960427 |                  10.960427 |                  1.401603 |                     0.0 |                      0.0 |     9.558824 |             0.0 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                      0.0 |                  9.585999 |                   9.585999 |                  1.261689 |                     0.0 |                      0.0 |      8.32431 |             0.0 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 0.002764 | 0.000032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |   0.0 | 0.000032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                      0.0 |                 10.870274 |                  10.870274 |                  1.498964 |                     0.0 |                      0.0 |      9.37131 |             0.0 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 0.031583 | 0.000366 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |   0.0 | 0.000366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                      0.0 |                  5.758296 |                   5.758296 |                  0.589077 |                     0.0 |                      0.0 |     5.169219 |             0.0 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 0.101111 |  0.00117 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |   0.0 |  0.00117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                      0.0 |                   5.53921 |                    5.53921 |                  0.708412 |                     0.0 |                      0.0 |     4.830799 |             0.0 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 0.168586 | 0.001951 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |   0.0 | 0.001951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                      0.0 |                  9.344591 |                   9.344591 |                  1.881434 |                     0.0 |                      0.0 |     7.463158 |             0.0 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.127614 | 0.001477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |   0.0 | 0.001477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                      0.0 |                  8.321079 |                   8.321079 |                  1.511299 |                     0.0 |                      0.0 |      6.80978 |             0.0 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.053743 | 0.000622 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |   0.0 | 0.000622 |

.. _lland_v4_acker_routing_daily:

:ref:`acre (routing) <lland_v1_acker_routing>`
----------------------------------------------

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test("lland_v4_acker_routing_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa)
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                      0.0 |                 12.492674 |                  12.492674 |                  1.877708 |                     0.0 |                      0.0 |    10.614966 |             0.0 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623966 |   0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 1.924005 | 0.022269 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  81.32045 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.022269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                      0.0 |                  8.382718 |                   8.382718 |                   0.96939 |                     0.0 |                      0.0 |     7.413327 |             0.0 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                98.302128 |               94.620659 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555713 | 0.094 | 1.697264 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.991667 | 0.023052 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 79.623186 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.023052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                      0.0 |                 10.795247 |                  10.795247 |                  1.386217 |                     0.0 |                      0.0 |      9.40903 |             0.0 |              0.0 |           9.40903 |             94.155751 |                 100.0 |               100.629116 |               95.574305 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286521 |   0.0 | 2.329549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.868696 | 0.021628 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 77.293637 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.021628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                      0.0 |                 17.664962 |                  17.664962 |                  3.108707 |                     0.0 |                      0.0 |    14.556255 |             0.0 |              0.0 |         14.556255 |             50.440581 |                 100.0 |               103.850807 |               96.853335 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847436 | 0.094 | 3.314416 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.803077 | 0.020869 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 73.979222 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                      0.0 |                 18.901295 |                  18.901295 |                  3.414868 |                     0.0 |                      0.0 |    15.486427 |             0.0 |              0.0 |         15.486427 |             47.077876 |                 100.0 |               108.496391 |               98.617204 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638616 |   0.0 | 3.732358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.768062 | 0.020464 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 70.246864 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                      0.0 |                 18.036211 |                  18.036211 |                  2.845741 |                     0.0 |                      0.0 |     15.19047 |             0.0 |              0.0 |          15.19047 |             40.352465 |                 100.0 |               113.828249 |              100.519671 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.829116 |   0.0 | 3.652462 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.749377 | 0.020247 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 66.594402 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                      0.0 |                 18.651721 |                  18.651721 |                  3.051087 |                     0.0 |                      0.0 |    15.600634 |             0.0 |              0.0 |         15.600634 |             37.166744 |                 100.0 |               119.168431 |              102.297453 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945284 |   0.0 | 4.128761 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.739407 | 0.020132 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 62.465641 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                      0.0 |                 17.267745 |                  17.267745 |                   2.45312 |                     0.0 |                      0.0 |    14.814625 |             0.0 |              0.0 |         14.814625 |             42.798069 |                 100.0 |               125.382697 |              104.231161 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.428125 |   0.0 | 3.518477 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.734087 |  0.02007 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.947163 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                      0.0 |                 17.877445 |                  17.877445 |                  2.688326 |                     0.0 |                      0.0 |    15.189119 |             0.0 |              0.0 |         15.189119 |             78.463126 |                 100.0 |               130.861333 |              105.804695 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036736 |   0.0 | 4.091355 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.731248 | 0.020038 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.855808 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                      0.0 |                  17.53355 |                   17.53355 |                   2.76572 |                     0.0 |                      0.0 |     14.76783 |             0.0 |              0.0 |          14.76783 |             74.333488 |                 100.0 |               137.498449 |               107.59179 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.494046 |   0.0 | 4.345849 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.729733 |  0.02002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 50.509959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                      0.0 |                 17.704758 |                  17.704758 |                  2.799361 |                     0.0 |                      0.0 |    14.905396 |             0.0 |              0.0 |         14.905396 |             47.077876 |                 100.0 |               144.953402 |                109.4495 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914762 |   0.0 | 4.548032 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728925 | 0.020011 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 45.961927 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                      0.0 |                 17.786952 |                  17.786952 |                  2.986484 |                     0.0 |                      0.0 |    14.800468 |             0.0 |              0.0 |         14.800468 |             56.493451 |                 100.0 |               153.359832 |              111.375646 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566718 | 0.094 | 4.554485 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728494 | 0.020006 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 41.407442 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                      0.0 |                 17.184083 |                  17.184083 |                  2.708378 |                     0.0 |                      0.0 |    14.475704 |             0.0 |              0.0 |         14.475704 |             88.271017 |                 100.0 |               162.655293 |                113.3188 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996364 |   0.0 | 4.137143 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728263 | 0.020003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 37.270299 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                      0.0 |                 14.968904 |                  14.968904 |                  2.239353 |                     0.0 |                      0.0 |    12.729551 |             0.0 |              0.0 |         12.729551 |             52.308751 |                 100.0 |               172.213709 |              115.125967 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983925 |   0.0 | 3.428814 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728141 | 0.020002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 33.841485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                      0.0 |                 14.206821 |                  14.206821 |                  2.021579 |                     0.0 |                      0.0 |    12.185243 |             0.0 |              0.0 |         12.185243 |             44.135508 |                 100.0 |               181.325786 |              116.675553 |  1.838 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470984 |   0.7 |  2.50575 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049891 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049891 | 1.740919 |  0.02015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 33.123844 | 0.049891 |   0.0 |   0.0 |   0.0 |       0.0 | 0.012919 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                      0.0 |                 14.881951 |                  14.881951 |                  2.487815 |                     0.0 |                      0.0 |    12.394137 |             0.0 |              0.0 |         12.394137 |            108.641252 |                 100.0 |               183.415178 |              116.867739 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.08134 |   0.0 | 2.836595 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.745318 |   0.0202 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 30.287249 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017318 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |   0.0202 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                      0.0 |                 15.465744 |                  15.465744 |                  2.520953 |                     0.0 |                      0.0 |     12.94479 |             0.0 |              0.0 |          12.94479 |             74.333488 |                 100.0 |               192.486718 |              118.223505 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175424 |   0.0 | 3.277008 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.038074 |      0.0 | 1.737173 | 0.020106 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 27.048315 |      0.0 |   0.0 |   0.0 |   0.0 | -0.038074 | 0.009241 |   0.0 |   0.0 |   0.0 | -0.000068 |  0.02 | 0.020106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                      0.0 |                 15.289825 |                  15.289825 |                  2.702703 |                     0.0 |                      0.0 |    12.587122 |             0.0 |              0.0 |         12.587122 |             70.616813 |                 100.0 |               204.997694 |              119.967773 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.738005 |   0.0 | 3.581497 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239328 |      0.0 | 1.732366 | 0.020051 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 23.706145 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239328 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |  0.02 | 0.020051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                      0.0 |                 13.113987 |                  13.113987 |                  2.307856 |                     0.0 |                      0.0 |    10.806131 |             0.0 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                221.73546 |              122.070817 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.867083 |   0.0 | 3.296906 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446997 |      0.0 | 1.728839 |  0.02001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 20.856236 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446997 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001793 |  0.02 |  0.02001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                      0.0 |                 15.350398 |                  15.350398 |                  2.688476 |                     0.0 |                      0.0 |    12.661922 |             0.0 |              0.0 |         12.661922 |             39.231563 |                 100.0 |               241.240755 |              124.200232 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.829385 |   0.0 | 3.892052 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.624079 |      0.0 | 1.725699 | 0.019973 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 17.588263 |      0.0 |   0.0 |   0.0 |   0.0 | -0.624079 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003705 |  0.02 | 0.019973 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                      0.0 |                 16.161308 |                  16.161308 |                  3.308913 |                     0.0 |                      0.0 |    12.852395 |             0.0 |              0.0 |         12.852395 |             56.493451 |                 100.0 |               274.870616 |              127.357125 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.32284 |   0.0 |  4.18324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827137 |      0.0 | 1.722458 | 0.019936 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 14.232159 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827137 | 0.000749 |   0.0 |   0.0 |   0.0 | -0.006292 |  0.02 | 0.019936 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                      0.0 |                  5.948207 |                   5.948207 |                   0.60663 |                     0.0 |                      0.0 |     5.341577 |             0.0 |              0.0 |          5.341577 |             67.254108 |                 100.0 |               337.482269 |              131.866514 |   1.18 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  2.56747 |   0.7 | 1.075772 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013304 |  0.0 |  0.0 | 0.0 | 1.035672 | 0.013304 | 1.722219 | 0.019933 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 15.358755 | 0.013304 |   0.0 |   0.0 |   0.0 | -1.035672 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009606 |  0.02 | 0.019933 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                      0.0 |                 11.290596 |                  11.290596 |                  1.641158 |                     0.0 |                      0.0 |     9.649438 |             0.0 |              0.0 |          9.649438 |             94.155751 |                 100.0 |               311.359555 |              129.743335 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374286 | 0.282 | 2.014946 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96567 |      0.0 | 1.719664 | 0.019904 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 14.309479 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96567 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013157 |  0.02 | 0.019904 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                      0.0 |                 14.219527 |                  14.219527 |                  1.868809 |                     0.0 |                      0.0 |    12.350718 |             0.0 |              0.0 |         12.350718 |             64.197103 |                 100.0 |               335.458668 |              131.071625 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573178 |   0.0 |  3.36219 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030867 |      0.0 | 1.713886 | 0.019837 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 11.978156 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030867 | 0.002572 |   0.0 |   0.0 |   0.0 | -0.016686 |  0.02 | 0.019837 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                      0.0 |                 10.960427 |                  10.960427 |                  1.401603 |                     0.0 |                      0.0 |     9.558824 |             0.0 |              0.0 |          9.558824 |             74.333488 |                 100.0 |               419.640495 |              135.153792 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467819 | 0.094 | 2.719425 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175726 |      0.0 | 1.708793 | 0.019778 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.28224 | 10.434457 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175726 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020579 |  0.02 | 0.019778 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                      0.0 |                  9.585999 |                   9.585999 |                  1.261689 |                     0.0 |                      0.0 |      8.32431 |             0.0 |              0.0 |           8.32431 |             61.405925 |                 100.0 |               522.198911 |              138.404806 |   9.17 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569973 |   0.7 | 1.597438 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104569 |  0.0 |  0.0 | 0.0 | 1.271645 | 0.104569 | 1.730764 | 0.020032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 19.174095 | 0.104569 |   0.0 |   0.0 |   0.0 | -1.271645 | 0.027653 |   0.0 |   0.0 |   0.0 |  -0.02489 |  0.02 | 0.020032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                      0.0 |                 10.870274 |                  10.870274 |                  1.498964 |                     0.0 |                      0.0 |      9.37131 |             0.0 |              0.0 |           9.37131 |             67.254108 |                 100.0 |               256.494997 |              123.965055 |   5.41 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.567075 |   0.7 | 1.722338 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090739 |  0.0 |  0.0 | 0.0 |   0.7286 | 0.090739 | 1.759583 | 0.020366 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 23.499618 | 0.090739 |   0.0 |   0.0 |   0.0 |   -0.7286 | 0.059966 |   0.0 |   0.0 |   0.0 | -0.028383 |  0.02 | 0.020366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                      0.0 |                  5.758296 |                   5.758296 |                  0.589077 |                     0.0 |                      0.0 |     5.169219 |             0.0 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                222.95161 |              120.113963 |   13.4 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621374 |   0.7 | 1.072031 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312931 |  0.0 |  0.0 | 0.0 |  0.45983 | 0.312931 | 1.829111 |  0.02117 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 35.974485 | 0.312931 |   0.0 |   0.0 |   0.0 |  -0.45983 | 0.131521 |   0.0 |   0.0 |   0.0 |  -0.03041 |  0.02 |  0.02117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                      0.0 |                   5.53921 |                    5.53921 |                  0.708412 |                     0.0 |                      0.0 |     4.830799 |             0.0 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 175.5055 |              113.153351 |  7.948 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027722 |   0.7 | 0.717235 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249156 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249156 | 1.896586 | 0.021951 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 42.956093 | 0.249156 |   0.0 |   0.0 |   0.0 |       0.0 |  0.19971 |   0.0 |   0.0 |   0.0 | -0.031124 |  0.02 | 0.021951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                      0.0 |                  9.344591 |                   9.344591 |                  1.881434 |                     0.0 |                      0.0 |     7.463158 |             0.0 |              0.0 |          7.463158 |             67.254108 |                 100.0 |               159.374248 |              110.061771 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393609 |   0.0 | 2.035968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.855614 | 0.021477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 40.920125 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158627 |   0.0 |   0.0 |   0.0 | -0.031013 |  0.02 | 0.021477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                      0.0 |                  8.321079 |                   8.321079 |                  1.511299 |                     0.0 |                      0.0 |      6.80978 |             0.0 |              0.0 |           6.80978 |             67.254108 |                 100.0 |               163.717417 |              110.731287 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.322101 |   0.0 | 2.018298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.781743 | 0.020622 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 38.901827 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084645 |   0.0 |   0.0 |   0.0 | -0.030902 |  0.02 | 0.020622 |

.. _lland_v4_acker_heavy_rain_daily:

acre (heavy rain)
-----------------

.. integration-test::

    >>> lnk(ACKER)
    >>> inlet.sequences.sim.series = 0.0
    >>> nied = inputs.nied.series.copy()
    >>> inputs.nied.series = 20.0
    >>> conditions_acker_heavy_rain = test(
    ...     "lland_v4_acker_heavy_rain_daily",
    ...     axis1=(inputs.nied, fluxes.qah),
    ...     axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo | evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada |       qdb |     qib1 |     qib2 |      qbb | qkap |      qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |       bowa |     qdgz1 |    qdgz2 |    qigz1 |    qigz2 |     qbgz |     qdga1 |    qdga2 |    qiga1 |    qiga2 |     qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 20.0 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 18.8 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             12.492674 |                      0.0 |                 12.492674 |                  12.492674 |                  1.877708 |                     0.0 |                      0.0 |    10.614966 |             0.0 |              0.0 |         10.614966 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623966 | 0.7 | 1.720207 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.217805 |      0.0 |      0.0 |      0.0 |  0.0 |  1.217805 |   0.31352 | 0.003629 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  87.161988 |  1.217805 |      0.0 |      0.0 |      0.0 |      0.0 |   0.31352 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.003629 |
    | 1997-02-08 | 20.0 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 | 18.8 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              8.382718 |                      0.0 |                  8.382718 |                   8.382718 |                   0.96939 |                     0.0 |                      0.0 |     7.413327 |             0.0 |              0.0 |          7.413327 |             88.271017 |                 100.0 |                90.407167 |               91.066405 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555713 | 0.7 | 1.294549 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.480868 |      0.0 |      0.0 |      0.0 |  0.0 |  1.480868 |  0.802993 | 0.009294 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  102.48657 |  1.480868 |      0.0 |      0.0 |      0.0 |      0.0 |  0.802993 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.009294 |
    | 1997-03-08 | 20.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 | 18.8 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             10.795247 |                      0.0 |                 10.795247 |                  10.795247 |                  1.386217 |                     0.0 |                      0.0 |      9.40903 |             0.0 |              0.0 |           9.40903 |             94.155751 |                 100.0 |                70.327061 |                80.36502 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286521 | 0.7 | 1.922457 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.761522 |      0.0 |      0.0 |      0.0 |  0.0 |  1.761522 |    1.1914 | 0.013789 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 116.902592 |  1.761522 |      0.0 |      0.0 |      0.0 |      0.0 |    1.1914 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.013789 |
    | 1997-04-08 | 20.0 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 | 18.8 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             17.664962 |                      0.0 |                 17.664962 |                  17.664962 |                  3.108707 |                     0.0 |                      0.0 |    14.556255 |             0.0 |              0.0 |         14.556255 |             50.440581 |                 100.0 |                51.993585 |               67.804653 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847436 | 0.7 | 3.397259 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.040913 |      0.0 |      0.0 |      0.0 |  0.0 |  2.040913 |  1.529227 | 0.017699 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  129.56442 |  2.040913 |      0.0 |      0.0 |      0.0 |      0.0 |  1.529227 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.017699 |
    | 1997-05-08 | 20.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 | 18.8 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             18.901295 |                      0.0 |                 18.901295 |                  18.901295 |                  3.414868 |                     0.0 |                      0.0 |    15.486427 |             0.0 |              0.0 |         15.486427 |             47.077876 |                 100.0 |                 48.85611 |               65.355461 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638616 | 0.7 | 3.917199 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.300279 | 0.101438 |      0.0 | 0.010164 |  0.0 |  2.300279 |  1.849046 | 0.021401 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  141.33534 |  2.300279 |      0.0 | 0.101438 |      0.0 | 0.010164 |  1.834644 |      0.0 | 0.014384 |      0.0 | 0.000018 |   0.0 | 0.021401 |
    | 1997-06-08 | 20.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 | 18.8 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             18.036211 |                      0.0 |                 18.036211 |                  18.036211 |                  2.845741 |                     0.0 |                      0.0 |     15.19047 |             0.0 |              0.0 |          15.19047 |             40.352465 |                 100.0 |                 48.85611 |               65.425338 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.829116 | 0.7 | 3.913663 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.554684 | 0.110653 |      0.0 | 0.021935 |  0.0 |  2.554684 |  2.156535 |  0.02496 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 152.834404 |  2.554684 |      0.0 | 0.110653 |      0.0 | 0.021935 |  2.117307 |      0.0 | 0.039153 |      0.0 | 0.000076 |   0.0 |  0.02496 |
    | 1997-07-08 | 20.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 | 18.8 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             18.651721 |                      0.0 |                 18.651721 |                  18.651721 |                  3.051087 |                     0.0 |                      0.0 |    15.600634 |             0.0 |              0.0 |         15.600634 |             37.166744 |                 100.0 |                 48.85611 |               65.496024 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945284 | 0.7 | 4.551539 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.817223 | 0.119656 |      0.0 | 0.033434 |  0.0 |  2.817223 |  2.448759 | 0.028342 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 163.412552 |  2.817223 |      0.0 | 0.119656 |      0.0 | 0.033434 |  2.388884 |      0.0 |   0.0597 |      0.0 | 0.000175 |   0.0 | 0.028342 |
    | 1997-08-08 | 20.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 | 18.8 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             17.267745 |                      0.0 |                 17.267745 |                  17.267745 |                   2.45312 |                     0.0 |                      0.0 |    14.814625 |             0.0 |              0.0 |         14.814625 |             42.798069 |                 100.0 |                 48.85611 |               65.567492 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.428125 | 0.7 | 3.767989 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.072608 | 0.127938 |      0.0 | 0.044013 |  0.0 |  3.072608 |   2.73175 | 0.031617 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 174.500005 |  3.072608 |      0.0 | 0.127938 |      0.0 | 0.044013 |  2.654405 |      0.0 | 0.077033 |      0.0 | 0.000313 |   0.0 | 0.031617 |
    | 1997-09-08 | 20.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 | 18.8 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             17.877445 |                      0.0 |                 17.877445 |                  17.877445 |                  2.688326 |                     0.0 |                      0.0 |    15.189119 |             0.0 |              0.0 |         15.189119 |             78.463126 |                 100.0 |                 48.85611 |               65.639719 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036736 | 0.7 | 4.121081 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.356591 | 0.136618 |      0.0 |   0.0551 |  0.0 |  3.356591 |  3.015033 | 0.034896 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 184.930615 |  3.356591 |      0.0 | 0.136618 |      0.0 |   0.0551 |   2.92256 |      0.0 | 0.091983 |      0.0 | 0.000489 |   0.0 | 0.034896 |
    | 1997-10-08 | 20.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 | 18.8 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |              17.53355 |                      0.0 |                  17.53355 |                   17.53355 |                   2.76572 |                     0.0 |                      0.0 |     14.76783 |             0.0 |              0.0 |          14.76783 |             74.333488 |                 100.0 |                 48.85611 |               65.712681 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.494046 | 0.7 | 4.450439 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.641259 | 0.144785 |      0.0 | 0.065531 |  0.0 |  3.641259 |  3.304149 | 0.038242 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 194.728603 |  3.641259 |      0.0 | 0.144785 |      0.0 | 0.065531 |  3.198274 |      0.0 | 0.105171 |      0.0 | 0.000703 |   0.0 | 0.038242 |
    | 1997-11-08 | 20.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 | 18.8 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             17.704758 |                      0.0 |                 17.704758 |                  17.704758 |                  2.799361 |                     0.0 |                      0.0 |    14.905396 |             0.0 |              0.0 |         14.905396 |             47.077876 |                 100.0 |                 48.85611 |               65.786357 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914762 | 0.7 | 4.993103 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.926571 | 0.152455 |      0.0 | 0.075329 |  0.0 |  3.926571 |  3.596218 | 0.041623 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.666265 | 203.681145 |  3.926571 |      0.0 | 0.152455 |      0.0 | 0.075329 |   3.47833 |      0.0 | 0.116935 |      0.0 | 0.000953 |   0.0 | 0.041623 |
    | 1997-12-08 | 20.0 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 | 18.8 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             17.786952 |                      0.0 |                 17.786952 |                  17.786952 |                  2.986484 |                     0.0 |                      0.0 |    14.800468 |             0.0 |              0.0 |         14.800468 |             56.493451 |                 100.0 |                 48.85611 |               65.860724 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566718 | 0.7 | 4.979876 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.204866 | 0.159465 | 0.009557 | 0.084281 |  0.0 |  4.204866 |  3.889124 | 0.045013 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |   212.3431 |  4.204866 |      0.0 | 0.159465 | 0.009557 | 0.084281 |  3.759031 |      0.0 | 0.127502 | 0.001355 | 0.001236 |   0.0 | 0.045013 |
    | 1997-13-08 | 20.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 | 18.8 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             17.184083 |                      0.0 |                 17.184083 |                  17.184083 |                  2.708378 |                     0.0 |                      0.0 |    14.475704 |             0.0 |              0.0 |         14.475704 |             88.271017 |                 100.0 |                 48.85611 |               65.935761 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996364 | 0.7 |  4.19851 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.492755 | 0.166246 | 0.045992 | 0.092943 |  0.0 |  4.492755 |  4.188438 | 0.048477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 221.446655 |  4.492755 |      0.0 | 0.166246 | 0.045992 | 0.092943 |  4.041079 |      0.0 | 0.137078 | 0.008732 | 0.001549 |   0.0 | 0.048477 |
    | 1997-14-08 | 20.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 | 18.8 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             14.968904 |                      0.0 |                 14.968904 |                  14.968904 |                  2.239353 |                     0.0 |                      0.0 |    12.729551 |             0.0 |              0.0 |         12.729551 |             52.308751 |                 100.0 |                 48.85611 |               66.011448 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983925 | 0.7 | 3.702096 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.818552 | 0.173373 | 0.100351 | 0.102047 |  0.0 |  4.818552 |  4.509936 | 0.052198 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 230.650235 |  4.818552 |      0.0 | 0.173373 | 0.100351 | 0.102047 |  4.335611 |      0.0 |  0.14595 | 0.026482 | 0.001893 |   0.0 | 0.052198 |
    | 1997-15-08 | 20.0 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 | 18.8 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             14.206821 |                      0.0 |                 14.206821 |                  14.206821 |                  2.021579 |                     0.0 |                      0.0 |    12.185243 |             0.0 |              0.0 |         12.185243 |             44.135508 |                 100.0 |                 48.85611 |               66.087764 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470984 | 0.7 | 3.155298 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.176929 | 0.180579 |  0.16802 |  0.11125 |  0.0 |  5.176929 |  4.865729 | 0.056316 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 239.958158 |  5.176929 |      0.0 | 0.180579 |  0.16802 |  0.11125 |  4.653113 |      0.0 | 0.154363 | 0.055986 | 0.002268 |   0.0 | 0.056316 |
    | 1997-16-08 | 20.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 | 18.8 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             14.881951 |                      0.0 |                 14.881951 |                  14.881951 |                  2.487815 |                     0.0 |                      0.0 |    12.394137 |             0.0 |              0.0 |         12.394137 |            108.641252 |                 100.0 |                 48.85611 |                66.16469 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.08134 | 0.7 | 2.708461 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.575829 | 0.187866 | 0.247347 | 0.120558 |  0.0 |  5.575829 |  5.262676 | 0.060911 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 249.218097 |  5.575829 |      0.0 | 0.187866 | 0.247347 | 0.120558 |   5.00011 |      0.0 | 0.162462 | 0.097429 | 0.002675 |   0.0 | 0.060911 |
    | 1997-17-08 | 20.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 | 18.8 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             15.465744 |                      0.0 |                 15.465744 |                  15.465744 |                  2.520953 |                     0.0 |                      0.0 |     12.94479 |             0.0 |              0.0 |          12.94479 |             74.333488 |                 100.0 |                 48.85611 |               66.242207 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175424 | 0.7 | 3.378706 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.018699 | 0.195116 | 0.335804 | 0.129818 |  0.0 |  6.018699 |  5.706462 | 0.066047 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 257.259955 |  6.018699 |      0.0 | 0.195116 | 0.335804 | 0.129818 |  5.382634 |      0.0 | 0.170336 | 0.150377 | 0.003114 |   0.0 | 0.066047 |
    | 1997-18-08 | 20.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 | 18.8 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             15.289825 |                      0.0 |                 15.289825 |                  15.289825 |                  2.702703 |                     0.0 |                      0.0 |    12.587122 |             0.0 |              0.0 |         12.587122 |             70.616813 |                 100.0 |                 48.85611 |               66.320298 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.738005 | 0.7 | 3.779831 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.451503 | 0.201412 | 0.419614 |  0.13786 |  0.0 |  6.451503 |  6.184438 | 0.071579 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 264.369735 |  6.451503 |      0.0 | 0.201412 | 0.419614 |  0.13786 |  5.790712 |      0.0 | 0.177908 | 0.212236 | 0.003582 |   0.0 | 0.071579 |
    | 1997-19-08 | 20.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 | 18.8 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             13.113987 |                      0.0 |                 13.113987 |                  13.113987 |                  2.307856 |                     0.0 |                      0.0 |    10.806131 |             0.0 |              0.0 |         10.806131 |             54.320626 |                 100.0 |                 48.85611 |               66.398945 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.867083 | 0.7 |  3.62847 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.883173 | 0.206978 | 0.498724 |  0.14497 |  0.0 |  6.883173 |  6.678482 | 0.077297 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 271.107419 |  6.883173 |      0.0 | 0.206978 | 0.498724 |  0.14497 |  6.210029 |      0.0 | 0.185032 | 0.279345 | 0.004076 |   0.0 | 0.077297 |
    | 1997-20-08 | 20.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 | 18.8 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             15.350398 |                      0.0 |                 15.350398 |                  15.350398 |                  2.688476 |                     0.0 |                      0.0 |    12.661922 |             0.0 |              0.0 |         12.661922 |             39.231563 |                 100.0 |                 48.85611 |                66.47813 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.829385 | 0.7 | 4.625035 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.348865 | 0.212253 | 0.577784 | 0.151707 |  0.0 |  7.348865 |  7.189836 | 0.083216 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 276.291774 |  7.348865 |      0.0 | 0.212253 | 0.577784 | 0.151707 |  6.643867 |      0.0 | 0.191695 | 0.349681 | 0.004593 |   0.0 | 0.083216 |
    | 1997-21-08 | 20.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 | 18.8 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             16.161308 |                      0.0 |                 16.161308 |                  16.161308 |                  3.308913 |                     0.0 |                      0.0 |    12.852395 |             0.0 |              0.0 |         12.852395 |             56.493451 |                 100.0 |                 48.85611 |               66.557838 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.32284 | 0.7 | 4.757012 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.757458 | 0.216312 | 0.641202 | 0.156892 |  0.0 |  7.757458 |  7.700952 | 0.089131 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 280.862897 |  7.757458 |      0.0 | 0.216312 | 0.641202 | 0.156892 |  7.077861 |      0.0 | 0.197811 |  0.42015 | 0.005129 |   0.0 | 0.089131 |
    | 1997-22-08 | 20.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 | 18.8 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              5.948207 |                      0.0 |                  5.948207 |                   5.948207 |                   0.60663 |                     0.0 |                      0.0 |     5.341577 |             0.0 |              0.0 |          5.341577 |             67.254108 |                 100.0 |                 48.85611 |               66.638052 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  2.56747 | 0.7 | 1.361228 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.166601 | 0.219891 | 0.698918 | 0.161463 |  0.0 |  8.166601 |  8.197047 | 0.094873 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 288.354796 |  8.166601 |      0.0 | 0.219891 | 0.698918 | 0.161463 |   7.50015 |      0.0 | 0.203305 | 0.487911 | 0.005681 |   0.0 | 0.094873 |
    | 1997-23-08 | 20.0 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 | 18.8 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             11.290596 |                      0.0 |                 11.290596 |                  11.290596 |                  1.641158 |                     0.0 |                      0.0 |     9.649438 |             0.0 |              0.0 |          9.649438 |             94.155751 |                 100.0 |                 48.85611 |               66.718758 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374286 | 0.7 | 2.097563 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.984192 | 0.225757 | 0.797025 | 0.168955 |  0.0 |  8.984192 |  8.795013 | 0.101794 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 294.181304 |  8.984192 |      0.0 | 0.225757 | 0.797025 | 0.168955 |  8.021462 |      0.0 | 0.208607 | 0.558691 | 0.006253 |   0.0 | 0.101794 |
    | 1997-24-08 | 20.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 | 18.8 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             14.219527 |                      0.0 |                 14.219527 |                  14.219527 |                  1.868809 |                     0.0 |                      0.0 |    12.350718 |             0.0 |              0.0 |         12.350718 |             64.197103 |                 100.0 |                 48.85611 |                66.79994 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573178 | 0.7 | 3.649927 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  9.834859 | 0.230318 | 0.876239 | 0.174781 |  0.0 |  9.834859 |   9.54435 | 0.110467 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 34.239011 | 297.515179 |  9.834859 |      0.0 | 0.230318 | 0.876239 | 0.174781 |   8.68947 |      0.0 | 0.213876 | 0.634158 | 0.006846 |   0.0 | 0.110467 |
    | 1997-25-08 | 20.0 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 | 18.8 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             10.960427 |                      0.0 |                 10.960427 |                  10.960427 |                  1.401603 |                     0.0 |                      0.0 |     9.558824 |             0.0 |              0.0 |          9.558824 |             74.333488 |                 100.0 |                 48.85611 |               66.881585 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467819 | 0.7 | 2.895981 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 10.482817 | 0.232928 | 0.922678 | 0.178115 |  0.0 | 10.482817 | 10.322599 | 0.119475 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  300.90266 | 10.482817 |      0.0 | 0.232928 | 0.922678 | 0.178115 |  9.390481 |      0.0 | 0.218677 | 0.705987 | 0.007454 |   0.0 | 0.119475 |
    | 1997-26-08 | 20.0 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 | 18.8 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |              9.585999 |                      0.0 |                  9.585999 |                   9.585999 |                  1.261689 |                     0.0 |                      0.0 |      8.32431 |             0.0 |              0.0 |           8.32431 |             61.405925 |                 100.0 |                 48.85611 |               66.963678 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569973 | 0.7 | 2.071555 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  11.39404 |  0.23558 | 0.970674 | 0.181503 |  0.0 |  11.39404 | 11.136685 | 0.128897 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 304.149308 |  11.39404 |      0.0 |  0.23558 | 0.970674 | 0.181503 | 10.134526 |      0.0 | 0.222894 | 0.771194 | 0.008071 |   0.0 | 0.128897 |
    | 1997-27-08 | 20.0 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 | 18.8 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             10.870274 |                      0.0 |                 10.870274 |                  10.870274 |                  1.498964 |                     0.0 |                      0.0 |      9.37131 |             0.0 |              0.0 |           9.37131 |             67.254108 |                 100.0 |                 48.85611 |               67.046207 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.567075 | 0.7 | 2.108966 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 13.249308 | 0.238122 | 1.017431 | 0.184749 |  0.0 | 13.249308 | 12.737269 | 0.147422 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 305.450732 |      12.0 | 1.249308 | 0.238122 | 1.017431 | 0.184749 | 10.877952 | 0.792359 | 0.226674 | 0.831586 | 0.008698 |   0.0 | 0.147422 |
    | 1997-28-08 | 20.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 | 18.8 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              5.758296 |                      0.0 |                  5.758296 |                   5.758296 |                  0.589077 |                     0.0 |                      0.0 |     5.169219 |             0.0 |              0.0 |          5.169219 |             54.320626 |                 100.0 |                 48.85611 |               67.129158 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621374 | 0.7 | 1.331712 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.550732 | 0.239141 | 1.036378 | 0.186051 |  0.0 | 14.550732 | 14.562531 | 0.168548 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 306.206718 |      12.0 | 2.550732 | 0.239141 | 1.036378 | 0.186051 | 11.401263 | 2.037672 | 0.229904 | 0.884361 | 0.009331 |   0.0 | 0.168548 |
    | 1997-29-08 | 20.0 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 | 18.8 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |               5.53921 |                      0.0 |                   5.53921 |                    5.53921 |                  0.708412 |                     0.0 |                      0.0 |     4.830799 |             0.0 |              0.0 |          4.830799 |             54.320626 |                 100.0 |                 48.85611 |               67.212519 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027722 | 0.7 | 0.881862 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 15.306718 | 0.239733 | 1.047438 | 0.186807 |  0.0 | 15.306718 |  15.83846 | 0.183316 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  306.64416 |      12.0 | 3.306718 | 0.239733 | 1.047438 | 0.186807 | 11.680507 |  2.98861 | 0.232477 |   0.9269 | 0.009966 |   0.0 | 0.183316 |
    | 1997-30-08 | 20.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 | 18.8 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |              9.344591 |                      0.0 |                  9.344591 |                   9.344591 |                  1.881434 |                     0.0 |                      0.0 |     7.463158 |             0.0 |              0.0 |          7.463158 |             67.254108 |                 100.0 |                 48.85611 |               67.296277 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393609 | 0.7 | 1.913447 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  15.74416 | 0.240076 | 1.053855 | 0.187244 |  0.0 |  15.74416 | 16.593262 | 0.192052 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 305.605378 |      12.0 |  3.74416 | 0.240076 | 1.053855 | 0.187244 | 11.829515 | 3.558369 | 0.234481 | 0.960296 |   0.0106 |   0.0 | 0.192052 |
    | 1997-31-08 | 20.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 | 18.8 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              8.321079 |                      0.0 |                  8.321079 |                   8.321079 |                  1.511299 |                     0.0 |                      0.0 |      6.80978 |             0.0 |              0.0 |           6.80978 |             67.254108 |                 100.0 |                 48.85611 |               67.380423 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.322101 | 0.7 | 1.882222 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.705378 | 0.239262 | 1.038637 | 0.186205 |  0.0 | 14.705378 | 16.209748 | 0.187613 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 305.653673 |      12.0 | 2.705378 | 0.239262 | 1.038637 | 0.186205 | 11.909027 | 3.070262 | 0.235874 | 0.983354 | 0.011231 |   0.0 | 0.187613 |

>>> inputs.nied.series = nied

.. _lland_v4_wasser_daily:

water
-----

.. integration-test::

    >>> lnk(WASSER)
    >>> surfaceresistance.wasser_aug = 0.0
    >>> cropheight.wasser_aug = 0.05
    >>> albedo.wasser_aug = 0.7
    >>> conditions_wasser = test(
    ...     "lland_v4_wasser_daily",
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                      0.0 |                  4.931319 |                   4.931319 |                  1.877708 |                     0.0 |                      0.0 |      3.05361 |             0.0 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 | 0.132232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                      0.0 |                  3.308967 |                   3.308967 |                   0.96939 |                     0.0 |                      0.0 |     2.339577 |             0.0 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                      0.0 |                  4.261282 |                   4.261282 |                  1.386217 |                     0.0 |                      0.0 |     2.875064 |             0.0 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                      0.0 |                  6.973011 |                   6.973011 |                  3.108707 |                     0.0 |                      0.0 |     3.864304 |             0.0 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                      0.0 |                  7.461037 |                   7.461037 |                  3.414868 |                     0.0 |                      0.0 |     4.046169 |             0.0 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                      0.0 |                  7.119557 |                   7.119557 |                  2.845741 |                     0.0 |                      0.0 |     4.273816 |             0.0 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                      0.0 |                  7.362521 |                   7.362521 |                  3.051087 |                     0.0 |                      0.0 |     4.311434 |             0.0 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                      0.0 |                  6.816215 |                   6.816215 |                   2.45312 |                     0.0 |                      0.0 |     4.363095 |             0.0 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                      0.0 |                  7.056886 |                   7.056886 |                  2.688326 |                     0.0 |                      0.0 |      4.36856 |             0.0 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                      0.0 |                  6.921138 |                   6.921138 |                   2.76572 |                     0.0 |                      0.0 |     4.155419 |             0.0 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                      0.0 |                   6.98872 |                    6.98872 |                  2.799361 |                     0.0 |                      0.0 |     4.189359 |             0.0 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                      0.0 |                  7.021165 |                   7.021165 |                  2.986484 |                     0.0 |                      0.0 |     4.034682 |             0.0 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                      0.0 |                   6.78319 |                    6.78319 |                  2.708378 |                     0.0 |                      0.0 |     4.074812 |             0.0 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                      0.0 |                  5.908778 |                   5.908778 |                  2.239353 |                     0.0 |                      0.0 |     3.669425 |             0.0 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                      0.0 |                  5.607956 |                   5.607956 |                  2.021579 |                     0.0 |                      0.0 |     3.586377 |             0.0 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 | 0.011672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                      0.0 |                  5.874454 |                   5.874454 |                  2.487815 |                     0.0 |                      0.0 |      3.38664 |             0.0 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                      0.0 |                  6.104899 |                   6.104899 |                  2.520953 |                     0.0 |                      0.0 |     3.583945 |             0.0 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                      0.0 |                  6.035457 |                   6.035457 |                  2.702703 |                     0.0 |                      0.0 |     3.332754 |             0.0 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                      0.0 |                  5.176574 |                   5.176574 |                  2.307856 |                     0.0 |                      0.0 |     2.868718 |             0.0 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                      0.0 |                  6.059368 |                   6.059368 |                  2.688476 |                     0.0 |                      0.0 |     3.370892 |             0.0 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                      0.0 |                  6.379464 |                   6.379464 |                  3.308913 |                     0.0 |                      0.0 |     3.070551 |             0.0 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                      0.0 |                  2.347976 |                   2.347976 |                   0.60663 |                     0.0 |                      0.0 |     1.741346 |             0.0 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 | 0.012428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                      0.0 |                  4.456814 |                   4.456814 |                  1.641158 |                     0.0 |                      0.0 |     2.815656 |             0.0 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |    0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                      0.0 |                  5.612971 |                   5.612971 |                  1.868809 |                     0.0 |                      0.0 |     3.744162 |             0.0 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                      0.0 |                  4.326484 |                   4.326484 |                  1.401603 |                     0.0 |                      0.0 |     2.924881 |             0.0 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                      0.0 |                  3.783947 |                   3.783947 |                  1.261689 |                     0.0 |                      0.0 |     2.522258 |             0.0 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |  0.10177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                      0.0 |                  4.290898 |                   4.290898 |                  1.498964 |                     0.0 |                      0.0 |     2.791934 |             0.0 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 | 0.058014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                      0.0 |                  2.273012 |                   2.273012 |                  0.589077 |                     0.0 |                      0.0 |     1.683934 |             0.0 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 | 0.153995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                      0.0 |                   2.18653 |                    2.18653 |                  0.708412 |                     0.0 |                      0.0 |     1.478119 |             0.0 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 | 0.093058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                      0.0 |                  3.688655 |                   3.688655 |                  1.881434 |                     0.0 |                      0.0 |     1.807221 |             0.0 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                      0.0 |                  3.284636 |                   3.284636 |                  1.511299 |                     0.0 |                      0.0 |     1.773338 |             0.0 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

.. _lland_v4_wasser_routing_daily:

water (routing)
---------------

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test("lland_v4_wasser_routing_daily",
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                      0.0 |                  4.931319 |                   4.931319 |                  1.877708 |                     0.0 |                      0.0 |      3.05361 |             0.0 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.152803 | 0.152232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.152232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                      0.0 |                  3.308967 |                   3.308967 |                   0.96939 |                     0.0 |                      0.0 |     2.339577 |             0.0 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.95655 | 0.011071 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.011071 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                      0.0 |                  4.261282 |                   4.261282 |                  1.386217 |                     0.0 |                      0.0 |     2.875064 |             0.0 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.635635 | 0.007357 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.007357 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                      0.0 |                  6.973011 |                   6.973011 |                  3.108707 |                     0.0 |                      0.0 |     3.864304 |             0.0 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.256106 | 0.002964 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002964 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                      0.0 |                  7.461037 |                   7.461037 |                  3.414868 |                     0.0 |                      0.0 |     4.046169 |             0.0 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                      0.0 |                  7.119557 |                   7.119557 |                  2.845741 |                     0.0 |                      0.0 |     4.273816 |             0.0 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                      0.0 |                  7.362521 |                   7.362521 |                  3.051087 |                     0.0 |                      0.0 |     4.311434 |             0.0 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                      0.0 |                  6.816215 |                   6.816215 |                   2.45312 |                     0.0 |                      0.0 |     4.363095 |             0.0 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                      0.0 |                  7.056886 |                   7.056886 |                  2.688326 |                     0.0 |                      0.0 |      4.36856 |             0.0 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                      0.0 |                  6.921138 |                   6.921138 |                   2.76572 |                     0.0 |                      0.0 |     4.155419 |             0.0 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                      0.0 |                   6.98872 |                    6.98872 |                  2.799361 |                     0.0 |                      0.0 |     4.189359 |             0.0 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                      0.0 |                  7.021165 |                   7.021165 |                  2.986484 |                     0.0 |                      0.0 |     4.034682 |             0.0 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |    1.822 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                      0.0 |                   6.78319 |                    6.78319 |                  2.708378 |                     0.0 |                      0.0 |     4.074812 |             0.0 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                      0.0 |                  5.908778 |                   5.908778 |                  2.239353 |                     0.0 |                      0.0 |     3.669425 |             0.0 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055264 |  0.00064 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00064 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                      0.0 |                  5.607956 |                   5.607956 |                  2.021579 |                     0.0 |                      0.0 |     3.586377 |             0.0 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.736458 | 0.031672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.031672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                      0.0 |                  5.874454 |                   5.874454 |                  2.487815 |                     0.0 |                      0.0 |      3.38664 |             0.0 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.444099 |  0.00514 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00514 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                      0.0 |                  6.104899 |                   6.104899 |                  2.520953 |                     0.0 |                      0.0 |     3.583945 |             0.0 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.217257 | 0.002515 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002515 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                      0.0 |                  6.035457 |                   6.035457 |                  2.702703 |                     0.0 |                      0.0 |     3.332754 |             0.0 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.133428 | 0.001544 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.001544 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                      0.0 |                  5.176574 |                   5.176574 |                  2.307856 |                     0.0 |                      0.0 |     2.868718 |             0.0 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.186433 | 0.002158 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002158 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                      0.0 |                  6.059368 |                   6.059368 |                  2.688476 |                     0.0 |                      0.0 |     3.370892 |             0.0 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                      0.0 |                  6.379464 |                   6.379464 |                  3.308913 |                     0.0 |                      0.0 |     3.070551 |             0.0 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                      0.0 |                  2.347976 |                   2.347976 |                   0.60663 |                     0.0 |                      0.0 |     1.741346 |             0.0 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.801812 | 0.032428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.032428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                      0.0 |                  4.456814 |                   4.456814 |                  1.641158 |                     0.0 |                      0.0 |     2.815656 |             0.0 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.92579 | 0.010715 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.010715 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                      0.0 |                  5.612971 |                   5.612971 |                  1.868809 |                     0.0 |                      0.0 |     3.744162 |             0.0 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055768 | 0.000645 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.000645 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                      0.0 |                  4.326484 |                   4.326484 |                  1.401603 |                     0.0 |                      0.0 |     2.924881 |             0.0 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.447898 | 0.005184 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.005184 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                      0.0 |                  3.783947 |                   3.783947 |                  1.261689 |                     0.0 |                      0.0 |     2.522258 |             0.0 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 10.520904 |  0.12177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.12177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                      0.0 |                  4.290898 |                   4.290898 |                  1.498964 |                     0.0 |                      0.0 |     2.791934 |             0.0 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   6.74044 | 0.078014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.078014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                      0.0 |                  2.273012 |                   2.273012 |                  0.589077 |                     0.0 |                      0.0 |     1.683934 |             0.0 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 15.033141 | 0.173995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.173995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                      0.0 |                   2.18653 |                    2.18653 |                  0.708412 |                     0.0 |                      0.0 |     1.478119 |             0.0 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  9.768224 | 0.113058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.113058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                      0.0 |                  3.688655 |                   3.688655 |                  1.881434 |                     0.0 |                      0.0 |     1.807221 |             0.0 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.82735 | 0.009576 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009576 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                      0.0 |                  3.284636 |                   3.284636 |                  1.511299 |                     0.0 |                      0.0 |     1.773338 |             0.0 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.815922 | 0.009444 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009444 |

.. _lland_v4_wasser_negq_daily:

water (negative runoff)
-----------------------

.. integration-test::

    >>> negq(True)
    >>> inlet.sequences.sim.series = 0.0
    >>> test("lland_v4_wasser_negq_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |        qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |    outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                      0.0 |                  4.931319 |                   4.931319 |                  1.877708 |                     0.0 |                      0.0 |      3.05361 |             0.0 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 |  0.132232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                      0.0 |                  3.308967 |                   3.308967 |                   0.96939 |                     0.0 |                      0.0 |     2.339577 |             0.0 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.77145 | -0.008929 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.008929 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                      0.0 |                  4.261282 |                   4.261282 |                  1.386217 |                     0.0 |                      0.0 |     2.875064 |             0.0 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.092365 | -0.012643 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.012643 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                      0.0 |                  6.973011 |                   6.973011 |                  3.108707 |                     0.0 |                      0.0 |     3.864304 |             0.0 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.471894 | -0.017036 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017036 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                      0.0 |                  7.461037 |                   7.461037 |                  3.414868 |                     0.0 |                      0.0 |     4.046169 |             0.0 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.729733 |  -0.02002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.02002 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                      0.0 |                  7.119557 |                   7.119557 |                  2.845741 |                     0.0 |                      0.0 |     4.273816 |             0.0 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826087 | -0.021135 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021135 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                      0.0 |                  7.362521 |                   7.362521 |                  3.051087 |                     0.0 |                      0.0 |     4.311434 |             0.0 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.041956 | -0.023634 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.023634 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                      0.0 |                  6.816215 |                   6.816215 |                   2.45312 |                     0.0 |                      0.0 |     4.363095 |             0.0 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.800053 | -0.020834 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020834 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                      0.0 |                  7.056886 |                   7.056886 |                  2.688326 |                     0.0 |                      0.0 |      4.36856 |             0.0 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |  1.82186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -1.82186 | -0.021086 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021086 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                      0.0 |                  6.921138 |                   6.921138 |                   2.76572 |                     0.0 |                      0.0 |     4.155419 |             0.0 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 | 1.891469 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.891469 | -0.021892 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021892 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                      0.0 |                   6.98872 |                    6.98872 |                  2.799361 |                     0.0 |                      0.0 |     4.189359 |             0.0 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 | 2.106575 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.106575 | -0.024382 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.024382 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                      0.0 |                  7.021165 |                   7.021165 |                  2.986484 |                     0.0 |                      0.0 |     4.034682 |             0.0 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 | 2.036929 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.942929 | -0.022488 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022488 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                      0.0 |                   6.78319 |                    6.78319 |                  2.708378 |                     0.0 |                      0.0 |     4.074812 |             0.0 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | 1.808559 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.808559 | -0.020932 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020932 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                      0.0 |                  5.908778 |                   5.908778 |                  2.239353 |                     0.0 |                      0.0 |     3.669425 |             0.0 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672736 |  -0.01936 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01936 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                      0.0 |                  5.607956 |                   5.607956 |                  2.021579 |                     0.0 |                      0.0 |     3.586377 |             0.0 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 |  0.011672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                      0.0 |                  5.874454 |                   5.874454 |                  2.487815 |                     0.0 |                      0.0 |      3.38664 |             0.0 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.283901 |  -0.01486 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01486 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                      0.0 |                  6.104899 |                   6.104899 |                  2.520953 |                     0.0 |                      0.0 |     3.583945 |             0.0 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.510743 | -0.017485 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017485 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                      0.0 |                  6.035457 |                   6.035457 |                  2.702703 |                     0.0 |                      0.0 |     3.332754 |             0.0 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.594572 | -0.018456 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.018456 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                      0.0 |                  5.176574 |                   5.176574 |                  2.307856 |                     0.0 |                      0.0 |     2.868718 |             0.0 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.541567 | -0.017842 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017842 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                      0.0 |                  6.059368 |                   6.059368 |                  2.688476 |                     0.0 |                      0.0 |     3.370892 |             0.0 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 | 1.939383 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.939383 | -0.022447 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022447 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                      0.0 |                  6.379464 |                   6.379464 |                  3.308913 |                     0.0 |                      0.0 |     3.070551 |             0.0 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 | 1.826811 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826811 | -0.021144 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021144 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                      0.0 |                  2.347976 |                   2.347976 |                   0.60663 |                     0.0 |                      0.0 |     1.741346 |             0.0 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 |  0.012428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                      0.0 |                  4.456814 |                   4.456814 |                  1.641158 |                     0.0 |                      0.0 |     2.815656 |             0.0 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.80221 | -0.009285 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.009285 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                      0.0 |                  5.612971 |                   5.612971 |                  1.868809 |                     0.0 |                      0.0 |     3.744162 |             0.0 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672232 | -0.019355 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.019355 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                      0.0 |                  4.326484 |                   4.326484 |                  1.401603 |                     0.0 |                      0.0 |     2.924881 |             0.0 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.280102 | -0.014816 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.014816 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                      0.0 |                  3.783947 |                   3.783947 |                  1.261689 |                     0.0 |                      0.0 |     2.522258 |             0.0 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |   0.10177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                      0.0 |                  4.290898 |                   4.290898 |                  1.498964 |                     0.0 |                      0.0 |     2.791934 |             0.0 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 |  0.058014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                      0.0 |                  2.273012 |                   2.273012 |                  0.589077 |                     0.0 |                      0.0 |     1.683934 |             0.0 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 |  0.153995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                      0.0 |                   2.18653 |                    2.18653 |                  0.708412 |                     0.0 |                      0.0 |     1.478119 |             0.0 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 |  0.093058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                      0.0 |                  3.688655 |                   3.688655 |                  1.881434 |                     0.0 |                      0.0 |     1.807221 |             0.0 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.90065 | -0.010424 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010424 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                      0.0 |                  3.284636 |                   3.284636 |                  1.511299 |                     0.0 |                      0.0 |     1.773338 |             0.0 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -0.912078 | -0.010556 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010556 |

.. _lland_v4_see_daily:

lakes
-----

.. integration-test::

    >>> lnk(SEE)
    >>> surfaceresistance.see_aug = 0.0
    >>> cropheight.see_aug = 0.05
    >>> albedo.see_aug = 0.7
    >>> negq(False)
    >>> test("lland_v4_see_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz | qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                      0.0 |                  4.931319 |                   4.931319 |                  1.877708 |                     0.0 |                      0.0 |      3.05361 |             0.0 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.020478 | 0.000237 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 11.424803 |   0.0 |   0.0 |   0.0 |   0.0 |  0.020478 |   0.0 | 0.000237 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                      0.0 |                  3.308967 |                   3.308967 |                   0.96939 |                     0.0 |                      0.0 |     2.339577 |             0.0 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039475 | 0.000457 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.77145 |   0.0 |   0.0 |   0.0 |   0.0 |  0.039475 |   0.0 | 0.000457 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                      0.0 |                  4.261282 |                   4.261282 |                  1.386217 |                     0.0 |                      0.0 |     2.875064 |             0.0 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035994 | 0.000417 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.092365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035994 |   0.0 | 0.000417 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                      0.0 |                  6.973011 |                   6.973011 |                  3.108707 |                     0.0 |                      0.0 |     3.864304 |             0.0 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.031272 | 0.000362 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.471894 |   0.0 |   0.0 |   0.0 |   0.0 |  0.031272 |   0.0 | 0.000362 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                      0.0 |                  7.461037 |                   7.461037 |                  3.414868 |                     0.0 |                      0.0 |     4.046169 |             0.0 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.025424 | 0.000294 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.729733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.025424 |   0.0 | 0.000294 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                      0.0 |                  7.119557 |                   7.119557 |                  2.845741 |                     0.0 |                      0.0 |     4.273816 |             0.0 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.018963 | 0.000219 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826087 |   0.0 |   0.0 |   0.0 |   0.0 |  0.018963 |   0.0 | 0.000219 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                      0.0 |                  7.362521 |                   7.362521 |                  3.051087 |                     0.0 |                      0.0 |     4.311434 |             0.0 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011966 | 0.000138 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.041956 |   0.0 |   0.0 |   0.0 |   0.0 |  0.011966 |   0.0 | 0.000138 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                      0.0 |                  6.816215 |                   6.816215 |                   2.45312 |                     0.0 |                      0.0 |     4.363095 |             0.0 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005041 | 0.000058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.800053 |   0.0 |   0.0 |   0.0 |   0.0 |  0.005041 |   0.0 | 0.000058 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                      0.0 |                  7.056886 |                   7.056886 |                  2.688326 |                     0.0 |                      0.0 |      4.36856 |             0.0 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 | 1.820396 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -1.82186 |   0.0 |   0.0 |   0.0 |   0.0 | -0.001465 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                      0.0 |                  6.921138 |                   6.921138 |                   2.76572 |                     0.0 |                      0.0 |     4.155419 |             0.0 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 | 1.883357 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.891469 |   0.0 |   0.0 |   0.0 |   0.0 | -0.008111 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                      0.0 |                   6.98872 |                    6.98872 |                  2.799361 |                     0.0 |                      0.0 |     4.189359 |             0.0 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 | 2.091331 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.106575 |   0.0 |   0.0 |   0.0 |   0.0 | -0.015244 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                      0.0 |                  7.021165 |                   7.021165 |                  2.986484 |                     0.0 |                      0.0 |     4.034682 |             0.0 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 | 2.014486 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.942929 |   0.0 |   0.0 |   0.0 |   0.0 | -0.022443 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                      0.0 |                   6.78319 |                    6.78319 |                  2.708378 |                     0.0 |                      0.0 |     4.074812 |             0.0 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | 1.779476 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.808559 |   0.0 |   0.0 |   0.0 |   0.0 | -0.029083 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                      0.0 |                  5.908778 |                   5.908778 |                  2.239353 |                     0.0 |                      0.0 |     3.669425 |             0.0 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.637521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672736 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035215 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                      0.0 |                  5.607956 |                   5.607956 |                  2.021579 |                     0.0 |                      0.0 |     3.586377 |             0.0 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.493266 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.008458 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036276 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                      0.0 |                  5.874454 |                   5.874454 |                  2.487815 |                     0.0 |                      0.0 |      3.38664 |             0.0 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.247259 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.283901 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036642 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                      0.0 |                  6.104899 |                   6.104899 |                  2.520953 |                     0.0 |                      0.0 |     3.583945 |             0.0 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.469226 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.510743 |   0.0 |   0.0 |   0.0 |   0.0 | -0.041517 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                      0.0 |                  6.035457 |                   6.035457 |                  2.702703 |                     0.0 |                      0.0 |     3.332754 |             0.0 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.547641 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.594572 |   0.0 |   0.0 |   0.0 |   0.0 | -0.046931 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                      0.0 |                  5.176574 |                   5.176574 |                  2.307856 |                     0.0 |                      0.0 |     2.868718 |             0.0 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.489186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.541567 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.05238 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                      0.0 |                  6.059368 |                   6.059368 |                  2.688476 |                     0.0 |                      0.0 |     3.370892 |             0.0 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 | 1.880954 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.939383 |   0.0 |   0.0 |   0.0 |   0.0 | -0.058428 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                      0.0 |                  6.379464 |                   6.379464 |                  3.308913 |                     0.0 |                      0.0 |     3.070551 |             0.0 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 | 1.761846 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826811 |   0.0 |   0.0 |   0.0 |   0.0 | -0.064965 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                      0.0 |                  2.347976 |                   2.347976 |                   0.60663 |                     0.0 |                      0.0 |     1.741346 |             0.0 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.740109 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.073812 |   0.0 |   0.0 |   0.0 |   0.0 | -0.066078 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                      0.0 |                  4.456814 |                   4.456814 |                  1.641158 |                     0.0 |                      0.0 |     2.815656 |             0.0 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 | 1.018853 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.80221 |   0.0 |   0.0 |   0.0 |   0.0 | -0.065357 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                      0.0 |                  5.612971 |                   5.612971 |                  1.868809 |                     0.0 |                      0.0 |     3.744162 |             0.0 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.602675 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672232 |   0.0 |   0.0 |   0.0 |   0.0 | -0.069557 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                      0.0 |                  4.326484 |                   4.326484 |                  1.401603 |                     0.0 |                      0.0 |     2.924881 |             0.0 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.299507 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.280102 |   0.0 |   0.0 |   0.0 |   0.0 | -0.074595 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                      0.0 |                  3.783947 |                   3.783947 |                  1.261689 |                     0.0 |                      0.0 |     2.522258 |             0.0 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.016237 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.792904 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.06086 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                      0.0 |                  4.290898 |                   4.290898 |                  1.498964 |                     0.0 |                      0.0 |     2.791934 |             0.0 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 | 1.061645 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |   5.01244 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035916 |   0.0 |      0.0 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                      0.0 |                  2.273012 |                   2.273012 |                  0.589077 |                     0.0 |                      0.0 |     1.683934 |             0.0 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.791893 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 13.305141 |   0.0 |   0.0 |   0.0 |   0.0 | -0.002966 |   0.0 |      0.0 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                      0.0 |                   2.18653 |                    2.18653 |                  0.708412 |                     0.0 |                      0.0 |     1.478119 |             0.0 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035276 | 0.000408 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.040224 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035276 |   0.0 | 0.000408 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                      0.0 |                  3.688655 |                   3.688655 |                  1.881434 |                     0.0 |                      0.0 |     1.807221 |             0.0 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047929 | 0.000555 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.90065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.047929 |   0.0 | 0.000555 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                      0.0 |                  3.284636 |                   3.284636 |                  1.511299 |                     0.0 |                      0.0 |     1.773338 |             0.0 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.04451 | 0.000515 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -0.912078 |   0.0 |   0.0 |   0.0 |   0.0 |   0.04451 |   0.0 | 0.000515 |

.. _lland_v4_fluss_daily:

streams
-------

.. integration-test::

    >>> lnk(FLUSS)
    >>> surfaceresistance.fluss_aug = 0.0
    >>> cropheight.fluss_aug = 0.05
    >>> albedo.fluss_aug = 0.7
    >>> test("lland_v4_fluss_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |       evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa |     qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |     qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | 0.0 | nan | 0.0 |              4.931319 |                      0.0 |                  4.931319 |                   4.931319 |                  1.877708 |                     0.0 |                      0.0 |      3.05361 |             0.0 |              0.0 |           3.05361 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 |  1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 11.424803 | 2.941276 | 0.034043 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | 11.424803 |      0.0 |   0.0 |   0.0 |  0.0 |  2.941276 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034043 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | 0.0 | nan | 0.0 |              3.308967 |                      0.0 |                  3.308967 |                   3.308967 |                   0.96939 |                     0.0 |                      0.0 |     2.339577 |             0.0 |              0.0 |          2.339577 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |   0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.77145 | 3.758016 | 0.043496 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.77145 |      0.0 |   0.0 |   0.0 |  0.0 |  3.758016 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043496 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | 0.0 | nan | 0.0 |              4.261282 |                      0.0 |                  4.261282 |                   4.261282 |                  1.386217 |                     0.0 |                      0.0 |     2.875064 |             0.0 |              0.0 |          2.875064 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 |  1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.092365 | 1.562905 | 0.018089 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.092365 |      0.0 |   0.0 |   0.0 |  0.0 |  1.562905 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018089 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | 0.0 | nan | 0.0 |              6.973011 |                      0.0 |                  6.973011 |                   6.973011 |                  3.108707 |                     0.0 |                      0.0 |     3.864304 |             0.0 |              0.0 |          3.864304 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 |  1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.471894 | 0.226808 | 0.002625 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.471894 |      0.0 |   0.0 |   0.0 |  0.0 |  0.226808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002625 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | 0.0 | nan | 0.0 |              7.461037 |                      0.0 |                  7.461037 |                   7.461037 |                  3.414868 |                     0.0 |                      0.0 |     4.046169 |             0.0 |              0.0 |          4.046169 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |  1.097905 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.729733 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.729733 |      0.0 |   0.0 |   0.0 |  0.0 | -0.631827 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | 0.0 | nan | 0.0 |              7.119557 |                      0.0 |                  7.119557 |                   7.119557 |                  2.845741 |                     0.0 |                      0.0 |     4.273816 |             0.0 |              0.0 |          4.273816 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |  0.657403 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826087 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826087 |      0.0 |   0.0 |   0.0 |  0.0 | -1.168684 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | 0.0 | nan | 0.0 |              7.362521 |                      0.0 |                  7.362521 |                   7.362521 |                  3.051087 |                     0.0 |                      0.0 |     4.311434 |             0.0 |              0.0 |          4.311434 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |  0.511091 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.041956 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.041956 |      0.0 |   0.0 |   0.0 |  0.0 | -1.530864 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | 0.0 | nan | 0.0 |              6.816215 |                      0.0 |                  6.816215 |                   6.816215 |                   2.45312 |                     0.0 |                      0.0 |     4.363095 |             0.0 |              0.0 |          4.363095 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |  0.093098 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.800053 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.800053 |      0.0 |   0.0 |   0.0 |  0.0 | -1.706955 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | 0.0 | nan | 0.0 |              7.056886 |                      0.0 |                  7.056886 |                   7.056886 |                  2.688326 |                     0.0 |                      0.0 |      4.36856 |             0.0 |              0.0 |           4.36856 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |  0.065871 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -1.82186 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -1.82186 |      0.0 |   0.0 |   0.0 |  0.0 | -1.755989 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | 0.0 | nan | 0.0 |              6.921138 |                      0.0 |                  6.921138 |                   6.921138 |                   2.76572 |                     0.0 |                      0.0 |     4.155419 |             0.0 |              0.0 |          4.155419 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |  0.086838 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.891469 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.891469 |      0.0 |   0.0 |   0.0 |  0.0 | -1.804631 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | 0.0 | nan | 0.0 |               6.98872 |                      0.0 |                   6.98872 |                    6.98872 |                  2.799361 |                     0.0 |                      0.0 |     4.189359 |             0.0 |              0.0 |          4.189359 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |  0.206065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.106575 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.106575 |      0.0 |   0.0 |   0.0 |  0.0 | -1.900509 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | 0.0 | nan | 0.0 |              7.021165 |                      0.0 |                  7.021165 |                   7.021165 |                  2.986484 |                     0.0 |                      0.0 |     4.034682 |             0.0 |              0.0 |          4.034682 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |  0.082443 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.942929 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.942929 |      0.0 |   0.0 |   0.0 |  0.0 | -1.954486 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | 0.0 | nan | 0.0 |               6.78319 |                      0.0 |                   6.78319 |                    6.78319 |                  2.708378 |                     0.0 |                      0.0 |     4.074812 |             0.0 |              0.0 |          4.074812 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | -0.105944 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.808559 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.808559 |      0.0 |   0.0 |   0.0 |  0.0 | -1.914503 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | 0.0 | nan | 0.0 |              5.908778 |                      0.0 |                  5.908778 |                   5.908778 |                  2.239353 |                     0.0 |                      0.0 |     3.669425 |             0.0 |              0.0 |          3.669425 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | -0.157388 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672736 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672736 |      0.0 |   0.0 |   0.0 |  0.0 | -1.830125 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | 0.0 | nan | 0.0 |              5.607956 |                      0.0 |                  5.607956 |                   5.607956 |                  2.021579 |                     0.0 |                      0.0 |     3.586377 |             0.0 |              0.0 |          3.586377 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 |  0.463086 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.008458 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.008458 |      0.0 |   0.0 |   0.0 |  0.0 | -1.066456 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | 0.0 | nan | 0.0 |              5.874454 |                      0.0 |                  5.874454 |                   5.874454 |                  2.487815 |                     0.0 |                      0.0 |      3.38664 |             0.0 |              0.0 |           3.38664 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 |  0.595002 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.283901 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.283901 |      0.0 |   0.0 |   0.0 |  0.0 | -0.688899 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | 0.0 | nan | 0.0 |              6.104899 |                      0.0 |                  6.104899 |                   6.104899 |                  2.520953 |                     0.0 |                      0.0 |     3.583945 |             0.0 |              0.0 |          3.583945 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 |  0.485942 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.510743 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.510743 |      0.0 |   0.0 |   0.0 |  0.0 | -1.024801 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | 0.0 | nan | 0.0 |              6.035457 |                      0.0 |                  6.035457 |                   6.035457 |                  2.702703 |                     0.0 |                      0.0 |     3.332754 |             0.0 |              0.0 |          3.332754 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 |  0.321551 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.594572 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.594572 |      0.0 |   0.0 |   0.0 |  0.0 |  -1.27302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | 0.0 | nan | 0.0 |              5.176574 |                      0.0 |                  5.176574 |                   5.176574 |                  2.307856 |                     0.0 |                      0.0 |     2.868718 |             0.0 |              0.0 |          2.868718 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 |  0.132224 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.541567 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.541567 |      0.0 |   0.0 |   0.0 |  0.0 | -1.409342 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | 0.0 | nan | 0.0 |              6.059368 |                      0.0 |                  6.059368 |                   6.059368 |                  2.688476 |                     0.0 |                      0.0 |     3.370892 |             0.0 |              0.0 |          3.370892 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |  0.365956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.939383 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.939383 |      0.0 |   0.0 |   0.0 |  0.0 | -1.573427 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | 0.0 | nan | 0.0 |              6.379464 |                      0.0 |                  6.379464 |                   6.379464 |                  3.308913 |                     0.0 |                      0.0 |     3.070551 |             0.0 |              0.0 |          3.070551 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |  0.111688 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826811 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826811 |      0.0 |   0.0 |   0.0 |  0.0 | -1.715123 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | 0.0 | nan | 0.0 |              2.347976 |                      0.0 |                  2.347976 |                   2.347976 |                   0.60663 |                     0.0 |                      0.0 |     1.741346 |             0.0 |              0.0 |          1.741346 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 |  -0.21427 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.073812 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.073812 |      0.0 |   0.0 |   0.0 |  0.0 | -1.020458 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | 0.0 | nan | 0.0 |              4.456814 |                      0.0 |                  4.456814 |                   4.456814 |                  1.641158 |                     0.0 |                      0.0 |     2.815656 |             0.0 |              0.0 |          2.815656 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  0.557521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.80221 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.80221 |      0.0 |   0.0 |   0.0 |  0.0 | -0.526689 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | 0.0 | nan | 0.0 |              5.612971 |                      0.0 |                  5.612971 |                   5.612971 |                  1.868809 |                     0.0 |                      0.0 |     3.744162 |             0.0 |              0.0 |          3.744162 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 |  0.793059 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672232 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672232 |      0.0 |   0.0 |   0.0 |  0.0 | -0.879173 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | 0.0 | nan | 0.0 |              4.326484 |                      0.0 |                  4.326484 |                   4.326484 |                  1.401603 |                     0.0 |                      0.0 |     2.924881 |             0.0 |              0.0 |          2.924881 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 |  0.226008 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.280102 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.280102 |      0.0 |   0.0 |   0.0 |  0.0 | -1.148095 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | 0.0 | nan | 0.0 |              3.783947 |                      0.0 |                  3.783947 |                   3.783947 |                  1.261689 |                     0.0 |                      0.0 |     2.522258 |             0.0 |              0.0 |          2.522258 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 |  1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.792904 |   1.3836 | 0.016014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.792904 |      0.0 |   0.0 |   0.0 |  0.0 |    1.3836 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016014 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | 0.0 | nan | 0.0 |              4.290898 |                      0.0 |                  4.290898 |                   4.290898 |                  1.498964 |                     0.0 |                      0.0 |     2.791934 |             0.0 |              0.0 |          2.791934 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |   1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |   5.01244 |  3.86595 | 0.044745 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   5.01244 |      0.0 |   0.0 |   0.0 |  0.0 |   3.86595 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044745 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | 0.0 | nan | 0.0 |              2.273012 |                      0.0 |                  2.273012 |                   2.273012 |                  0.589077 |                     0.0 |                      0.0 |     1.683934 |             0.0 |              0.0 |          1.683934 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 |  0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 13.305141 | 7.027354 | 0.081335 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |      12.0 | 1.305141 |   0.0 |   0.0 |  0.0 |  6.199583 |  0.82777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081335 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | 0.0 | nan | 0.0 |               2.18653 |                      0.0 |                   2.18653 |                    2.18653 |                  0.708412 |                     0.0 |                      0.0 |     1.478119 |             0.0 |              0.0 |          1.478119 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 |  0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.040224 |  8.32407 | 0.096343 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.040224 |      0.0 |   0.0 |   0.0 |  0.0 |  7.885403 | 0.438667 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096343 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | 0.0 | nan | 0.0 |              3.688655 |                      0.0 |                  3.688655 |                   3.688655 |                  1.881434 |                     0.0 |                      0.0 |     1.807221 |             0.0 |              0.0 |          1.807221 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |   0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.90065 | 5.691378 | 0.065872 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.90065 |      0.0 |   0.0 |   0.0 |  0.0 |  5.655812 | 0.035566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065872 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | 0.0 | nan | 0.0 |              3.284636 |                      0.0 |                  3.284636 |                   3.284636 |                  1.511299 |                     0.0 |                      0.0 |     1.773338 |             0.0 |              0.0 |          1.773338 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 |  0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -0.912078 | 2.597892 | 0.030068 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -0.912078 |      0.0 |   0.0 |   0.0 |  0.0 |  2.595008 | 0.002884 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030068 |

.. _lland_v4_vers_daily:

sealed surfaces
---------------

.. integration-test::

    >>> lnk(VERS)
    >>> surfaceresistance.vers_aug = 500
    >>> cropheight.vers_aug = 5.0
    >>> albedo.vers_aug = 0.10
    >>> lai.vers_aug = 10.0
    >>> test("lland_v4_vers_daily",
    ...      axis1=(inputs.nied, fluxes.qah, states.bowa))
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |   evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |    qdb | qib1 | qib2 | qbb | qkap |   qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn | bowa |  qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.793956 |                      0.0 |                 14.793956 |                  14.793956 |                  1.877708 |                     0.0 |                      0.0 |    12.916248 |             0.0 |              0.0 |         12.916248 |             16.038785 |                   nan |                    500.0 |                   500.0 | 10.502 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.049585 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 10.502 | 10.502 |  0.0 |  0.0 | 0.0 |  0.0 | 10.502 | 2.703704 | 0.031293 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   32.4284 |  0.0 | 10.502 |   0.0 |   0.0 |   0.0 |  0.0 | 2.703704 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031293 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.926902 |                      0.0 |                  9.926902 |                   9.926902 |                   0.96939 |                     0.0 |                      0.0 |     8.957512 |             0.0 |              0.0 |          8.957512 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.020895 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 3.637039 | 0.042095 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.282605 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 3.637039 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042095 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.783845 |                      0.0 |                 12.783845 |                  12.783845 |                  1.386217 |                     0.0 |                      0.0 |    11.397628 |             0.0 |              0.0 |         11.397628 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.335547 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.940764 | 0.022463 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.946184 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.940764 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022463 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.919034 |                      0.0 |                 20.919034 |                  20.919034 |                  3.108707 |                     0.0 |                      0.0 |    17.810327 |             0.0 |              0.0 |         17.810327 |              13.74753 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 12.230738 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.035613 | 0.011986 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.164858 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.035613 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011986 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |             22.383112 |                      0.0 |                 22.383112 |                  22.383112 |                  3.414868 |                     0.0 |                      0.0 |    18.968244 |             0.0 |              0.0 |         18.968244 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 14.271962 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.552615 | 0.006396 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.824543 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.552615 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006396 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             21.358671 |                      0.0 |                 21.358671 |                  21.358671 |                  2.845741 |                     0.0 |                      0.0 |     18.51293 |             0.0 |              0.0 |          18.51293 |             10.998024 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 14.791696 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.294881 | 0.003413 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342053 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.294881 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003413 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             22.087564 |                      0.0 |                 22.087564 |                  22.087564 |                  3.051087 |                     0.0 |                      0.0 |    19.036477 |             0.0 |              0.0 |         19.036477 |             10.129759 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 18.203529 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.157352 | 0.001821 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.159866 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157352 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001821 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             20.448645 |                      0.0 |                 20.448645 |                  20.448645 |                   2.45312 |                     0.0 |                      0.0 |    17.995525 |             0.0 |              0.0 |         17.995525 |             11.664571 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.456812 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.083965 | 0.000972 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.616338 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083965 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000972 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |             21.170659 |                      0.0 |                 21.170659 |                  21.170659 |                  2.688326 |                     0.0 |                      0.0 |    18.482332 |             0.0 |              0.0 |         18.482332 |             21.385046 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  11.84053 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.044804 | 0.000519 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.737786 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044804 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000519 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.763414 |                      0.0 |                 20.763414 |                  20.763414 |                   2.76572 |                     0.0 |                      0.0 |    17.997695 |             0.0 |              0.0 |         17.997695 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.388795 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.023908 | 0.000277 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.284252 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.023908 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000277 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |              20.96616 |                      0.0 |                  20.96616 |                   20.96616 |                  2.799361 |                     0.0 |                      0.0 |    18.166799 |             0.0 |              0.0 |         18.166799 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 18.009869 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.012758 | 0.000148 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.012758 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000148 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             21.063496 |                      0.0 |                 21.063496 |                  21.063496 |                  2.986484 |                     0.0 |                      0.0 |    18.077012 |             0.0 |              0.0 |         18.077012 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 16.851837 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.006808 | 0.000079 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.439999 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000079 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             20.349571 |                      0.0 |                 20.349571 |                  20.349571 |                  2.708378 |                     0.0 |                      0.0 |    17.641193 |             0.0 |              0.0 |         17.641193 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.954049 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.003633 | 0.000042 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.604721 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003633 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000042 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.726334 |                      0.0 |                 17.726334 |                  17.726334 |                  2.239353 |                     0.0 |                      0.0 |    15.486981 |             0.0 |              0.0 |         15.486981 |             14.256697 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.005903 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001938 | 0.000022 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.707203 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001938 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000022 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.823868 |                      0.0 |                 16.823868 |                  16.823868 |                  2.021579 |                     0.0 |                      0.0 |    14.802289 |             0.0 |              0.0 |         14.802289 |             12.029088 |                   nan |                    500.0 |                   500.0 |  0.538 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.728768 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.538 |  0.538 |  0.0 |  0.0 | 0.0 |  0.0 |  0.538 | 0.139541 | 0.001615 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.894036 |  0.0 |  0.538 |   0.0 |   0.0 |   0.0 |  0.0 | 0.139541 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001615 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.623363 |                      0.0 |                 17.623363 |                  17.623363 |                  2.487815 |                     0.0 |                      0.0 |    15.135549 |             0.0 |              0.0 |         15.135549 |             29.610064 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   7.56081 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.186871 | 0.002163 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.084102 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.186871 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002163 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             18.314696 |                      0.0 |                 18.314696 |                  18.314696 |                  2.520953 |                     0.0 |                      0.0 |    15.793743 |             0.0 |              0.0 |         15.793743 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  10.39177 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.099717 | 0.001154 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.451334 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.099717 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001154 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             18.106372 |                      0.0 |                 18.106372 |                  18.106372 |                  2.702703 |                     0.0 |                      0.0 |    15.403668 |             0.0 |              0.0 |         15.403668 |             19.246541 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 12.298568 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 |  0.05321 | 0.000616 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.767029 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05321 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000616 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.529721 |                      0.0 |                 15.529721 |                  15.529721 |                  2.307856 |                     0.0 |                      0.0 |    13.221865 |             0.0 |              0.0 |         13.221865 |             14.805032 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.759102 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.028393 | 0.000329 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.006003 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.028393 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000329 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             18.178103 |                      0.0 |                 18.178103 |                  18.178103 |                  2.688476 |                     0.0 |                      0.0 |    15.489627 |             0.0 |              0.0 |         15.489627 |             10.692523 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 19.233472 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.015151 | 0.000175 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.50595 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.015151 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000175 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             19.138391 |                      0.0 |                 19.138391 |                  19.138391 |                  3.308913 |                     0.0 |                      0.0 |    15.829478 |             0.0 |              0.0 |         15.829478 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  17.26069 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.008085 | 0.000094 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.193828 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.008085 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000094 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              7.043929 |                      0.0 |                  7.043929 |                   7.043929 |                   0.60663 |                     0.0 |                      0.0 |     6.437299 |             0.0 |              0.0 |          6.437299 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.830469 |  1.88 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.004314 |  0.00005 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.917693 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004314 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00005 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.370443 |                      0.0 |                 13.370443 |                  13.370443 |                  1.641158 |                     0.0 |                      0.0 |    11.729284 |             0.0 |              0.0 |         11.729284 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.371493 | 0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.002302 | 0.000027 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   33.2096 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000027 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.838913 |                      0.0 |                 16.838913 |                  16.838913 |                  1.868809 |                     0.0 |                      0.0 |    14.970104 |             0.0 |              0.0 |         14.970104 |             17.496856 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.504016 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001228 | 0.000014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001228 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000014 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |             12.979453 |                      0.0 |                 12.979453 |                  12.979453 |                  1.401603 |                     0.0 |                      0.0 |     11.57785 |             0.0 |              0.0 |          11.57785 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.461825 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.000655 | 0.000008 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000655 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000008 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.351841 |                      0.0 |                 11.351841 |                  11.351841 |                  1.261689 |                     0.0 |                      0.0 |    10.090152 |             0.0 |              0.0 |         10.090152 |             16.736123 |                   nan |                    500.0 |                   500.0 |   7.87 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.472563 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   7.87 |   7.87 |  0.0 |  0.0 | 0.0 |  0.0 |   7.87 | 2.026454 | 0.023454 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.63093 |  0.0 |   7.87 |   0.0 |   0.0 |   0.0 |  0.0 | 2.026454 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023454 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.872693 |                      0.0 |                 12.872693 |                  12.872693 |                  1.498964 |                     0.0 |                      0.0 |    11.373729 |             0.0 |              0.0 |         11.373729 |             18.330039 |                   nan |                    500.0 |                   500.0 |   4.11 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.960765 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   4.11 |   4.11 |  0.0 |  0.0 | 0.0 |  0.0 |   4.11 |  3.78382 | 0.043794 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.001403 |  0.0 |   4.11 |   0.0 |   0.0 |   0.0 |  0.0 |  3.78382 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043794 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |              6.819035 |                      0.0 |                  6.819035 |                   6.819035 |                  0.589077 |                     0.0 |                      0.0 |     6.229957 |             0.0 |              0.0 |          6.229957 |             14.805032 |                   nan |                    500.0 |                   500.0 |   12.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.075158 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   12.1 |   12.1 |  0.0 |  0.0 | 0.0 |  0.0 |   12.1 | 6.030624 | 0.069799 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.368818 |  0.0 |   12.0 |   0.1 |   0.0 |   0.0 |  0.0 | 5.967201 | 0.063424 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069799 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.559591 |                      0.0 |                  6.559591 |                   6.559591 |                  0.708412 |                     0.0 |                      0.0 |     5.851179 |             0.0 |              0.0 |          5.851179 |             14.805032 |                   nan |                    500.0 |                   500.0 |  6.648 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.534236 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  6.648 |  6.648 |  0.0 |  0.0 | 0.0 |  0.0 |  6.648 | 7.436588 | 0.086072 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   30.7039 |  0.0 |  6.648 |   0.0 |   0.0 |   0.0 |  0.0 | 7.402977 | 0.033611 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086072 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             11.065964 |                      0.0 |                 11.065964 |                  11.065964 |                  1.881434 |                     0.0 |                      0.0 |      9.18453 |             0.0 |              0.0 |           9.18453 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.752367 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 5.342084 |  0.06183 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.382521 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 5.339359 | 0.002725 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06183 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.853909 |                      0.0 |                  9.853909 |                   9.853909 |                  1.511299 |                     0.0 |                      0.0 |      8.34261 |             0.0 |              0.0 |           8.34261 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.711096 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 2.849362 | 0.032979 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.782565 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849141 | 0.000221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032979 |

.. _lland_v4_nadelw_summer_daily:

conifers (summer)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> surfaceresistance.nadelw = 56.0
    >>> cropheight.nadelw = 10.0
    >>> albedo.nadelw = 0.12
    >>> lai.nadelw = 11.0
    >>> conditions_nadelw_summer = test(
    ...     "lland_v4_nadelw_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah, states.bowa),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |     g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |             86.2 |       2.4 |               100.7 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |             1.579769 |                  1.579769 |      99.120231 |   1.205601 |         0.307429 | 4.507941 | 20.198726 |         0.971564 |                  36.46712 |                15.690785 |                     15.690785 |            16.437729 | 100.0 |       16.437729 |               16.437729 | -0.25 | 10.933333 |   -2.6784 |             14.465201 |                      nan |                  2.686395 |                  14.465201 |                  1.877708 |                     0.0 |                      0.0 |    12.587493 |             0.0 |              0.0 |         12.587493 |             39.166667 |                 100.0 |               211.364559 |              122.747678 | 10.302 |  0.0 |       49.0 |        0.42 | 4.32684 |     0.0 |      0.0 |             nan | 4.32684 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.686428 |   2.2 | 1.110404 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 10.302 | 0.657815 |  0.0 |  0.0 | 0.0 |      0.0 | 0.657815 | 0.169352 |  0.00196 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   32.4284 | 80.533782 | 0.657815 |   0.0 |   0.0 |   0.0 |       0.0 | 0.169352 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.00196 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |             85.8 |       1.6 |               101.3 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                 1.977914 |                      1.977914 |                         0.0 |                          0.0 |                      0.125103 |                           0.125103 |              1.69705 |                   1.69705 |       99.60295 |   1.207284 |         0.302714 | 4.538923 | 20.167744 |         0.971855 |                 36.236444 |                15.628821 |                     15.628821 |            11.029892 | 100.0 |       11.029892 |               11.029892 | -0.25 |    19.028 |  0.895795 |              9.706305 |                      nan |                  1.802599 |                   9.706305 |                   0.96939 |                     0.0 |                      0.0 |     8.736914 |             0.0 |              0.0 |          8.736914 |                 58.75 |                 100.0 |               181.457925 |               115.35055 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.312002 | 0.094 | 1.809101 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.227814 | 0.002637 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.282605 | 78.724681 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.227814 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.002637 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |             82.4 |       1.5 |               101.6 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 2.187158 |                      2.187158 |                         0.0 |                          0.0 |                      0.136594 |                           0.136594 |             1.802218 |                  1.802218 |      99.797782 |   1.203775 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                15.566134 |                     15.566134 |            14.204272 | 100.0 |       14.204272 |               14.204272 | -0.25 | 15.208683 | -1.913579 |             12.499759 |                      nan |                  2.321384 |                  12.499759 |                  1.386217 |                     0.0 |                      0.0 |    11.113542 |             0.0 |              0.0 |         11.113542 |             62.666667 |                 100.0 |               184.734569 |              115.873944 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   4.24805 |   0.0 | 2.495048 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.121564 | 0.001407 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.946184 | 76.229633 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.001407 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |             77.6 |       2.8 |               101.7 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 2.041337 |                      2.041337 |                         0.0 |                          0.0 |                      0.128603 |                           0.128603 |             1.584077 |                  1.584077 |     100.115923 |    1.21051 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                15.502754 |                     15.502754 |            23.243371 | 100.0 |       23.243371 |               23.243371 | -0.25 | 20.753946 |  1.531325 |             20.454167 |                      nan |                  3.798631 |                  20.454167 |                  3.108707 |                     0.0 |                      0.0 |     17.34546 |             0.0 |              0.0 |          17.34546 |             33.571429 |                 100.0 |               209.622626 |              121.261768 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.655528 | 0.094 |  3.24309 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.064868 | 0.000751 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.164858 | 72.986543 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000751 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |             74.9 |       3.0 |               101.8 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.133112 |                      2.133112 |                         0.0 |                          0.0 |                      0.133641 |                           0.133641 |             1.597701 |                  1.597701 |     100.202299 |   1.208738 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                15.438712 |                     15.438712 |            24.870125 | 100.0 |       24.870125 |               24.870125 | -0.25 | 14.816194 | -1.909685 |              21.88571 |                      nan |                  4.064489 |                   21.88571 |                  3.414868 |                     0.0 |                      0.0 |    18.470842 |             0.0 |              0.0 |         18.470842 |             31.333333 |                 100.0 |               223.524712 |              123.686516 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.731502 |   0.0 | 3.618105 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.034614 | 0.000401 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.824543 | 69.368437 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000401 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |             77.9 |       3.5 |               101.8 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           1.242466 |          3.5 |                 2.327535 |                      2.327535 |                         0.0 |                          0.0 |                      0.144219 |                           0.144219 |              1.81315 |                   1.81315 |       99.98685 |   1.201995 |         0.282971 | 4.666316 |  20.04035 |         0.973104 |                 35.273753 |                15.374034 |                     15.374034 |            23.731856 | 100.0 |       23.731856 |               23.731856 | -0.25 | 20.348477 |  0.232491 |             20.884034 |                      nan |                  3.878463 |                  20.884034 |                  2.845741 |                     0.0 |                      0.0 |    18.038293 |             0.0 |              0.0 |         18.038293 |             26.857143 |                 100.0 |               217.806222 |              122.172235 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.967691 |   0.0 | 3.538589 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.018471 | 0.000214 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342053 | 65.829848 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.018471 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000214 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |             73.5 |       3.8 |               101.9 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             1.808493 |                  1.808493 |     100.091507 |   1.199521 |         0.277822 | 4.698958 | 20.007708 |         0.973436 |                  35.02347 |                 15.30875 |                      15.30875 |            24.541738 | 100.0 |       24.541738 |               24.541738 | -0.25 | 18.740175 | -1.067813 |             21.596729 |                      nan |                  4.010821 |                  21.596729 |                  3.051087 |                     0.0 |                      0.0 |    18.545642 |             0.0 |              0.0 |         18.545642 |             24.736842 |                 100.0 |                249.89699 |              127.503607 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.546527 |   0.0 | 3.943139 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.009856 | 0.000114 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.159866 | 61.886709 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.009856 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000114 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |             80.1 |       3.3 |               101.9 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           1.171468 |          3.3 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             1.983047 |                  1.983047 |      99.916953 |   1.198331 |         0.272591 | 4.731891 | 19.974776 |         0.973776 |                 34.769499 |                15.242885 |                     15.242885 |            22.720717 | 100.0 |       22.720717 |               22.720717 | -0.25 | 21.466219 |  0.293528 |             19.994231 |                      nan |                  3.713214 |                  19.994231 |                   2.45312 |                     0.0 |                      0.0 |    17.541111 |             0.0 |              0.0 |         17.541111 |             28.484848 |                 100.0 |               234.560208 |              124.470702 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.375029 |   0.0 | 3.447428 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.005259 | 0.000061 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.616338 | 58.439281 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.005259 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000061 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |             72.9 |       1.8 |               101.9 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 2.696965 |                      2.696965 |                         0.0 |                          0.0 |                      0.163997 |                           0.163997 |             1.966088 |                  1.966088 |      99.933912 |   1.192729 |         0.267279 |   4.7651 | 19.941567 |         0.974124 |                 34.511921 |                15.176467 |                     15.176467 |            23.522954 | 100.0 |       23.522954 |               23.522954 | -0.25 | 19.654459 | -1.371448 |               20.7002 |                      nan |                  3.844323 |                    20.7002 |                  2.688326 |                     0.0 |                      0.0 |    18.011873 |             0.0 |              0.0 |         18.011873 |             52.222222 |                 100.0 |                290.49321 |              132.398807 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.801371 |   0.0 | 4.238315 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.002806 | 0.000032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.737786 | 54.200966 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002806 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000032 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |             68.5 |       1.9 |               102.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.813915 |                      2.813915 |                         0.0 |                          0.0 |                      0.170178 |                           0.170178 |             1.927532 |                  1.927532 |     100.072468 |   1.191257 |         0.261887 | 4.798573 | 19.908094 |         0.974479 |                 34.250819 |                15.109521 |                     15.109521 |             23.07046 | 100.0 |        23.07046 |                23.07046 | -0.25 |  23.39262 |  0.203534 |             20.302005 |                      nan |                  3.770372 |                  20.302005 |                   2.76572 |                     0.0 |                      0.0 |    17.536286 |             0.0 |              0.0 |         17.536286 |             49.473684 |                 100.0 |               348.466656 |              138.318012 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.457407 |   0.0 | 4.453782 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.001498 | 0.000017 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.284252 | 49.747184 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001498 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000017 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |             68.6 |       3.0 |               102.1 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 2.830978 |                      2.830978 |                         0.0 |                          0.0 |                      0.171077 |                           0.171077 |             1.942051 |                  1.942051 |     100.157949 |   1.191966 |         0.256418 | 4.832298 | 19.874369 |         0.974842 |                 33.986274 |                15.042071 |                     15.042071 |            23.295734 | 100.0 |       23.295734 |               23.295734 | -0.25 | 21.880839 | -0.632013 |             20.500246 |                      nan |                  3.807188 |                  20.500246 |                  2.799361 |                     0.0 |                      0.0 |    17.700884 |             0.0 |              0.0 |         17.700884 |             31.333333 |                 100.0 |               368.725681 |              139.570111 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.478761 |   0.0 | 4.442334 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000799 | 0.000009 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.666265 |  45.30485 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000799 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000009 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |             66.0 |       2.5 |               102.2 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |             1.879766 |                  1.879766 |     100.320234 |   1.193016 |         0.250873 | 4.866263 | 19.840404 |         0.975213 |                  33.71837 |                14.974141 |                     14.974141 |            23.403884 | 100.0 |       23.403884 |               23.403884 | -0.25 | 23.154216 | -0.023734 |             20.595418 |                      nan |                  3.824863 |                  20.595418 |                  2.986484 |                     0.0 |                      0.0 |    17.608935 |             0.0 |              0.0 |         17.608935 |                  37.6 |                 100.0 |               419.711234 |              142.937942 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.985603 | 0.094 | 4.546047 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000426 | 0.000005 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.439999 | 40.758803 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000426 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000005 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |             69.0 |       1.6 |               101.9 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                  2.84813 |                       2.84813 |                         0.0 |                          0.0 |                       0.17198 |                            0.17198 |              1.96521 |                   1.96521 |       99.93479 |    1.18911 |         0.245254 | 4.900456 | 19.806211 |          0.97559 |                 33.447192 |                14.905755 |                     14.905755 |            22.610635 | 100.0 |       22.610635 |               22.610635 | -0.25 | 22.399998 | -0.414721 |             19.897359 |                      nan |                  3.695224 |                  19.897359 |                  2.708378 |                     0.0 |                      0.0 |    17.188981 |             0.0 |              0.0 |         17.188981 |                 58.75 |                 100.0 |               411.253273 |              141.649314 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.775599 |   0.0 | 4.347844 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000228 | 0.000003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.604721 | 36.410959 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000228 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |             75.6 |       2.7 |               101.5 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 2.600118 |                      2.600118 |                         0.0 |                          0.0 |                       0.15885 |                            0.15885 |             1.965689 |                  1.965689 |      99.534311 |   1.190432 |         0.239562 | 4.934866 |   19.7718 |         0.975975 |                 33.172825 |                14.836934 |                     14.836934 |            19.695927 | 100.0 |       19.695927 |               19.695927 | -0.25 | 22.949068 |  0.647517 |             17.332416 |                      nan |                  3.218877 |                  17.332416 |                  2.239353 |                     0.0 |                      0.0 |    15.093063 |             0.0 |              0.0 |         15.093063 |             34.814815 |                 100.0 |               357.548831 |              136.707247 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.857065 |   0.0 | 3.459696 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.000121 | 0.000001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.707203 | 32.951263 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000121 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000001 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |             81.4 |       3.2 |               101.6 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           1.135969 |          3.2 |                 2.460535 |                      2.460535 |                         0.0 |                          0.0 |                      0.151385 |                           0.151385 |             2.002875 |                  2.002875 |      99.597125 |   1.195095 |         0.233799 | 4.969483 | 19.737183 |         0.976367 |                 32.895357 |                  14.7677 |                       14.7677 |            18.693186 | 100.0 |       18.693186 |               18.693186 | -0.25 | 19.957345 | -0.436832 |             16.450004 |                      nan |                  3.055001 |                  16.450004 |                  2.021579 |                     0.0 |                      0.0 |    14.428425 |             0.0 |              0.0 |         14.428425 |                29.375 |                 100.0 |               333.908335 |              133.915514 |  0.338 |  0.0 |       49.0 |        0.42 | 0.14196 |     0.0 |      0.0 |             nan | 0.14196 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.162052 |   2.2 | 1.984501 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.338 | 0.008727 |  0.0 |  0.0 | 0.0 |      0.0 | 0.008727 | 0.002312 | 0.000027 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.894036 | 31.296034 | 0.008727 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002312 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000027 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |             79.0 |       1.3 |               101.9 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 2.270468 |                      2.270468 |                         0.0 |                          0.0 |                      0.141127 |                           0.141127 |             1.793669 |                  1.793669 |     100.106331 |   1.204916 |         0.227967 | 5.004296 | 19.702371 |         0.976766 |                 32.614875 |                14.698075 |                     14.698075 |            19.581515 | 100.0 |       19.581515 |               19.581515 | -0.25 |  20.58012 |  0.559934 |             17.231733 |                      nan |                  3.200179 |                  17.231733 |                  2.487815 |                     0.0 |                      0.0 |    14.743918 |             0.0 |              0.0 |         14.743918 |             72.307692 |                 100.0 |               354.695699 |              135.231698 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.274274 |   0.0 | 3.081998 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.003057 | 0.000035 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.084102 | 28.214037 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003057 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 | 0.000035 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |             75.9 |       1.9 |               101.9 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 2.490975 |                      2.490975 |                         0.0 |                          0.0 |                      0.153018 |                           0.153018 |              1.89065 |                   1.89065 |      100.00935 |   1.198338 |         0.222067 | 5.039294 | 19.667372 |         0.977172 |                 32.331467 |                14.628078 |                     14.628078 |            20.349663 | 100.0 |       20.349663 |               20.349663 | -0.25 |  17.88034 | -1.617232 |             17.907703 |                      nan |                  3.325716 |                  17.907703 |                  2.520953 |                     0.0 |                      0.0 |     15.38675 |             0.0 |              0.0 |          15.38675 |             49.473684 |                 100.0 |               400.426126 |               138.20342 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.725557 |   0.0 | 3.429544 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.166895 |      0.0 | 0.001332 | 0.000015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.451334 | 24.951387 |      0.0 |   0.0 |   0.0 |   0.0 | -0.166895 | 0.001631 |   0.0 |   0.0 |   0.0 | -0.000299 |   0.0 | 0.000015 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |             70.0 |       2.0 |               101.8 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 2.584275 |                      2.584275 |                         0.0 |                          0.0 |                      0.158006 |                           0.158006 |             1.808993 |                  1.808993 |      99.991007 |   1.195081 |         0.216102 | 5.074469 | 19.632198 |         0.977584 |                 32.045225 |                14.557728 |                     14.557728 |            20.118191 | 100.0 |       20.118191 |               20.118191 | -0.25 | 22.437779 |  0.434305 |             17.704008 |                      nan |                  3.287887 |                  17.704008 |                  2.702703 |                     0.0 |                      0.0 |    15.001304 |             0.0 |              0.0 |         15.001304 |                  47.0 |                 100.0 |               491.324867 |              142.949964 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   7.53971 |   0.0 | 3.733858 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.369623 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.767029 | 21.587152 |      0.0 |   0.0 |   0.0 |   0.0 | -0.369623 |  0.00087 |   0.0 |   0.0 |   0.0 | -0.001259 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |             70.3 |       2.6 |               101.9 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |             1.761941 |                  1.761941 |     100.138059 |   1.198506 |         0.210072 |  5.10981 | 19.596856 |         0.978003 |                 31.756238 |                14.487046 |                     14.487046 |            17.255246 | 100.0 |       17.255246 |               17.255246 | -0.25 | 20.156763 | -0.488974 |             15.184616 |                      nan |                      2.82 |                  15.184616 |                  2.307856 |                     0.0 |                      0.0 |     12.87676 |             0.0 |              0.0 |          12.87676 |             36.153846 |                 100.0 |               525.389413 |              143.816091 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.832014 |   0.0 |  3.37772 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.578662 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.006003 | 18.788095 |      0.0 |   0.0 |   0.0 |   0.0 | -0.578662 | 0.000464 |   0.0 |   0.0 |   0.0 | -0.002954 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |             69.0 |       3.6 |               101.9 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           1.277965 |          3.6 |                 2.713408 |                      2.713408 |                         0.0 |                          0.0 |                      0.164868 |                           0.164868 |             1.872252 |                  1.872252 |     100.027748 |   1.192743 |          0.20398 | 5.145309 | 19.561357 |         0.978429 |                   31.4646 |                14.416048 |                     14.416048 |            20.197892 | 100.0 |       20.197892 |               20.197892 | -0.25 | 20.953343 | -0.749947 |             17.774145 |                      nan |                  3.300913 |                  17.774145 |                  2.688476 |                     0.0 |                      0.0 |    15.085669 |             0.0 |              0.0 |         15.085669 |             26.111111 |                 100.0 |               629.540329 |              146.963401 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.542903 |   0.0 | 3.855581 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.752584 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.50595 | 15.685099 |      0.0 |   0.0 |   0.0 |   0.0 | -0.752584 | 0.000248 |   0.0 |   0.0 |   0.0 | -0.005328 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |             62.1 |       2.5 |               101.7 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                 2.729939 |                      2.729939 |                         0.0 |                          0.0 |                      0.165743 |                           0.165743 |             1.695292 |                  1.695292 |     100.004708 |   1.190771 |         0.197828 | 5.180957 | 19.525709 |          0.97886 |                 31.170402 |                14.344752 |                     14.344752 |            21.264879 | 100.0 |       21.264879 |               21.264879 | -0.25 | 22.619833 |  0.062121 |             18.713094 |                      nan |                  3.475289 |                  18.713094 |                  3.308913 |                     0.0 |                      0.0 |    15.404181 |             0.0 |              0.0 |         15.404181 |                  37.6 |                 100.0 |               884.502325 |              152.120949 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.796451 |   0.0 |  4.29127 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.945392 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.193828 | 12.339221 |      0.0 |   0.0 |   0.0 |   0.0 | -0.945392 | 0.000132 |   0.0 |   0.0 |   0.0 | -0.008351 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |             84.5 |       2.1 |               101.5 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.284619 |                      2.284619 |                         0.0 |                          0.0 |                      0.141895 |                           0.141895 |             1.930503 |                  1.930503 |      99.569497 |   1.199129 |         0.191617 | 5.216746 | 19.489921 |         0.979299 |                  30.87374 |                14.273175 |                     14.273175 |             7.826588 | 100.0 |        7.826588 |                7.826588 | -0.25 | 21.579428 |  1.026136 |              6.887397 |                      nan |                  1.279088 |                   6.887397 |                   0.60663 |                     0.0 |                      0.0 |     6.280768 |             0.0 |              0.0 |          6.280768 |             44.761905 |                 100.0 |               696.001395 |              147.247892 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.397063 |  1.88 | 0.698098 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.153291 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.917693 | 12.794413 |      0.0 |   0.0 |   0.0 |   0.0 | -1.153291 | 0.000071 |   0.0 |   0.0 |   0.0 | -0.012081 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |             83.6 |       1.5 |               101.5 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                  2.38585 |                       2.38585 |                         0.0 |                          0.0 |                      0.147368 |                           0.147368 |             1.994571 |                  1.994571 |      99.505429 |   1.195981 |          0.18535 | 5.252667 |    19.454 |         0.979743 |                 30.574709 |                14.201333 |                     14.201333 |            14.856047 | 100.0 |       14.856047 |               14.856047 | -0.25 | 17.325643 | -1.541907 |             13.073322 |                      nan |                  2.427903 |                  13.073322 |                  1.641158 |                     0.0 |                      0.0 |    11.432163 |             0.0 |              0.0 |         11.432163 |             62.666667 |                 100.0 |               667.674311 |              145.668066 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.351914 | 0.282 | 2.206487 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.125007 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   33.2096 | 11.712934 |      0.0 |   0.0 |   0.0 |   0.0 | -1.125007 | 0.000038 |   0.0 |   0.0 |   0.0 | -0.016118 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |             76.5 |       2.2 |               101.1 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 3.006626 |                      3.006626 |                         0.0 |                          0.0 |                      0.180287 |                           0.180287 |             2.300069 |                  2.300069 |      98.799931 |   1.174651 |         0.179027 | 5.288712 | 19.417955 |         0.980193 |                 30.273404 |                14.129243 |                     14.129243 |            18.709904 | 100.0 |       18.709904 |               18.709904 | -0.25 | 21.631999 | -1.279412 |             16.464715 |                      nan |                  3.057733 |                  16.464715 |                  1.868809 |                     0.0 |                      0.0 |    14.595906 |             0.0 |              0.0 |         14.595906 |             42.727273 |                 100.0 |               938.453811 |              150.290353 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.236094 |   0.0 | 3.494179 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.192206 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 34.239011 |  9.410961 |      0.0 |   0.0 |   0.0 |   0.0 | -1.192206 |  0.00002 |   0.0 |   0.0 |   0.0 | -0.020212 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |             76.6 |       1.9 |               100.8 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  2.93526 |                       2.93526 |                         0.0 |                          0.0 |                      0.176554 |                           0.176554 |             2.248409 |                  2.248409 |      98.551591 |   1.172942 |         0.172651 | 5.324874 | 19.381792 |         0.980649 |                 29.969924 |                14.056918 |                     14.056918 |            14.421614 | 100.0 |       14.421614 |               14.421614 | -0.25 | 25.063372 |  0.706772 |              12.69102 |                      nan |                  2.356904 |                   12.69102 |                  1.401603 |                     0.0 |                      0.0 |    11.289418 |             0.0 |              0.0 |         11.289418 |             49.473684 |                 100.0 |              1353.717528 |               153.70968 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.830325 | 0.094 | 2.901447 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.335241 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  33.28224 |  7.844755 |      0.0 |   0.0 |   0.0 |   0.0 | -1.335241 | 0.000011 |   0.0 |   0.0 |   0.0 | -0.024667 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |             84.2 |       2.3 |               100.6 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 2.506318 |                      2.506318 |                         0.0 |                          0.0 |                       0.15384 |                            0.15384 |              2.11032 |                   2.11032 |       98.48968 |   1.181557 |         0.166225 | 5.361147 |  19.34552 |         0.981111 |                 29.664367 |                13.984374 |                     13.984374 |            12.613157 | 100.0 |       12.613157 |               12.613157 | -0.25 | 21.874132 |   0.40131 |             11.099578 |                      nan |                   2.06135 |                  11.099578 |                  1.261689 |                     0.0 |                      0.0 |     9.837889 |             0.0 |              0.0 |          9.837889 |             40.869565 |                 100.0 |              1694.004243 |              154.756666 |   7.67 |  0.0 |       49.0 |        0.42 |  3.2214 |     0.0 |      0.0 |             nan |  3.2214 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.656548 |   2.2 | 1.080941 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   7.67 | 0.067757 |  0.0 |  0.0 | 0.0 | 1.432559 | 0.067757 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.63093 | 15.798616 | 0.067757 |   0.0 |   0.0 |   0.0 | -1.432559 |  0.01745 |   0.0 |   0.0 |   0.0 | -0.029537 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |             85.8 |       2.1 |               100.5 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.475714 |                      2.475714 |                         0.0 |                          0.0 |                        0.1522 |                             0.1522 |             2.124163 |                  2.124163 |      98.375837 |   1.181114 |         0.159748 | 5.397522 | 19.309145 |         0.981578 |                 29.356832 |                13.911624 |                     13.911624 |            14.302992 | 100.0 |       14.302992 |               14.302992 | -0.25 | 19.703098 | -0.620474 |             12.586633 |                      nan |                  2.337518 |                  12.586633 |                  1.498964 |                     0.0 |                      0.0 |    11.087669 |             0.0 |              0.0 |         11.087669 |             44.761905 |                 100.0 |               514.362016 |              137.508257 |   3.91 |  0.0 |       49.0 |        0.42 |  1.6422 |     0.0 |      0.0 |             nan |  1.6422 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.603863 |   2.2 | 1.172642 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   3.91 | 0.052955 |  0.0 |  0.0 | 0.0 | 0.938338 | 0.052955 | 0.003425 |  0.00004 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 33.001403 | 19.421358 | 0.052955 |   0.0 |   0.0 |   0.0 | -0.938338 | 0.037102 |   0.0 |   0.0 |   0.0 | -0.033677 |   0.0 |  0.00004 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |             86.5 |       2.6 |               100.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 2.371153 |                      2.371153 |                         0.0 |                          0.0 |                      0.146575 |                           0.146575 |             2.051047 |                  2.051047 |      97.948953 |   1.178323 |         0.153225 | 5.433993 | 19.272674 |         0.982051 |                 29.047419 |                13.838682 |                     13.838682 |             7.576705 | 100.0 |        7.576705 |                7.576705 | -0.25 | 20.938011 |  0.382585 |                6.6675 |                      nan |                   1.23825 |                     6.6675 |                  0.589077 |                     0.0 |                      0.0 |     6.078423 |             0.0 |              0.0 |          6.078423 |             36.153846 |                 100.0 |               423.391318 |               132.52466 |   11.9 |  0.0 |       49.0 |        0.42 |   4.998 |     0.0 |      0.0 |             nan |   4.998 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.480236 |   2.2 | 0.567552 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   11.9 |   0.2321 |  0.0 |  0.0 | 0.0 | 0.713236 |   0.2321 | 0.054101 | 0.000626 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.368818 | 31.234942 |   0.2321 |   0.0 |   0.0 |   0.0 | -0.713236 | 0.090616 |   0.0 |   0.0 |   0.0 | -0.036515 |   0.0 | 0.000626 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |             89.6 |       2.6 |               100.4 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                 1.832679 |                      1.832679 |                         0.0 |                          0.0 |                      0.117031 |                           0.117031 |              1.64208 |                   1.64208 |       98.75792 |   1.201704 |         0.146656 | 5.470553 | 19.236114 |         0.982529 |                  28.73623 |                13.765561 |                     13.765561 |             7.288435 | 100.0 |        7.288435 |                7.288435 | -0.25 | 18.829395 |  1.414918 |              6.413823 |                      nan |                  1.191138 |                   6.413823 |                  0.708412 |                     0.0 |                      0.0 |     5.705411 |             0.0 |              0.0 |          5.705411 |             36.153846 |                 100.0 |               347.072847 |              127.000616 |  6.448 |  0.0 |       49.0 |        0.42 | 2.70816 |     0.0 |      0.0 |             nan | 2.70816 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  2.685024 |   2.2 | 0.203792 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  6.448 | 0.173145 |  0.0 |  0.0 | 0.0 |      0.0 | 0.173145 | 0.103764 | 0.001201 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   30.7039 | 37.306005 | 0.173145 |   0.0 |   0.0 |   0.0 |       0.0 | 0.141425 |   0.0 |   0.0 |   0.0 | -0.037661 |   0.0 | 0.001201 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |             78.3 |       2.1 |               101.1 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 1.940693 |                      1.940693 |                         0.0 |                          0.0 |                      0.123042 |                           0.123042 |             1.519563 |                  1.519563 |      99.580437 |   1.206937 |         0.140044 | 5.507197 |  19.19947 |         0.983012 |                 28.423368 |                13.692273 |                     13.692273 |            12.295515 | 100.0 |       12.295515 |               12.295515 | -0.25 | 13.279667 |  -1.92862 |             10.820053 |                      nan |                  2.009438 |                  10.820053 |                  1.881434 |                     0.0 |                      0.0 |      8.93862 |             0.0 |              0.0 |           8.93862 |             44.761905 |                 100.0 |               346.875642 |              126.428035 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.533079 |   0.0 | 2.160946 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.074117 | 0.000858 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.382521 | 35.145059 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.111643 |   0.0 |   0.0 |   0.0 | -0.037526 |   0.0 | 0.000858 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |             78.5 |       2.1 |               101.4 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 2.093348 |                      2.093348 |                         0.0 |                          0.0 |                      0.131461 |                           0.131461 |             1.643278 |                  1.643278 |      99.756722 |   1.204994 |          0.13339 | 5.543918 | 19.162749 |           0.9835 |                 28.108935 |                13.618831 |                     13.618831 |            10.948788 | 100.0 |       10.948788 |               10.948788 | -0.25 | 18.875069 |  0.349956 |              9.634933 |                      nan |                  1.789345 |                   9.634933 |                  1.511299 |                     0.0 |                      0.0 |     8.123635 |             0.0 |              0.0 |          8.123635 |             44.761905 |                 100.0 |               345.936719 |              125.805103 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   4.43981 |   0.0 | 2.165797 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 0.022183 | 0.000257 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.782565 | 32.979262 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.059574 |   0.0 |   0.0 |   0.0 | -0.037392 |   0.0 | 0.000257 |

.. _lland_v4_acker_winter_daily:

acre (winter)
-------------

>>> pub.timegrids = "2010-12-01", "2011-01-01", "1d"
>>> lnk(ACKER)
>>> lai.acker_dec = 0.3
>>> cropheight.acker_dec = 0.05
>>> albedo.acker_dec = 0.1
>>> wg2z.dec = 0.7416
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.stinz, 0.0),
...     (states.sinz, 0.0),
...     (states.esnowinz, 0.0),
...     (states.asinz, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 0.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     0.0, 11.1, 0.4, 1.2, 6.4, 0.0, 0.0, 0.0, 0.7, 3.3, 9.4, 0.2, 2.9, 0.0,
...     0.0, 6.7, 0.0, 0.1, 0.0, 0.2, 0.2, 2.4, 7.0, 0.2, 0.0, 1.2, 0.9, 0.0,
...     0.0, 0.2, 0.9)
>>> inputs.teml.series = (
...     -5.8, -3.0, -5.6, -3.8, 0.2, -1.1, -2.9, -2.9, -2.1, -3.1, 5.1, 0.2,
...     -3.5, -3.9, -6.3, -3.5, -7.3, -5.8, -7.9, -5.7, -5.6, -4.8, -1.1,
...     -2.9, -9.1, -8.6, -5.3, -7.7, -9.2, -7.4, 1.7)
>>> inputs.sunshineduration.series = (
...     3.5, 0.0, 0.1, 3.7, 0.0, 0.1, 0.0, 0.0, 4.4, 0.1, 0.2, 0.7, 0.0, 1.2,
...     3.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.8, 0.0, 0.0, 0.0, 3.2, 4.3, 0.0, 3.4,
...     0.0, 0.0, 4.0)
>>> inputs.windspeed.series = (
...     6.9, 5.2, 2.0, 4.3, 4.0, 2.4, 1.9, 1.8, 2.7, 3.0, 7.4, 5.7, 2.2, 3.1,
...     2.8, 3.7, 0.9, 1.3, 1.9, 2.4, 2.3, 4.6, 8.6, 7.6, 3.8, 1.5, 1.3, 3.0,
...     2.7, 1.4, 4.2)
>>> inputs.relativehumidity.series = (
...     70.3, 85.0, 95.9, 90.6, 96.5, 99.0, 99.2, 99.1, 96.9, 95.0, 90.6,
...     78.7, 90.5, 88.1, 92.8, 98.1, 96.6, 97.7, 95.8, 97.5, 95.0, 93.8,
...     94.9, 85.4, 87.2, 92.7, 97.8, 95.4, 94.9, 96.0, 96.5)
>>> inputs.atmosphericpressure.series = (
...     101.5, 100.7, 100.5, 100.8, 99.6, 99.8, 100.4, 100.6, 100.9, 101.8,
...     100.4, 101.2, 102.0, 102.5, 102.3, 100.0, 99.1, 99.7, 99.7, 99.7,
...     100.4, 100.7, 100.4, 100.8, 101.8, 101.9, 101.7, 102.2, 102.4, 102.1,
...     101.6)
>>> inlet.sequences.sim.series = 0.0

.. integration-test::

    >>> conditions_acker_winter = test(
    ...     "lland_v4_acker_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |         tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo | evi |     evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr | wlatinz |  wlatsnow | wsensinz | wsenssnow | wsurfinz |     wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                         0.0 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |        0.0 |   3.00672 |              2.483872 |                      0.0 |                  2.483872 |                   2.483872 |                   4.01525 |                     0.0 |                      0.0 |    -1.531378 |             0.0 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.808179 | 0.0 | 0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                         0.0 |                     0.381881 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |    -7.5504 | -0.665713 |              0.106229 |                      0.0 |                  0.106229 |                   0.106229 |                  1.277835 |                     0.0 |                 2.406897 |    -1.171606 |             0.0 |        -2.300668 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 | 10.374 |  10.374 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.779906 | 0.0 |     0.0 | -0.168357 | -0.065045 |    -6.266232 |     0.875479 |       0.0 |       0.0 | 0.406773 |      0.0 |     0.0 | -0.449083 |      0.0 |  -2.44648 |      0.0 | -0.594896 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.542357 | 10.542357 | -0.135862 | 0.084415 | -0.857807 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                         0.0 |                     0.357079 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 |  -2.859357 |  2.216454 |              0.200535 |                      0.0 |                  0.200535 |                   0.200535 |                  1.329454 |                     0.0 |                 3.241865 |    -1.128919 |             0.0 |         -3.04133 |         -1.128919 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |  0.376 |   0.376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.039557 | 0.0 |     0.0 | -0.073821 | -0.004401 |    -7.134925 |      0.86088 |       0.0 |       0.0 | 0.487468 |      0.0 |     0.0 | -0.196913 |      0.0 | -0.605411 |      0.0 |  2.239006 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.992177 | 10.992177 | -0.162814 |   0.1418 | -2.332662 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                         0.0 |                     0.323611 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 |  -7.775538 |  0.321724 |              0.408995 |                      0.0 |                  0.408995 |                   0.408995 |                  3.727317 |                     0.0 |                 4.828741 |    -3.318322 |             0.0 |        -4.419745 |         -3.318322 |             49.536202 |                 100.0 |               112.102145 |               99.524551 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092108 | 0.0 |     0.0 | -0.402208 | -0.008959 |    -8.396148 |     0.851175 |       0.0 |       0.0 | 0.636779 |      0.0 |     0.0 |  -1.07287 |      0.0 |  -2.98424 |      0.0 |  0.362635 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.522386 | 12.522386 | -0.212684 | 0.183432 | -1.912786 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                         0.0 |                     0.491255 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 |  -6.375953 | -1.777027 |              0.126225 |                      0.0 |                  0.126225 |                   0.126225 |                  1.140917 |                     0.0 |                 1.372429 |    -1.014692 |             0.0 |        -1.246204 |         -1.014692 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |  6.016 |  2.7072 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.134777 | 0.0 |     0.0 | -0.434119 |  0.003902 |    -2.948046 |     0.846742 |       0.0 |       0.0 | 0.403221 | 0.403221 |     0.0 |  -1.15799 |      0.0 | -1.939347 |      0.0 | -1.851133 | 0.977326 |  0.28655 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 15.998885 | 18.972504 |       0.0 | 0.203476 |  0.605841 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                         0.0 |                     0.515157 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |        0.0 |  1.198162 |               0.26017 |                      0.0 |                   0.26017 |                    0.26017 |                  1.217715 |                     0.0 |                 2.425437 |    -0.957545 |             0.0 |        -2.165267 |         -0.957545 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000586 | 0.0 |     0.0 | -0.123565 |       0.0 |    -2.311269 |     0.813448 |       0.0 |       0.0 | 0.317541 | 0.317541 |     0.0 | -0.329604 |      0.0 | -0.531443 |      0.0 |  1.304221 | 0.994413 | 0.296657 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.422692 | 19.096069 |       0.0 | 0.378868 |  0.149279 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                         0.0 |                     0.462365 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |        0.0 |  1.946783 |              0.171372 |                      0.0 |                  0.171372 |                   0.171372 |                  1.193664 |                     0.0 |                 2.797224 |    -1.022292 |             0.0 |        -2.625852 |         -1.022292 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.015911 | 0.0 |     0.0 |  -0.06757 |       0.0 |    -3.755368 |     0.787669 |       0.0 |       0.0 | 0.511768 | 0.511768 |     0.0 | -0.180239 |      0.0 | -0.327899 |      0.0 |  2.117714 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.994381 | 19.163639 |       0.0 |  0.55426 | -1.055904 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                         0.0 |                     0.422362 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 |   -3.51968 |  0.741623 |              0.186256 |                      0.0 |                  0.186256 |                   0.186256 |                  1.194208 |                     0.0 |                 2.345793 |    -1.007952 |             0.0 |        -2.159537 |         -1.007952 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.012641 | 0.0 |     0.0 |  -0.16266 |       0.0 |    -4.950281 |     0.767118 |       0.0 |       0.0 | 0.661033 | 0.661033 |     0.0 | -0.433888 |      0.0 | -0.763241 |      0.0 |  0.962408 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.805273 | 19.326299 |       0.0 | 0.729652 | -1.055927 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                         0.0 |                     0.389718 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 |  -3.519758 |  1.286849 |              0.707155 |                      0.0 |                  0.707155 |                   0.707155 |                  4.005891 |                     0.0 |                 5.077714 |    -3.298735 |             0.0 |        -4.370559 |         -3.298735 |             78.890988 |                 100.0 |               112.102145 |               99.513532 |  0.658 |   0.658 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.235007 | 0.0 |     0.0 | -0.365518 | -0.002888 |    -6.002106 |     0.754993 |       0.0 |       0.0 | 0.807885 | 0.807885 |     0.0 |    -0.975 |      0.0 | -1.841764 |      0.0 |  1.553795 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.623632 | 20.349817 |       0.0 |  0.85302 | -1.601177 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                         0.0 |                     0.402698 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 |  -5.337255 |  0.123184 |               0.31535 |                      0.0 |                   0.31535 |                    0.31535 |                  1.286731 |                     0.0 |                 2.472681 |    -0.971381 |             0.0 |        -2.157331 |         -0.971381 |             71.001889 |                 100.0 |               112.102145 |               99.511814 |  3.102 |   3.102 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.137189 | 0.0 |     0.0 | -0.195734 | -0.020098 |    -5.574878 |       0.7659 |       0.0 |       0.0 | 0.843525 | 0.726185 |     0.0 | -0.522111 |      0.0 | -1.250397 |      0.0 |  0.384823 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.647551 | 23.647551 | -0.039192 | 0.741291 |  -0.98276 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                         0.0 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 |  -3.275868 |  -1.69821 |               0.34539 |                      0.0 |                   0.34539 |                    0.34539 |                   1.16859 |                     0.0 |                -0.065197 |      -0.8232 |             0.0 |         0.410586 |           -0.8232 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |  8.836 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.033005 | 0.0 |     0.0 | -1.215856 |  0.188672 |          0.0 |     0.748447 | 21.462327 | 21.462327 |      0.0 |      0.0 |     0.0 | -3.243236 |      0.0 | -5.063324 |      0.0 | -8.717146 |  0.94547 | 0.268174 | 30.457573 | 9.778899 |  0.0 |  0.0 | 0.0 |  0.0 | 9.778899 | 2.517544 | 0.029138 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.267017 |  3.241834 |       0.0 | 0.927231 |  1.457049 | 92.124934 | 9.778899 |   0.0 |   0.0 |   0.0 |  0.0 | 2.517544 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029138 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                         0.0 |                     0.517645 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |        0.0 |  1.164579 |              0.407516 |                      0.0 |                  0.407516 |                   0.407516 |                  1.733999 |                     0.0 |                 3.137362 |    -1.326483 |             0.0 |        -2.729846 |         -1.326483 |             37.369415 |                 100.0 |                83.817217 |               98.586605 |  0.188 |  0.0846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.429349 | 0.0 |     0.0 |  0.158718 |  0.000066 |    -2.246487 |     0.735371 |       0.0 |       0.0 | 0.061533 | 0.061533 |     0.0 |  0.423374 |      0.0 | -1.968023 |      0.0 |  1.185197 |   0.9613 | 0.277229 |  0.067022 | 0.022382 |  0.0 |  0.0 | 0.0 |  0.0 | 0.022382 | 3.392378 | 0.039264 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.299254 |  3.204093 |       0.0 | 1.094821 |  1.034071 | 92.169574 | 0.022382 |   0.0 |   0.0 |   0.0 |  0.0 | 3.392378 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039264 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                         0.0 |                     0.432007 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |        0.0 |  2.412276 |              0.193285 |                      0.0 |                  0.193285 |                   0.193285 |                  1.254261 |                     0.0 |                 3.092635 |    -1.060976 |             0.0 |        -2.899351 |         -1.060976 |             96.820757 |                 100.0 |                83.758368 |               98.584459 |  2.726 |   2.726 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.193889 | 0.0 |     0.0 |  0.014973 | -0.019941 |    -4.653311 |     0.749803 |       0.0 |       0.0 | 0.199105 | 0.199105 |     0.0 |  0.039939 |      0.0 | -0.480453 |      0.0 |  2.458837 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 1.814887 | 0.021006 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  5.211168 |   5.91512 |       0.0 | 0.911369 | -0.636606 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.814887 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021006 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                         0.0 |                     0.383823 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |  -2.122018 |    2.1142 |              0.448703 |                      0.0 |                  0.448703 |                   0.448703 |                  2.107297 |                     0.0 |                 4.010007 |    -1.658594 |             0.0 |        -3.561304 |         -1.658594 |             68.711505 |                 100.0 |                83.758368 |               98.584802 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.281648 | 0.0 |     0.0 | -0.065079 |       0.0 |    -6.200337 |     0.735952 |       0.0 |       0.0 | 0.256899 | 0.256899 |     0.0 | -0.173596 |      0.0 | -1.187703 |      0.0 |  2.200005 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.968444 | 0.011209 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  5.528228 |    5.9802 |       0.0 | 1.086761 | -2.009206 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.968444 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011209 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                         0.0 |                     0.282608 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 |  -6.697353 |  1.768945 |              0.639262 |                      0.0 |                  0.639262 |                   0.639262 |                  3.397964 |                     0.0 |                 4.986924 |    -2.758702 |             0.0 |        -4.347662 |         -2.758702 |             76.073452 |                 100.0 |                83.758368 |               98.585088 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.028526 | 0.0 |     0.0 | -0.225845 |       0.0 |   -10.109669 |     0.724248 |       0.0 |       0.0 | 0.406999 | 0.406999 |     0.0 | -0.602429 |      0.0 |  -1.84035 |      0.0 |  1.904882 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.516772 | 0.005981 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.159374 |  6.206044 |       0.0 | 1.262153 | -3.036551 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.516772 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.005981 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                         0.0 |                     0.357858 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -10.121835 | -1.562975 |              0.163977 |                      0.0 |                  0.163977 |                   0.163977 |                  1.211777 |                     0.0 |                 1.915561 |      -1.0478 |             0.0 |        -1.751584 |           -1.0478 |             57.569099 |                 100.0 |                83.758368 |               98.585317 |  6.298 |   6.298 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035293 | 0.0 |     0.0 |  -0.40193 |  -0.04607 |    -7.106837 |     0.785252 |       0.0 |       0.0 | 0.558149 | 0.046671 |     0.0 | -1.072127 |      0.0 |  -2.10208 |      0.0 | -1.422623 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.275755 | 0.003192 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.905974 | 12.905974 | -0.170834 | 0.573017 | -0.731975 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.275755 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003192 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                         0.0 |                     0.326115 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 |  -2.439918 |  3.036773 |              0.169175 |                      0.0 |                  0.169175 |                   0.169175 |                  1.287913 |                     0.0 |                 3.600107 |    -1.118738 |             0.0 |        -3.430932 |         -1.118738 |            236.672963 |                 100.0 |                83.758368 |               98.585489 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001256 | 0.0 |     0.0 | -0.025948 |       0.0 |    -8.297891 |     0.777934 |       0.0 |       0.0 | 0.670129 |      0.0 |     0.0 | -0.069214 |      0.0 | -0.271956 |      0.0 |  3.089762 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.147146 | 0.001703 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.931922 | 12.931922 | -0.223823 | 0.632681 | -3.027148 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.147146 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001703 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                         0.0 |                     0.283712 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -10.090494 | -0.015431 |              0.172921 |                      0.0 |                  0.172921 |                   0.172921 |                  1.257034 |                     0.0 |                 2.130829 |    -1.084113 |             0.0 |        -1.957908 |         -1.084113 |            163.850513 |                 100.0 |                83.758368 |               98.585603 |  0.094 |   0.094 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002818 | 0.0 |     0.0 | -0.215326 | -0.001139 |   -10.060727 |     0.772676 |       0.0 |       0.0 | 0.820045 |      0.0 |     0.0 | -0.574371 |      0.0 | -1.350035 |      0.0 |  0.033501 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.078519 | 0.000909 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.241248 | 13.241248 | -0.273895 | 0.678422 | -2.270117 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.078519 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000909 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                         0.0 |                     0.272279 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 |  -7.567057 |  1.559857 |              0.176614 |                      0.0 |                  0.176614 |                   0.176614 |                  1.301361 |                     0.0 |                 3.118579 |    -1.124747 |             0.0 |        -2.941965 |         -1.124747 |            112.108246 |                 100.0 |                83.758368 |               98.585661 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.032531 | 0.0 |     0.0 | -0.126518 |       0.0 |    -10.57604 |      0.76766 |       0.0 |       0.0 | 0.876298 |      0.0 |     0.0 | -0.337479 |      0.0 |  -1.02584 |      0.0 |  1.578645 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.041898 | 0.000485 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.367765 | 13.367765 | -0.292684 | 0.724518 | -3.088374 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041898 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000485 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                         0.0 |                     0.293041 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -10.294579 | -0.332219 |              0.314734 |                      0.0 |                  0.314734 |                   0.314734 |                  1.393316 |                     0.0 |                 2.435899 |    -1.078582 |             0.0 |        -2.121165 |         -1.078582 |             88.752361 |                 100.0 |                83.758368 |               98.585661 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020944 | 0.0 |     0.0 |  -0.27746 |  -0.00224 |    -9.653725 |     0.764448 |       0.0 |       0.0 | 0.818877 |      0.0 |     0.0 |  -0.74011 |      0.0 | -1.734692 |      0.0 | -0.353637 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.022357 | 0.000259 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.833225 | 13.833225 | -0.273505 | 0.755386 | -2.014555 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.022357 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000259 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                         0.0 |                     0.312843 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 |  -6.715183 |  1.094569 |              0.369069 |                      0.0 |                  0.369069 |                   0.369069 |                  1.820398 |                     0.0 |                 3.345756 |     -1.45133 |             0.0 |        -2.976687 |          -1.45133 |             92.611159 |                 100.0 |                83.758368 |               98.585603 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034932 | 0.0 |     0.0 |  -0.19466 |   -0.0022 |     -8.82662 |     0.761089 |       0.0 |       0.0 | 0.774426 |      0.0 |     0.0 | -0.519245 |      0.0 |  -1.37992 |      0.0 |  1.077522 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.01193 | 0.000138 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 14.215885 | 14.215885 | -0.258658 | 0.788873 | -2.367524 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.01193 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000138 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                         0.0 |                     0.344223 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 |  -7.891745 | -0.147882 |              0.171942 |                      0.0 |                  0.171942 |                   0.171942 |                  1.259735 |                     0.0 |                 2.598961 |    -1.087793 |             0.0 |        -2.427019 |         -1.087793 |              46.30558 |                 100.0 |                83.758368 |               98.585489 |  2.256 |   2.256 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.238205 | 0.0 |     0.0 | -0.254479 | -0.022632 |    -7.606479 |     0.774424 |       0.0 |       0.0 | 0.784019 |      0.0 |     0.0 | -0.678809 |      0.0 |  -1.91552 |      0.0 |  -0.16731 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.006366 | 0.000074 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.726363 | 16.726363 | -0.261862 | 0.662934 | -1.478042 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006366 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000074 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                         0.0 |                     0.487656 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 |  -4.926805 | -0.974852 |              0.143863 |                      0.0 |                  0.143863 |                   0.143863 |                  1.179808 |                     0.0 |                 2.170447 |    -1.035945 |             0.0 |        -2.026585 |         -1.035945 |             24.768101 |                 100.0 |                83.758368 |               98.585317 |   6.58 |  5.0995 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461634 | 0.0 |     0.0 | -0.349728 | -0.018542 |    -3.046303 |     0.811736 |       0.0 |       0.0 | 0.472584 | 0.472584 |     0.0 | -0.932883 |      0.0 | -2.191115 |      0.0 | -1.097414 | 0.991077 |  0.29467 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.003397 | 0.000039 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.633051 | 23.656092 |       0.0 | 0.389279 |  0.238411 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003397 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000039 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                         0.0 |                     0.451022 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |        0.0 |  2.117452 |              0.162963 |                      0.0 |                  0.162963 |                   0.162963 |                  1.273508 |                     0.0 |                 3.158372 |    -1.110545 |             0.0 |        -2.995409 |         -1.110545 |             28.027061 |                 100.0 |                83.758368 |               98.585088 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.132438 | 0.0 |     0.0 |   0.20077 | -0.001139 |    -4.084591 |     0.787447 |       0.0 |       0.0 | 0.635671 | 0.635671 |     0.0 |  0.535543 |      0.0 | -1.202325 |      0.0 |  2.328626 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.001813 | 0.000021 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.259214 | 23.643322 |       0.0 | 0.555966 | -1.137441 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001813 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000021 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                         0.0 |                     0.260819 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 |  -3.791471 |  3.795094 |              0.558983 |                      0.0 |                  0.558983 |                   0.558983 |                  3.730222 |                     0.0 |                  6.17159 |    -3.171238 |             0.0 |        -5.612606 |         -3.171238 |             56.054123 |                 100.0 |                83.758368 |               98.584802 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146109 | 0.0 |     0.0 | -0.024146 |       0.0 |   -11.112254 |     0.766938 |       0.0 |       0.0 | 1.670829 | 0.384108 |     0.0 | -0.064407 |      0.0 | -1.195048 |      0.0 |  4.353151 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000967 | 0.000011 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.667468 | 23.667468 | -0.429765 | 0.731358 | -4.190935 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000967 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000011 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                         0.0 |                     0.181906 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -13.969784 |  0.796549 |               0.64132 |                      0.0 |                   0.64132 |                    0.64132 |                  4.423241 |                     0.0 |                 4.808909 |    -3.781921 |             0.0 |        -4.167588 |         -3.781921 |            142.003778 |                 100.0 |                83.758368 |               98.584459 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  -0.16868 | 0.0 |     0.0 | -0.253474 | -0.020275 |   -15.506336 |       0.7704 |       0.0 |       0.0 | 2.405921 |      0.0 |     0.0 | -0.676129 |      0.0 | -2.341372 |      0.0 |  1.150087 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000516 | 0.000006 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 25.048942 | 25.048942 | -0.803578 | 0.699031 | -4.245884 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000516 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000006 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                         0.0 |                     0.253736 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -14.152947 | -1.399016 |              0.176319 |                      0.0 |                  0.176319 |                   0.176319 |                  1.247575 |                     0.0 |                  1.36221 |    -1.071255 |             0.0 |         -1.18589 |         -1.071255 |            163.850513 |                 100.0 |                83.758368 |               98.584061 |  0.846 |   0.846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002438 | 0.0 |     0.0 | -0.310389 | -0.009371 |   -11.454227 |     0.773367 |       0.0 |       0.0 | 1.856011 |      0.0 |     0.0 | -0.827947 |      0.0 | -1.950001 |      0.0 | -1.592057 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000275 | 0.000003 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.205331 | 26.205331 | -0.619908 | 0.672265 | -2.105268 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000275 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000003 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                         0.0 |                     0.255303 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 |  -7.017558 |  2.260351 |              0.579004 |                      0.0 |                  0.579004 |                   0.579004 |                    3.6896 |                     0.0 |                 5.329433 |    -3.110597 |             0.0 |         -4.75043 |         -3.110597 |             71.001889 |                 100.0 |                83.758368 |               98.583607 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.122157 | 0.0 |     0.0 | -0.235684 |       0.0 |   -11.377804 |     0.768775 |       0.0 |       0.0 | 1.865726 |      0.0 |     0.0 | -0.628675 |      0.0 | -1.858158 |      0.0 |  2.263596 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000147 | 0.000002 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.441015 | 26.441015 | -0.623153 |  0.71405 | -3.624019 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000147 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000002 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                         0.0 |                     0.225291 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -12.080062 |  0.434923 |              0.185809 |                      0.0 |                  0.185809 |                   0.185809 |                  1.325064 |                     0.0 |                 2.990028 |    -1.139254 |             0.0 |        -2.804219 |         -1.139254 |             78.890988 |                 100.0 |                83.758368 |               98.583098 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.072905 | 0.0 |     0.0 | -0.196129 |       0.0 |   -12.919034 |     0.764389 |       0.0 |       0.0 | 2.137509 |      0.0 |     0.0 | -0.523164 |      0.0 | -1.755355 |      0.0 |  0.525699 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000078 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.637144 | 26.637144 | -0.713928 | 0.755957 | -3.317342 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000078 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                         0.0 |                     0.247548 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -11.057806 |  0.364008 |               0.18917 |                      0.0 |                   0.18917 |                    0.18917 |                  1.292505 |                     0.0 |                 2.434873 |    -1.103335 |             0.0 |        -2.245703 |         -1.103335 |            152.146905 |                 100.0 |                83.758368 |               98.582534 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018935 | 0.0 |     0.0 | -0.190974 | -0.002908 |   -11.759981 |     0.762015 |       0.0 |       0.0 | 1.974006 |      0.0 |     0.0 | -0.509415 |      0.0 | -1.429798 |      0.0 |   0.30649 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000042 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 27.016118 | 27.016118 | -0.659318 | 0.779513 |  -2.93975 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000042 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                         0.0 |                     0.446985 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 |  -9.799165 | -2.900775 |              0.669518 |                      0.0 |                  0.669518 |                   0.669518 |                  3.475519 |                     0.0 |                 3.601606 |    -2.806001 |             0.0 |        -2.932088 |         -2.806001 |             50.715635 |                 100.0 |                83.758368 |               98.581916 |  0.846 | 0.06345 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.128252 | 0.0 |     0.0 | -0.924858 |  0.005795 |    -4.203534 |     0.758312 |       0.0 |       0.0 | 0.753524 | 0.753524 |     0.0 | -2.467012 |      0.0 | -3.767696 |      0.0 | -3.302621 | 0.973704 |  0.28443 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000022 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 28.756986 | 28.786976 |       0.0 | 0.817531 |  0.702625 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000022 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

.. _lland_v4_nadelw_winter_daily:

conifers (winter)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> conditions_nadelw_winter = test(
    ...     "lland_v4_nadelw_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |  nbes |    sbes | snowintmax | snowintrate |  nbesinz |  sbesinz |  wniedinz | actualalbedoinz |  wadainz | schmpotinz |  schminz | gefrpotinz |  gefrinz |   evsinz |      evpo | evi |       evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr |  wlatinz |  wlatsnow |  wsensinz | wsenssnow |  wsurfinz |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa | inzp |    stinz |     sinz |  esnowinz |    asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |           2.449434 |          6.9 |                 0.395811 |                      0.395811 |                         0.0 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.428675 |                      nan |                   0.45104 |                   2.428675 |                   4.01525 |                     0.0 |                      0.0 |    -1.586575 |             0.0 |              0.0 |         -1.586575 |             13.623188 |                 100.0 |             27827.355778 |              101.116829 |   0.0 |     0.0 |       24.5 |        0.42 |      0.0 |      0.0 |       0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  4.142976 | 0.0 |  0.674992 |       0.0 |       0.0 |          nan |         0.12 |       0.0 |       0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |            1.84595 |          5.2 |                 0.489349 |                      0.489349 |                    0.489349 |                     0.401405 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -1.002332 |              0.110651 |                 0.086501 |                  0.020549 |                   0.110651 |                  1.277835 |                1.549088 |                 0.318115 |    -1.167185 |       -1.462587 |        -0.297566 |         -1.167185 |             18.076923 |                 100.0 |              2039.690308 |               99.509579 | 8.234 |   8.234 |       24.5 |        0.42 |  3.45828 |  3.45828 | -0.021683 |        0.875479 |      0.0 |        0.0 |      0.0 |   0.165137 |      0.0 | 0.207278 |  1.795102 | 0.0 |  0.390029 | -0.041065 | -0.033983 |    -5.616888 |     0.870297 |       0.0 |       0.0 | 0.312445 | 0.312445 | 0.552905 | -0.109538 |  -1.98202 | -1.119987 |  0.033472 | -0.931959 |      1.0 |      0.3 |  1.40472 | 0.479795 |  0.0 |  0.0 | 0.0 |  0.0 | 0.479795 | 0.123521 |  0.00143 |  2.2 | 3.251002 | 3.251002 | -0.055156 | 0.084415 |  5.118761 |  6.870344 |       0.0 | 0.104114 | -0.521188 | 71.859905 | 0.479795 |   0.0 |   0.0 |   0.0 |  0.0 | 0.123521 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00143 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 0.401924 |                      0.401924 |                    0.401924 |                     0.419541 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -1.737292 |  1.711227 |              0.238392 |                 0.161562 |                  0.044273 |                   0.238392 |                  1.329454 |                1.536251 |                  1.53282 |    -1.091062 |       -1.374689 |        -1.488547 |         -1.091062 |                  47.0 |                 100.0 |             27877.434071 |              100.748969 | 0.376 |   0.376 |       24.5 |    0.429753 | 0.161587 | 0.161587 | -0.001891 |        0.862355 |      0.0 |        0.0 |      0.0 |   0.232496 |      0.0 |  0.02946 |  0.133789 | 0.0 |   0.05307 |  0.060955 | -0.004401 |    -5.038271 |     0.834617 |       0.0 |       0.0 | 0.288967 | 0.288967 | 0.078583 |  0.162595 | -1.432665 |  0.152199 |  0.020607 |  1.803342 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.166162 | 0.001923 |  2.2 | 3.383129 | 3.383129 | -0.077654 | 0.135729 |  5.574848 |  7.185389 |       0.0 | 0.261933 | -1.490815 | 71.806835 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166162 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001923 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           1.526459 |          4.3 |                 0.460812 |                      0.460812 |                    0.460812 |                      0.40538 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 | -4.969383 |  0.268903 |              0.515262 |                 0.328588 |                  0.095691 |                   0.515262 |                  3.727317 |                 3.01792 |                 1.204876 |    -3.212055 |       -2.689332 |        -1.109184 |         -3.212055 |             21.860465 |                 100.0 |              3326.630538 |               99.760835 | 1.128 |   1.128 |       24.5 |    0.430149 | 0.485209 | 0.485209 | -0.003854 |        0.853164 |      0.0 |        0.0 |      0.0 |   0.279301 |      0.0 | 0.109697 |   0.59622 | 0.0 |  0.147151 | -0.030683 | -0.008959 |      -5.4881 |     0.812507 |       0.0 |       0.0 | 0.357702 | 0.357702 |  0.29261 | -0.081845 | -2.970163 | -0.647923 |  0.011779 |  0.379417 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.088666 | 0.001026 |  2.2 | 3.758641 | 3.758641 | -0.093287 | 0.174653 |  6.599609 |  8.344072 |       0.0 | 0.384574 | -1.018118 | 71.659685 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088666 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001026 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           1.419962 |          4.0 |                 0.619748 |                      0.619748 |                    0.619748 |                      0.53632 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 | -3.393727 | -0.842315 |              0.163632 |                   0.1048 |                  0.030389 |                   0.163632 |                  1.140917 |                1.101652 |                 0.373462 |    -0.977285 |       -0.996852 |        -0.343073 |         -0.977285 |                  23.5 |                 100.0 |               760.725643 |               96.608883 | 6.016 |  2.7072 |       49.0 |    0.431276 | 2.594556 |  1.16755 |  0.001683 |        0.843734 |      0.0 |        0.0 |      0.0 |   0.146071 | 0.146071 | 0.052825 |  0.347144 | 0.0 |  0.098873 | -0.150348 |  0.001624 |     -1.76889 |     0.801324 |       0.0 |       0.0 | 0.167697 | 0.167697 | 0.140908 | -0.401045 | -1.180575 | -0.726709 | -0.042816 | -0.784681 | 0.978825 | 0.287429 | 2.720931 | 0.898881 |  0.0 |  0.0 | 0.0 |  0.0 | 0.898881 | 0.278727 | 0.003226 |  2.2 | 5.030087 | 6.300372 |       0.0 | 0.217465 |  8.414261 | 11.789488 |       0.0 | 0.456151 |  0.565797 | 73.382862 | 0.898881 |   0.0 |   0.0 |   0.0 |  0.0 | 0.278727 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003226 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.563475 |                      0.563475 |                    0.563475 |                     0.547577 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 |       0.0 |  0.771412 |              0.309511 |                 0.214458 |                  0.057481 |                   0.309511 |                  1.217715 |                1.267082 |                 1.041344 |    -0.908204 |       -1.052624 |        -0.983863 |         -0.908204 |             39.166667 |                 100.0 |               984.035419 |               97.341917 |   0.0 |     0.0 |     47.775 |    0.438901 |      0.0 |      0.0 |       0.0 |        0.811154 |      0.0 |        0.0 |      0.0 |   0.216736 | 0.216736 | 0.010803 |  0.039077 | 0.0 |  0.015065 | -0.019678 |       0.0 |    -1.488064 |     0.778069 |       0.0 |       0.0 | 0.141309 | 0.141309 | 0.028817 | -0.052491 | -1.009051 | -0.112763 |   0.07239 |  0.818609 | 0.979941 | 0.288085 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.336546 | 0.003895 |  2.2 | 5.237827 | 6.289569 |       0.0 | 0.392857 |   8.56985 | 11.809167 |       0.0 | 0.631543 |  0.535985 | 73.367797 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.336546 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003895 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 0.493024 |                      0.493024 |                    0.493024 |                     0.503649 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  1.355371 |              0.194246 |                 0.140734 |                  0.036074 |                   0.194246 |                  1.193664 |                1.302701 |                 1.329569 |    -0.999418 |       -1.161967 |        -1.293495 |         -0.999418 |             49.473684 |                 100.0 |              1826.303334 |               98.592457 |   0.0 |     0.0 |     25.725 |    0.438869 |      0.0 |      0.0 |       0.0 |        0.785861 |      0.0 |        0.0 |      0.0 |   0.318812 | 0.318812 | 0.006923 |  0.003186 | 0.0 |   0.00136 |  0.025574 |       0.0 |    -2.614528 |     0.759328 |       0.0 |       0.0 | 0.246371 | 0.246371 | 0.018468 |  0.068217 | -1.073951 |  0.075947 |  0.106483 |  1.437659 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.179585 | 0.002079 |  2.2 | 5.550522 | 6.282645 |       0.0 | 0.568249 |  8.797129 | 11.783593 |       0.0 | 0.806935 | -0.077786 | 73.366437 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.179585 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002079 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 0.493024 |                      0.493024 |                    0.493024 |                     0.499981 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 | -0.259288 |  1.271724 |              0.204821 |                 0.152617 |                  0.038038 |                   0.204821 |                  1.194208 |                1.296371 |                 1.292208 |    -0.989386 |       -1.143754 |         -1.25417 |         -0.989386 |             52.222222 |                 100.0 |              1826.786758 |               98.473511 |   0.0 |     0.0 |     25.725 |    0.438848 |      0.0 |      0.0 |       0.0 |        0.765656 |      0.0 |        0.0 |      0.0 |    0.30591 |  0.30591 | 0.007645 |  0.007313 | 0.0 |  0.003217 |  0.019633 |       0.0 |     -2.71246 |     0.743904 |       0.0 |       0.0 |  0.25086 |  0.25086 | 0.020393 |  0.052369 | -1.061974 |  0.048973 |  0.102174 |  1.355512 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095828 | 0.001109 |  2.2 | 5.849306 |    6.275 |       0.0 | 0.743641 |  9.032914 |  11.76396 |       0.0 | 0.982327 | -0.607911 |  73.36322 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.095828 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001109 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.523311 |                      0.523311 |                    0.523311 |                     0.479099 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 | -2.026369 |  0.651172 |              0.762275 |                 0.584095 |                  0.141565 |                   0.762275 |                  4.005891 |                2.997012 |                 1.407262 |    -3.243616 |       -2.412918 |        -1.265697 |         -3.243616 |             34.814815 |                 100.0 |              1330.759988 |               97.768001 | 0.658 |   0.658 |     35.525 |    0.438825 | 0.288747 | 0.288747 | -0.001267 |        0.751475 |      0.0 |        0.0 |      0.0 |   0.429708 | 0.425694 | 0.032679 | -0.141032 | 0.0 | -0.049786 | -0.056382 | -0.002888 |    -3.282488 |     0.735895 |       0.0 |       0.0 | 0.317377 | 0.317377 | 0.087169 | -0.150395 | -2.357831 | -0.361013 |  0.142255 |  0.754288 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.051135 | 0.000592 |  2.2 | 6.531068 | 6.531068 |  -0.00134 | 0.892179 |   9.76366 | 12.478342 |       0.0 | 1.087538 | -0.517483 | 73.413006 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.051135 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000592 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.485698 |                      0.485698 |                    0.485698 |                     0.472835 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 | -1.724943 |  0.898292 |              0.335543 |                 0.274606 |                  0.062315 |                   0.335543 |                  1.286731 |                1.538807 |                 1.132833 |    -0.951188 |       -1.264201 |        -1.070518 |         -0.951188 |             31.333333 |                 100.0 |              2038.131987 |                98.43038 | 3.102 |   3.102 |       24.5 |    0.439593 | 1.363618 | 1.363618 | -0.008835 |        0.749654 |      0.0 |        0.0 |      0.0 |    0.35279 |      0.0 | 0.051278 |  0.335027 | 0.0 |  0.108496 |  0.024117 | -0.020098 |    -3.457759 |      0.75091 |       0.0 |       0.0 | 0.425639 | 0.425639 | 0.136782 |   0.06433 | -1.293327 | -0.114491 |  0.107657 |  1.020357 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.027286 | 0.000316 |  2.2 | 7.843408 | 7.843408 | -0.117832 | 0.913105 | 11.909218 | 15.556225 |       0.0 | 0.898615 | -0.674175 |  73.30451 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.027286 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000316 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           2.626929 |          7.4 |                 0.879017 |                      0.879017 |                    0.879017 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 | -2.247249 | -1.164974 |              0.361674 |                 0.290086 |                  0.067168 |                   0.361674 |                   1.16859 |               -0.053089 |                -1.000249 |    -0.806915 |        0.343174 |         1.067417 |         -0.806915 |             12.702703 |                 100.0 |                395.63381 |               92.725399 | 8.836 |     0.0 |       49.0 |     0.44353 | 3.919033 |      0.0 |  0.072015 |        0.740541 | 9.934016 |   6.740382 | 6.740382 |        0.0 |      0.0 | 0.292189 |  2.302592 | 0.0 |   0.47469 | -0.656354 |  0.031473 |          0.0 |     0.736587 | 13.227652 | 11.909218 |      0.0 |      0.0 | 0.779398 | -1.750792 | -2.733328 | -2.733328 | -2.297104 | -5.551537 | 0.953877 | 0.272965 | 25.04858 | 8.141336 |  0.0 |  0.0 | 0.0 |  0.0 | 8.141336 | 2.110519 | 0.024427 |  2.2 | 1.074292 | 1.536237 |       0.0 | 1.025195 |       0.0 |       0.0 |       0.0 | 1.078021 |  1.232399 | 89.737064 | 8.141336 |   0.0 |   0.0 |   0.0 |  0.0 | 2.110519 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024427 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |           2.023445 |          5.7 |                 0.619748 |                      0.619748 |                    0.619748 |                     0.556667 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |  0.655768 |              0.423006 |                 0.340599 |                  0.078558 |                   0.423006 |                  1.733999 |                1.682182 |                 0.848577 |    -1.310993 |       -1.341582 |        -0.770019 |         -1.310993 |             16.491228 |                 100.0 |               626.784063 |               95.143255 | 0.188 |  0.0846 |       49.0 |    0.424609 | 0.079826 | 0.035922 |  0.000028 |        0.728381 | 0.028458 |        0.0 |      0.0 |   0.065172 | 0.065172 | 0.394139 |  3.266178 | 0.0 |  0.727186 |   0.06961 |  0.000023 |    -1.264985 |     0.725312 |       0.0 |       0.0 | 0.000717 | 0.000717 | 1.051347 |  0.548948 | -2.371134 | -0.662937 |  0.021795 |  0.656031 | 0.950665 | 0.271129 |  0.11839 | 0.038672 |  0.0 |  0.0 | 0.0 |  0.0 | 0.038672 | 2.837223 | 0.032838 |  2.2 | 0.883584 | 1.193466 |       0.0 | 1.196942 |       0.0 |       0.0 |       0.0 | 1.244993 |  1.318231 | 89.089596 | 0.038672 |   0.0 |   0.0 |   0.0 |  0.0 | 2.837223 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032838 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 0.471336 |                      0.471336 |                    0.471336 |                     0.480482 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |  1.681712 |              0.199159 |                 0.169759 |                  0.036987 |                   0.199159 |                  1.254261 |                1.373602 |                 1.387288 |    -1.055102 |       -1.203843 |        -1.350301 |         -1.055102 |             42.727273 |                 100.0 |              2096.444323 |               98.227989 | 2.726 |   2.726 |       24.5 |     0.42358 |  1.15468 |  1.15468 | -0.008446 |        0.730139 |      0.0 |        0.0 |      0.0 |   0.140414 | 0.140414 | 0.082949 |  0.461865 | 0.0 |  0.180585 |  0.099892 | -0.012921 |    -3.244043 |     0.742199 |       0.0 |       0.0 | 0.059373 | 0.059373 | 0.221262 |  0.266457 | -1.386653 |  0.071863 |  0.038452 |  1.688621 | 0.985849 | 0.291569 | 0.479013 | 0.165355 |  0.0 |  0.0 | 0.0 |  0.0 | 0.165355 | 1.564624 | 0.018109 |  2.2 | 2.101716 | 2.265198 |       0.0 | 1.170338 |  1.558199 |  2.147095 |       0.0 | 1.003832 |  0.378119 | 89.222669 | 0.165355 |   0.0 |   0.0 |   0.0 |  0.0 | 1.564624 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018109 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |            1.10047 |          3.1 |                 0.457351 |                      0.457351 |                    0.457351 |                     0.464226 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |  1.919136 |              0.459587 |                 0.388432 |                  0.085352 |                   0.459587 |                  2.107297 |                2.124006 |                 1.610792 |     -1.64771 |       -1.735573 |         -1.52544 |          -1.64771 |             30.322581 |                 100.0 |              2867.176818 |               98.540199 |   0.0 |     0.0 |       24.5 |    0.426796 |      0.0 |      0.0 |       0.0 |        0.719287 |      0.0 |        0.0 |      0.0 |   0.152651 | 0.152651 | 0.116681 |  0.744745 | 0.0 |  0.233111 |   0.13142 |       0.0 |    -3.702038 |     0.729547 |       0.0 |       0.0 | 0.063425 | 0.063425 | 0.311242 |  0.350557 | -1.995829 |  0.064324 |  0.050986 |   1.94032 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.86945 | 0.010063 |  2.2 | 2.138244 | 2.148516 |       0.0 |  1.34573 |  1.522367 |  2.015675 |       0.0 | 1.179224 | -0.799417 | 88.989558 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.86945 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010063 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 0.380889 |                      0.380889 |                    0.380889 |                        0.383 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 | -2.664725 |   1.84732 |              0.651938 |                 0.547543 |                  0.121074 |                   0.651938 |                  3.397964 |                2.906053 |                 1.817561 |    -2.746025 |        -2.35851 |        -1.696487 |         -2.746025 |             33.571429 |                 100.0 |             21992.280362 |                99.38023 |   0.0 |     0.0 |       24.5 |    0.426446 |      0.0 |      0.0 |       0.0 |        0.709944 |      0.0 |        0.0 |      0.0 |   0.191547 | 0.010273 |  0.05613 |  0.181934 | 0.0 |  0.058593 |  0.060451 |       0.0 |    -6.228229 |      0.71878 |       0.0 |       0.0 | 0.097845 | 0.097845 | 0.149725 |   0.16125 | -2.444259 |  0.022264 |  0.063977 |  1.880001 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.463948 |  0.00537 |  2.2 | 2.092386 | 2.092386 | -0.060546 | 1.521122 |  1.571621 |  1.955224 |       0.0 | 1.354616 | -1.905138 | 88.930965 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.463948 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00537 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           1.313465 |          3.7 |                 0.471336 |                      0.471336 |                    0.471336 |                     0.410402 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 |  -6.35046 | -0.530555 |              0.167414 |                 0.167583 |                  0.031091 |                   0.167414 |                  1.211777 |                1.478916 |                 0.545843 |    -1.044364 |       -1.311334 |        -0.514752 |         -1.044364 |             25.405405 |                 100.0 |              2063.672198 |               98.054118 | 6.298 |   6.298 |       24.5 |    0.426277 | 2.684694 | 2.684694 | -0.019639 |        0.730476 |      0.0 |        0.0 |      0.0 |   0.220502 |      0.0 | 0.020939 |  0.122248 | 0.0 |  0.034207 | -0.121535 | -0.033779 |    -5.327013 |     0.780752 |       0.0 |       0.0 | 0.321711 | 0.321711 | 0.055854 |  -0.32419 | -1.373725 | -0.647445 | -0.006537 | -0.456882 |      1.0 |      0.3 | 0.838777 | 0.296086 |  0.0 |  0.0 | 0.0 |  0.0 | 0.296086 | 0.323794 | 0.003748 |  2.2 |  4.75614 |  4.75614 | -0.073648 | 1.165301 |  5.596902 |  7.535982 |       0.0 |  0.60918 | -0.632983 |  89.43945 | 0.296086 |   0.0 |   0.0 |   0.0 |  0.0 | 0.323794 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003748 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.319491 |          0.9 |                 0.352531 |                      0.352531 |                    0.352531 |                     0.386966 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -2.109943 |  2.065506 |              0.181685 |                 0.169432 |                  0.033742 |                   0.181685 |                  1.287913 |                1.486763 |                 1.781161 |    -1.106227 |       -1.317331 |        -1.747419 |         -1.106227 |            104.444444 |                 100.0 |             21842.761675 |               99.304205 |   0.0 |     0.0 |       24.5 |    0.434268 |      0.0 |      0.0 |       0.0 |        0.726874 |      0.0 |        0.0 |      0.0 |   0.396187 |      0.0 | 0.017159 |  0.025356 | 0.0 |   0.01472 |  0.066454 |       0.0 |     -6.09433 |     0.761513 |       0.0 |       0.0 | 0.361574 | 0.361574 |  0.04577 |  0.177262 | -1.304423 |   0.26159 |  0.058679 |  2.186272 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.234645 | 0.002716 |  2.2 | 4.738981 | 4.738981 | -0.132327 | 1.220283 |  5.905932 |  7.469528 |       0.0 | 0.784572 | -1.956889 |  89.42473 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.234645 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002716 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 0.395811 |                      0.395811 |                    0.395811 |                     0.359898 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -6.522963 |  0.264688 |              0.192732 |                 0.170406 |                  0.035793 |                   0.192732 |                  1.257034 |                1.397056 |                  0.83768 |    -1.064302 |        -1.22665 |        -0.801887 |         -1.064302 |             72.307692 |                 100.0 |             21847.650467 |               99.282763 | 0.094 |   0.094 |       24.5 |    0.434217 | 0.040816 | 0.040816 | -0.000495 |         0.72489 |      0.0 |        0.0 |      0.0 |   0.335942 |      0.0 | 0.014212 |   0.02794 | 0.0 |  0.013992 | -0.041852 | -0.001139 |     -7.03355 |     0.746632 |       0.0 |       0.0 | 0.403732 | 0.403732 | 0.037908 | -0.111639 | -1.285175 | -0.291853 | -0.020617 |  0.398395 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.125209 | 0.001449 |  2.2 | 4.765586 | 4.765586 | -0.112205 | 1.251767 |  6.398057 |  7.605381 |       0.0 | 0.948871 | -1.479977 | 89.410738 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.125209 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001449 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                  0.33643 |                       0.33643 |                     0.33643 |                     0.342668 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -4.933257 |  1.415912 |              0.202742 |                 0.171871 |                  0.037652 |                   0.202742 |                  1.301361 |                1.663046 |                 1.436791 |    -1.098618 |       -1.491175 |        -1.399139 |         -1.098618 |             49.473684 |                 100.0 |             21852.297704 |               99.272053 |   0.0 |     0.0 |       24.5 |    0.434297 |      0.0 |      0.0 |       0.0 |        0.722332 |      0.0 |        0.0 |      0.0 |   0.407467 |      0.0 | 0.024803 |  0.111648 | 0.0 |  0.044875 |  0.035754 |       0.0 |    -7.664568 |     0.733287 |       0.0 |       0.0 | 0.422854 | 0.422854 | 0.066161 |  0.095372 | -1.533447 |  0.062634 |   0.02389 |  1.557145 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.066813 | 0.000773 |  2.2 | 4.740783 | 4.740783 | -0.136094 | 1.293705 |  6.788844 |  7.569627 |       0.0 | 1.124263 | -2.154289 | 89.365863 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.066813 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000773 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 0.398857 |                      0.398857 |                    0.398857 |                     0.356052 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -7.180962 | -0.004626 |              0.369164 |                 0.303726 |                  0.068559 |                   0.369164 |                  1.393316 |                1.678188 |                 0.786605 |    -1.024152 |       -1.374462 |        -0.718046 |         -1.024152 |             39.166667 |                 100.0 |             21867.203851 |               99.272145 | 0.188 |   0.188 |       24.5 |    0.434222 | 0.081634 | 0.081634 | -0.000973 |        0.720843 |      0.0 |        0.0 |      0.0 |   0.323065 |      0.0 | 0.019118 |  0.091926 | 0.0 |  0.032936 | -0.062951 |  -0.00224 |     -7.17204 |     0.723712 |       0.0 |       0.0 | 0.386984 | 0.386984 | 0.050996 | -0.167918 |  -1.45462 | -0.427741 | -0.029163 |  0.122387 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.035652 | 0.000413 |  2.2 | 4.803299 | 4.803299 | -0.107904 | 1.318854 |  7.341288 |  7.820577 |       0.0 | 1.270895 | -1.408063 | 89.332927 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.035652 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000413 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 0.401924 |                      0.401924 |                    0.401924 |                     0.381677 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 | -4.693544 |  0.818871 |              0.439582 |                 0.353546 |                  0.081637 |                   0.439582 |                  1.820398 |                1.947522 |                 1.203744 |    -1.380816 |       -1.593975 |        -1.122107 |         -1.380816 |             40.869565 |                 100.0 |             21878.145617 |               99.282952 | 0.188 |   0.188 |       24.5 |     0.43441 | 0.081669 | 0.081669 | -0.000956 |         0.71894 |      0.0 |        0.0 |      0.0 |   0.351632 |      0.0 | 0.037879 |  0.167697 | 0.0 |  0.061761 | -0.000284 |   -0.0022 |    -6.273157 |     0.715443 |       0.0 |       0.0 | 0.336462 | 0.336462 | 0.101039 | -0.000758 | -1.686429 | -0.192301 |  0.008585 |  0.929049 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.019024 |  0.00022 |  2.2 | 4.847089 | 4.847089 | -0.117445 | 1.351802 |  7.784356 |  8.008861 |       0.0 | 1.414855 | -1.485335 | 89.271166 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.019024 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00022 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |           1.632956 |          4.6 |                 0.427219 |                      0.427219 |                    0.427219 |                     0.397947 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 | -4.951115 |  0.403685 |              0.208423 |                  0.17121 |                  0.038707 |                   0.208423 |                  1.259735 |                1.677522 |                 0.947266 |    -1.051312 |       -1.506312 |        -0.908559 |         -1.051312 |             20.434783 |                 100.0 |             15534.805818 |               99.249277 | 2.256 |   2.256 |       24.5 |    0.434541 | 0.980325 | 0.980325 | -0.009835 |        0.724156 |      0.0 |        0.0 |      0.0 |   0.331372 |      0.0 | 0.069651 |  0.581511 | 0.0 |  0.134803 | -0.007322 | -0.022632 |    -5.729829 |     0.726564 |       0.0 |       0.0 | 0.411379 | 0.411379 | 0.185792 | -0.019531 | -1.708705 | -0.370574 | -0.016601 |  0.518454 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.010152 | 0.000117 |  2.2 | 5.757763 | 5.757763 | -0.110678 | 1.263645 |  9.478166 | 10.272183 |       0.0 | 1.225144 |  -1.14742 | 89.136363 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.010152 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000117 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           3.052917 |          8.6 |                 0.563475 |                      0.563475 |                    0.563475 |                     0.506045 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -3.824733 | -0.660372 |              0.179764 |                 0.162269 |                  0.033385 |                   0.179764 |                  1.179808 |                1.494702 |                 0.631821 |    -1.000044 |       -1.332433 |        -0.598436 |         -1.000044 |             10.930233 |                 100.0 |               799.197627 |               95.891339 |  6.58 |  5.0995 |     47.775 |    0.437273 | 2.877258 | 2.229875 | -0.008108 |        0.739219 |      0.0 |        0.0 |      0.0 |   0.227608 | 0.227608 | 0.112787 |   1.08753 | 0.0 |  0.168717 | -0.112612 | -0.018542 |    -2.550866 |     0.764755 |       0.0 |       0.0 | 0.341128 | 0.341128 | 0.300853 | -0.300388 | -1.676051 | -0.863027 | -0.042765 | -0.564978 | 0.990473 | 0.294311 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.005417 | 0.000063 |  2.2 | 8.107942 | 8.522235 |       0.0 |  1.04255 |  12.77371 | 16.964796 |       0.0 |  0.75239 |  0.254552 | 88.967645 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.005417 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000063 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           2.697927 |          7.6 |                 0.493024 |                      0.493024 |                    0.493024 |                     0.486417 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 |       0.0 |  1.596813 |              0.191867 |                 0.170002 |                  0.035632 |                   0.191867 |                  1.273508 |                 1.73851 |                 1.239779 |     -1.08164 |       -1.568508 |        -1.204146 |          -1.08164 |             12.368421 |                 100.0 |              1521.393947 |               97.598044 | 0.188 |   0.188 |     25.725 |    0.445567 | 0.083767 | 0.083767 | -0.000508 |        0.727695 |      0.0 |        0.0 |      0.0 |   0.384253 | 0.384253 | 0.259203 |  2.589104 | 0.0 |  0.423025 |  0.235411 | -0.001139 |    -3.080272 |     0.749746 |       0.0 |       0.0 | 0.413281 | 0.413281 | 0.691411 |  0.627948 | -2.132086 | -0.098385 |  0.127833 |  1.733709 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.002891 | 0.000033 |  2.2 | 8.317664 | 8.346798 |       0.0 | 1.207507 | 13.108811 | 16.917384 |       0.0 |  0.91203 |  -0.60066 | 88.544621 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002891 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000033 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                  0.30617 |                       0.30617 |                     0.30617 |                     0.342808 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -2.002202 |  2.932659 |              0.633417 |                 0.552327 |                  0.117635 |                   0.633417 |                  3.730222 |                3.411688 |                 2.385189 |    -3.096805 |       -2.859361 |        -2.267554 |         -3.096805 |             24.736842 |                 100.0 |             22140.352821 |                99.43473 |   0.0 |     0.0 |       24.5 |     0.44504 |      0.0 |      0.0 |       0.0 |        0.717192 |      0.0 |        0.0 |      0.0 |    0.88702 | 0.029135 | 0.092902 |  0.606615 | 0.0 |  0.152005 |  0.179753 |       0.0 |    -7.659337 |     0.735904 |       0.0 |       0.0 | 0.993951 | 0.993951 | 0.247811 |  0.479482 | -2.810907 |  0.517603 |  0.296265 |  3.264638 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.001542 | 0.000018 |  2.2 | 8.253897 | 8.253897 | -0.286534 | 1.382899 | 13.952916 | 16.737631 |       0.0 | 1.087422 | -2.791719 | 88.392616 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001542 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000018 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 0.318472 |                      0.318472 |                    0.318472 |                     0.283049 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -9.305731 |  0.406622 |              0.743102 |                 0.638178 |                  0.138005 |                   0.743102 |                  4.423241 |                3.066536 |                 1.390615 |     -3.68014 |       -2.428359 |         -1.25261 |          -3.68014 |             62.666667 |                 100.0 |             22190.985309 |               99.499302 | 1.128 |   1.128 |       24.5 |    0.444762 | 0.501691 | 0.501691 | -0.009017 |        0.719417 |      0.0 |        0.0 |      0.0 |    1.02325 |      0.0 | 0.037798 | -0.088483 | 0.0 | -0.040207 | -0.019793 | -0.020275 |   -10.090109 |     0.733961 |       0.0 |       0.0 | 1.336187 | 1.336187 | 0.100824 | -0.052798 | -2.482969 | -0.367179 |  0.046214 |  0.832634 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000823 |  0.00001 |  2.2 |  8.71779 |  8.71779 | -0.341765 | 1.343461 | 15.933045 | 17.885425 |       0.0 |  1.11465 | -2.456741 | 88.432823 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000823 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00001 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                  0.41125 |                       0.41125 |                     0.41125 |                     0.353179 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -8.189136 |  -0.47318 |              0.210607 |                 0.176845 |                  0.039113 |                   0.210607 |                  1.247575 |                1.298494 |                 0.548706 |    -1.036967 |       -1.121649 |        -0.509593 |         -1.036967 |             72.307692 |                 100.0 |             22177.590146 |               99.574122 | 0.846 |   0.846 |       24.5 |    0.446153 | 0.377446 | 0.377446 | -0.004181 |         0.72085 |      0.0 |        0.0 |      0.0 |   0.614373 |      0.0 | 0.014124 |  0.030222 | 0.0 |  0.015188 | -0.076529 | -0.009371 |    -7.276367 |     0.729295 |       0.0 |       0.0 | 0.959301 | 0.959301 | 0.037675 | -0.204137 | -1.300069 | -0.467601 | -0.140746 | -0.162144 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000439 | 0.000005 |  2.2 | 9.081112 | 9.081112 | -0.205201 | 1.318735 | 17.431829 | 18.807953 |       0.0 | 1.183026 | -1.241961 | 88.417635 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000439 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000005 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 0.341723 |                      0.341723 |                    0.341723 |                       0.3491 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 | -4.139871 |  1.703424 |              0.704736 |                  0.57427 |                   0.13088 |                   0.704736 |                    3.6896 |                3.135462 |                 1.929791 |    -2.984864 |       -2.561192 |        -1.798911 |         -2.984864 |             31.333333 |                 100.0 |             22182.649921 |               99.659657 |   0.0 |     0.0 |       24.5 |    0.447243 |      0.0 |      0.0 |       0.0 |        0.718361 |      0.0 |        0.0 |      0.0 |   0.891611 |      0.0 | 0.033192 |  0.016723 | 0.0 |   0.00504 |  0.048768 |       0.0 |    -7.425797 |     0.718564 |       0.0 |       0.0 | 0.938098 | 0.938098 | 0.088537 |  0.130086 | -2.557131 |  0.087751 |  0.092598 |  2.016749 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000234 | 0.000003 |  2.2 |  9.04792 |  9.04792 | -0.297798 | 1.362006 | 18.322295 | 18.759185 |       0.0 | 1.358418 | -2.203785 | 88.412595 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000234 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000003 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 0.303761 |                      0.303761 |                    0.303761 |                     0.302164 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -7.345951 |  0.995721 |               0.22924 |                 0.181962 |                  0.042573 |                    0.22924 |                  1.325064 |                1.855986 |                 1.355808 |    -1.095824 |       -1.674024 |        -1.313235 |         -1.095824 |             34.814815 |                 100.0 |             22184.328927 |               99.755736 |   0.0 |     0.0 |       24.5 |    0.447144 |      0.0 |      0.0 |       0.0 |        0.716645 |      0.0 |        0.0 |      0.0 |   0.849376 |      0.0 | 0.031207 |  0.206774 | 0.0 |  0.065641 |  0.027988 |       0.0 |    -9.266708 |     0.709319 |       0.0 |       0.0 | 1.113192 | 0.436891 | 0.083242 |  0.074657 | -1.771373 | -0.020366 | -0.014106 |  1.367527 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000125 | 0.000001 |  2.2 | 9.016714 | 9.016714 | -0.283692 | 1.392793 | 18.731197 | 18.731197 | -0.225885 |  1.53381 | -2.457906 | 88.346954 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000125 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           0.496987 |          1.4 |                 0.349801 |                      0.349801 |                    0.349801 |                     0.315114 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 |  -8.19302 |   0.28083 |              0.232353 |                 0.184136 |                  0.043151 |                   0.232353 |                  1.292505 |                 1.53206 |                 0.852417 |    -1.060152 |       -1.347924 |        -0.809266 |         -1.060152 |             67.142857 |                 100.0 |             22206.199872 |               99.862463 | 0.188 |   0.188 |       24.5 |     0.44705 | 0.084045 | 0.084045 |   -0.0013 |        0.715513 |      0.0 |        0.0 |      0.0 |   0.760586 |      0.0 | 0.022296 |  0.078235 | 0.0 |   0.03716 | -0.032976 | -0.002908 |    -8.734745 |     0.707688 |       0.0 |       0.0 | 1.038685 | 0.084045 | 0.059473 | -0.087962 | -1.438352 | -0.322345 | -0.030956 |  0.398959 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000067 | 0.000001 |  2.2 | 9.078463 | 9.078463 | -0.254036 | 1.413544 | 18.952173 | 18.952173 |  -0.31885 | 1.567484 | -1.997136 | 88.309794 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000067 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |            1.49096 |          4.2 |                 0.690794 |                      0.690794 |                    0.690794 |                     0.515834 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 |  -6.65712 | -2.262048 |              0.815688 |                 0.645945 |                  0.151485 |                   0.815688 |                  3.475519 |                 2.76841 |                 0.193942 |    -2.659831 |       -2.122465 |        -0.042457 |         -2.659831 |             22.380952 |                 100.0 |                469.95008 |               94.018076 | 0.846 | 0.06345 |       49.0 |    0.447235 | 0.378361 | 0.028377 |  0.002592 |        0.713641 |      0.0 |        0.0 |      0.0 |   0.290571 | 0.290571 | 0.060446 |  0.116771 | 0.0 |  0.033693 | -0.376966 |  0.005795 |    -2.293601 |     0.705547 |       0.0 |       0.0 | 0.295823 | 0.295823 | 0.161237 | -1.005539 | -2.438096 | -1.513217 | -0.154393 | -2.476298 | 0.962331 | 0.277824 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000036 |      0.0 |  2.2 | 9.337344 | 9.396378 |       0.0 | 1.448677 | 19.650228 | 20.175139 |       0.0 | 1.613099 |  1.006512 | 88.276101 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000036 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

.. _lland_v4_acker_winter_ktschnee_daily:

acre (snow surface temperature)
-------------------------------

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> conditions_acker_winter_ktschnee = test(
    ...     "lland_v4_acker_winter_ktschnee_daily",
    ...     axis1=(inputs.nied, fluxes.wada),
    ...     axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |    sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |  evi |      evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr | wlatinz |  wlatsnow | wsensinz | wsenssnow | wsurfinz |     wsurf |      sff |      fvg |      wada |       qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |     qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |             70.3 |       6.9 |               101.5 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 0.395811 |                      0.395811 |                         0.0 |                          0.0 |                      0.030356 |                           0.030356 |             0.278255 |                  0.278255 |     101.221745 |   1.321192 |        -0.389384 | 8.532819 | 16.173848 |         1.028719 |                  5.761386 |                 7.641029 |                      7.641029 |             2.759858 | 100.0 |        2.759858 |                2.759858 | 0.7416 |       0.0 |   3.00672 |              2.483872 |                      0.0 |                  2.483872 |                   2.483872 |                   4.01525 |                     0.0 |                      0.0 |    -1.531378 |             0.0 |              0.0 |         -1.531378 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.808179 |  0.0 |  0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  -2.26512 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |             85.0 |       5.2 |               100.7 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 0.489349 |                      0.489349 |                         0.0 |                     0.387464 |                      0.036626 |                           0.036626 |             0.415947 |                  0.415947 |     100.284053 |   1.296511 |        -0.391536 | 8.548511 | 16.158156 |         1.028994 |                  5.687365 |                 7.609644 |                      7.609644 |             0.853105 | 100.0 |        0.853105 |                0.853105 | 0.7416 |   -7.5504 | -0.391921 |              0.106229 |                      0.0 |                  0.106229 |                   0.106229 |                  1.277835 |                     0.0 |                  2.47724 |    -1.171606 |             0.0 |        -2.371011 |         -1.171606 |             40.962628 |                 100.0 |               112.102145 |               99.530051 | 10.374 |  10.374 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.779906 |  0.0 |      0.0 | -0.140764 | -0.065045 |     -6.07759 |     0.875479 |       0.0 |       0.0 | 0.441058 |      0.0 |     0.0 |  -0.37548 |      0.0 | -2.305184 |      0.0 | -0.309653 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.514764 | 10.514764 | -0.147313 | 0.084415 | -1.131599 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |             95.9 |       2.0 |               100.5 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 0.401924 |                      0.401924 |                         0.0 |                     0.324336 |                      0.030771 |                           0.030771 |             0.385445 |                  0.385445 |     100.114555 |   1.306656 |        -0.393572 | 8.563421 | 16.143245 |          1.02926 |                  5.617593 |                 7.579824 |                      7.579824 |             1.441454 | 100.0 |        1.441454 |                1.441454 | 0.7416 | -3.771996 |  1.092655 |              0.199691 |                      0.0 |                  0.199691 |                   0.199691 |                  1.329454 |                     0.0 |                 2.790301 |    -1.129764 |             0.0 |         -2.59061 |         -1.129764 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |  0.376 |   0.376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.039468 |  0.0 |      0.0 | -0.159032 | -0.004401 |    -8.367644 |     0.861466 |       0.0 |       0.0 | 0.400697 |      0.0 |     0.0 | -0.424211 |      0.0 | -1.091625 |      0.0 |  1.074774 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 11.049796 | 11.049796 | -0.133833 | 0.139381 | -1.482654 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |             90.6 |       4.3 |               100.8 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 0.460812 |                      0.460812 |                         0.0 |                     0.331298 |                       0.03473 |                            0.03473 |             0.417496 |                  0.417496 |     100.382504 |   1.301648 |        -0.395492 | 8.577535 | 16.129131 |         1.029518 |                  5.552051 |                 7.551596 |                      7.551596 |             2.748162 | 100.0 |        2.748162 |                2.748162 | 0.7416 |  -4.94218 |   0.73294 |              0.411175 |                      0.0 |                  0.411175 |                   0.411175 |                  3.727317 |                     0.0 |                 4.937862 |    -3.316142 |             0.0 |        -4.526687 |         -3.316142 |             49.536202 |                 100.0 |               112.102145 |               99.524551 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092386 |  0.0 |      0.0 | -0.369279 | -0.008959 |    -8.096584 |     0.850382 |       0.0 |       0.0 | 0.484345 |      0.0 |     0.0 | -0.985032 |      0.0 | -2.789736 |      0.0 |   0.75192 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.547075 | 12.547075 | -0.161771 | 0.186969 | -1.473994 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |             96.5 |       4.0 |                99.6 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 0.619748 |                      0.619748 |                         0.0 |                     0.523186 |                      0.045128 |                           0.045128 |             0.598057 |                  0.598057 |      99.001943 |   1.266438 |        -0.397294 | 8.590839 | 16.115828 |         1.029767 |                  5.490726 |                 7.524989 |                      7.524989 |             0.823609 | 100.0 |        0.823609 |                0.823609 | 0.7416 | -4.913313 | -0.666543 |              0.128375 |                      0.0 |                  0.128375 |                   0.128375 |                  1.140917 |                     0.0 |                 1.700543 |    -1.012542 |             0.0 |        -1.572169 |         -1.012542 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |  6.016 |  2.7072 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135099 |  0.0 |      0.0 | -0.304329 |  0.003902 |     -2.10322 |     0.844132 |       0.0 |       0.0 |  0.49672 |  0.49672 |     0.0 | -0.811782 |      0.0 | -1.418894 |      0.0 | -0.658507 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.009222 | 18.867404 |       0.0 | 0.215598 | -0.065851 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |             99.0 |       2.4 |                99.8 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.563475 |                      0.563475 |                         0.0 |                     0.493464 |                      0.041489 |                           0.041489 |             0.557841 |                  0.557841 |      99.242159 |   1.275245 |        -0.398979 | 8.603319 | 16.103347 |         1.030007 |                  5.433602 |                 7.500028 |                      7.500028 |             1.394624 | 100.0 |        1.394624 |                1.394624 | 0.7416 | -0.219502 |  0.594934 |              0.262946 |                      0.0 |                  0.262946 |                   0.262946 |                  1.217715 |                     0.0 |                 2.201607 |    -0.954768 |             0.0 |        -1.938661 |         -0.954768 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000956 |  0.0 |      0.0 | -0.186363 |       0.0 |    -2.888059 |     0.811457 |       0.0 |       0.0 | 0.185939 | 0.185939 |     0.0 | -0.497115 |      0.0 | -0.784508 |      0.0 |  0.657038 | 0.996975 | 0.298188 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.355129 | 19.053767 |       0.0 |  0.39099 |  0.080815 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |             99.2 |       1.9 |               100.4 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 0.493024 |                      0.493024 |                         0.0 |                     0.429691 |                      0.036869 |                           0.036869 |             0.489079 |                  0.489079 |      99.910921 |   1.291808 |        -0.400545 | 8.614964 | 16.091703 |         1.030238 |                  5.380663 |                 7.476739 |                      7.476739 |             0.807099 | 100.0 |        0.807099 |                0.807099 | 0.7416 |       0.0 |  1.062327 |              0.172638 |                      0.0 |                  0.172638 |                   0.172638 |                  1.193664 |                     0.0 |                 2.428099 |    -1.021026 |             0.0 |        -2.255461 |         -1.021026 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.015761 |  0.0 |      0.0 | -0.150213 |       0.0 |    -4.724078 |       0.7861 |       0.0 |       0.0 | 0.279046 | 0.279046 |     0.0 | -0.400685 |      0.0 | -0.699247 |      0.0 |  1.155529 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.765312 |  19.20398 |       0.0 | 0.566382 | -0.239912 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |             99.1 |       1.8 |               100.6 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 0.493024 |                      0.493024 |                         0.0 |                     0.424215 |                      0.036869 |                           0.036869 |             0.488586 |                  0.488586 |     100.111414 |   1.294388 |        -0.401992 | 8.625761 | 16.080906 |         1.030461 |                  5.331896 |                 7.455145 |                      7.455145 |             0.799784 | 100.0 |        0.799784 |                0.799784 | 0.7416 | -0.799708 |  0.901798 |              0.187269 |                      0.0 |                  0.187269 |                   0.187269 |                  1.194208 |                     0.0 |                 2.367544 |    -1.006938 |             0.0 |        -2.180275 |         -1.006938 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.012523 |  0.0 |      0.0 | -0.158109 |       0.0 |     -4.89276 |      0.76585 |       0.0 |       0.0 | 0.344018 | 0.344018 |     0.0 | -0.421747 |      0.0 | -0.741828 |      0.0 |    1.0167 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.250194 | 19.362088 |       0.0 | 0.741774 |  -0.40011 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |             96.9 |       2.7 |               100.9 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.523311 |                      0.523311 |                         0.0 |                     0.381225 |                      0.038864 |                           0.038864 |             0.507088 |                  0.507088 |     100.392912 |   1.294334 |        -0.403321 | 8.635699 | 16.070967 |         1.030674 |                  5.287288 |                 7.435268 |                      7.435268 |             2.886262 | 100.0 |        2.886262 |                2.886262 | 0.7416 | -1.333701 |   1.08274 |              0.710165 |                      0.0 |                  0.710165 |                   0.710165 |                  4.005891 |                     0.0 |                 4.970876 |    -3.295726 |             0.0 |        -4.260711 |         -3.295726 |             78.890988 |                 100.0 |               112.102145 |               99.513532 |  0.658 |   0.658 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.234616 |  0.0 |      0.0 | -0.391967 | -0.002888 |    -6.288533 |      0.75395 |       0.0 |       0.0 | 0.474106 | 0.474106 |     0.0 | -1.045552 |      0.0 | -1.976956 |      0.0 |  1.238204 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 18.742201 | 20.412055 |       0.0 | 0.864457 | -0.741251 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |             95.0 |       3.0 |               101.8 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.485698 |                      0.485698 |                         0.0 |                     0.412172 |                      0.036384 |                           0.036384 |             0.461413 |                  0.461413 |     101.338587 |   1.310959 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                  7.41713 |                       7.41713 |             1.347076 | 100.0 |        1.347076 |                1.347076 | 0.7416 | -2.470836 |  0.568272 |               0.31656 |                      0.0 |                   0.31656 |                    0.31656 |                  1.286731 |                     0.0 |                 2.587081 |    -0.970171 |             0.0 |         -2.27052 |         -0.970171 |             71.001889 |                 100.0 |               112.102145 |               99.511814 |  3.102 |   3.102 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.137342 |  0.0 |      0.0 | -0.164151 | -0.020098 |    -5.270657 |     0.765002 |       0.0 |       0.0 | 0.562244 | 0.562244 |     0.0 | -0.437864 |      0.0 | -1.096694 |      0.0 |  0.735963 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.562863 | 23.678206 |       0.0 | 0.749982 | -0.567922 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |             90.6 |       7.4 |               100.4 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 0.879017 |                      0.879017 |                         0.0 |                      0.61078 |                      0.061412 |                           0.061412 |              0.79639 |                   0.79639 |       99.60361 |   1.253213 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 7.400749 |                      7.400749 |             1.373029 | 100.0 |        1.373029 |                1.373029 | 0.7416 | -1.893074 |  -0.98137 |              0.346666 |                      0.0 |                  0.346666 |                   0.346666 |                   1.16859 |                     0.0 |                -0.065197 |    -0.821924 |             0.0 |         0.411862 |         -0.821924 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |  8.836 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.033243 | 0.06 | 0.338985 | -1.215856 |  0.183349 |          0.0 |     0.747518 | 23.713776 | 22.562863 |      0.0 |      0.0 |     0.0 | -3.243236 |      0.0 | -5.063324 |      0.0 | -8.718422 | 0.956772 | 0.274624 | 33.730062 | 11.067544 |  0.0 |  0.0 | 0.0 |  0.0 | 11.067544 | 2.849301 | 0.032978 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |  0.93825 |  1.155047 | 93.769793 | 11.067544 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849301 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032978 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |             78.7 |       5.7 |               101.2 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 0.619748 |                      0.619748 |                         0.0 |                      0.50499 |                      0.045128 |                           0.045128 |             0.487741 |                  0.487741 |     100.712259 |    1.28736 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 7.386143 |                      7.386143 |             1.539952 | 100.0 |        1.539952 |                1.539952 | 0.7416 |       0.0 |  0.607505 |              0.409117 |                      0.0 |                  0.409117 |                   0.409117 |                  1.733999 |                     0.0 |                 3.008114 |    -1.324882 |             0.0 |        -2.598997 |         -1.324882 |             37.369415 |                 100.0 |                81.653408 |               98.490732 |  0.128 |  0.0576 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.429595 | 0.06 | 0.519628 |  0.082368 |  0.000045 |    -2.578858 |     0.734331 |       0.0 |       0.0 | 0.000786 | 0.000786 |     0.0 |  0.244207 |      0.0 | -2.235392 |      0.0 |  0.607812 | 0.951754 | 0.271751 |  0.045632 |  0.015062 |  0.0 |  0.0 | 0.0 |  0.0 |  0.015062 | 3.836775 | 0.044407 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 1.109403 |  1.289142 | 93.280735 |  0.015062 |   0.0 |   0.0 |   0.0 |  0.0 | 3.836775 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044407 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |             90.5 |       2.2 |               102.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 0.471336 |                      0.471336 |                         0.0 |                     0.397748 |                      0.035431 |                           0.035431 |             0.426559 |                  0.426559 |     101.573441 |   1.315658 |        -0.407436 | 8.666669 | 16.039997 |         1.031437 |                  5.150203 |                 7.373328 |                      7.373328 |              0.77253 | 100.0 |         0.77253 |                 0.77253 | 0.7416 |       0.0 |  1.354389 |              0.192007 |                      0.0 |                  0.192007 |                   0.192007 |                  1.254261 |                     0.0 |                 2.684327 |    -1.062254 |             0.0 |         -2.49232 |         -1.062254 |             96.820757 |                 100.0 |                82.295785 |                98.52021 |  2.666 |   2.666 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.193737 |  0.0 |      0.0 | -0.079193 | -0.019502 |    -5.736347 |     0.751457 |       0.0 |       0.0 | 0.043585 |      0.0 |     0.0 | -0.211243 |      0.0 | -0.931632 |      0.0 |  1.349445 | 0.974687 | 0.285005 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 2.050493 | 0.023733 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.745193 |  2.745193 | -0.014557 | 0.892385 |  0.676353 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.050493 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023733 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |             88.1 |       3.1 |               102.5 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                 0.457351 |                      0.457351 |                         0.0 |                     0.368122 |                      0.034499 |                           0.034499 |             0.402926 |                  0.402926 |     102.097074 |   1.324197 |        -0.408164 | 8.672175 | 16.034492 |         1.031605 |                  5.126218 |                 7.362317 |                      7.362317 |             1.699321 | 100.0 |        1.699321 |                1.699321 | 0.7416 |       0.0 |   1.58696 |              0.435319 |                      0.0 |                  0.435319 |                   0.435319 |                  2.107297 |                     0.0 |                 3.808798 |    -1.671978 |             0.0 |         -3.37348 |         -1.671978 |             68.711505 |                 100.0 |                82.295785 |               98.520648 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.280006 |  0.0 |      0.0 | -0.118568 |       0.0 |    -6.741743 |     0.743828 |       0.0 |       0.0 | 0.052586 |      0.0 |     0.0 | -0.316273 |      0.0 |  -1.46724 |      0.0 |  1.589966 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 1.094166 | 0.012664 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.863761 |  2.863761 | -0.017564 |  0.98328 | -0.169006 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.094166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012664 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |             92.8 |       2.8 |               102.3 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 0.380889 |                      0.380889 |                         0.0 |                     0.288325 |                       0.02934 |                            0.02934 |             0.353465 |                  0.353465 |     101.946535 |    1.33374 |         -0.40877 | 8.676772 | 16.029895 |         1.031763 |                  5.106331 |                 7.353123 |                      7.353123 |             2.318249 | 100.0 |        2.318249 |                2.318249 | 0.7416 | -0.563355 |  2.184142 |              0.609046 |                      0.0 |                  0.609046 |                   0.609046 |                  3.397964 |                     0.0 |                 5.076808 |    -2.788918 |             0.0 |        -4.467762 |         -2.788918 |             76.073452 |                 100.0 |                82.295785 |               98.521014 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.031797 |  0.0 |      0.0 | -0.207622 |       0.0 |    -9.857997 |     0.737282 |       0.0 |       0.0 | 0.085596 |      0.0 |     0.0 | -0.553821 |      0.0 | -1.718774 |      0.0 |  2.195167 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.583859 | 0.006758 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.071382 |  3.071382 | -0.028589 | 1.068532 | -1.611548 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.583859 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006758 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |             98.1 |       3.7 |               100.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 0.471336 |                      0.471336 |                         0.0 |                     0.393632 |                      0.035431 |                           0.035431 |             0.462381 |                  0.462381 |      99.537619 |   1.289645 |        -0.409256 | 8.680456 | 16.026211 |         1.031912 |                  5.090537 |                 7.345755 |                      7.345755 |             0.763581 | 100.0 |        0.763581 |                0.763581 | 0.7416 | -5.371828 |  0.099152 |              0.153408 |                      0.0 |                  0.153408 |                   0.153408 |                  1.211777 |                     0.0 |                   2.3744 |     -1.05837 |             0.0 |        -2.220993 |          -1.05837 |             57.569099 |                 100.0 |                82.295785 |               98.521307 |  6.298 |   6.298 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.03395 |  0.0 |      0.0 | -0.264364 |  -0.04607 |    -5.871957 |     0.799094 |       0.0 |       0.0 | 0.326157 |      0.0 |     0.0 | -0.705177 |      0.0 | -1.382387 |      0.0 |  0.133429 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.311554 | 0.003606 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.633746 |  9.633746 | -0.108937 | 0.471314 |   -0.9691 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.311554 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003606 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |             96.6 |       0.9 |                99.1 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                 0.352531 |                      0.352531 |                         0.0 |                     0.279984 |                      0.027395 |                           0.027395 |             0.340545 |                  0.340545 |      98.759455 |   1.296889 |         -0.40962 | 8.683222 | 16.023444 |         1.032051 |                  5.078832 |                 7.340222 |                      7.340222 |             0.761825 | 100.0 |        0.761825 |                0.761825 | 0.7416 | -3.230334 |  1.637841 |              0.159896 |                      0.0 |                  0.159896 |                   0.159896 |                  1.287913 |                     0.0 |                 2.905551 |    -1.128017 |             0.0 |        -2.745655 |         -1.128017 |            236.672963 |                 100.0 |                82.295785 |               98.521527 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001988 |  0.0 |      0.0 | -0.108899 |       0.0 |   -10.226692 |     0.790114 |       0.0 |       0.0 | 0.385193 |      0.0 |     0.0 | -0.290482 |      0.0 | -0.797614 |      0.0 |  1.657559 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.166248 | 0.001924 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.742645 |  9.742645 | -0.128655 |  0.53573 | -1.865341 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166248 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001924 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |             97.7 |       1.3 |                99.7 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 0.395811 |                      0.395811 |                         0.0 |                     0.303297 |                      0.030356 |                           0.030356 |             0.386708 |                  0.386708 |      99.313292 |   1.297204 |        -0.409863 | 8.685068 | 16.021598 |         1.032182 |                  5.071211 |                  7.33653 |                       7.33653 |             0.760682 | 100.0 |        0.760682 |                0.760682 | 0.7416 | -6.217804 |  0.693349 |              0.165025 |                      0.0 |                  0.165025 |                   0.165025 |                  1.257034 |                     0.0 |                 2.432501 |     -1.09201 |             0.0 |        -2.267476 |          -1.09201 |            163.850513 |                 100.0 |                82.295785 |               98.521673 |  0.094 |   0.094 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003569 |  0.0 |      0.0 | -0.174381 | -0.001139 |    -9.219359 |     0.783057 |       0.0 |       0.0 | 0.465048 |      0.0 |     0.0 | -0.465153 |      0.0 | -1.083443 |      0.0 |  0.718881 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.088712 | 0.001027 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.011026 | 10.011026 | -0.155326 | 0.590452 |  -1.81709 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088712 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001027 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |             95.8 |       1.9 |                99.7 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                  0.33643 |                       0.33643 |                         0.0 |                     0.262456 |                      0.026283 |                           0.026283 |               0.3223 |                    0.3223 |        99.3777 |   1.307793 |        -0.409985 | 8.685992 | 16.020675 |         1.032302 |                  5.067673 |                 7.334683 |                      7.334683 |             0.760151 | 100.0 |        0.760151 |                0.760151 | 0.7416 | -6.056967 |  1.162854 |              0.169935 |                      0.0 |                  0.169935 |                   0.169935 |                  1.301361 |                     0.0 |                 2.956416 |    -1.131426 |             0.0 |        -2.786481 |         -1.131426 |            112.108246 |                 100.0 |                82.295785 |               98.521747 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.031907 |  0.0 |      0.0 | -0.151363 |       0.0 |   -11.034367 |     0.776446 |       0.0 |       0.0 | 0.519952 |      0.0 |     0.0 | -0.403752 |      0.0 | -1.201537 |      0.0 |  1.181192 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.047338 | 0.000548 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.162388 | 10.162388 | -0.173664 | 0.645375 | -2.238344 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047338 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000548 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |             97.5 |       2.4 |                99.7 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 0.398857 |                      0.398857 |                         0.0 |                     0.309875 |                      0.030563 |                           0.030563 |             0.388886 |                  0.388886 |      99.311114 |   1.296708 |        -0.409985 | 8.685992 | 16.020675 |         1.032413 |                  5.068219 |                 7.334683 |                      7.334683 |             1.336154 | 100.0 |        1.336154 |                1.336154 | 0.7416 | -7.461147 |  0.350194 |              0.304781 |                      0.0 |                  0.304781 |                   0.304781 |                  1.393316 |                     0.0 |                 2.690071 |    -1.088535 |             0.0 |         -2.38529 |         -1.088535 |             88.752361 |                 100.0 |                82.295785 |               98.521747 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019864 |  0.0 |      0.0 | -0.228727 |  -0.00224 |    -8.947567 |     0.771897 |       0.0 |       0.0 | 0.526991 |      0.0 |     0.0 | -0.610119 |      0.0 | -1.424866 |      0.0 |  0.350305 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |  0.02526 | 0.000292 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.579116 | 10.579116 | -0.176015 | 0.685419 | -1.846938 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02526 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000292 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |             95.0 |       2.3 |               100.4 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 0.401924 |                      0.401924 |                         0.0 |                     0.303538 |                      0.030771 |                           0.030771 |             0.381828 |                  0.381828 |     100.018172 |   1.305372 |        -0.409863 | 8.685068 | 16.021598 |         1.032515 |                  5.072848 |                  7.33653 |                       7.33653 |             1.544792 | 100.0 |        1.544792 |                1.544792 | 0.7416 |  -6.15646 |  0.723866 |               0.35908 |                      0.0 |                   0.35908 |                    0.35908 |                  1.820398 |                     0.0 |                 3.207565 |    -1.461319 |             0.0 |        -2.848486 |         -1.461319 |             92.611159 |                 100.0 |                82.295785 |               98.521673 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.03385 |  0.0 |      0.0 | -0.220917 |   -0.0022 |    -9.209318 |     0.767555 |       0.0 |       0.0 | 0.508871 |      0.0 |     0.0 | -0.589284 |      0.0 | -1.543588 |      0.0 |  0.715613 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.013479 | 0.000156 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.988032 | 10.988032 | -0.169963 |  0.72551 | -1.829204 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.013479 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000156 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |             93.8 |       4.6 |               100.7 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 0.427219 |                      0.427219 |                         0.0 |                     0.352891 |                      0.032479 |                           0.032479 |             0.400731 |                  0.400731 |     100.299269 |   1.305282 |         -0.40962 | 8.683222 | 16.023444 |         1.032607 |                  5.081565 |                 7.340222 |                      7.340222 |             0.762235 | 100.0 |        0.762235 |                0.762235 | 0.7416 | -6.097346 |  0.284988 |              0.167996 |                      0.0 |                  0.167996 |                   0.167996 |                  1.259735 |                     0.0 |                 2.716442 |    -1.091739 |             0.0 |        -2.548446 |         -1.091739 |              46.30558 |                 100.0 |                82.295785 |               98.521527 |  2.256 |   2.256 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.237721 |  0.0 |      0.0 | -0.215444 | -0.022632 |     -7.28687 |     0.779601 |       0.0 |       0.0 | 0.550873 |      0.0 |     0.0 | -0.574686 |      0.0 | -1.697376 |      0.0 |  0.276384 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.007193 | 0.000083 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.459476 | 13.459476 | -0.183992 | 0.618694 | -1.372592 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.007193 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000083 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |             94.9 |       8.6 |               100.4 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 0.563475 |                      0.563475 |                         0.0 |                     0.498371 |                      0.041489 |                           0.041489 |             0.534738 |                  0.534738 |      99.865262 |    1.28304 |        -0.409256 | 8.680456 | 16.026211 |         1.032689 |                  5.094372 |                 7.345755 |                      7.345755 |             0.764156 | 100.0 |        0.764156 |                0.764156 | 0.7416 | -4.575305 | -0.411111 |              0.141857 |                      0.0 |                  0.141857 |                   0.141857 |                  1.179808 |                     0.0 |                 2.282858 |    -1.037951 |             0.0 |        -2.141001 |         -1.037951 |             24.768101 |                 100.0 |                82.295785 |               98.521307 |   6.58 |  5.0995 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461334 |  0.0 |      0.0 | -0.270137 | -0.018542 |    -2.755662 |     0.814361 |       0.0 |       0.0 | 0.509438 | 0.509438 |     0.0 | -0.720577 |      0.0 | -1.863916 |      0.0 | -0.443493 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.003838 | 0.000044 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.325461 | 20.309613 |       0.0 | 0.373379 |  -0.21988 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003838 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000044 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |             85.4 |       7.6 |               100.8 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 0.493024 |                      0.493024 |                         0.0 |                     0.428703 |                      0.036869 |                           0.036869 |             0.421042 |                  0.421042 |     100.378958 |   1.297296 |         -0.40877 | 8.676772 | 16.029895 |         1.032762 |                  5.111276 |                 7.353123 |                      7.353123 |             0.766691 | 100.0 |        0.766691 |                0.766691 | 0.7416 | -0.732934 |  0.888297 |              0.161369 |                      0.0 |                  0.161369 |                   0.161369 |                  1.273508 |                     0.0 |                 2.903665 |    -1.112139 |             0.0 |        -2.742297 |         -1.112139 |             28.027061 |                 100.0 |                82.295785 |               98.521014 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.132218 |  0.0 |      0.0 |  0.051303 | -0.001139 |     -4.75437 |     0.789526 |       0.0 |       0.0 | 0.328908 | 0.328908 |     0.0 |  0.136848 |      0.0 | -1.882132 |      0.0 |  0.997013 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.002048 | 0.000024 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.792706 | 20.446311 |       0.0 | 0.540148 | -0.366577 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002048 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000024 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |             87.2 |       3.8 |               101.8 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                  0.30617 |                       0.30617 |                         0.0 |                     0.233412 |                      0.024175 |                           0.024175 |             0.266981 |                  0.266981 |     101.533019 |   1.341718 |        -0.408164 | 8.672175 | 16.034492 |         1.032825 |                  5.132282 |                 7.362317 |                      7.362317 |             2.398433 | 100.0 |        2.398433 |                2.398433 | 0.7416 | -1.221924 |  2.508159 |               0.55496 |                      0.0 |                   0.55496 |                    0.55496 |                  3.730222 |                     0.0 |                 5.691454 |    -3.175262 |             0.0 |        -5.136494 |         -3.175262 |             56.054123 |                 100.0 |                82.295785 |               98.520648 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.145715 |  0.0 |      0.0 | -0.131539 |       0.0 |   -12.484807 |     0.768616 |       0.0 |       0.0 | 0.800223 | 0.653605 |     0.0 | -0.350874 |      0.0 | -2.010188 |      0.0 |  2.775433 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.001093 | 0.000013 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.57785 |  20.57785 |  -0.04897 |  0.71554 | -2.133136 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001093 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000013 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |             92.7 |       1.5 |               101.9 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 0.318472 |                      0.318472 |                         0.0 |                     0.197901 |                      0.025034 |                           0.025034 |             0.295223 |                  0.295223 |     101.604777 |   1.340358 |        -0.407436 | 8.666669 | 16.039997 |         1.032879 |                  5.157401 |                 7.373328 |                      7.373328 |             2.793204 | 100.0 |        2.793204 |                2.793204 | 0.7416 | -7.110453 |  1.528737 |              0.655326 |                      0.0 |                  0.655326 |                   0.655326 |                  4.423241 |                     0.0 |                 5.150105 |    -3.767915 |             0.0 |        -4.494778 |         -3.767915 |            142.003778 |                 100.0 |                82.295785 |                98.52021 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.167474 |  0.0 |      0.0 | -0.217696 | -0.020275 |   -14.494219 |     0.765385 |       0.0 |       0.0 | 1.366309 |      0.0 |     0.0 | -0.580693 |      0.0 | -1.998246 |      0.0 |  1.915839 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000583 | 0.000007 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.923545 | 21.923545 | -0.456347 |  0.74626 | -2.920273 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000583 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000007 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |             97.8 |       1.3 |               101.7 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                  0.41125 |                       0.41125 |                         0.0 |                     0.296322 |                      0.031402 |                           0.031402 |             0.402202 |                  0.402202 |     101.297798 |   1.320718 |        -0.406588 | 8.660262 | 16.046405 |         1.032922 |                  5.186639 |                 7.386143 |                      7.386143 |             0.777996 | 100.0 |        0.777996 |                0.777996 | 0.7416 | -9.734244 | -0.048282 |              0.181176 |                      0.0 |                  0.181176 |                   0.181176 |                  1.247575 |                     0.0 |                 2.051458 |    -1.066399 |             0.0 |        -1.870283 |         -1.066399 |            163.850513 |                 100.0 |                82.295785 |               98.519701 |  0.846 |   0.846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001963 |  0.0 |      0.0 | -0.221357 | -0.009371 |    -9.513339 |     0.767125 |       0.0 |       0.0 | 1.373666 |      0.0 |     0.0 | -0.590459 |      0.0 |  -1.33502 |      0.0 | -0.055196 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000311 | 0.000004 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.990902 | 22.990902 | -0.458804 | 0.729578 | -2.130391 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000311 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000004 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |             95.4 |       3.0 |               102.2 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 0.341723 |                      0.341723 |                         0.0 |                     0.235849 |                      0.026649 |                           0.026649 |             0.326003 |                  0.326003 |     101.873997 |   1.339598 |        -0.405619 | 8.652959 | 16.053708 |         1.032956 |                   5.22001 |                 7.400749 |                      7.400749 |             2.504073 | 100.0 |        2.504073 |                2.504073 | 0.7416 | -7.101304 |  1.239908 |              0.595257 |                      0.0 |                  0.595257 |                   0.595257 |                    3.6896 |                     0.0 |                 4.987128 |    -3.094344 |             0.0 |        -4.391872 |         -3.094344 |             71.001889 |                 100.0 |                82.295785 |                98.51912 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.120498 |  0.0 |      0.0 | -0.300538 |       0.0 |   -12.357122 |     0.762285 |       0.0 |       0.0 | 1.365727 |      0.0 |     0.0 |  -0.80167 |      0.0 | -2.352945 |      0.0 |  1.237257 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000166 | 0.000002 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  23.29144 |  23.29144 | -0.456153 | 0.776804 | -2.628699 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000002 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |             94.9 |       2.7 |               102.4 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 0.303761 |                      0.303761 |                         0.0 |                     0.232742 |                      0.024006 |                           0.024006 |              0.28827 |                   0.28827 |      102.11173 |   1.350039 |         -0.40453 | 8.644768 | 16.061898 |         1.032981 |                  5.257524 |                  7.41713 |                       7.41713 |             0.788629 | 100.0 |        0.788629 |                0.788629 | 0.7416 | -8.762331 |  0.857693 |              0.191133 |                      0.0 |                  0.191133 |                   0.191133 |                  1.325064 |                     0.0 |                 3.128101 |    -1.133931 |             0.0 |        -2.936968 |         -1.133931 |             78.890988 |                 100.0 |                82.295785 |               98.518469 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073401 |  0.0 |      0.0 | -0.172927 |       0.0 |   -12.520123 |     0.757639 |       0.0 |       0.0 |  1.51821 |      0.0 |     0.0 | -0.461273 |      0.0 | -1.567073 |      0.0 |  0.908622 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000089 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.464367 | 23.464367 | -0.507082 | 0.824618 | -2.744792 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000089 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |             96.0 |       1.4 |               102.1 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 0.349801 |                      0.349801 |                         0.0 |                     0.253668 |                      0.027207 |                           0.027207 |             0.335809 |                  0.335809 |     101.764191 |   1.336727 |        -0.403321 | 8.635699 | 16.070967 |         1.032995 |                  5.299194 |                 7.435268 |                      7.435268 |             0.794879 | 100.0 |        0.794879 |                0.794879 | 0.7416 | -9.149307 |  0.546932 |              0.194815 |                      0.0 |                  0.194815 |                   0.194815 |                  1.292505 |                     0.0 |                 2.540901 |     -1.09769 |             0.0 |        -2.346087 |          -1.09769 |            152.146905 |                 100.0 |                82.295785 |               98.517748 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019443 |  0.0 |      0.0 | -0.177732 | -0.002908 |   -11.457537 |     0.754913 |       0.0 |       0.0 | 1.510287 |      0.0 |     0.0 | -0.474092 |      0.0 | -1.330616 |      0.0 |  0.541379 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000047 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |   23.8301 |   23.8301 | -0.504436 | 0.853891 | -2.550124 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000047 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |             96.5 |       4.2 |               101.6 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 0.690794 |                      0.690794 |                         0.0 |                     0.494668 |                      0.049663 |                           0.049663 |             0.666616 |                  0.666616 |     100.933384 |   1.284547 |        -0.401992 | 8.625761 | 16.080906 |            1.033 |                  5.345035 |                 7.455145 |                      7.455145 |             2.770177 | 100.0 |        2.770177 |                2.770177 | 0.7416 | -8.500415 |  -1.23571 |              0.689458 |                      0.0 |                  0.689458 |                   0.689458 |                  3.475519 |                     0.0 |                 4.119365 |    -2.786061 |             0.0 |        -3.429907 |         -2.786061 |             50.715635 |                 100.0 |                82.295785 |               98.516958 |  0.846 | 0.06345 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.125078 |  0.0 |      0.0 | -0.724068 |  0.005795 |    -2.855473 |     0.751114 |       0.0 |       0.0 | 0.974534 |  0.78255 |     0.0 | -1.931414 |      0.0 |  -2.90735 |      0.0 | -1.408857 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000025 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 25.400167 | 25.400167 | -0.064123 | 0.896287 | -0.572814 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000025 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> ktschnee(inf)

hourly simulation
_________________

>>> pub.timegrids = "1997-08-03", "1997-08-06", "1h"
>>> for parameter in control:
...     if parameter.TIME == True:
...         parameter.value = parameter.value/24
...     if parameter.TIME == False:
...         parameter.value = parameter.value*24
>>> for parameter in fixed:
...     parameter.restore()
>>> parameterstep("1h")
>>> test = IntegrationTest(land)

.. _lland_v4_acker_summer_hourly:

acre (summer)
-------------

>>> inputs.nied.series = 0.0
>>> inputs.teml.series = (
...     16.9, 16.6, 16.4, 16.3, 16.0, 15.9, 16.0, 16.6, 17.4, 19.0, 20.3,
...     21.4, 21.3, 21.8, 22.9, 22.7, 22.5, 21.9, 21.4, 20.7, 19.4, 17.8,
...     17.0, 16.4, 15.9, 15.4, 15.2, 14.9, 14.3, 14.1, 14.3, 15.4, 16.8,
...     18.2, 19.5, 20.3, 21.0, 21.8, 21.9, 21.8, 22.2, 21.8, 20.9, 19.7,
...     17.9, 15.8, 14.8, 14.1, 13.4, 12.8, 12.3, 12.0, 11.7, 11.6, 12.6,
...     14.8, 17.1, 19.4, 20.6, 21.4, 22.5, 23.1, 23.4, 23.8, 24.0, 24.0,
...     23.7, 22.8, 21.3, 19.3, 18.3, 17.4)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5, 0.7, 0.8, 0.5,
...     0.4, 0.5, 0.5, 0.3, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.9, 1.0, 1.0, 0.9, 0.8, 0.9, 0.8, 0.9, 0.9, 0.9, 1.0,
...     1.0, 1.0, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.9,
...     1.0, 0.9, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     0.8, 0.8, 0.8, 0.8, 0.8, 0.6, 0.9, 0.9, 0.9, 1.3, 1.5, 1.2, 1.3, 1.5,
...     1.9, 1.9, 2.3, 2.4, 2.5, 2.5, 2.2, 1.7, 1.7, 2.3, 2.3, 2.2, 2.3, 2.0,
...     2.3, 2.5, 2.4, 2.5, 2.5, 2.9, 3.1, 3.3, 3.3, 3.2, 2.5, 2.9, 3.6, 4.2,
...     4.2, 3.6, 3.2, 2.2, 2.2, 2.1, 1.8, 1.8, 1.6, 1.2, 1.9, 1.3, 1.6, 1.9,
...     3.2, 4.0, 4.6, 4.1, 4.1, 4.4, 4.6, 4.5, 4.7, 4.5, 4.1, 3.8, 2.5, 1.9,
...     2.2, 2.7)
>>> inputs.relativehumidity.series = (
...     95.1, 94.9, 95.9, 96.7, 97.2, 97.5, 97.7, 97.4, 96.8, 86.1, 76.8,
...     71.8, 67.5, 66.1, 63.4, 62.4, 61.1, 62.1, 67.0, 74.5, 81.2, 86.9,
...     90.1, 90.9, 88.0, 89.2, 88.8, 89.7, 93.0, 93.6, 94.6, 93.3, 86.9,
...     78.9, 75.8, 73.4, 68.3, 61.3, 60.6, 58.8, 57.2, 57.2, 59.9, 66.1,
...     71.7, 80.6, 85.9, 89.3, 91.4, 93.1, 94.6, 95.6, 96.2, 95.7, 95.1,
...     96.3, 89.9, 79.2, 73.4, 67.3, 62.2, 59.6, 55.4, 52.1, 49.5, 48.1,
...     49.4, 57.0, 65.9, 73.0, 76.7, 81.8)
>>> inputs.atmosphericpressure.series = (
...     101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.5, 101.6, 101.6,
...     101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6, 101.6,
...     101.6, 101.6, 101.7, 101.7, 101.7, 101.6, 101.6, 101.6, 101.6, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.7,
...     101.7, 101.7, 101.7, 101.7, 101.7, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8, 101.8, 101.8, 101.8, 101.9, 101.9, 101.9, 101.9, 101.9,
...     101.9, 101.9, 101.9, 101.9, 101.8, 101.8, 101.8, 101.8, 101.8, 101.8,
...     101.8, 101.8)
>>> inlet.sequences.sim.series = 0.0

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2,
...        0.1, 0.2, 0.1, 0.2, 0.2, 0.3, 0.0, 0.0,
...        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.2, 0.5,
...       0.8, 1.1, 1.2, 1.4, 1.9, 1.6, 0.9, 0.8,
...       0.6, 0.4, 0.2, 0.1, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [13.2, 13.2, 13.1, 12.6, 12.7, 13.0, 13.5, 14.8,
...       16.2, 17.7, 18.8, 19.4, 20.4, 21.0, 21.5, 21.2,
...       20.4, 20.7, 20.2, 19.7, 19.0, 18.0, 17.5, 17.1]),
...     (logs.loggedrelativehumidity,
...      [95.1, 94.5, 94.8, 96.4, 96.6, 97.1, 97.1, 96.7,
...       92.2, 88.5, 81.1, 76.5, 75.1, 70.8, 68.9, 69.2,
...       75.0, 74.0, 77.4, 81.4, 85.3, 90.1, 92.3, 93.8]),
...     (logs.loggedwindspeed2m,
...      [0.8, 1.0, 1.2, 1.3, 0.9, 1.1, 1.3, 1.3,
...       1.9, 2.2, 1.8, 2.3, 2.4, 2.5, 2.4, 2.5,
...       2.1, 2.2, 1.7, 1.7, 1.3, 1.3, 0.7, 0.8]))

.. integration-test::

    >>> test("lland_v4_acker_summer_hourly",
    ...      axis1=(fluxes.evb, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                         0.0 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.901196 |                     0.0 |                      0.0 |     -0.03755 |             0.0 |              0.0 |           7.08166 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002965 | -0.002965 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.256857 | 0.071349 | 0.002965 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.256857 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.071349 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                         0.0 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.901844 |                     0.0 |                      0.0 |    -0.037577 |             0.0 |              0.0 |          7.081012 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002685 | -0.002685 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.250222 | 0.069506 |  0.00565 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.250222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069506 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                         0.0 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.902355 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |            7.0805 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004107 | -0.004107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.243758 | 0.067711 | 0.009757 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.243758 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.067711 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                         0.0 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.902364 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          7.080492 |            176.542034 |                 100.0 |               100.629116 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005232 | -0.005232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.237462 | 0.065962 | 0.014989 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.237462 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065962 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                         0.0 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.005318 |                      0.0 |                  0.005318 |                   7.982856 |                  0.902341 |                     0.0 |                      0.0 |     -0.03228 |             0.0 |              0.0 |          7.080514 |            176.542034 |                 100.0 |               100.629116 |               92.771119 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004819 | -0.004819 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.231328 | 0.064258 | 0.019808 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.231328 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064258 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                         0.0 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.060007 |                      0.0 |                  0.060007 |                   7.982856 |                  0.901482 |                     0.0 |                      0.0 |     0.022445 |             0.0 |              0.0 |          7.081373 |            235.389378 |                 100.0 |               100.629116 |              100.447767 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005073 |  0.005073 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.225353 | 0.062598 | 0.014735 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 79.623186 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.225353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062598 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                         0.0 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.156653 |                      0.0 |                  0.156653 |                   7.982856 |                   0.89977 |                     0.0 |                      0.0 |     0.119163 |             0.0 |              0.0 |          7.083086 |            156.926252 |                 100.0 |               100.629116 |              100.447767 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026292 |  0.014735 | 0.008821 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.219531 | 0.060981 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 79.614365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.219531 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060981 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                         0.0 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.299135 |                      0.0 |                  0.299135 |                   7.982856 |                  0.896826 |                     0.0 |                      0.0 |     0.261767 |             0.0 |              0.0 |          7.086029 |            156.926252 |                 100.0 |               100.641253 |               100.45639 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.057456 |       0.0 | 0.044017 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.213861 | 0.059406 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 79.570349 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.213861 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.059406 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                         0.0 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.467717 |                      0.0 |                  0.467717 |                   7.982856 |                  0.891371 |                     0.0 |                      0.0 |     0.430576 |             0.0 |              0.0 |          7.091484 |            156.926252 |                 100.0 |               100.701823 |              100.499412 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096099 |       0.0 | 0.073977 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.208337 | 0.057871 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 79.496372 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.208337 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.057871 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                         0.0 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.852983 |                      0.0 |                  0.852983 |                    8.15863 |                  0.922604 |                     0.0 |                      0.0 |     0.814541 |             0.0 |              0.0 |          7.236026 |            108.641252 |                 100.0 |               100.803646 |              100.571701 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222763 |       0.0 | 0.161018 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.202955 | 0.056376 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 79.335354 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.202955 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056376 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                         0.0 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.372332 |                      0.0 |                  1.372332 |                   8.334404 |                  0.955299 |                     0.0 |                      0.0 |     1.332528 |             0.0 |              0.0 |          7.379106 |             94.155751 |                 100.0 |               101.025386 |              100.728981 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.389923 |       0.0 | 0.275567 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.197713 |  0.05492 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 79.059786 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.197713 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05492 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                         0.0 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.682691 |                      0.0 |                  1.682691 |                   8.773841 |                  1.039981 |                     0.0 |                      0.0 |     1.639359 |             0.0 |              0.0 |          7.733859 |            117.694689 |                 100.0 |               101.405231 |              100.997942 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.46628 |       0.0 | 0.348858 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.192605 | 0.053502 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 78.710929 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.192605 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053502 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                         0.0 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.714225 |                      0.0 |                  1.714225 |                   9.301164 |                  1.142661 |                     0.0 |                      0.0 |     1.666614 |             0.0 |              0.0 |          8.158503 |            108.641252 |                 100.0 |               101.886758 |              101.338062 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.491066 |       0.0 | 0.360852 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18763 |  0.05212 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 78.350077 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18763 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05212 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                         0.0 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.357421 |                      0.0 |                  1.357421 |                   9.652713 |                  1.211985 |                     0.0 |                      0.0 |     1.306922 |             0.0 |              0.0 |          8.440727 |             94.155751 |                 100.0 |               102.385622 |              101.689443 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424802 |       0.0 |  0.30342 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.182784 | 0.050773 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 78.046657 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.182784 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050773 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                         0.0 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.147809 |                      0.0 |                  1.147809 |                   9.828487 |                   1.24753 |                     0.0 |                      0.0 |     1.095829 |             0.0 |              0.0 |          8.580957 |             74.333488 |                 100.0 |               102.805714 |              101.984567 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.415458 |       0.0 | 0.282559 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.178062 | 0.049462 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.764098 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.178062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.049462 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                         0.0 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |               1.06126 |                      0.0 |                   1.06126 |                  10.180036 |                  1.316893 |                     0.0 |                      0.0 |     1.006389 |             0.0 |              0.0 |          8.863143 |             74.333488 |                 100.0 |               103.197445 |              102.259131 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.393508 |       0.0 | 0.266848 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.173463 | 0.048184 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06233 |  77.49725 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.173463 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.048184 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                         0.0 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.761967 |                      0.0 |                  0.761967 |                  10.443698 |                  1.370145 |                     0.0 |                      0.0 |     0.704878 |             0.0 |              0.0 |          9.073553 |             61.405925 |                 100.0 |               103.567861 |              102.518193 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.342513 |       0.0 |   0.2192 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.168982 | 0.046939 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 77.278049 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.168982 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046939 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                         0.0 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.375211 |                      0.0 |                  0.375211 |                   10.70736 |                  1.421011 |                     0.0 |                      0.0 |     0.316002 |             0.0 |              0.0 |          9.286348 |             58.847345 |                 100.0 |                103.87248 |               102.73083 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.236583 |       0.0 | 0.148202 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.164617 | 0.045727 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 77.129847 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.164617 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045727 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                         0.0 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.139757 |                      0.0 |                  0.139757 |                  10.795247 |                  1.433696 |                     0.0 |                      0.0 |     0.080019 |             0.0 |              0.0 |           9.36155 |             56.493451 |                 100.0 |               104.078611 |              102.874509 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.156446 |       0.0 | 0.096042 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.160365 | 0.044546 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 77.033805 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.160365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044546 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                         0.0 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.037299 |                      0.0 |                  0.037299 |                  10.795247 |                  1.423756 |                     0.0 |                      0.0 |    -0.022024 |             0.0 |              0.0 |          9.371491 |             56.493451 |                 100.0 |                104.21227 |              102.967584 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.09619 |       0.0 |   0.0585 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.156222 | 0.043395 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 76.975305 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.156222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043395 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                         0.0 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000509 |                      0.0 |                  0.000509 |                  10.795247 |                  1.412078 |                     0.0 |                      0.0 |    -0.058328 |             0.0 |              0.0 |          9.383169 |             64.197103 |                 100.0 |               104.293713 |               89.533534 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04938 |       0.0 | 0.032417 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.152187 | 0.042274 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 76.942888 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.152187 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042274 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                         0.0 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                  1.401564 |                     0.0 |                      0.0 |    -0.058399 |             0.0 |              0.0 |          9.393683 |             83.078604 |                 100.0 |               104.338854 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019258 |       0.0 | 0.013442 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.148256 | 0.041182 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 76.929445 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.148256 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041182 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                         0.0 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                  1.391158 |                     0.0 |                      0.0 |    -0.057965 |             0.0 |              0.0 |          9.404089 |             83.078604 |                 100.0 |               104.357574 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010298 |       0.0 | 0.007136 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.144426 | 0.040118 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.248458 |  76.92231 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.144426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040118 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                         0.0 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                   1.38284 |                     0.0 |                      0.0 |    -0.057618 |             0.0 |              0.0 |          9.412407 |             61.405925 |                 100.0 |               104.367512 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014661 |       0.0 | 0.009245 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.140696 | 0.039082 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.913065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039082 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                         0.0 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                   1.39654 |                     0.0 |                      0.0 |    -0.058189 |             0.0 |              0.0 |          9.343179 |             61.405925 |                 100.0 |               104.380388 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023989 |       0.0 | 0.015041 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.137061 | 0.038073 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.898024 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.137061 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038073 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                         0.0 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                  1.406633 |                     0.0 |                      0.0 |     -0.05861 |             0.0 |              0.0 |          9.333085 |             64.197103 |                 100.0 |               104.401338 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018182 |       0.0 |   0.0115 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.133521 | 0.037089 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.886524 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.133521 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037089 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                         0.0 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                  1.418069 |                     0.0 |                      0.0 |    -0.059086 |             0.0 |              0.0 |           9.32165 |             61.405925 |                 100.0 |               104.417357 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02055 |       0.0 |  0.01278 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.130072 | 0.036131 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 76.873744 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.130072 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036131 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                         0.0 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                   1.43017 |                     0.0 |                      0.0 |     -0.05959 |             0.0 |              0.0 |          9.309548 |             70.616813 |                 100.0 |               104.435159 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01363 |       0.0 | 0.008835 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.126712 | 0.035198 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.864909 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.126712 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035198 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                         0.0 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.004352 |                      0.0 |                  0.004352 |                  10.739718 |                  1.440702 |                     0.0 |                      0.0 |    -0.055678 |             0.0 |              0.0 |          9.299016 |             61.405925 |                 100.0 |               104.447467 |               93.267161 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007255 |       0.0 | 0.004359 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.123439 | 0.034289 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.853319 |  76.86055 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.123439 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034289 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                         0.0 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.054586 |                      0.0 |                  0.054586 |                  10.739718 |                  1.451305 |                     0.0 |                      0.0 |    -0.005885 |             0.0 |              0.0 |          9.288414 |             56.493451 |                 100.0 |               104.453539 |              103.135415 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018162 |       0.0 | 0.010113 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.120251 | 0.033403 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.850437 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.120251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033403 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                         0.0 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.172329 |                      0.0 |                  0.172329 |                   10.82738 |                  1.478175 |                     0.0 |                      0.0 |     0.110738 |             0.0 |              0.0 |          9.349205 |             58.847345 |                 100.0 |               104.467628 |              103.145209 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04093 |       0.0 | 0.023217 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.117144 |  0.03254 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.827221 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.117144 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03254 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                         0.0 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |              0.781884 |                      0.0 |                  0.781884 |                  11.616331 |                   1.64464 |                     0.0 |                      0.0 |     0.713358 |             0.0 |              0.0 |          9.971692 |             56.493451 |                 100.0 |               104.499975 |              103.167691 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.18972 |       0.0 | 0.107491 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.114118 |   0.0317 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.719729 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114118 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0317 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                         0.0 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.300216 |                      0.0 |                  1.300216 |                  12.492945 |                   1.83472 |                     0.0 |                      0.0 |      1.22377 |             0.0 |              0.0 |         10.658225 |             56.493451 |                 100.0 |               104.649787 |               103.27176 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.343734 |       0.0 | 0.198364 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.111171 | 0.030881 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.521365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.111171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030881 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                         0.0 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.679922 |                      0.0 |                  1.679922 |                  13.194235 |                  1.989515 |                     0.0 |                      0.0 |     1.597026 |             0.0 |              0.0 |          11.20472 |             48.701251 |                 100.0 |               104.926453 |              103.463719 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.494183 |       0.0 | 0.274461 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.108299 | 0.030083 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.246905 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.108299 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030083 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                         0.0 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              1.761455 |                      0.0 |                  1.761455 |                   13.54488 |                   2.06621 |                     0.0 |                      0.0 |     1.675363 |             0.0 |              0.0 |          11.47867 |             45.559235 |                 100.0 |               105.309691 |              103.729123 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.550449 |       0.0 | 0.303114 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.105502 | 0.029306 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 75.943791 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.105502 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029306 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                         0.0 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |               1.64232 |                      0.0 |                   1.64232 |                  13.632542 |                  2.087634 |                     0.0 |                      0.0 |     1.555335 |             0.0 |              0.0 |         11.544908 |             42.798069 |                 100.0 |               105.733535 |              104.021981 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.548384 |       0.0 | 0.297669 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.102776 | 0.028549 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 75.646122 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.102776 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028549 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                         0.0 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.697727 |                      0.0 |                  1.697727 |                  13.720203 |                  2.106001 |                     0.0 |                      0.0 |     1.609977 |             0.0 |              0.0 |         11.614202 |             42.798069 |                 100.0 |               106.150385 |              104.309322 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.597559 |       0.0 | 0.327471 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.100122 | 0.027812 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.318651 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.100122 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027812 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                         0.0 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |              1.640072 |                      0.0 |                  1.640072 |                  13.983187 |                  2.166369 |                     0.0 |                      0.0 |     1.549806 |             0.0 |              0.0 |         11.816818 |             44.135508 |                 100.0 |               106.609685 |              104.625144 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.621801 |       0.0 | 0.348686 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.097535 | 0.027093 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 74.969965 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.097535 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027093 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                         0.0 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |              1.678676 |                      0.0 |                  1.678676 |                  14.421494 |                  2.265756 |                     0.0 |                      0.0 |      1.58427 |             0.0 |              0.0 |         12.155738 |             56.493451 |                 100.0 |               107.099579 |              104.961105 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.589183 |       0.0 | 0.361258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.095016 | 0.026393 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 74.608707 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.095016 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026393 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                         0.0 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.414805 |                      0.0 |                  1.414805 |                  14.772139 |                  2.348437 |                     0.0 |                      0.0 |     1.316953 |             0.0 |              0.0 |         12.423701 |             48.701251 |                 100.0 |               107.608064 |              105.308836 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.551424 |       0.0 | 0.320065 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.092562 | 0.025712 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.139377 | 74.288642 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.092562 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025712 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                         0.0 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.033761 |                      0.0 |                  1.033761 |                  15.122784 |                  2.428674 |                     0.0 |                      0.0 |     0.932566 |             0.0 |              0.0 |          12.69411 |             39.231563 |                 100.0 |               108.059369 |              105.616633 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.505761 |       0.0 | 0.270996 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.090171 | 0.025047 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.017646 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.090171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025047 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                         0.0 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.690032 |                      0.0 |                  0.690032 |                  15.736413 |                  2.564427 |                     0.0 |                      0.0 |     0.583181 |             0.0 |              0.0 |         13.171986 |             33.627054 |                 100.0 |               108.442082 |               105.87704 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.444603 |       0.0 | 0.220974 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.087841 |   0.0244 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 73.796672 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.087841 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0244 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                         0.0 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.351786 |                      0.0 |                  0.351786 |                  16.525365 |                  2.744116 |                     0.0 |                      0.0 |     0.237448 |             0.0 |              0.0 |         13.781249 |             33.627054 |                 100.0 |               108.754562 |              106.089245 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.32478 |       0.0 | 0.158718 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.085572 |  0.02377 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 73.637953 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.085572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02377 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                         0.0 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.115111 |                      0.0 |                  0.115111 |                  17.401978 |                  2.950104 |                     0.0 |                      0.0 |     -0.00781 |             0.0 |              0.0 |         14.451875 |             39.231563 |                 100.0 |               108.979235 |              106.241591 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.183091 |       0.0 | 0.093928 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.083362 | 0.023156 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 73.544026 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083362 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023156 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                         0.0 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002224 |                      0.0 |                  0.002224 |                  17.664962 |                  3.034376 |                     0.0 |                      0.0 |    -0.124208 |             0.0 |              0.0 |         14.630586 |             44.135508 |                 100.0 |               109.112286 |               89.357093 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.100631 |       0.0 | 0.057096 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.081209 | 0.022558 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.30697 |  73.48693 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.081209 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022558 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                         0.0 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.060678 |                     0.0 |                      0.0 |    -0.127528 |             0.0 |              0.0 |         14.604285 |             64.197103 |                 100.0 |               109.193197 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029981 |       0.0 |  0.01905 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079111 | 0.021975 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.235572 |  73.46788 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079111 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021975 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                         0.0 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.084324 |                     0.0 |                      0.0 |    -0.128514 |             0.0 |              0.0 |         14.580638 |             64.197103 |                 100.0 |               109.220198 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012064 |       0.0 |  0.00758 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077068 | 0.021408 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.147714 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077068 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021408 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                         0.0 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.103639 |                     0.0 |                      0.0 |    -0.129318 |             0.0 |              0.0 |         14.561323 |             67.254108 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000111 | -0.000111 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075077 | 0.020855 | 0.000111 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.051062 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020855 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                         0.0 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.124611 |                     0.0 |                      0.0 |    -0.130192 |             0.0 |              0.0 |         14.465269 |             78.463126 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009107 | -0.009107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073138 | 0.020316 | 0.009218 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.946249 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073138 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020316 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                         0.0 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.135507 |                     0.0 |                      0.0 |    -0.130646 |             0.0 |              0.0 |         14.454373 |             78.463126 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.013348 | -0.013348 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.071248 | 0.019791 | 0.022566 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.836022 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.071248 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019791 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                         0.0 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.144898 |                     0.0 |                      0.0 |    -0.131037 |             0.0 |              0.0 |         14.444982 |             88.271017 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.017799 | -0.017799 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.069408 |  0.01928 | 0.040365 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.722932 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.069408 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01928 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                         0.0 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                   3.15411 |                     0.0 |                      0.0 |    -0.131421 |             0.0 |              0.0 |          14.43577 |            117.694689 |                 100.0 |               109.230942 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.020794 | -0.020794 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.067615 | 0.018782 | 0.061159 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.611504 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.067615 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018782 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                         0.0 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004056 |                      0.0 |                  0.004056 |                   17.58988 |                  3.166293 |                     0.0 |                      0.0 |    -0.127873 |             0.0 |              0.0 |         14.423587 |             74.333488 |                 100.0 |               109.230942 |               93.746935 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019617 | -0.019617 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.065868 | 0.018297 | 0.080776 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.065868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018297 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                         0.0 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.057827 |                      0.0 |                  0.057827 |                   17.58988 |                  3.179808 |                     0.0 |                      0.0 |    -0.074665 |             0.0 |              0.0 |         14.410072 |            108.641252 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   -0.0093 |   -0.0093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064167 | 0.017824 | 0.090076 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.397486 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064167 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017824 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                         0.0 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.344417 |                      0.0 |                  0.344417 |                  18.027018 |                  3.288825 |                     0.0 |                      0.0 |     0.207383 |             0.0 |              0.0 |         14.738193 |             88.271017 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.05232 |   0.05232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06251 | 0.017364 | 0.037756 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.32245 |   73.4603 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017364 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                         0.0 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.837093 |                      0.0 |                  0.837093 |                  18.027018 |                  3.287363 |                     0.0 |                      0.0 |     0.700119 |             0.0 |              0.0 |         14.739656 |             74.333488 |                 100.0 |               109.230942 |               106.41204 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.166735 |  0.037756 | 0.079061 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.060895 | 0.016915 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 73.381239 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060895 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016915 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                         0.0 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.398382 |                      0.0 |                  1.398382 |                  18.027018 |                  3.279458 |                     0.0 |                      0.0 |     1.261738 |             0.0 |              0.0 |         14.747561 |             44.135508 |                 100.0 |               109.343038 |              106.487872 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.358814 |       0.0 | 0.185653 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.059322 | 0.016478 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 73.195586 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.059322 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016478 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                         0.0 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.722365 |                      0.0 |                  1.722365 |                  17.939591 |                  3.250753 |                     0.0 |                      0.0 |     1.586917 |             0.0 |              0.0 |         14.688838 |             35.308407 |                 100.0 |               109.606454 |              106.665883 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.541756 |       0.0 | 0.263073 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05779 | 0.016053 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 72.932513 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05779 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016053 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                         0.0 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.095274 |                      0.0 |                  2.095274 |                  18.027018 |                  3.266966 |                     0.0 |                      0.0 |     1.959151 |             0.0 |              0.0 |         14.760052 |             30.702962 |                 100.0 |               109.980185 |              106.917992 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.716789 |       0.0 | 0.332173 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056297 | 0.015638 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.50337 |  72.60034 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056297 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015638 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                         0.0 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.098911 |                      0.0 |                  2.098911 |                  18.201874 |                  3.309837 |                     0.0 |                      0.0 |     1.961001 |             0.0 |              0.0 |         14.892037 |             34.447226 |                 100.0 |               110.452867 |                107.2361 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.746035 |       0.0 | 0.369918 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054843 | 0.015234 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 72.230422 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054843 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015234 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                         0.0 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.014418 |                      0.0 |                  2.014418 |                  18.289301 |                  3.330376 |                     0.0 |                      0.0 |     1.875652 |             0.0 |              0.0 |         14.958925 |             34.447226 |                 100.0 |               110.980311 |              107.590074 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.771401 |       0.0 | 0.389258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.053426 | 0.014841 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 71.841164 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.053426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014841 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                         0.0 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.101265 |                      0.0 |                  2.101265 |                  18.464157 |                  3.363175 |                     0.0 |                      0.0 |     1.961133 |             0.0 |              0.0 |         15.100981 |             32.098552 |                 100.0 |               111.536545 |              107.962247 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.840981 |       0.0 | 0.414609 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052046 | 0.014457 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 71.426555 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052046 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014457 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                         0.0 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.013503 |                      0.0 |                  2.013503 |                  18.551584 |                   3.38194 |                     0.0 |                      0.0 |     1.872589 |             0.0 |              0.0 |         15.169644 |             30.702962 |                 100.0 |               112.130402 |              108.358325 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.865172 |       0.0 | 0.419297 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.050702 | 0.014084 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 71.007258 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.050702 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014084 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                         0.0 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.697346 |                      0.0 |                  1.697346 |                  18.639012 |                  3.400029 |                     0.0 |                      0.0 |     1.555678 |             0.0 |              0.0 |         15.238983 |              31.38525 |                 100.0 |               112.732468 |              108.758548 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.798412 |       0.0 | 0.393038 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049392 |  0.01372 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.111043 |  70.61422 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049392 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01372 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                         0.0 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.231461 |                      0.0 |                  1.231461 |                   18.72644 |                   3.42173 |                     0.0 |                      0.0 |     1.088889 |             0.0 |              0.0 |         15.304709 |             30.049708 |                 100.0 |               113.298222 |              109.133419 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.702711 |       0.0 | 0.339609 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048116 | 0.013366 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 70.274611 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013366 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                         0.0 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |              0.748608 |                      0.0 |                  0.748608 |                   18.72644 |                  3.423912 |                     0.0 |                      0.0 |     0.605945 |             0.0 |              0.0 |         15.302528 |              31.38525 |                 100.0 |               113.788174 |              109.457117 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.55931 |       0.0 | 0.275426 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046873 |  0.01302 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 69.999185 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046873 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01302 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                         0.0 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |              0.364895 |                      0.0 |                  0.364895 |                   18.72644 |                   3.42458 |                     0.0 |                      0.0 |     0.222204 |             0.0 |              0.0 |          15.30186 |             34.447226 |                 100.0 |               114.186294 |                109.7195 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.406398 |       0.0 | 0.207367 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.045662 | 0.012684 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 69.791818 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.045662 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012684 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                         0.0 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.110333 |                      0.0 |                  0.110333 |                   18.72644 |                  3.420162 |                     0.0 |                      0.0 |    -0.032173 |             0.0 |              0.0 |         15.306277 |             37.166744 |                 100.0 |               114.486496 |              109.916969 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.257469 |       0.0 | 0.133625 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044483 | 0.012356 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 69.658193 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044483 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012356 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                         0.0 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002372 |                      0.0 |                  0.002372 |                  18.901295 |                  3.445356 |                     0.0 |                      0.0 |    -0.141184 |             0.0 |              0.0 |         15.455939 |             56.493451 |                 100.0 |               114.680153 |                89.03222 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102212 |       0.0 | 0.066116 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043334 | 0.012037 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 69.592077 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043334 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012037 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                         0.0 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.434698 |                     0.0 |                      0.0 |    -0.143112 |             0.0 |              0.0 |         15.466597 |             74.333488 |                 100.0 |               114.776034 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.046259 |       0.0 | 0.031824 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042215 | 0.011726 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 69.560253 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011726 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                         0.0 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.427556 |                     0.0 |                      0.0 |    -0.142815 |             0.0 |              0.0 |         15.473739 |             64.197103 |                 100.0 |               114.822199 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.042514 |       0.0 | 0.027773 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041124 | 0.011423 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 69.532479 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041124 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011423 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                         0.0 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.418333 |                     0.0 |                      0.0 |    -0.142431 |             0.0 |              0.0 |         15.482962 |             52.308751 |                 100.0 |               114.862497 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035583 |       0.0 | 0.021518 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040062 | 0.011128 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.445391 | 69.510962 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011128 |

.. _lland_v4_acker_heavy_rain_hourly:

acre (heavy rain)
-----------------

.. integration-test::

    >>> inputs.nied.series = 20.0/24.0
    >>> test("lland_v4_acker_heavy_rain_hourly",
    ...      axis1=(inputs.nied, fluxes.qah),
    ...      axis2=states.bowa,
    ...      use_conditions=conditions_acker_heavy_rain)
    |                date |     nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |     nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |     wada |      qdb |     qib1 | qib2 |      qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |       bowa |    qdgz1 | qdgz2 |    qigz1 | qigz2 |     qbgz |    qdga1 | qdga2 |    qiga1 | qiga2 |     qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.833333 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 | 0.783333 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                         0.0 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.901196 |                     0.0 |                      0.0 |     -0.03755 |             0.0 |              0.0 |           7.08166 |            176.542034 |                 100.0 |                70.327061 |               87.719298 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002965 | -0.002965 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.007403 |      0.0 |  0.0 |      0.0 |  0.0 | 0.007403 |  0.80139 | 0.222608 | 0.702965 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.308721 |   102.5625 | 0.007403 |   0.0 |      0.0 |   0.0 |      0.0 |  0.80139 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.222608 |
    | 1997-08-03 01:00:00 | 0.833333 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 | 0.783333 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                         0.0 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.901844 |                     0.0 |                      0.0 |    -0.037577 |             0.0 |              0.0 |          7.081012 |            176.542034 |                 100.0 |                70.229321 |               87.719298 | 0.786298 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002685 | -0.002685 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.786298 | 0.070172 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070172 | 0.781695 | 0.217137 | 0.702685 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 103.278626 | 0.070172 |   0.0 |      0.0 |   0.0 |      0.0 | 0.781695 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.217137 |
    | 1997-08-03 02:00:00 | 0.833333 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 | 0.783333 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                         0.0 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.902355 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |            7.0805 |            176.542034 |                 100.0 |                  69.3082 |               87.719298 | 0.786019 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004107 | -0.004107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.786019 | 0.070722 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070722 | 0.763323 | 0.212034 | 0.704107 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 103.993923 | 0.070722 |   0.0 |      0.0 |   0.0 |      0.0 | 0.763323 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.212034 |
    | 1997-08-03 03:00:00 | 0.833333 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 | 0.783333 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                         0.0 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                      0.0 |                       0.0 |                   7.982856 |                  0.902364 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          7.080492 |            176.542034 |                 100.0 |                68.389393 |               87.719298 | 0.787441 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005232 | -0.005232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.787441 | 0.071427 |      0.0 |  0.0 |      0.0 |  0.0 | 0.071427 | 0.745441 | 0.207067 | 0.705232 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 104.709937 | 0.071427 |   0.0 |      0.0 |   0.0 |      0.0 | 0.745441 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.207067 |
    | 1997-08-03 04:00:00 | 0.833333 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 | 0.783333 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                         0.0 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.005318 |                      0.0 |                  0.005318 |                   7.982856 |                  0.902341 |                     0.0 |                      0.0 |     -0.03228 |             0.0 |              0.0 |          7.080514 |            176.542034 |                 100.0 |                67.470886 |                81.46212 | 0.788565 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004819 | -0.004819 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.788565 | 0.072109 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072109 |  0.72804 | 0.202233 | 0.704819 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 105.426394 | 0.072109 |   0.0 |      0.0 |   0.0 |      0.0 |  0.72804 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.202233 |
    | 1997-08-03 05:00:00 | 0.833333 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 | 0.783333 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                         0.0 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.060007 |                      0.0 |                  0.060007 |                   7.982856 |                  0.901482 |                     0.0 |                      0.0 |     0.022445 |             0.0 |              0.0 |          7.081373 |            235.389378 |                 100.0 |                66.553006 |               73.619304 | 0.788152 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005073 |  0.005073 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.788152 | 0.072652 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072652 | 0.711104 | 0.197529 | 0.694927 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 106.141894 | 0.072652 |   0.0 |      0.0 |   0.0 |      0.0 | 0.711104 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.197529 |
    | 1997-08-03 06:00:00 | 0.833333 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 | 0.783333 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                         0.0 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.156653 |                      0.0 |                  0.156653 |                   7.982856 |                   0.89977 |                     0.0 |                      0.0 |     0.119163 |             0.0 |              0.0 |          7.083086 |            156.926252 |                 100.0 |                65.637518 |               72.818235 |  0.77826 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026292 |  0.026292 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.77826 | 0.072311 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072311 | 0.694607 | 0.192946 | 0.673708 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 106.847843 | 0.072311 |   0.0 |      0.0 |   0.0 |      0.0 | 0.694607 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.192946 |
    | 1997-08-03 07:00:00 | 0.833333 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 | 0.783333 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                         0.0 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.299135 |                      0.0 |                  0.299135 |                   7.982856 |                  0.896826 |                     0.0 |                      0.0 |     0.261767 |             0.0 |              0.0 |          7.086029 |            156.926252 |                 100.0 |                64.735367 |               72.024255 | 0.757042 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.057456 |  0.057456 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.757042 | 0.070885 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070885 | 0.678514 | 0.188476 | 0.642544 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.365978 |    107.534 | 0.070885 |   0.0 |      0.0 |   0.0 |      0.0 | 0.678514 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.188476 |
    | 1997-08-03 08:00:00 | 0.833333 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 | 0.783333 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                         0.0 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.467717 |                      0.0 |                  0.467717 |                   7.982856 |                  0.891371 |                     0.0 |                      0.0 |     0.430576 |             0.0 |              0.0 |          7.091484 |            156.926252 |                 100.0 |                63.859549 |               71.249063 | 0.725878 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096099 |  0.096099 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.725878 | 0.068473 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068473 | 0.662788 | 0.184108 | 0.603901 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 108.191405 | 0.068473 |   0.0 |      0.0 |   0.0 |      0.0 | 0.662788 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.184108 |
    | 1997-08-03 09:00:00 | 0.833333 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 | 0.783333 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                         0.0 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.852983 |                      0.0 |                  0.852983 |                    8.15863 |                  0.922604 |                     0.0 |                      0.0 |     0.814541 |             0.0 |              0.0 |          7.236026 |            108.641252 |                 100.0 |                63.021373 |               70.503105 | 0.687235 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222763 |  0.222763 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.687235 | 0.065285 |      0.0 |  0.0 |      0.0 |  0.0 | 0.065285 | 0.647395 | 0.179832 | 0.477237 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 108.813354 | 0.065285 |   0.0 |      0.0 |   0.0 |      0.0 | 0.647395 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.179832 |
    | 1997-08-03 10:00:00 | 0.833333 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 | 0.783333 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                         0.0 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.372332 |                      0.0 |                  1.372332 |                   8.334404 |                  0.955299 |                     0.0 |                      0.0 |     1.332528 |             0.0 |              0.0 |          7.379106 |             94.155751 |                 100.0 |                62.229232 |               69.794418 |  0.56057 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.389923 |  0.389923 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.56057 | 0.053582 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053582 | 0.632206 | 0.175613 | 0.310077 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 109.320342 | 0.053582 |   0.0 |      0.0 |   0.0 |      0.0 | 0.632206 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.175613 |
    | 1997-08-03 11:00:00 | 0.833333 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 | 0.783333 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                         0.0 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.682691 |                      0.0 |                  1.682691 |                   8.773841 |                  1.039981 |                     0.0 |                      0.0 |     1.639359 |             0.0 |              0.0 |          7.733859 |            117.694689 |                 100.0 |                61.584098 |               69.214576 |  0.39341 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.46628 |   0.46628 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.39341 | 0.037782 |      0.0 |  0.0 |      0.0 |  0.0 | 0.037782 | 0.617055 | 0.171404 |  0.23372 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 109.675971 | 0.037782 |   0.0 |      0.0 |   0.0 |      0.0 | 0.617055 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.171404 |
    | 1997-08-03 12:00:00 | 0.833333 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 | 0.783333 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                         0.0 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.714225 |                      0.0 |                  1.714225 |                   9.301164 |                  1.142661 |                     0.0 |                      0.0 |     1.666614 |             0.0 |              0.0 |          8.158503 |            108.641252 |                 100.0 |                61.131877 |               68.806682 | 0.317053 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.491066 |  0.491066 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.317053 | 0.030555 |      0.0 |  0.0 |      0.0 |  0.0 | 0.030555 | 0.601998 | 0.167222 | 0.208934 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 109.962469 | 0.030555 |   0.0 |      0.0 |   0.0 |      0.0 | 0.601998 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.167222 |
    | 1997-08-03 13:00:00 | 0.833333 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 | 0.783333 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                         0.0 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.357421 |                      0.0 |                  1.357421 |                   9.652713 |                  1.211985 |                     0.0 |                      0.0 |     1.306922 |             0.0 |              0.0 |          8.440727 |             94.155751 |                 100.0 |                60.767747 |               68.477376 | 0.292267 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424802 |  0.424802 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.292267 | 0.028251 |      0.0 |  0.0 |      0.0 |  0.0 | 0.028251 | 0.587207 | 0.163113 | 0.275198 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 110.226486 | 0.028251 |   0.0 |      0.0 |   0.0 |      0.0 | 0.587207 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.163113 |
    | 1997-08-03 14:00:00 | 0.833333 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 | 0.783333 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                         0.0 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.147809 |                      0.0 |                  1.147809 |                   9.828487 |                   1.24753 |                     0.0 |                      0.0 |     1.095829 |             0.0 |              0.0 |          8.580957 |             74.333488 |                 100.0 |                60.432333 |               68.173354 | 0.358532 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.415458 |  0.415458 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.358532 | 0.034766 |      0.0 |  0.0 |      0.0 |  0.0 | 0.034766 | 0.572854 | 0.159126 | 0.284542 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 110.550251 | 0.034766 |   0.0 |      0.0 |   0.0 |      0.0 | 0.572854 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.159126 |
    | 1997-08-03 15:00:00 | 0.833333 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 | 0.783333 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                         0.0 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |               1.06126 |                      0.0 |                   1.06126 |                  10.180036 |                  1.316893 |                     0.0 |                      0.0 |     1.006389 |             0.0 |              0.0 |          8.863143 |             74.333488 |                 100.0 |                  60.0212 |               67.799797 | 0.367875 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.393508 |  0.393508 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.367875 | 0.035799 |      0.0 |  0.0 |      0.0 |  0.0 | 0.035799 | 0.558968 | 0.155269 | 0.306492 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 110.882327 | 0.035799 |   0.0 |      0.0 |   0.0 |      0.0 | 0.558968 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.155269 |
    | 1997-08-03 16:00:00 | 0.833333 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 | 0.783333 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                         0.0 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.761967 |                      0.0 |                  0.761967 |                  10.443698 |                  1.370145 |                     0.0 |                      0.0 |     0.704878 |             0.0 |              0.0 |          9.073553 |             61.405925 |                 100.0 |                59.599725 |               67.415809 | 0.389826 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.342513 |  0.342513 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.389826 | 0.038075 |      0.0 |  0.0 |      0.0 |  0.0 | 0.038075 | 0.545483 | 0.151523 | 0.357487 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 111.234078 | 0.038075 |   0.0 |      0.0 |   0.0 |      0.0 | 0.545483 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.151523 |
    | 1997-08-03 17:00:00 | 0.833333 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 | 0.783333 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                         0.0 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.375211 |                      0.0 |                  0.375211 |                   10.70736 |                  1.421011 |                     0.0 |                      0.0 |     0.316002 |             0.0 |              0.0 |          9.286348 |             58.847345 |                 100.0 |                59.153511 |               67.008135 | 0.440821 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.236583 |  0.236583 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.440821 |  0.04323 |      0.0 |  0.0 |      0.0 |  0.0 |  0.04323 | 0.532443 | 0.147901 | 0.463417 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 111.631669 |  0.04323 |   0.0 |      0.0 |   0.0 |      0.0 | 0.532443 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.147901 |
    | 1997-08-03 18:00:00 | 0.833333 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 | 0.783333 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                         0.0 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.139757 |                      0.0 |                  0.139757 |                  10.795247 |                  1.433696 |                     0.0 |                      0.0 |     0.080019 |             0.0 |              0.0 |           9.36155 |             56.493451 |                 100.0 |                58.649429 |               66.546166 |  0.54675 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.156446 |  0.156446 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.54675 | 0.053875 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053875 | 0.519945 | 0.144429 | 0.543554 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.280179 | 112.124544 | 0.053875 |   0.0 |      0.0 |   0.0 |      0.0 | 0.519945 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.144429 |
    | 1997-08-03 19:00:00 | 0.833333 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 | 0.783333 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                         0.0 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.037299 |                      0.0 |                  0.037299 |                  10.795247 |                  1.423756 |                     0.0 |                      0.0 |    -0.022024 |             0.0 |              0.0 |          9.371491 |             56.493451 |                 100.0 |                58.024955 |               65.971756 | 0.626887 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.09619 |   0.09619 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.626887 | 0.062121 |      0.0 |  0.0 |      0.0 |  0.0 | 0.062121 | 0.508013 | 0.141115 |  0.60381 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.31655 |  112.68931 | 0.062121 |   0.0 |      0.0 |   0.0 |      0.0 | 0.508013 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.141115 |
    | 1997-08-03 20:00:00 | 0.833333 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 | 0.783333 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                         0.0 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000509 |                      0.0 |                  0.000509 |                  10.795247 |                  1.412078 |                     0.0 |                      0.0 |    -0.058328 |             0.0 |              0.0 |          9.383169 |             64.197103 |                 100.0 |                 57.30995 |               83.797685 | 0.687143 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04938 |   0.04938 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.687143 | 0.068523 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068523 | 0.496578 | 0.137938 |  0.65062 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 113.307931 | 0.068523 |   0.0 |      0.0 |   0.0 |      0.0 | 0.496578 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.137938 |
    | 1997-08-03 21:00:00 | 0.833333 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 | 0.783333 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                         0.0 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                  1.401564 |                     0.0 |                      0.0 |    -0.058399 |             0.0 |              0.0 |          9.393683 |             83.078604 |                 100.0 |                56.527431 |               87.719298 | 0.733954 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019258 |  0.019258 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.733954 | 0.073689 |      0.0 |  0.0 |      0.0 |  0.0 | 0.073689 | 0.485588 | 0.134886 | 0.680742 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 113.968195 | 0.073689 |   0.0 |      0.0 |   0.0 |      0.0 | 0.485588 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.134886 |
    | 1997-08-03 22:00:00 | 0.833333 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 | 0.783333 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                         0.0 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                  1.391158 |                     0.0 |                      0.0 |    -0.057965 |             0.0 |              0.0 |          9.404089 |             83.078604 |                 100.0 |                 55.69299 |               87.719298 | 0.764075 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010298 |  0.010298 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.764075 | 0.077261 |      0.0 |  0.0 |      0.0 |  0.0 | 0.077261 | 0.474995 | 0.131943 | 0.689702 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.248458 |  114.65501 | 0.077261 |   0.0 |      0.0 |   0.0 |      0.0 | 0.474995 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.131943 |
    | 1997-08-03 23:00:00 | 0.833333 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 | 0.783333 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                         0.0 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                      0.0 |                       0.0 |                  10.795247 |                   1.38284 |                     0.0 |                      0.0 |    -0.057618 |             0.0 |              0.0 |          9.412407 |             61.405925 |                 100.0 |                 54.82581 |               87.719298 | 0.773036 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014661 |  0.014661 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.773036 | 0.078736 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078736 |  0.46474 | 0.129094 | 0.685339 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 115.349309 | 0.078736 |   0.0 |      0.0 |   0.0 |      0.0 |  0.46474 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.129094 |
    | 1997-08-04 00:00:00 | 0.833333 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 | 0.783333 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                         0.0 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                   1.39654 |                     0.0 |                      0.0 |    -0.058189 |             0.0 |              0.0 |          9.343179 |             61.405925 |                 100.0 |                53.950005 |               87.719298 | 0.768673 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023989 |  0.023989 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.768673 | 0.078861 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078861 | 0.454771 | 0.126325 | 0.676011 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 116.039121 | 0.078861 |   0.0 |      0.0 |   0.0 |      0.0 | 0.454771 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.126325 |
    | 1997-08-04 01:00:00 | 0.833333 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 | 0.783333 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                         0.0 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                  1.406633 |                     0.0 |                      0.0 |     -0.05861 |             0.0 |              0.0 |          9.333085 |             64.197103 |                 100.0 |                53.080668 |               87.719298 | 0.759344 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018182 |  0.018182 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.759344 | 0.078462 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078462 | 0.445055 | 0.123627 | 0.681818 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 116.720003 | 0.078462 |   0.0 |      0.0 |   0.0 |      0.0 | 0.445055 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.123627 |
    | 1997-08-04 02:00:00 | 0.833333 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 | 0.783333 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                         0.0 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                  1.418069 |                     0.0 |                      0.0 |    -0.059086 |             0.0 |              0.0 |           9.32165 |             61.405925 |                 100.0 |                52.223358 |               87.719298 | 0.765151 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02055 |   0.02055 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.765151 | 0.079625 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079625 | 0.435601 |    0.121 |  0.67945 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.998611 | 117.405529 | 0.079625 |   0.0 |      0.0 |   0.0 |      0.0 | 0.435601 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 |    0.121 |
    | 1997-08-04 03:00:00 | 0.833333 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 | 0.783333 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                         0.0 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                      0.0 |                       0.0 |                  10.739718 |                   1.43017 |                     0.0 |                      0.0 |     -0.05959 |             0.0 |              0.0 |          9.309548 |             70.616813 |                 100.0 |                 51.36096 |               87.719298 | 0.762784 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01363 |   0.01363 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.762784 | 0.079941 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079941 |  0.42641 | 0.118447 |  0.68637 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 118.088372 | 0.079941 |   0.0 |      0.0 |   0.0 |      0.0 |  0.42641 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.118447 |
    | 1997-08-04 04:00:00 | 0.833333 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 | 0.783333 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                         0.0 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.004352 |                      0.0 |                  0.004352 |                  10.739718 |                  1.440702 |                     0.0 |                      0.0 |    -0.055678 |             0.0 |              0.0 |          9.299016 |             61.405925 |                 100.0 |                50.502684 |               73.397763 | 0.769703 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007255 |  0.007255 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.769703 | 0.081237 |      0.0 |  0.0 |      0.0 |  0.0 | 0.081237 | 0.417477 | 0.115966 | 0.692745 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 118.776838 | 0.081237 |   0.0 |      0.0 |   0.0 |      0.0 | 0.417477 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.115966 |
    | 1997-08-04 05:00:00 | 0.833333 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 | 0.783333 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                         0.0 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.054586 |                      0.0 |                  0.054586 |                  10.739718 |                  1.451305 |                     0.0 |                      0.0 |    -0.005885 |             0.0 |              0.0 |          9.288414 |             56.493451 |                 100.0 |                49.638078 |               58.024051 | 0.776079 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018162 |  0.018162 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.776079 | 0.082492 |      0.0 |  0.0 |      0.0 |  0.0 | 0.082492 | 0.408808 | 0.113558 | 0.681838 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 119.470425 | 0.082492 |   0.0 |      0.0 |   0.0 |      0.0 | 0.408808 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.113558 |
    | 1997-08-04 06:00:00 | 0.833333 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 | 0.783333 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                         0.0 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.172329 |                      0.0 |                  0.172329 |                   10.82738 |                  1.478175 |                     0.0 |                      0.0 |     0.110738 |             0.0 |              0.0 |          9.349205 |             58.847345 |                 100.0 |                 48.85611 |               57.260179 | 0.765171 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04093 |   0.04093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.765171 | 0.081905 | 0.003897 |  0.0 | 0.000003 |  0.0 | 0.081905 | 0.400397 | 0.111221 |  0.65907 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.712161 |  120.14979 | 0.081905 |   0.0 | 0.003897 |   0.0 | 0.000003 | 0.400372 |   0.0 | 0.000025 |   0.0 |      0.0 |   0.0 | 0.111221 |
    | 1997-08-04 07:00:00 | 0.833333 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 | 0.783333 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                         0.0 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |              0.781884 |                      0.0 |                  0.781884 |                  11.616331 |                   1.64464 |                     0.0 |                      0.0 |     0.713358 |             0.0 |              0.0 |          9.971692 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.742403 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.18972 |   0.18972 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.742403 |  0.08001 | 0.003919 |  0.0 | 0.000031 |  0.0 |  0.08001 | 0.392197 | 0.108943 |  0.51028 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 120.808233 |  0.08001 |   0.0 | 0.003919 |   0.0 | 0.000031 | 0.392121 |   0.0 | 0.000076 |   0.0 |      0.0 |   0.0 | 0.108943 |
    | 1997-08-04 08:00:00 | 0.833333 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 | 0.783333 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                         0.0 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.300216 |                      0.0 |                  1.300216 |                  12.492945 |                   1.83472 |                     0.0 |                      0.0 |      1.22377 |             0.0 |              0.0 |         10.658225 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.593614 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.343734 |  0.343734 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.593614 | 0.064359 | 0.003941 |  0.0 | 0.000059 |  0.0 | 0.064359 | 0.383982 | 0.106662 | 0.356266 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 121.333488 | 0.064359 |   0.0 | 0.003941 |   0.0 | 0.000059 | 0.383856 |   0.0 | 0.000126 |   0.0 |      0.0 |   0.0 | 0.106662 |
    | 1997-08-04 09:00:00 | 0.833333 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 | 0.783333 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                         0.0 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.679922 |                      0.0 |                  1.679922 |                  13.194235 |                  1.989515 |                     0.0 |                      0.0 |     1.597026 |             0.0 |              0.0 |          11.20472 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.439599 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.494183 |  0.494183 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.439599 | 0.047881 | 0.003958 |  0.0 | 0.000081 |  0.0 | 0.047881 | 0.375565 | 0.104324 | 0.205817 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 121.721168 | 0.047881 |   0.0 | 0.003958 |   0.0 | 0.000081 | 0.375389 |   0.0 | 0.000176 |   0.0 |      0.0 |   0.0 | 0.104324 |
    | 1997-08-04 10:00:00 | 0.833333 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 | 0.783333 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                         0.0 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              1.761455 |                      0.0 |                  1.761455 |                   13.54488 |                   2.06621 |                     0.0 |                      0.0 |     1.675363 |             0.0 |              0.0 |          11.47867 |             45.559235 |                 100.0 |                 48.85611 |               57.260179 | 0.289151 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.550449 |  0.550449 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.289151 | 0.031596 | 0.003971 |  0.0 | 0.000097 |  0.0 | 0.031596 | 0.366943 | 0.101929 | 0.149551 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.749357 | 121.974655 | 0.031596 |   0.0 | 0.003971 |   0.0 | 0.000097 | 0.366718 |   0.0 | 0.000225 |   0.0 |      0.0 |   0.0 | 0.101929 |
    | 1997-08-04 11:00:00 | 0.833333 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 | 0.783333 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                         0.0 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |               1.64232 |                      0.0 |                   1.64232 |                  13.632542 |                  2.087634 |                     0.0 |                      0.0 |     1.555335 |             0.0 |              0.0 |         11.544908 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 | 0.232884 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.548384 |  0.548384 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.232884 | 0.025506 | 0.003979 |  0.0 | 0.000107 |  0.0 | 0.025506 | 0.358256 | 0.099516 | 0.151616 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 122.177947 | 0.025506 |   0.0 | 0.003979 |   0.0 | 0.000107 | 0.357982 |   0.0 | 0.000274 |   0.0 |      0.0 |   0.0 | 0.099516 |
    | 1997-08-04 12:00:00 | 0.833333 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 | 0.783333 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                         0.0 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.697727 |                      0.0 |                  1.697727 |                  13.720203 |                  2.106001 |                     0.0 |                      0.0 |     1.609977 |             0.0 |              0.0 |         11.614202 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 |  0.23495 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.597559 |  0.597559 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.23495 | 0.025785 | 0.003986 |  0.0 | 0.000116 |  0.0 | 0.025785 |  0.34972 | 0.097144 | 0.102441 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 122.383011 | 0.025785 |   0.0 | 0.003986 |   0.0 | 0.000116 | 0.349398 |   0.0 | 0.000322 |   0.0 |      0.0 |   0.0 | 0.097144 |
    | 1997-08-04 13:00:00 | 0.833333 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 | 0.783333 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                         0.0 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |              1.640072 |                      0.0 |                  1.640072 |                  13.983187 |                  2.166369 |                     0.0 |                      0.0 |     1.549806 |             0.0 |              0.0 |         11.816818 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.185775 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.621801 |  0.621801 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.185775 | 0.020426 | 0.003992 |  0.0 | 0.000124 |  0.0 | 0.020426 | 0.341339 | 0.094816 | 0.078199 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 122.544243 | 0.020426 |   0.0 | 0.003992 |   0.0 | 0.000124 | 0.340969 |   0.0 |  0.00037 |   0.0 |      0.0 |   0.0 | 0.094816 |
    | 1997-08-04 14:00:00 | 0.833333 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 | 0.783333 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                         0.0 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |              1.678676 |                      0.0 |                  1.678676 |                  14.421494 |                  2.265756 |                     0.0 |                      0.0 |      1.58427 |             0.0 |              0.0 |         12.155738 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.161533 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.589183 |  0.589183 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.161533 | 0.017787 | 0.003998 |  0.0 | 0.000131 |  0.0 | 0.017787 | 0.333072 |  0.09252 | 0.110817 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.063657 |  122.68386 | 0.017787 |   0.0 | 0.003998 |   0.0 | 0.000131 | 0.332655 |   0.0 | 0.000417 |   0.0 |      0.0 |   0.0 |  0.09252 |
    | 1997-08-04 15:00:00 | 0.833333 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 | 0.783333 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                         0.0 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.414805 |                      0.0 |                  1.414805 |                  14.772139 |                  2.348437 |                     0.0 |                      0.0 |     1.316953 |             0.0 |              0.0 |         12.423701 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 |  0.19415 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.551424 |  0.551424 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.19415 | 0.021412 | 0.004002 |  0.0 | 0.000137 |  0.0 | 0.021412 | 0.325032 | 0.090287 | 0.148576 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.139377 |  122.85246 | 0.021412 |   0.0 | 0.004002 |   0.0 | 0.000137 | 0.324569 |   0.0 | 0.000463 |   0.0 |      0.0 |   0.0 | 0.090287 |
    | 1997-08-04 16:00:00 | 0.833333 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 | 0.783333 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                         0.0 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.033761 |                      0.0 |                  1.033761 |                  15.122784 |                  2.428674 |                     0.0 |                      0.0 |     0.932566 |             0.0 |              0.0 |          12.69411 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.231909 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.505761 |  0.505761 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.231909 | 0.025623 | 0.004008 |  0.0 | 0.000144 |  0.0 | 0.025623 | 0.317302 | 0.088139 | 0.194239 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 123.054595 | 0.025623 |   0.0 | 0.004008 |   0.0 | 0.000144 | 0.316793 |   0.0 | 0.000509 |   0.0 |      0.0 |   0.0 | 0.088139 |
    | 1997-08-04 17:00:00 | 0.833333 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 | 0.783333 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                         0.0 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.690032 |                      0.0 |                  0.690032 |                  15.736413 |                  2.564427 |                     0.0 |                      0.0 |     0.583181 |             0.0 |              0.0 |         13.171986 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.277572 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.444603 |  0.444603 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.277572 | 0.030737 | 0.004014 |  0.0 | 0.000152 |  0.0 | 0.030737 | 0.309893 | 0.086081 | 0.255397 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.282872 | 123.297264 | 0.030737 |   0.0 | 0.004014 |   0.0 | 0.000152 | 0.309338 |   0.0 | 0.000555 |   0.0 |      0.0 |   0.0 | 0.086081 |
    | 1997-08-04 18:00:00 | 0.833333 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 | 0.783333 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                         0.0 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.351786 |                      0.0 |                  0.351786 |                  16.525365 |                  2.744116 |                     0.0 |                      0.0 |     0.237448 |             0.0 |              0.0 |         13.781249 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 |  0.33873 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.32478 |   0.32478 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.33873 |  0.03761 | 0.004022 |  0.0 | 0.000162 |  0.0 |  0.03761 |  0.30283 |  0.08412 |  0.37522 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.323368 |   123.5942 |  0.03761 |   0.0 | 0.004022 |   0.0 | 0.000162 |  0.30223 |   0.0 |   0.0006 |   0.0 |      0.0 |   0.0 |  0.08412 |
    | 1997-08-04 19:00:00 | 0.833333 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 | 0.783333 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                         0.0 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.115111 |                      0.0 |                  0.115111 |                  17.401978 |                  2.950104 |                     0.0 |                      0.0 |     -0.00781 |             0.0 |              0.0 |         14.451875 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.458554 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.183091 |  0.183091 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.458554 | 0.051092 | 0.004032 |  0.0 | 0.000175 |  0.0 | 0.051092 | 0.296214 | 0.082282 | 0.516909 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 123.997455 | 0.051092 |   0.0 | 0.004032 |   0.0 | 0.000175 |  0.29557 |   0.0 | 0.000644 |   0.0 |      0.0 |   0.0 | 0.082282 |
    | 1997-08-04 20:00:00 | 0.833333 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 | 0.783333 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                         0.0 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002224 |                      0.0 |                  0.002224 |                  17.664962 |                  3.034376 |                     0.0 |                      0.0 |    -0.124208 |             0.0 |              0.0 |         14.630586 |             44.135508 |                 100.0 |                 48.85611 |               83.091125 | 0.600242 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.100631 |  0.100631 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.600242 | 0.067188 | 0.004045 |  0.0 | 0.000192 |  0.0 | 0.067188 | 0.290152 | 0.080598 | 0.599369 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 124.526272 | 0.067188 |   0.0 | 0.004045 |   0.0 | 0.000192 | 0.289464 |   0.0 | 0.000689 |   0.0 |      0.0 |   0.0 | 0.080598 |
    | 1997-08-04 21:00:00 | 0.833333 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 | 0.783333 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                         0.0 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.060678 |                     0.0 |                      0.0 |    -0.127528 |             0.0 |              0.0 |         14.604285 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.682702 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029981 |  0.029981 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.682702 | 0.076847 | 0.004062 |  0.0 | 0.000214 |  0.0 | 0.076847 |  0.28458 |  0.07905 | 0.670019 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 125.127851 | 0.076847 |   0.0 | 0.004062 |   0.0 | 0.000214 | 0.283847 |   0.0 | 0.000732 |   0.0 |      0.0 |   0.0 |  0.07905 |
    | 1997-08-04 22:00:00 | 0.833333 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 | 0.783333 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                         0.0 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.084324 |                     0.0 |                      0.0 |    -0.128514 |             0.0 |              0.0 |         14.580638 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.753353 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012064 |  0.012064 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.753353 | 0.085331 | 0.004082 |  0.0 | 0.000239 |  0.0 | 0.085331 | 0.279386 | 0.077607 | 0.687936 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.147714 | 125.791553 | 0.085331 |   0.0 | 0.004082 |   0.0 | 0.000239 |  0.27861 |   0.0 | 0.000776 |   0.0 |      0.0 |   0.0 | 0.077607 |
    | 1997-08-04 23:00:00 | 0.833333 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 | 0.783333 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                         0.0 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                      0.0 |                       0.0 |                  17.664962 |                  3.103639 |                     0.0 |                      0.0 |    -0.129318 |             0.0 |              0.0 |         14.561323 |             67.254108 |                 100.0 |                 48.85611 |               87.719298 | 0.771269 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000111 | -0.000111 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.771269 | 0.087938 | 0.004103 |  0.0 | 0.000266 |  0.0 | 0.087938 | 0.274471 | 0.076242 | 0.700111 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.051062 | 126.470515 | 0.087938 |   0.0 | 0.004103 |   0.0 | 0.000266 | 0.273652 |   0.0 | 0.000819 |   0.0 |      0.0 |   0.0 | 0.076242 |
    | 1997-08-05 00:00:00 | 0.833333 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 | 0.783333 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                         0.0 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.124611 |                     0.0 |                      0.0 |    -0.130192 |             0.0 |              0.0 |         14.465269 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.783444 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009107 | -0.009107 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.783444 | 0.089926 | 0.004126 |  0.0 | 0.000295 |  0.0 | 0.089926 | 0.269742 | 0.074928 | 0.709107 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.946249 | 127.159612 | 0.089926 |   0.0 | 0.004126 |   0.0 | 0.000295 |  0.26888 |   0.0 | 0.000862 |   0.0 |      0.0 |   0.0 | 0.074928 |
    | 1997-08-05 01:00:00 | 0.833333 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 | 0.783333 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                         0.0 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.135507 |                     0.0 |                      0.0 |    -0.130646 |             0.0 |              0.0 |         14.454373 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.792441 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.013348 | -0.013348 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.792441 | 0.091575 | 0.004148 |  0.0 | 0.000323 |  0.0 | 0.091575 | 0.265184 | 0.073662 | 0.713348 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.836022 | 127.856006 | 0.091575 |   0.0 | 0.004148 |   0.0 | 0.000323 | 0.264279 |   0.0 | 0.000904 |   0.0 |      0.0 |   0.0 | 0.073662 |
    | 1997-08-05 02:00:00 | 0.833333 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 | 0.783333 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                         0.0 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                  3.144898 |                     0.0 |                      0.0 |    -0.131037 |             0.0 |              0.0 |         14.444982 |             88.271017 |                 100.0 |                 48.85611 |               87.719298 | 0.796681 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.017799 | -0.017799 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.796681 | 0.092692 | 0.004171 |  0.0 | 0.000352 |  0.0 | 0.092692 |  0.26078 | 0.072439 | 0.717799 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.722932 | 128.555473 | 0.092692 |   0.0 | 0.004171 |   0.0 | 0.000352 | 0.259832 |   0.0 | 0.000947 |   0.0 |      0.0 |   0.0 | 0.072439 |
    | 1997-08-05 03:00:00 | 0.833333 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 | 0.783333 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                         0.0 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                      0.0 |                       0.0 |                   17.58988 |                   3.15411 |                     0.0 |                      0.0 |    -0.131421 |             0.0 |              0.0 |          14.43577 |            117.694689 |                 100.0 |                 48.85611 |               87.719298 | 0.801132 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.020794 | -0.020794 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.801132 | 0.093844 | 0.004194 |  0.0 | 0.000381 |  0.0 | 0.093844 | 0.256519 | 0.071255 | 0.720794 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.611504 | 129.258186 | 0.093844 |   0.0 | 0.004194 |   0.0 | 0.000381 |  0.25553 |   0.0 | 0.000989 |   0.0 | 0.000001 |   0.0 | 0.071255 |
    | 1997-08-05 04:00:00 | 0.833333 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 | 0.783333 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                         0.0 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004056 |                      0.0 |                  0.004056 |                   17.58988 |                  3.166293 |                     0.0 |                      0.0 |    -0.127873 |             0.0 |              0.0 |         14.423587 |             74.333488 |                 100.0 |                 48.85611 |               73.423526 | 0.804127 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019617 | -0.019617 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.804127 | 0.094836 | 0.004217 |  0.0 | 0.000411 |  0.0 | 0.094836 | 0.252397 |  0.07011 | 0.719617 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |  129.96285 | 0.094836 |   0.0 | 0.004217 |   0.0 | 0.000411 | 0.251366 |   0.0 |  0.00103 |   0.0 | 0.000001 |   0.0 |  0.07011 |
    | 1997-08-05 05:00:00 | 0.833333 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 | 0.783333 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                         0.0 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.057827 |                      0.0 |                  0.057827 |                   17.58988 |                  3.179808 |                     0.0 |                      0.0 |    -0.074665 |             0.0 |              0.0 |         14.410072 |            108.641252 |                 100.0 |                 48.85611 |               57.260179 |  0.80295 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   -0.0093 |   -0.0093 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.80295 | 0.095339 |  0.00424 |  0.0 |  0.00044 |  0.0 | 0.095339 | 0.248402 | 0.069001 |   0.7093 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.397486 | 130.665781 | 0.095339 |   0.0 |  0.00424 |   0.0 |  0.00044 |  0.24733 |   0.0 | 0.001072 |   0.0 | 0.000001 |   0.0 | 0.069001 |
    | 1997-08-05 06:00:00 | 0.833333 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 | 0.783333 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                         0.0 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.344417 |                      0.0 |                  0.344417 |                  18.027018 |                  3.288825 |                     0.0 |                      0.0 |     0.207383 |             0.0 |              0.0 |         14.738193 |             88.271017 |                 100.0 |                 48.85611 |               57.260179 | 0.792634 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.05232 |   0.05232 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.792634 | 0.094745 | 0.004262 |  0.0 | 0.000469 |  0.0 | 0.094745 |  0.24451 | 0.067919 |  0.64768 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 131.358937 | 0.094745 |   0.0 | 0.004262 |   0.0 | 0.000469 | 0.243396 |   0.0 | 0.001113 |   0.0 | 0.000001 |   0.0 | 0.067919 |
    | 1997-08-05 07:00:00 | 0.833333 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 | 0.783333 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                         0.0 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.837093 |                      0.0 |                  0.837093 |                  18.027018 |                  3.287363 |                     0.0 |                      0.0 |     0.700119 |             0.0 |              0.0 |         14.739656 |             74.333488 |                 100.0 |                 48.85611 |               57.260179 | 0.731013 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.166735 |  0.166735 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.731013 | 0.087936 | 0.004285 |  0.0 | 0.000498 |  0.0 | 0.087936 | 0.240623 |  0.06684 | 0.533265 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 131.997231 | 0.087936 |   0.0 | 0.004285 |   0.0 | 0.000498 | 0.239468 |   0.0 | 0.001154 |   0.0 | 0.000001 |   0.0 |  0.06684 |
    | 1997-08-05 08:00:00 | 0.833333 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 | 0.783333 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                         0.0 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.398382 |                      0.0 |                  1.398382 |                  18.027018 |                  3.279458 |                     0.0 |                      0.0 |     1.261738 |             0.0 |              0.0 |         14.747561 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.616599 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.358814 |  0.358814 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.616599 | 0.074588 | 0.004306 |  0.0 | 0.000525 |  0.0 | 0.074588 | 0.236577 | 0.065716 | 0.341186 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 132.534411 | 0.074588 |   0.0 | 0.004306 |   0.0 | 0.000525 |  0.23538 |   0.0 | 0.001195 |   0.0 | 0.000001 |   0.0 | 0.065716 |
    | 1997-08-05 09:00:00 | 0.833333 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 | 0.783333 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                         0.0 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.722365 |                      0.0 |                  1.722365 |                  17.939591 |                  3.250753 |                     0.0 |                      0.0 |     1.586917 |             0.0 |              0.0 |         14.688838 |             35.308407 |                 100.0 |                 48.85611 |               57.260179 | 0.424519 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.541756 |  0.541756 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.424519 | 0.051574 | 0.004323 |  0.0 | 0.000547 |  0.0 | 0.051574 | 0.232165 |  0.06449 | 0.158244 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 132.902486 | 0.051574 |   0.0 | 0.004323 |   0.0 | 0.000547 | 0.230928 |   0.0 | 0.001236 |   0.0 | 0.000001 |   0.0 |  0.06449 |
    | 1997-08-05 10:00:00 | 0.833333 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 | 0.783333 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                         0.0 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.095274 |                      0.0 |                  2.095274 |                  18.027018 |                  3.266966 |                     0.0 |                      0.0 |     1.959151 |             0.0 |              0.0 |         14.760052 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.241578 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.716789 |       0.7 | 0.010363 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.241578 | 0.029429 | 0.004335 |  0.0 | 0.000563 |  0.0 | 0.029429 | 0.227285 | 0.063135 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.50337 | 133.099374 | 0.029429 |   0.0 | 0.004335 |   0.0 | 0.000563 | 0.226008 |   0.0 | 0.001276 |   0.0 | 0.000001 |   0.0 | 0.063135 |
    | 1997-08-05 11:00:00 | 0.833333 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 | 0.783333 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                         0.0 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.098911 |                      0.0 |                  2.098911 |                  18.201874 |                  3.309837 |                     0.0 |                      0.0 |     1.961001 |             0.0 |              0.0 |         14.892037 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.746035 |       0.7 | 0.029836 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 | 0.221998 | 0.061666 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 133.137795 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.220681 |   0.0 | 0.001316 |   0.0 | 0.000001 |   0.0 | 0.061666 |
    | 1997-08-05 12:00:00 | 0.833333 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 | 0.783333 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                         0.0 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.014418 |                      0.0 |                  2.014418 |                  18.289301 |                  3.330376 |                     0.0 |                      0.0 |     1.875652 |             0.0 |              0.0 |         14.958925 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.771401 |       0.7 | 0.046898 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010167 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010167 | 0.216599 | 0.060166 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 133.159147 | 0.010167 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.215243 |   0.0 | 0.001355 |   0.0 | 0.000001 |   0.0 | 0.060166 |
    | 1997-08-05 13:00:00 | 0.833333 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 | 0.783333 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                         0.0 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.101265 |                      0.0 |                  2.101265 |                  18.464157 |                  3.363175 |                     0.0 |                      0.0 |     1.961133 |             0.0 |              0.0 |         15.100981 |             32.098552 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.840981 |       0.7 | 0.091225 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010169 | 0.004344 |  0.0 | 0.000573 |  0.0 | 0.010169 | 0.211341 | 0.058706 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 133.136169 | 0.010169 |   0.0 | 0.004344 |   0.0 | 0.000573 | 0.209946 |   0.0 | 0.001394 |   0.0 | 0.000001 |   0.0 | 0.058706 |
    | 1997-08-05 14:00:00 | 0.833333 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 | 0.783333 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                         0.0 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.013503 |                      0.0 |                  2.013503 |                  18.551584 |                   3.38194 |                     0.0 |                      0.0 |     1.872589 |             0.0 |              0.0 |         15.169644 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.865172 |       0.7 | 0.105748 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010167 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010167 | 0.206219 | 0.057283 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 133.098671 | 0.010167 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.204785 |   0.0 | 0.001432 |   0.0 | 0.000001 |   0.0 | 0.057283 |
    | 1997-08-05 15:00:00 | 0.833333 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 | 0.783333 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                         0.0 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.697346 |                      0.0 |                  1.697346 |                  18.639012 |                  3.400029 |                     0.0 |                      0.0 |     1.555678 |             0.0 |              0.0 |         15.238983 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.798412 |       0.7 | 0.063779 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 |  0.20123 | 0.055897 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 133.103149 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.199758 |   0.0 |  0.00147 |   0.0 | 0.000001 |   0.0 | 0.055897 |
    | 1997-08-05 16:00:00 | 0.833333 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 | 0.783333 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                         0.0 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.231461 |                      0.0 |                  1.231461 |                   18.72644 |                   3.42173 |                     0.0 |                      0.0 |     1.088889 |             0.0 |              0.0 |         15.304709 |             30.049708 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.702711 |       0.7 | 0.001736 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 |  0.19637 | 0.054547 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 133.169669 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.194861 |   0.0 | 0.001508 |   0.0 | 0.000002 |   0.0 | 0.054547 |
    | 1997-08-05 17:00:00 | 0.833333 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 | 0.783333 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                         0.0 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |              0.748608 |                      0.0 |                  0.748608 |                   18.72644 |                  3.423912 |                     0.0 |                      0.0 |     0.605945 |             0.0 |              0.0 |         15.302528 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.55931 |   0.55931 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 |  0.01017 | 0.004344 |  0.0 | 0.000574 |  0.0 |  0.01017 | 0.191636 | 0.053232 |  0.14069 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 133.237914 |  0.01017 |   0.0 | 0.004344 |   0.0 | 0.000574 |  0.19009 |   0.0 | 0.001544 |   0.0 | 0.000002 |   0.0 | 0.053232 |
    | 1997-08-05 18:00:00 | 0.833333 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 | 0.783333 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                         0.0 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |              0.364895 |                      0.0 |                  0.364895 |                   18.72644 |                   3.42458 |                     0.0 |                      0.0 |     0.222204 |             0.0 |              0.0 |          15.30186 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.224023 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.406398 |  0.406398 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.224023 | 0.027375 | 0.004346 |  0.0 | 0.000577 |  0.0 | 0.027375 | 0.187248 | 0.052013 | 0.293602 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.438786 |  133.42964 | 0.027375 |   0.0 | 0.004346 |   0.0 | 0.000577 | 0.185665 |   0.0 | 0.001581 |   0.0 | 0.000002 |   0.0 | 0.052013 |
    | 1997-08-05 19:00:00 | 0.833333 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 | 0.783333 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                         0.0 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.110333 |                      0.0 |                  0.110333 |                   18.72644 |                  3.420162 |                     0.0 |                      0.0 |    -0.032173 |             0.0 |              0.0 |         15.306277 |             37.166744 |                 100.0 |                 48.85611 |               57.260179 | 0.376935 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.257469 |  0.257469 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.376935 | 0.046172 | 0.004353 |  0.0 | 0.000585 |  0.0 | 0.046172 | 0.183439 | 0.050955 | 0.442531 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 133.755465 | 0.046172 |   0.0 | 0.004353 |   0.0 | 0.000585 | 0.181821 |   0.0 | 0.001617 |   0.0 | 0.000002 |   0.0 | 0.050955 |
    | 1997-08-05 20:00:00 | 0.833333 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 | 0.783333 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                         0.0 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002372 |                      0.0 |                  0.002372 |                  18.901295 |                  3.445356 |                     0.0 |                      0.0 |    -0.141184 |             0.0 |              0.0 |         15.455939 |             56.493451 |                 100.0 |                 48.85611 |               84.453771 | 0.525865 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102212 |  0.102212 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.525865 | 0.064653 | 0.004363 |  0.0 | 0.000598 |  0.0 | 0.064653 | 0.180211 | 0.050059 | 0.597788 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 134.211715 | 0.064653 |   0.0 | 0.004363 |   0.0 | 0.000598 | 0.178556 |   0.0 | 0.001652 |   0.0 | 0.000002 |   0.0 | 0.050059 |
    | 1997-08-05 21:00:00 | 0.833333 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 | 0.783333 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                         0.0 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.434698 |                     0.0 |                      0.0 |    -0.143112 |             0.0 |              0.0 |         15.466597 |             74.333488 |                 100.0 |                 48.85611 |               87.719298 | 0.681121 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.046259 |  0.046259 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.681121 | 0.084155 | 0.004378 |  0.0 | 0.000617 |  0.0 | 0.084155 | 0.177557 | 0.049321 | 0.653741 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 134.803686 | 0.084155 |   0.0 | 0.004378 |   0.0 | 0.000617 | 0.175867 |   0.0 | 0.001688 |   0.0 | 0.000002 |   0.0 | 0.049321 |
    | 1997-08-05 22:00:00 | 0.833333 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 | 0.783333 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                         0.0 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.427556 |                     0.0 |                      0.0 |    -0.142815 |             0.0 |              0.0 |         15.473739 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.737074 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.042514 |  0.042514 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.737074 | 0.091595 | 0.004397 |  0.0 | 0.000642 |  0.0 | 0.091595 |  0.17532 |   0.0487 | 0.657486 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 135.444126 | 0.091595 |   0.0 | 0.004397 |   0.0 | 0.000642 | 0.173595 |   0.0 | 0.001723 |   0.0 | 0.000002 |   0.0 |   0.0487 |
    | 1997-08-05 23:00:00 | 0.833333 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 | 0.783333 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                         0.0 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                      0.0 |                       0.0 |                  18.901295 |                  3.418333 |                     0.0 |                      0.0 |    -0.142431 |             0.0 |              0.0 |         15.482962 |             52.308751 |                 100.0 |                 48.85611 |               87.719298 | 0.740819 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035583 |  0.035583 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.740819 | 0.092614 | 0.004418 |  0.0 | 0.000669 |  0.0 | 0.092614 |  0.17325 | 0.048125 | 0.664417 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.445391 | 136.087245 | 0.092614 |   0.0 | 0.004418 |   0.0 | 0.000669 |  0.17149 |   0.0 | 0.001758 |   0.0 | 0.000002 |   0.0 | 0.048125 |

.. _lland_v4_water_hourly:

water
-----

.. integration-test::

    >>> lnk(WASSER)
    >>> inputs.nied.series = 0.05
    >>> test("lland_v4_water_hourly",
    ...      axis1=(fluxes.nkor, fluxes.evpo, fluxes.evi, fluxes.qah),
    ...      use_conditions=conditions_wasser)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.05 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 | 0.047 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                 1.928423 |                      1.975309 |                         0.0 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.151127 |                  0.901196 |                     0.0 |                      0.0 |     -0.03755 |             0.0 |              0.0 |          2.249931 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.036205 | 0.036205 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010795 | 0.002999 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002999 |
    | 1997-08-03 01:00:00 | 0.05 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 | 0.047 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                 1.892018 |                      1.970628 |                         0.0 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.151127 |                  0.901844 |                     0.0 |                      0.0 |    -0.037577 |             0.0 |              0.0 |          2.249283 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.036073 | 0.036073 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010927 | 0.003035 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003035 |
    | 1997-08-03 02:00:00 | 0.05 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 | 0.047 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 1.868084 |                       1.96233 |                         0.0 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.151127 |                  0.902355 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          2.248772 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.035724 | 0.035724 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011276 | 0.003132 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003132 |
    | 1997-08-03 03:00:00 | 0.05 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 | 0.047 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                 1.856216 |                      1.948396 |                         0.0 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.151127 |                  0.902364 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          2.248763 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.035169 | 0.035169 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011831 | 0.003286 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003286 |
    | 1997-08-03 04:00:00 | 0.05 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 | 0.047 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                 1.821009 |                      1.929443 |                         0.0 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | 0.0 | nan | 0.0 |              0.002099 |                      0.0 |                  0.002099 |                   3.151127 |                  0.902341 |                     0.0 |                      0.0 |    -0.035498 |             0.0 |              0.0 |          2.248786 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.034397 | 0.034397 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.012603 | 0.003501 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003501 |
    | 1997-08-03 05:00:00 | 0.05 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 | 0.047 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                 1.809403 |                      1.907618 |                         0.0 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | 0.0 | nan | 0.0 |              0.023687 |                      0.0 |                  0.023687 |                   3.151127 |                  0.901482 |                     0.0 |                      0.0 |    -0.013875 |             0.0 |              0.0 |          2.249645 |            355.009444 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.033483 | 0.033483 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013517 | 0.003755 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003755 |
    | 1997-08-03 06:00:00 | 0.05 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 | 0.047 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                 1.821009 |                      1.884011 |                         0.0 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | 0.0 | nan | 0.0 |              0.061837 |                      0.0 |                  0.061837 |                   3.151127 |                   0.89977 |                     0.0 |                      0.0 |     0.024346 |             0.0 |              0.0 |          2.251357 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.032456 | 0.032456 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.014544 |  0.00404 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00404 |
    | 1997-08-03 07:00:00 | 0.05 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 | 0.047 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                 1.892018 |                      1.865111 |                         0.0 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | 0.0 | nan | 0.0 |              0.118079 |                      0.0 |                  0.118079 |                   3.151127 |                  0.896826 |                     0.0 |                      0.0 |     0.080712 |             0.0 |              0.0 |          2.254301 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.031557 | 0.031557 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015443 |  0.00429 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00429 |
    | 1997-08-03 08:00:00 | 0.05 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 | 0.047 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                 1.990459 |                      1.846377 |                         0.0 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | 0.0 | nan | 0.0 |              0.184625 |                      0.0 |                  0.184625 |                   3.151127 |                  0.891371 |                     0.0 |                      0.0 |     0.147485 |             0.0 |              0.0 |          2.259756 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.030534 | 0.030534 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016466 | 0.004574 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004574 |
    | 1997-08-03 09:00:00 | 0.05 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 | 0.047 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                 2.200854 |                      1.834142 |                         0.0 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | 0.0 | nan | 0.0 |              0.336704 |                      0.0 |                  0.336704 |                   3.220512 |                  0.922604 |                     0.0 |                      0.0 |     0.298262 |             0.0 |              0.0 |          2.297908 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.030275 | 0.030275 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016725 | 0.004646 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004646 |
    | 1997-08-03 10:00:00 | 0.05 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 | 0.047 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                  2.38585 |                      1.830729 |                         0.0 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | 0.0 | nan | 0.0 |               0.54171 |                      0.0 |                   0.54171 |                   3.289897 |                  0.955299 |                     0.0 |                      0.0 |     0.501906 |             0.0 |              0.0 |          2.334598 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.030379 | 0.030379 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.016621 | 0.004617 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004617 |
    | 1997-08-03 11:00:00 | 0.05 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 | 0.047 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                 2.552842 |                      1.835607 |                         0.0 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | 0.0 | nan | 0.0 |               0.66422 |                      0.0 |                   0.66422 |                   3.463358 |                  1.039981 |                     0.0 |                      0.0 |     0.620888 |             0.0 |              0.0 |          2.423377 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.031338 | 0.031338 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015662 | 0.004351 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004351 |
    | 1997-08-03 12:00:00 | 0.05 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 | 0.047 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 2.537251 |                      1.844905 |                         0.0 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | 0.0 | nan | 0.0 |              0.676668 |                      0.0 |                  0.676668 |                   3.671512 |                  1.142661 |                     0.0 |                      0.0 |     0.629057 |             0.0 |              0.0 |          2.528851 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.032665 | 0.032665 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.014335 | 0.003982 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003982 |
    | 1997-08-03 13:00:00 | 0.05 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 | 0.047 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                 2.616045 |                      1.859671 |                         0.0 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | 0.0 | nan | 0.0 |              0.535824 |                      0.0 |                  0.535824 |                   3.810281 |                  1.211985 |                     0.0 |                      0.0 |     0.485325 |             0.0 |              0.0 |          2.598296 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.033843 | 0.033843 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013157 | 0.003655 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003655 |
    | 1997-08-03 14:00:00 | 0.05 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 | 0.047 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                 2.796942 |                       1.88551 |                         0.0 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | 0.0 | nan | 0.0 |              0.453083 |                      0.0 |                  0.453083 |                   3.879666 |                   1.24753 |                     0.0 |                      0.0 |     0.401102 |             0.0 |              0.0 |          2.632136 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.034948 | 0.034948 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.012052 | 0.003348 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003348 |
    | 1997-08-03 15:00:00 | 0.05 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 | 0.047 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                 2.763263 |                      1.918249 |                         0.0 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | 0.0 | nan | 0.0 |              0.418918 |                      0.0 |                  0.418918 |                   4.018435 |                  1.316893 |                     0.0 |                      0.0 |     0.364048 |             0.0 |              0.0 |          2.701542 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.036637 | 0.036637 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010363 | 0.002879 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002879 |
    | 1997-08-03 16:00:00 | 0.05 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 | 0.047 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                 2.729939 |                      1.957676 |                         0.0 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | 0.0 | nan | 0.0 |              0.300777 |                      0.0 |                  0.300777 |                   4.122512 |                  1.370145 |                     0.0 |                      0.0 |     0.243687 |             0.0 |              0.0 |          2.752367 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.038445 | 0.038445 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.008555 | 0.002376 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002376 |
    | 1997-08-03 17:00:00 | 0.05 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 | 0.047 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                 2.632058 |                      2.001493 |                         0.0 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | 0.0 | nan | 0.0 |               0.14811 |                      0.0 |                   0.14811 |                   4.226589 |                  1.421011 |                     0.0 |                      0.0 |     0.088901 |             0.0 |              0.0 |          2.805578 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.040342 | 0.040342 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.006658 |  0.00185 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00185 |
    | 1997-08-03 18:00:00 | 0.05 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 | 0.047 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                 2.552842 |                      2.046164 |                         0.0 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | 0.0 | nan | 0.0 |              0.055167 |                      0.0 |                  0.055167 |                   4.261282 |                  1.433696 |                     0.0 |                      0.0 |     -0.00457 |             0.0 |              0.0 |          2.827585 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.041848 | 0.041848 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005152 | 0.001431 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001431 |
    | 1997-08-03 19:00:00 | 0.05 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 | 0.047 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                 2.445437 |                      2.089517 |                         0.0 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | 0.0 | nan | 0.0 |              0.014723 |                      0.0 |                  0.014723 |                   4.261282 |                  1.423756 |                     0.0 |                      0.0 |      -0.0446 |             0.0 |              0.0 |          2.837526 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.043061 | 0.043061 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.003939 | 0.001094 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001094 |
    | 1997-08-03 20:00:00 | 0.05 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 | 0.047 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                 2.256393 |                      2.126995 |                         0.0 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | 0.0 | nan | 0.0 |              0.000201 |                      0.0 |                  0.000201 |                   4.261282 |                  1.412078 |                     0.0 |                      0.0 |    -0.058636 |             0.0 |              0.0 |          2.849203 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.043962 | 0.043962 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.003038 | 0.000844 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000844 |
    | 1997-08-03 21:00:00 | 0.05 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 | 0.047 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                 2.041337 |                      2.153241 |                         0.0 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.261282 |                  1.401564 |                     0.0 |                      0.0 |    -0.058399 |             0.0 |              0.0 |          2.859717 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.044514 | 0.044514 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.002486 |  0.00069 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00069 |
    | 1997-08-03 22:00:00 | 0.05 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 | 0.047 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                 1.940693 |                      2.174668 |                         0.0 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.261282 |                  1.391158 |                     0.0 |                      0.0 |    -0.057965 |             0.0 |              0.0 |          2.870124 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.044975 | 0.044975 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.002025 | 0.000563 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000563 |
    | 1997-08-03 23:00:00 | 0.05 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 | 0.047 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 1.868084 |                      2.192855 |                         0.0 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.261282 |                   1.38284 |                     0.0 |                      0.0 |    -0.057618 |             0.0 |              0.0 |          2.878441 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.045525 | 0.045525 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.001475 |  0.00041 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00041 |
    | 1997-08-04 00:00:00 | 0.05 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 | 0.047 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                 1.809403 |                      2.187158 |                         0.0 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.239363 |                   1.39654 |                     0.0 |                      0.0 |    -0.058189 |             0.0 |              0.0 |          2.842823 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.04554 |  0.04554 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.00146 | 0.000406 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000406 |
    | 1997-08-04 01:00:00 | 0.05 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 | 0.047 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                 1.752342 |                      2.180337 |                         0.0 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.239363 |                  1.406633 |                     0.0 |                      0.0 |     -0.05861 |             0.0 |              0.0 |           2.83273 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.045787 | 0.045787 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.001213 | 0.000337 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000337 |
    | 1997-08-04 02:00:00 | 0.05 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 | 0.047 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                 1.729962 |                      2.173535 |                         0.0 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.239363 |                  1.418069 |                     0.0 |                      0.0 |    -0.059086 |             0.0 |              0.0 |          2.821294 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.046085 | 0.046085 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000915 | 0.000254 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000254 |
    | 1997-08-04 03:00:00 | 0.05 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 | 0.047 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                 1.696861 |                      2.165623 |                         0.0 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   4.239363 |                   1.43017 |                     0.0 |                      0.0 |     -0.05959 |             0.0 |              0.0 |          2.809192 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.046319 | 0.046319 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000681 | 0.000189 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000189 |
    | 1997-08-04 04:00:00 | 0.05 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 | 0.047 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                 1.632319 |                       2.15605 |                         0.0 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | 0.0 | nan | 0.0 |              0.001718 |                      0.0 |                  0.001718 |                   4.239363 |                  1.440702 |                     0.0 |                      0.0 |    -0.058312 |             0.0 |              0.0 |           2.79866 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.046528 | 0.046528 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000472 | 0.000131 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000131 |
    | 1997-08-04 05:00:00 | 0.05 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 | 0.047 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                 1.611289 |                      2.145953 |                         0.0 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | 0.0 | nan | 0.0 |              0.021547 |                      0.0 |                  0.021547 |                   4.239363 |                  1.451305 |                     0.0 |                      0.0 |    -0.038924 |             0.0 |              0.0 |          2.788058 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.046795 | 0.046795 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000205 | 0.000057 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000057 |
    | 1997-08-04 06:00:00 | 0.05 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 | 0.047 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                 1.632319 |                      2.136456 |                         0.0 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | 0.0 | nan | 0.0 |              0.068025 |                      0.0 |                  0.068025 |                   4.273966 |                  1.478175 |                     0.0 |                      0.0 |     0.006434 |             0.0 |              0.0 |          2.795791 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.047182 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 07:00:00 | 0.05 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 | 0.047 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                 1.752342 |                      2.129774 |                         0.0 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | 0.0 | nan | 0.0 |              0.308638 |                      0.0 |                  0.308638 |                   4.585394 |                   1.64464 |                     0.0 |                      0.0 |     0.240112 |             0.0 |              0.0 |          2.940754 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.0492 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 08:00:00 | 0.05 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 | 0.047 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                  1.91622 |                       2.12644 |                         0.0 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | 0.0 | nan | 0.0 |              0.513243 |                      0.0 |                  0.513243 |                   4.931426 |                   1.83472 |                     0.0 |                      0.0 |     0.436797 |             0.0 |              0.0 |          3.096706 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.051571 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 09:00:00 | 0.05 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 | 0.047 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                 2.093348 |                      2.122001 |                         0.0 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | 0.0 | nan | 0.0 |              0.663127 |                      0.0 |                  0.663127 |                   5.208251 |                  1.989515 |                     0.0 |                      0.0 |     0.580231 |             0.0 |              0.0 |          3.218735 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.053462 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 10:00:00 | 0.05 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 | 0.047 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                 2.270468 |                      2.117571 |                         0.0 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | 0.0 | nan | 0.0 |              0.695311 |                      0.0 |                  0.695311 |                   5.346663 |                   2.06621 |                     0.0 |                      0.0 |     0.609219 |             0.0 |              0.0 |          3.280453 |             68.711505 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054455 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 11:00:00 | 0.05 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 | 0.047 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                  2.38585 |                      2.111492 |                         0.0 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | 0.0 | nan | 0.0 |              0.648284 |                      0.0 |                  0.648284 |                   5.381266 |                  2.087634 |                     0.0 |                      0.0 |     0.561299 |             0.0 |              0.0 |          3.293633 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054881 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 12:00:00 | 0.05 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 | 0.047 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                 2.490975 |                      2.109837 |                         0.0 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | 0.0 | nan | 0.0 |              0.670155 |                      0.0 |                  0.670155 |                    5.41587 |                  2.106001 |                     0.0 |                      0.0 |     0.582405 |             0.0 |              0.0 |          3.309868 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.055387 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 13:00:00 | 0.05 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 | 0.047 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                 2.616045 |                      2.109837 |                         0.0 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | 0.0 | nan | 0.0 |              0.647397 |                      0.0 |                  0.647397 |                   5.519679 |                  2.166369 |                     0.0 |                      0.0 |     0.557131 |             0.0 |              0.0 |           3.35331 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.056372 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 14:00:00 | 0.05 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 | 0.047 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                 2.632058 |                      2.104328 |                         0.0 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | 0.0 | nan | 0.0 |              0.662635 |                      0.0 |                  0.662635 |                   5.692695 |                  2.265756 |                     0.0 |                      0.0 |     0.568229 |             0.0 |              0.0 |          3.426939 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.057374 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 15:00:00 | 0.05 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 | 0.047 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                 2.616045 |                      2.099381 |                         0.0 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | 0.0 | nan | 0.0 |              0.558476 |                      0.0 |                  0.558476 |                   5.831107 |                  2.348437 |                     0.0 |                      0.0 |     0.460624 |             0.0 |              0.0 |           3.48267 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.058283 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 16:00:00 | 0.05 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 | 0.047 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                 2.680609 |                      2.097734 |                         0.0 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | 0.0 | nan | 0.0 |              0.408064 |                      0.0 |                  0.408064 |                    5.96952 |                  2.428674 |                     0.0 |                      0.0 |     0.306869 |             0.0 |              0.0 |          3.540846 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.05932 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 17:00:00 | 0.05 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 | 0.047 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                 2.616045 |                      2.097185 |                         0.0 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | 0.0 | nan | 0.0 |              0.272381 |                      0.0 |                  0.272381 |                   6.211742 |                  2.564427 |                     0.0 |                      0.0 |      0.16553 |             0.0 |              0.0 |          3.647315 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.061051 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 18:00:00 | 0.05 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 | 0.047 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 2.475714 |                      2.094444 |                         0.0 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | 0.0 | nan | 0.0 |              0.138863 |                      0.0 |                  0.138863 |                    6.52317 |                  2.744116 |                     0.0 |                      0.0 |     0.024525 |             0.0 |              0.0 |          3.779055 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.063121 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 19:00:00 | 0.05 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 | 0.047 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                 2.298847 |                       2.08897 |                         0.0 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | 0.0 | nan | 0.0 |              0.045438 |                      0.0 |                  0.045438 |                   6.869202 |                  2.950104 |                     0.0 |                      0.0 |    -0.077483 |             0.0 |              0.0 |          3.919098 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.0652 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 20:00:00 | 0.05 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 | 0.047 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                 2.054233 |                      2.080783 |                         0.0 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | 0.0 | nan | 0.0 |              0.000878 |                      0.0 |                  0.000878 |                   6.973011 |                  3.034376 |                     0.0 |                      0.0 |    -0.125554 |             0.0 |              0.0 |          3.938635 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.06592 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 21:00:00 | 0.05 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 | 0.047 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                 1.797863 |                       2.06991 |                         0.0 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.973011 |                  3.060678 |                     0.0 |                      0.0 |    -0.127528 |             0.0 |              0.0 |          3.912334 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.065866 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 22:00:00 | 0.05 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 | 0.047 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                 1.685952 |                      2.058007 |                         0.0 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.973011 |                  3.084324 |                     0.0 |                      0.0 |    -0.128514 |             0.0 |              0.0 |          3.888687 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.065746 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 23:00:00 | 0.05 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 | 0.047 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                 1.611289 |                      2.045628 |                         0.0 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.973011 |                  3.103639 |                     0.0 |                      0.0 |    -0.129318 |             0.0 |              0.0 |          3.869372 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.065422 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 00:00:00 | 0.05 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 | 0.047 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                 1.539538 |                      2.032245 |                         0.0 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.943374 |                  3.124611 |                     0.0 |                      0.0 |    -0.130192 |             0.0 |              0.0 |          3.818763 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.064471 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 01:00:00 | 0.05 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 | 0.047 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                 1.480281 |                      2.018409 |                         0.0 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.943374 |                  3.135507 |                     0.0 |                      0.0 |    -0.130646 |             0.0 |              0.0 |          3.807867 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.063963 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 02:00:00 | 0.05 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 | 0.047 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                 1.432438 |                      2.003073 |                         0.0 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.943374 |                  3.144898 |                     0.0 |                      0.0 |    -0.131037 |             0.0 |              0.0 |          3.798476 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.063323 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 03:00:00 | 0.05 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 | 0.047 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 1.404387 |                       1.98784 |                         0.0 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   6.943374 |                   3.15411 |                     0.0 |                      0.0 |    -0.131421 |             0.0 |              0.0 |          3.789264 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062667 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 04:00:00 | 0.05 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 | 0.047 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                 1.376819 |                      1.974268 |                         0.0 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | 0.0 | nan | 0.0 |              0.001601 |                      0.0 |                  0.001601 |                   6.943374 |                  3.166293 |                     0.0 |                      0.0 |    -0.130328 |             0.0 |              0.0 |          3.777081 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062183 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 05:00:00 | 0.05 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 | 0.047 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 1.367736 |                      1.961295 |                         0.0 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | 0.0 | nan | 0.0 |              0.022826 |                      0.0 |                  0.022826 |                   6.943374 |                  3.179808 |                     0.0 |                      0.0 |    -0.109666 |             0.0 |              0.0 |          3.763566 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.061583 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 06:00:00 | 0.05 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 | 0.047 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                 1.460978 |                      1.952516 |                         0.0 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | 0.0 | nan | 0.0 |              0.135954 |                      0.0 |                  0.135954 |                   7.115928 |                  3.288825 |                     0.0 |                      0.0 |     -0.00108 |             0.0 |              0.0 |          3.827103 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062021 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 07:00:00 | 0.05 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 | 0.047 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                 1.685952 |                      1.949426 |                         0.0 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | 0.0 | nan | 0.0 |              0.330431 |                      0.0 |                  0.330431 |                   7.115928 |                  3.287363 |                     0.0 |                      0.0 |     0.193458 |             0.0 |              0.0 |          3.828566 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.061775 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 08:00:00 | 0.05 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 | 0.047 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                 1.953031 |                       1.95097 |                         0.0 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | 0.0 | nan | 0.0 |              0.551993 |                      0.0 |                  0.551993 |                   7.115928 |                  3.279458 |                     0.0 |                      0.0 |     0.415349 |             0.0 |              0.0 |          3.836471 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.061896 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 09:00:00 | 0.05 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 | 0.047 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                 2.256393 |                      1.957159 |                         0.0 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | 0.0 | nan | 0.0 |              0.679881 |                      0.0 |                  0.679881 |                   7.081417 |                  3.250753 |                     0.0 |                      0.0 |     0.544433 |             0.0 |              0.0 |          3.830665 |             53.251417 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062093 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 10:00:00 | 0.05 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 | 0.047 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                  2.43042 |                      1.962848 |                         0.0 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | 0.0 | nan | 0.0 |              0.827082 |                      0.0 |                  0.827082 |                   7.115928 |                  3.266966 |                     0.0 |                      0.0 |     0.690958 |             0.0 |              0.0 |          3.848962 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062733 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 11:00:00 | 0.05 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 | 0.047 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                 2.552842 |                      1.968551 |                         0.0 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | 0.0 | nan | 0.0 |              0.828518 |                      0.0 |                  0.828518 |                    7.18495 |                  3.309837 |                     0.0 |                      0.0 |     0.690608 |             0.0 |              0.0 |          3.875113 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.063478 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 12:00:00 | 0.05 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 | 0.047 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                 2.729939 |                      1.976351 |                         0.0 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | 0.0 | nan | 0.0 |              0.795165 |                      0.0 |                  0.795165 |                   7.219461 |                  3.330376 |                     0.0 |                      0.0 |     0.656399 |             0.0 |              0.0 |          3.889085 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.064117 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 13:00:00 | 0.05 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 | 0.047 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                 2.830978 |                      1.983133 |                         0.0 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | 0.0 | nan | 0.0 |              0.829447 |                      0.0 |                  0.829447 |                   7.288483 |                  3.363175 |                     0.0 |                      0.0 |     0.689314 |             0.0 |              0.0 |          3.925307 |             48.410379 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.064896 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 14:00:00 | 0.05 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 | 0.047 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                 2.882708 |                      1.990983 |                         0.0 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | 0.0 | nan | 0.0 |              0.794804 |                      0.0 |                  0.794804 |                   7.322994 |                   3.38194 |                     0.0 |                      0.0 |      0.65389 |             0.0 |              0.0 |          3.941053 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.065782 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 15:00:00 | 0.05 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 | 0.047 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                 2.952961 |                      2.001493 |                         0.0 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | 0.0 | nan | 0.0 |              0.670005 |                      0.0 |                  0.670005 |                   7.357505 |                  3.400029 |                     0.0 |                      0.0 |     0.528337 |             0.0 |              0.0 |          3.957476 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.066681 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 16:00:00 | 0.05 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 | 0.047 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                 2.988644 |                      2.010993 |                         0.0 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | 0.0 | nan | 0.0 |              0.486103 |                      0.0 |                  0.486103 |                   7.392016 |                   3.42173 |                     0.0 |                      0.0 |     0.343531 |             0.0 |              0.0 |          3.970285 |             45.320355 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.06748 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 17:00:00 | 0.05 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 | 0.047 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                 2.988644 |                      2.022657 |                         0.0 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | 0.0 | nan | 0.0 |              0.295503 |                      0.0 |                  0.295503 |                   7.392016 |                  3.423912 |                     0.0 |                      0.0 |      0.15284 |             0.0 |              0.0 |          3.968104 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.068039 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 18:00:00 | 0.05 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 | 0.047 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                  2.93526 |                      2.037589 |                         0.0 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | 0.0 | nan | 0.0 |              0.144038 |                      0.0 |                  0.144038 |                   7.392016 |                   3.42458 |                     0.0 |                      0.0 |     0.001347 |             0.0 |              0.0 |          3.967436 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.068633 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 19:00:00 | 0.05 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 | 0.047 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                 2.780058 |                      2.054233 |                         0.0 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | 0.0 | nan | 0.0 |              0.043553 |                      0.0 |                  0.043553 |                   7.392016 |                  3.420162 |                     0.0 |                      0.0 |    -0.098954 |             0.0 |              0.0 |          3.971854 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.069308 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 20:00:00 | 0.05 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 | 0.047 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 2.537251 |                      2.072623 |                         0.0 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | 0.0 | nan | 0.0 |              0.000936 |                      0.0 |                  0.000936 |                   7.461037 |                  3.445356 |                     0.0 |                      0.0 |     -0.14262 |             0.0 |              0.0 |          4.015681 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.070109 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 21:00:00 | 0.05 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 | 0.047 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                 2.242395 |                      2.091705 |                         0.0 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   7.461037 |                  3.434698 |                     0.0 |                      0.0 |    -0.143112 |             0.0 |              0.0 |          4.026339 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.070708 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 22:00:00 | 0.05 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 | 0.047 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                  2.10653 |                       2.11094 |                         0.0 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   7.461037 |                  3.427556 |                     0.0 |                      0.0 |    -0.142815 |             0.0 |              0.0 |          4.033482 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.071405 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 23:00:00 | 0.05 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 | 0.047 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                 1.990459 |                      2.129218 |                         0.0 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   7.461037 |                  3.418333 |                     0.0 |                      0.0 |    -0.142431 |             0.0 |              0.0 |          4.042705 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.07218 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

    >>> inputs.nied.series = 0.0

.. _lland_v4_nadelw_summer_hourly:

conifers (summer)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v4_nadelw_summer_hourly",
    ...      axis1=(fluxes.evi, fluxes.evb),
    ...      axis2=states.inzp,
    ...      use_conditions=conditions_nadelw_summer)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |         g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |      ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |             95.1 |       0.8 |               101.5 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.283992 |          0.8 |                 1.928423 |                      1.975309 |                         0.0 |                          0.0 |                      0.122362 |                           0.124959 |              1.83393 |                  1.696461 |       99.66607 |    1.21073 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.208683 | -0.036532 |                   0.0 |                      nan |                       0.0 |                   9.243306 |                  0.901196 |                     0.0 |                      0.0 |     -0.03755 |             0.0 |              0.0 |          8.342111 |                 117.5 |                 100.0 |               170.875609 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001025 | -0.001025 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.221929 | 0.061647 | 0.001025 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.308721 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.221929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061647 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |             94.9 |       0.8 |               101.5 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.283992 |          0.8 |                 1.892018 |                      1.970628 |                         0.0 |                          0.0 |                      0.120339 |                             0.1247 |             1.795525 |                  1.694576 |      99.704475 |   1.212158 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.295735 | -0.028172 |                   0.0 |                      nan |                       0.0 |                   9.243306 |                  0.901844 |                     0.0 |                      0.0 |    -0.037577 |             0.0 |              0.0 |          8.341463 |                 117.5 |                 100.0 |               173.430933 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000655 | -0.000655 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.216196 | 0.060055 |  0.00168 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.326476 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.216196 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060055 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |             95.9 |       0.8 |               101.5 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.283992 |          0.8 |                 1.868084 |                       1.96233 |                         0.0 |                          0.0 |                      0.119006 |                           0.124241 |             1.791493 |                  1.692183 |      99.708507 |   1.213014 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 |  15.35492 | -0.022574 |                   0.0 |                      nan |                       0.0 |                   9.243306 |                  0.902355 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          8.340951 |                 117.5 |                 100.0 |               173.240523 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002635 | -0.002635 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.210612 | 0.058503 | 0.004315 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.338633 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.210612 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058503 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |             96.7 |       0.8 |               101.5 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.283992 |          0.8 |                 1.856216 |                      1.948396 |                         0.0 |                          0.0 |                      0.118345 |                           0.123469 |             1.794961 |                  1.689422 |      99.705039 |   1.213417 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            10.503757 |       0.0 |             0.0 |                     0.0 | -0.010417 | 15.395444 | -0.019538 |                   0.0 |                      nan |                       0.0 |                   9.243306 |                  0.902364 |                     0.0 |                      0.0 |    -0.037598 |             0.0 |              0.0 |          8.340943 |                 117.5 |                 100.0 |               172.677019 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004198 | -0.004198 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.205172 | 0.056992 | 0.008513 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.347755 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.205172 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056992 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |             97.2 |       0.8 |               101.5 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.283992 |          0.8 |                 1.821009 |                      1.929443 |                         0.0 |                          0.0 |                      0.116379 |                           0.122418 |              1.77002 |                   1.68569 |       99.72998 |    1.21479 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                 0.429734 |                     15.566134 |            10.503757 |  0.086656 |        0.007799 |                0.006997 | -0.010417 | 15.425849 | -0.012402 |              0.006158 |                      nan |                  0.001144 |                   9.243306 |                  0.902341 |                     0.0 |                      0.0 |     -0.03144 |             0.0 |              0.0 |          8.340965 |                 117.5 |                 100.0 |                174.22122 |               93.330305 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003798 | -0.003798 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.199872 |  0.05552 | 0.012311 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.34974 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.199872 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05552 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |             97.5 |       0.6 |               101.5 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.212994 |          0.6 |                 1.809403 |                      1.907618 |                         0.0 |                          0.0 |                       0.11573 |                           0.121206 |             1.764168 |                  1.682599 |      99.735832 |   1.215237 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  0.968392 |        0.087159 |                0.078957 | -0.010417 | 15.432466 | -0.010099 |              0.069482 |                      nan |                  0.012904 |                   9.243306 |                  0.901482 |                     0.0 |                      0.0 |      0.03192 |             0.0 |              0.0 |          8.341824 |            156.666667 |                 100.0 |               174.539602 |              172.004443 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008314 |  0.008314 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.194709 | 0.054086 | 0.003996 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.349422 | 78.724681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.194709 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.054086 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |             97.7 |       0.9 |               101.5 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.319491 |          0.9 |                 1.821009 |                      1.884011 |                         0.0 |                          0.0 |                      0.116379 |                           0.119893 |             1.779125 |                   1.68038 |      99.720875 |   1.214748 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |    2.4897 |        0.224084 |                0.206122 | -0.010417 | 15.431406 | -0.012282 |              0.181388 |                      nan |                  0.033686 |                   9.243306 |                   0.89977 |                     0.0 |                      0.0 |     0.143897 |             0.0 |              0.0 |          8.343537 |            104.444444 |                 100.0 |               173.411109 |              170.929929 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034791 |  0.003996 | 0.017757 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.189679 | 0.052689 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.351287 | 78.706923 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.189679 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052689 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |             97.4 |       0.9 |               101.5 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.319491 |          0.9 |                 1.892018 |                      1.865111 |                         0.0 |                          0.0 |                      0.120339 |                           0.118841 |             1.842826 |                  1.680931 |      99.657174 |   1.211943 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  4.679733 |        0.421196 |                0.393598 | -0.010417 | 15.437623 | -0.025107 |              0.346366 |                      nan |                  0.064325 |                   9.243306 |                  0.896826 |                     0.0 |                      0.0 |     0.308999 |             0.0 |              0.0 |           8.34648 |            104.444444 |                 100.0 |                169.26657 |              166.979573 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073429 |       0.0 | 0.043065 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18478 | 0.051328 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.365978 | 78.663858 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18478 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.051328 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |             96.8 |       0.9 |               101.6 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.319491 |          0.9 |                 1.990459 |                      1.846377 |                         0.0 |                          0.0 |                      0.125796 |                           0.117796 |             1.926764 |                  1.685281 |      99.673236 |   1.209425 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.503757 |  7.219274 |        0.649765 |                0.615417 | -0.010417 | 15.486592 |  -0.04133 |              0.541567 |                      nan |                  0.100577 |                   9.243306 |                  0.891371 |                     0.0 |                      0.0 |     0.504426 |             0.0 |              0.0 |          8.351935 |            104.444444 |                 100.0 |               164.539821 |              162.466484 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.121304 |       0.0 | 0.072626 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.180007 | 0.050002 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.39689 | 78.591232 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.180007 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050002 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |             86.1 |       1.3 |               101.6 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           0.461488 |          1.3 |                 2.200854 |                      1.834142 |                         0.0 |                          0.0 |                      0.137341 |                           0.117113 |             1.894936 |                  1.687258 |      99.705064 |   1.202945 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             10.73504 |  9.442017 |         1.18975 |                1.122345 | -0.010417 | 15.589635 | -0.073664 |              0.987664 |                      nan |                  0.183423 |                   9.446835 |                  0.922604 |                     0.0 |                      0.0 |     0.949222 |             0.0 |              0.0 |          8.524231 |             72.307692 |                 100.0 |               175.701188 |              173.109975 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.278008 |       0.0 | 0.144385 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.175357 |  0.04871 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.460138 | 78.446847 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.175357 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04871 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |             76.8 |       1.5 |               101.6 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           0.532486 |          1.5 |                  2.38585 |                      1.830729 |                         0.0 |                          0.0 |                      0.147368 |                           0.116922 |             1.832333 |                  1.688695 |      99.767667 |   1.197896 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            10.966322 | 10.709164 |         1.92774 |                  1.8057 | -0.010417 | 15.800459 |  -0.09719 |              1.589016 |                      nan |                  0.295103 |                   9.650363 |                  0.955299 |                     0.0 |                      0.0 |     1.549212 |             0.0 |              0.0 |          8.695064 |             62.666667 |                 100.0 |                 197.9318 |              194.171796 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.484138 |       0.0 | 0.227779 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.170827 | 0.047452 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.546911 | 78.219068 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.170827 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047452 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |             71.8 |       1.2 |               101.6 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.425988 |          1.2 |                 2.552842 |                      1.835607 |                         0.0 |                          0.0 |                      0.156328 |                           0.117195 |             1.832941 |                   1.69067 |      99.767059 |    1.19342 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            11.544527 | 10.773868 |        2.327265 |                2.214067 | -0.010417 | 16.089704 | -0.114702 |              1.948379 |                      nan |                  0.361842 |                  10.159184 |                  1.039981 |                     0.0 |                      0.0 |     1.905047 |             0.0 |              0.0 |          9.119203 |             78.333333 |                 100.0 |               224.347905 |              218.964094 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.579474 |       0.0 | 0.287656 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.166415 | 0.046226 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.651197 | 77.931412 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166415 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046226 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |             67.5 |       1.3 |               101.6 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           0.461488 |          1.3 |                 2.537251 |                      1.844905 |                         0.0 |                          0.0 |                      0.155495 |                           0.117714 |             1.712644 |                  1.692316 |      99.887356 |   1.194363 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.238374 | 10.364714 |        2.425457 |                 2.25556 | -0.010417 | 16.437323 | -0.105034 |              1.984892 |                      nan |                  0.368623 |                  10.769769 |                  1.142661 |                     0.0 |                      0.0 |     1.937282 |             0.0 |              0.0 |          9.627108 |             72.307692 |                 100.0 |               245.282288 |              238.432816 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.611134 |       0.0 |  0.27972 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.162116 | 0.045032 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.745814 | 77.651692 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.162116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045032 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |             66.1 |       1.5 |               101.6 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           0.532486 |          1.5 |                 2.616045 |                      1.859671 |                         0.0 |                          0.0 |                      0.159698 |                           0.118538 |             1.729206 |                  1.694238 |      99.870794 |   1.192265 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            12.700938 | 10.872489 |         1.95714 |                 1.78608 | -0.010417 | 16.752713 | -0.109021 |              1.571751 |                      nan |                  0.291897 |                  11.176825 |                  1.211985 |                     0.0 |                      0.0 |     1.521251 |             0.0 |              0.0 |           9.96484 |             62.666667 |                 100.0 |               259.971088 |              251.999948 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.532257 |       0.0 | 0.222625 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.157929 | 0.043869 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.844419 | 77.429067 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043869 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |             63.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           0.674482 |          1.9 |                 2.796942 |                       1.88551 |                         0.0 |                          0.0 |                      0.169283 |                           0.119977 |             1.773261 |                  1.698059 |      99.826739 |   1.187639 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |             12.93222 | 10.461143 |        1.694785 |                1.510276 | -0.010417 | 17.081396 | -0.125682 |              1.329043 |                      nan |                  0.246822 |                  11.380354 |                   1.24753 |                     0.0 |                      0.0 |     1.277062 |             0.0 |              0.0 |         10.132824 |             49.473684 |                 100.0 |               297.295394 |              286.132852 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.526288 |       0.0 | 0.183586 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.153849 | 0.042736 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.959684 | 77.245482 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.153849 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042736 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |             62.4 |       1.9 |               101.6 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           0.674482 |          1.9 |                 2.763263 |                      1.918249 |                         0.0 |                          0.0 |                      0.167505 |                           0.121797 |             1.724276 |                  1.703725 |      99.875724 |    1.18866 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.394785 |  8.865382 |        1.595844 |                1.396394 | -0.010417 | 17.465613 | -0.113063 |              1.228827 |                      nan |                  0.228211 |                   11.78741 |                  1.316893 |                     0.0 |                      0.0 |     1.173957 |             0.0 |              0.0 |         10.470517 |             49.473684 |                 100.0 |               302.773995 |              291.102238 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.500938 |       0.0 | 0.172023 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.149875 | 0.041632 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06233 | 77.073458 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.149875 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041632 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |             61.1 |       2.3 |               101.6 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |           0.816478 |          2.3 |                 2.729939 |                      1.957676 |                         0.0 |                          0.0 |                      0.165743 |                           0.123983 |             1.667993 |                  1.709704 |      99.932007 |   1.189715 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            13.741708 |  6.474205 |        1.165412 |                1.002589 | -0.010417 | 17.807767 | -0.101352 |              0.882278 |                      nan |                  0.163852 |                  12.092703 |                  1.370145 |                     0.0 |                      0.0 |     0.825189 |             0.0 |              0.0 |         10.722558 |             40.869565 |                 100.0 |               310.898811 |              298.452854 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.446874 |       0.0 | 0.133947 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.146004 | 0.040557 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.153266 | 76.939511 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146004 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040557 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |             62.1 |       2.4 |               101.6 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           0.851977 |          2.4 |                 2.632058 |                      2.001493 |                         0.0 |                          0.0 |                       0.16055 |                           0.126406 |             1.634508 |                  1.718782 |      99.965492 |   1.192283 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.088631 |  3.977658 |         0.57281 |                0.493699 | -0.010417 | 18.110885 | -0.081845 |              0.434455 |                      nan |                  0.080685 |                  12.397995 |                  1.421011 |                     0.0 |                      0.0 |     0.375246 |             0.0 |              0.0 |         10.976984 |             39.166667 |                 100.0 |               291.447874 |              280.817391 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.321464 |       0.0 | 0.096503 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.142232 | 0.039509 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.224694 | 76.843008 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.142232 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039509 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |             67.0 |       2.5 |               101.6 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           0.887476 |          2.5 |                 2.552842 |                      2.046164 |                         0.0 |                          0.0 |                      0.156328 |                           0.128868 |             1.710404 |                  1.731481 |      99.889596 |   1.193968 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  1.972911 |        0.213084 |                0.183891 | -0.010417 | 18.348979 | -0.065902 |              0.161824 |                      nan |                  0.030053 |                  12.499759 |                  1.433696 |                     0.0 |                      0.0 |     0.102086 |             0.0 |              0.0 |         11.066063 |                  37.6 |                 100.0 |               252.719347 |              245.311527 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222652 |       0.0 | 0.070699 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.138558 | 0.038488 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.280179 |  76.77231 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.138558 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038488 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |             74.5 |       2.5 |               101.6 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           0.887476 |          2.5 |                 2.445437 |                      2.089517 |                         0.0 |                          0.0 |                      0.150575 |                           0.131251 |              1.82185 |                  1.748925 |       99.77815 |   1.196313 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      1.0 |                     15.566134 |            14.204272 |  0.634046 |        0.057067 |                0.049078 | -0.010417 | 18.533931 | -0.046787 |              0.043189 |                      nan |                  0.008021 |                  12.499759 |                  1.423756 |                     0.0 |                      0.0 |    -0.016134 |             0.0 |              0.0 |         11.076003 |                  37.6 |                 100.0 |               212.737062 |              208.098179 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.143053 |       0.0 | 0.049883 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.134979 | 0.037494 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.31655 | 76.722426 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.134979 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037494 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |             81.2 |       2.2 |               101.6 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           0.780979 |          2.2 |                 2.256393 |                      2.126995 |                         0.0 |                          0.0 |                      0.140363 |                           0.133306 |             1.832191 |                  1.766824 |      99.767809 |   1.201582 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                   0.1364 |                     15.566134 |            14.204272 |  0.008648 |        0.000778 |                0.000669 | -0.010417 | 18.655165 | -0.016088 |              0.000589 |                      nan |                  0.000109 |                  12.499759 |                  1.412078 |                     0.0 |                      0.0 |    -0.058248 |             0.0 |              0.0 |         11.087681 |             42.727273 |                 100.0 |               190.515534 |               75.961268 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.077666 |       0.0 | 0.047501 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.131493 | 0.036526 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.322221 | 76.674926 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.131493 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036526 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |             86.9 |       1.7 |               101.6 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           0.603484 |          1.7 |                 2.041337 |                      2.153241 |                         0.0 |                          0.0 |                      0.128603 |                           0.134742 |             1.773922 |                  1.781538 |      99.826078 |   1.208454 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.674071 |   0.01888 |                   0.0 |                      nan |                       0.0 |                  12.499759 |                  1.401564 |                     0.0 |                      0.0 |    -0.058399 |             0.0 |              0.0 |         11.098195 |             55.294118 |                 100.0 |               185.552232 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033668 |       0.0 | 0.022649 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.128096 | 0.035582 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.292925 | 76.652277 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.128096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035582 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |             90.1 |       1.7 |               101.7 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           0.603484 |          1.7 |                 1.940693 |                      2.174668 |                         0.0 |                          0.0 |                      0.123042 |                           0.135912 |             1.748564 |                  1.795279 |      99.951436 |   1.213101 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.576416 |  0.034051 |                   0.0 |                      nan |                       0.0 |                  12.499759 |                  1.391158 |                     0.0 |                      0.0 |    -0.057965 |             0.0 |              0.0 |         11.108602 |             55.294118 |                 100.0 |               184.347096 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020838 |       0.0 | 0.013901 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.124787 | 0.034663 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.248458 | 76.638376 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.124787 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034663 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |             90.9 |       2.3 |               101.7 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |           0.816478 |          2.3 |                 1.868084 |                      2.192855 |                         0.0 |                          0.0 |                      0.119006 |                           0.136905 |             1.698088 |                  1.806456 |     100.001912 |   1.215845 |         0.297909 | 4.570266 |   20.1364 |         0.972155 |                 36.001693 |                      0.0 |                     15.566134 |            14.204272 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.428192 |  0.043809 |                   0.0 |                      nan |                       0.0 |                  12.499759 |                   1.38284 |                     0.0 |                      0.0 |    -0.057618 |             0.0 |              0.0 |         11.116919 |             40.869565 |                 100.0 |               187.258263 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.027773 |       0.0 | 0.016657 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.121564 | 0.033768 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.194232 | 76.621719 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033768 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |             88.0 |       2.3 |               101.7 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |           0.816478 |          2.3 |                 1.809403 |                      2.187158 |                         0.0 |                          0.0 |                       0.11573 |                           0.136594 |             1.592275 |                  1.795292 |     100.107725 |    1.21843 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.24744 |  0.050705 |                   0.0 |                      nan |                       0.0 |                  12.435463 |                   1.39654 |                     0.0 |                      0.0 |    -0.058189 |             0.0 |              0.0 |         11.038924 |             40.869565 |                 100.0 |               196.850637 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.041216 |       0.0 |  0.02456 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.118424 | 0.032895 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.133111 | 76.597159 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.118424 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032895 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |             89.2 |       2.2 |               101.6 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           0.780979 |          2.2 |                 1.752342 |                      2.180337 |                         0.0 |                          0.0 |                      0.112529 |                           0.136222 |             1.563089 |                  1.784515 |     100.036911 |   1.219467 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.043702 |  0.057104 |                   0.0 |                      nan |                       0.0 |                  12.435463 |                  1.406633 |                     0.0 |                      0.0 |     -0.05861 |             0.0 |              0.0 |          11.02883 |             42.727273 |                 100.0 |               198.635637 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.032732 |       0.0 | 0.019695 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.115365 | 0.032046 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.06559 | 76.577464 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.115365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032046 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |             88.8 |       2.3 |               101.6 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |           0.816478 |          2.3 |                 1.729962 |                      2.173535 |                         0.0 |                          0.0 |                      0.111271 |                            0.13585 |             1.536206 |                  1.772518 |     100.063794 |   1.220436 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.818633 |  0.056562 |                   0.0 |                      nan |                       0.0 |                  12.435463 |                  1.418069 |                     0.0 |                      0.0 |    -0.059086 |             0.0 |              0.0 |         11.017395 |             40.869565 |                 100.0 |               201.155121 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.036288 |       0.0 | 0.021425 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.112385 | 0.031218 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.998611 |  76.55604 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.112385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031218 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |             89.7 |       2.0 |               101.6 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           0.709981 |          2.0 |                 1.696861 |                      2.165623 |                         0.0 |                          0.0 |                      0.109405 |                           0.135418 |             1.522084 |                  1.759749 |     100.077916 |   1.221771 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            14.131208 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.595369 |   0.05822 |                   0.0 |                      nan |                       0.0 |                  12.435463 |                   1.43017 |                     0.0 |                      0.0 |     -0.05959 |             0.0 |              0.0 |         11.005293 |                  47.0 |                 100.0 |               202.122033 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.025988 |       0.0 | 0.016073 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.109482 | 0.030412 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.929974 | 76.539966 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.109482 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030412 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |             93.0 |       2.3 |               101.6 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |           0.816478 |          2.3 |                 1.632319 |                       2.15605 |                         0.0 |                          0.0 |                      0.105752 |                           0.134895 |             1.518057 |                  1.748197 |     100.081943 |    1.22434 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.398044 |                     15.502754 |            14.131208 |   0.07485 |        0.006692 |                0.005726 | -0.010417 |  17.36658 |  0.066238 |              0.005039 |                      nan |                  0.000936 |                  12.435463 |                  1.440702 |                     0.0 |                      0.0 |    -0.054991 |             0.0 |              0.0 |         10.994761 |             40.869565 |                 100.0 |               201.760902 |               93.626696 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.016861 |       0.0 | 0.008591 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.106654 | 0.029626 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.853319 | 76.531375 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.106654 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029626 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |             93.6 |       2.5 |               101.7 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           0.887476 |          2.5 |                 1.611289 |                      2.145953 |                         0.0 |                          0.0 |                      0.104558 |                           0.134343 |             1.508166 |                  1.736523 |     100.191834 |   1.226451 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.131208 |  0.938714 |        0.083928 |                0.071823 | -0.010417 | 17.111064 |  0.065039 |              0.063204 |                      nan |                  0.011738 |                  12.435463 |                  1.451305 |                     0.0 |                      0.0 |     0.002733 |             0.0 |              0.0 |         10.984159 |                  37.6 |                 100.0 |               202.699444 |              198.665144 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029878 |       0.0 | 0.009282 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.103899 | 0.028861 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.777864 | 76.522093 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.103899 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028861 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |             94.6 |       2.4 |               101.7 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           0.851977 |          2.4 |                 1.632319 |                      2.136456 |                         0.0 |                          0.0 |                      0.105752 |                           0.133824 |             1.544174 |                  1.726078 |     100.155826 |   1.225432 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            14.246552 |  2.455474 |        0.263445 |                0.226749 | -0.010417 | 16.859546 |  0.055286 |              0.199539 |                      nan |                  0.037057 |                  12.536966 |                  1.478175 |                     0.0 |                      0.0 |     0.137948 |             0.0 |              0.0 |         11.058791 |             39.166667 |                 100.0 |               199.052037 |              195.228329 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.056009 |       0.0 | 0.018144 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.101215 | 0.028115 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.712161 | 76.503949 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.101215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028115 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |             93.3 |       2.5 |               101.7 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           0.887476 |          2.5 |                 1.752342 |                      2.129774 |                         0.0 |                          0.0 |                      0.112529 |                           0.133458 |             1.634935 |                  1.717041 |     100.065065 |   1.220347 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            15.284647 |  4.650125 |        1.164115 |                1.028795 | -0.010417 | 16.640536 |  0.026796 |               0.90534 |                      nan |                  0.168134 |                  13.450489 |                   1.64464 |                     0.0 |                      0.0 |     0.836813 |             0.0 |              0.0 |         11.805849 |                  37.6 |                 100.0 |               191.292078 |              187.900211 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.234858 |       0.0 | 0.077952 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.098601 | 0.027389 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.674949 | 76.425997 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.098601 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027389 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |             86.9 |       2.5 |               101.7 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           0.887476 |          2.5 |                  1.91622 |                       2.12644 |                         0.0 |                          0.0 |                      0.121684 |                           0.133275 |             1.665196 |                  1.705582 |     100.034804 |   1.214317 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            16.438085 |  7.209034 |        1.933622 |                1.710811 | -0.010417 | 16.516495 | -0.006124 |              1.505514 |                      nan |                  0.279595 |                  14.465515 |                   1.83472 |                     0.0 |                      0.0 |     1.429067 |             0.0 |              0.0 |         12.630795 |                  37.6 |                 100.0 |               192.882798 |              189.404207 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424426 |       0.0 | 0.144453 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.096054 | 0.026682 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.670656 | 76.281544 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.096054 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026682 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |             78.9 |       2.9 |               101.7 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           1.029472 |          2.9 |                 2.093348 |                      2.122001 |                         0.0 |                          0.0 |                      0.131461 |                           0.133032 |             1.651651 |                  1.695656 |     100.048349 |   1.208543 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.360836 |  9.458797 |        2.537058 |                2.210424 | -0.010417 | 16.502185 | -0.036673 |              1.945173 |                      nan |                  0.361246 |                  15.277535 |                  1.989515 |                     0.0 |                      0.0 |     1.862277 |             0.0 |              0.0 |          13.28802 |             32.413793 |                 100.0 |               203.810564 |              199.711145 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.613278 |       0.0 | 0.189597 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.093572 | 0.025992 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.696912 | 76.091948 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.093572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025992 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |             75.8 |       3.1 |               101.7 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |            1.10047 |          3.1 |                 2.270468 |                      2.117571 |                         0.0 |                          0.0 |                      0.141127 |                            0.13279 |             1.721014 |                  1.691233 |      99.978986 |   1.202862 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.822211 | 10.747438 |         2.69052 |                2.317703 | -0.010417 | 16.589706 | -0.062862 |              2.039579 |                      nan |                  0.378779 |                  15.683546 |                   2.06621 |                     0.0 |                      0.0 |     1.953487 |             0.0 |              0.0 |         13.617336 |             30.322581 |                 100.0 |               207.855185 |              203.514925 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.685505 |       0.0 | 0.206707 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.091155 | 0.025321 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.749357 |  75.88524 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.091155 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025321 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |             73.4 |       3.3 |               101.8 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           1.171468 |          3.3 |                  2.38585 |                      2.111492 |                         0.0 |                          0.0 |                      0.147368 |                           0.132457 |             1.751214 |                  1.687786 |     100.048786 |   1.200635 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            17.937555 | 10.819272 |        2.515039 |                2.160947 | -0.010417 | 16.764525 | -0.076366 |              1.901633 |                      nan |                   0.35316 |                  15.785048 |                  2.087634 |                     0.0 |                      0.0 |     1.814648 |             0.0 |              0.0 |         13.697414 |             28.484848 |                 100.0 |               216.999554 |              212.092863 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.688399 |       0.0 | 0.197407 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.088801 | 0.024667 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.815307 | 75.687833 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088801 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024667 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |             68.3 |       3.3 |               101.8 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           1.171468 |          3.3 |                 2.490975 |                      2.109837 |                         0.0 |                          0.0 |                      0.153018 |                           0.132366 |             1.701336 |                  1.687166 |     100.098664 |   1.198001 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.052899 | 10.408076 |        2.605564 |                2.233851 | -0.010417 | 16.984357 | -0.086738 |              1.965789 |                      nan |                  0.365075 |                  15.886551 |                  2.106001 |                     0.0 |                      0.0 |     1.878039 |             0.0 |              0.0 |          13.78055 |             28.484848 |                 100.0 |               245.428784 |              238.568502 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.754061 |       0.0 | 0.202161 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.086507 |  0.02403 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.891628 | 75.485672 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.086507 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02403 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |             61.3 |       3.2 |               101.8 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           1.135969 |          3.2 |                 2.616045 |                      2.109837 |                         0.0 |                          0.0 |                      0.159698 |                           0.132366 |             1.603636 |                  1.682947 |     100.196364 |   1.195188 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             18.39893 | 10.917404 |         2.53785 |                2.157989 | -0.010417 | 17.238761 | -0.098523 |               1.89903 |                      nan |                  0.352677 |                  16.191059 |                  2.166369 |                     0.0 |                      0.0 |     1.808765 |             0.0 |              0.0 |          14.02469 |                29.375 |                 100.0 |               301.593569 |              290.032403 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.791859 |       0.0 | 0.191425 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.084272 | 0.023409 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.979734 | 75.294247 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.084272 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023409 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |             60.6 |       2.5 |               101.8 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           0.887476 |          2.5 |                 2.632058 |                      2.104328 |                         0.0 |                          0.0 |                       0.16055 |                           0.132064 |             1.595027 |                  1.676097 |     100.204973 |   1.194821 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            18.975649 | 10.494383 |         2.62717 |                2.208784 | -0.010417 | 17.532448 | -0.094339 |               1.94373 |                      nan |                  0.360978 |                  16.698571 |                  2.265756 |                     0.0 |                      0.0 |     1.849324 |             0.0 |              0.0 |         14.432816 |                  37.6 |                 100.0 |               310.085747 |              297.718286 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.744343 |       0.0 | 0.208898 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.082096 | 0.022804 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.063657 | 75.085349 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.082096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022804 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |             58.8 |       2.9 |               101.7 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           1.029472 |          2.9 |                 2.616045 |                      2.099381 |                         0.0 |                          0.0 |                      0.159698 |                           0.131792 |             1.538235 |                  1.669008 |     100.161765 |   1.194298 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            19.437025 |  8.874215 |        2.221576 |                1.861585 | -0.010417 | 17.812189 | -0.086137 |              1.638195 |                      nan |                  0.304236 |                  17.104582 |                  2.348437 |                     0.0 |                      0.0 |     1.540344 |             0.0 |              0.0 |         14.756144 |             32.413793 |                 100.0 |               324.688901 |              310.877147 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.707659 |       0.0 |  0.17376 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079975 | 0.022215 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.139377 | 74.911589 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079975 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022215 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |             57.2 |       3.6 |               101.7 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           1.277965 |          3.6 |                 2.680609 |                      2.097734 |                         0.0 |                          0.0 |                      0.163129 |                           0.131702 |             1.533308 |                   1.66429 |     100.166692 |   1.192703 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |              19.8984 |   6.45697 |        1.616442 |                1.360212 | -0.010417 |  18.06459 | -0.089325 |              1.196986 |                      nan |                  0.222297 |                  17.510592 |                  2.428674 |                     0.0 |                      0.0 |     1.095792 |             0.0 |              0.0 |         15.081918 |             26.111111 |                 100.0 |               351.951333 |              335.249533 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.668848 |       0.0 |  0.13417 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077909 | 0.021641 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.218285 | 74.777419 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077909 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021641 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |             57.2 |       4.2 |               101.7 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |            1.49096 |          4.2 |                 2.616045 |                      2.097185 |                         0.0 |                          0.0 |                      0.159698 |                           0.131672 |             1.496378 |                  1.659573 |     100.203622 |   1.194485 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            20.705807 |  3.945159 |        1.058179 |                0.907937 | -0.010417 | 18.327617 | -0.075003 |              0.798984 |                      nan |                  0.148383 |                   18.22111 |                  2.564427 |                     0.0 |                      0.0 |     0.692133 |             0.0 |              0.0 |         15.656683 |             22.380952 |                 100.0 |               341.504263 |              325.939578 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.608166 |       0.0 | 0.109849 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075897 | 0.021082 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.282872 |  74.66757 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075897 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021082 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |             59.9 |       4.2 |               101.7 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |            1.49096 |          4.2 |                 2.475714 |                      2.094444 |                         0.0 |                          0.0 |                        0.1522 |                           0.131522 |             1.482953 |                  1.651207 |     100.217047 |   1.198201 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |            21.743901 |  1.939642 |        0.520255 |                0.462876 | -0.010417 | 18.542906 | -0.050913 |              0.407331 |                      nan |                  0.075647 |                  19.134633 |                  2.744116 |                     0.0 |                      0.0 |     0.292993 |             0.0 |              0.0 |         16.390518 |             22.380952 |                 100.0 |               298.906719 |              287.595486 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461333 |       0.0 | 0.089872 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073936 | 0.020538 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.323368 | 74.577699 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073936 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020538 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |             66.1 |       3.6 |               101.7 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           1.277965 |          3.6 |                 2.298847 |                       2.08897 |                         0.0 |                          0.0 |                      0.142666 |                           0.131221 |             1.519538 |                   1.63958 |     100.180462 |   1.202947 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      1.0 |                     15.502754 |             22.89734 |   0.60531 |        0.162358 |                0.151461 | -0.010417 | 18.677895 | -0.022077 |              0.133286 |                      nan |                  0.024753 |                  20.149659 |                  2.950104 |                     0.0 |                      0.0 |     0.010365 |             0.0 |              0.0 |         17.199555 |             26.111111 |                 100.0 |               249.928887 |              242.732867 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.272979 |       0.0 | 0.065761 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.072026 | 0.020007 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.335029 | 74.511937 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.072026 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020007 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |             71.7 |       3.2 |               101.7 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           1.135969 |          3.2 |                 2.054233 |                      2.080783 |                         0.0 |                          0.0 |                      0.129312 |                           0.130772 |             1.472885 |                  1.624918 |     100.227115 |   1.210597 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                 0.104711 |                     15.502754 |            23.243371 |  0.005136 |        0.003091 |                0.002926 | -0.010417 | 18.716764 |  0.017642 |              0.002575 |                      nan |                  0.000478 |                  20.454167 |                  3.034376 |                     0.0 |                      0.0 |    -0.123857 |             0.0 |              0.0 |         17.419791 |                29.375 |                 100.0 |               232.156834 |               74.878388 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.160746 |       0.0 | 0.083418 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.070166 | 0.019491 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  32.30697 | 74.428519 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.070166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019491 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |             80.6 |       2.2 |               101.8 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           0.780979 |          2.2 |                 1.797863 |                       2.06991 |                         0.0 |                          0.0 |                      0.115083 |                           0.130174 |             1.449077 |                  1.610994 |     100.350923 |    1.22071 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 | 18.623235 |  0.060982 |                   0.0 |                      nan |                       0.0 |                  20.454167 |                  3.060678 |                     0.0 |                      0.0 |    -0.127528 |             0.0 |              0.0 |         17.393489 |             42.727273 |                 100.0 |               219.836505 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.056353 |       0.0 | 0.034084 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.068353 | 0.018987 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.235572 | 74.394435 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.068353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018987 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |             85.9 |       2.2 |               101.8 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           0.780979 |          2.2 |                 1.685952 |                      2.058007 |                         0.0 |                          0.0 |                      0.108789 |                            0.12952 |             1.448232 |                  1.598128 |     100.351768 |   1.224953 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.38524 |  0.077441 |                   0.0 |                      nan |                       0.0 |                  20.454167 |                  3.084324 |                     0.0 |                      0.0 |    -0.128514 |             0.0 |              0.0 |         17.369842 |             42.727273 |                 100.0 |               216.481047 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.030452 |       0.0 | 0.018184 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.066588 | 0.018497 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.147714 |  74.37625 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.066588 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018497 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |             89.3 |       2.1 |               101.8 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |            0.74548 |          2.1 |                 1.611289 |                      2.045628 |                         0.0 |                          0.0 |                      0.104558 |                           0.128839 |             1.438881 |                  1.587151 |     100.361119 |   1.227981 |         0.293016 | 4.601956 | 20.104711 |         0.972464 |                 35.762943 |                      0.0 |                     15.502754 |            23.243371 |       0.0 |             0.0 |                     0.0 | -0.010417 |  18.09238 |  0.086235 |                   0.0 |                      nan |                       0.0 |                  20.454167 |                  3.103639 |                     0.0 |                      0.0 |    -0.129318 |             0.0 |              0.0 |         17.350527 |             44.761905 |                 100.0 |               216.446595 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012642 |       0.0 | 0.007611 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064868 | 0.018019 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.051062 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018019 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |             91.4 |       1.8 |               101.8 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           0.638983 |          1.8 |                 1.539538 |                      2.032245 |                         0.0 |                          0.0 |                      0.100468 |                           0.128102 |             1.407137 |                  1.579647 |     100.392863 |   1.231127 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.770207 |  0.094396 |                   0.0 |                      nan |                       0.0 |                   20.36723 |                  3.124611 |                     0.0 |                      0.0 |    -0.130192 |             0.0 |              0.0 |         17.242619 |             52.222222 |                 100.0 |               219.889739 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000942 | -0.000942 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.063192 | 0.017553 | 0.000942 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.946249 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.063192 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017553 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |             93.1 |       1.8 |               101.8 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           0.638983 |          1.8 |                 1.480281 |                      2.018409 |                         0.0 |                          0.0 |                      0.097071 |                           0.127339 |             1.378142 |                  1.572172 |     100.421858 |   1.233843 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  17.42083 |   0.09981 |                   0.0 |                      nan |                       0.0 |                   20.36723 |                  3.135507 |                     0.0 |                      0.0 |    -0.130646 |             0.0 |              0.0 |         17.231723 |             52.222222 |                 100.0 |               223.677492 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007156 | -0.007156 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06156 |   0.0171 | 0.008098 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.836022 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06156 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0171 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |             94.6 |       1.6 |               101.8 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           0.567985 |          1.6 |                 1.432438 |                      2.003073 |                         0.0 |                          0.0 |                      0.094315 |                           0.126493 |             1.355086 |                  1.565068 |     100.444914 |   1.236111 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 | 17.053408 |  0.102674 |                   0.0 |                      nan |                       0.0 |                   20.36723 |                  3.144898 |                     0.0 |                      0.0 |    -0.131037 |             0.0 |              0.0 |         17.222331 |                 58.75 |                 100.0 |               227.175049 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  -0.01403 |  -0.01403 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05997 | 0.016658 | 0.022128 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.722932 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05997 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016658 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |             95.6 |       1.2 |               101.8 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.425988 |          1.2 |                 1.404387 |                       1.98784 |                         0.0 |                          0.0 |                      0.092693 |                           0.125652 |             1.342594 |                  1.558052 |     100.457406 |   1.237469 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            23.144579 |       0.0 |             0.0 |                     0.0 | -0.010417 |  16.67644 |  0.101011 |                   0.0 |                      nan |                       0.0 |                   20.36723 |                   3.15411 |                     0.0 |                      0.0 |    -0.131421 |             0.0 |              0.0 |          17.21312 |             78.333333 |                 100.0 |               229.328233 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019316 | -0.019316 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.058421 | 0.016228 | 0.041443 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.611504 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.058421 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016228 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |             96.2 |       1.9 |               101.8 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           0.674482 |          1.9 |                 1.376819 |                      1.974268 |                         0.0 |                          0.0 |                      0.091096 |                           0.124901 |               1.3245 |                  1.550047 |       100.4755 |   1.238856 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.366022 |                     15.438712 |            23.144579 |  0.063733 |         0.00566 |                0.005336 | -0.010417 | 16.305014 |  0.099468 |              0.004696 |                      nan |                  0.000872 |                   20.36723 |                  3.166293 |                     0.0 |                      0.0 |    -0.127233 |             0.0 |              0.0 |         17.200937 |             49.473684 |                 100.0 |               232.312366 |               93.057374 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.016434 | -0.016434 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056912 | 0.015809 | 0.057877 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.501619 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056912 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015809 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |             95.7 |       1.3 |               101.8 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           0.461488 |          1.3 |                 1.367736 |                      1.961295 |                         0.0 |                          0.0 |                      0.090568 |                           0.124183 |             1.308923 |                  1.541578 |     100.491077 |   1.239363 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.144579 |  0.908598 |        0.080684 |                0.076088 | -0.010417 | 15.938731 |  0.093717 |              0.066957 |                      nan |                  0.012435 |                   20.36723 |                  3.179808 |                     0.0 |                      0.0 |    -0.065535 |             0.0 |              0.0 |         17.187422 |             72.307692 |                 100.0 |               234.493634 |              228.419105 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005056 | -0.005056 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.055442 |   0.0154 | 0.062933 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.397486 |  74.36864 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.055442 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0154 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |             95.1 |       1.6 |               101.8 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           0.567985 |          1.6 |                 1.460978 |                      1.952516 |                         0.0 |                          0.0 |                       0.09596 |                           0.123697 |              1.38939 |                  1.535084 |      100.41061 |   1.234655 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  2.420659 |        0.472904 |                 0.45318 | -0.010417 |  15.59162 |  0.064619 |              0.398799 |                      nan |                  0.074063 |                   20.87339 |                  3.288825 |                     0.0 |                      0.0 |     0.261764 |             0.0 |              0.0 |         17.584565 |                 58.75 |                 100.0 |               222.635333 |              217.364459 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.07078 |  0.062933 | 0.002912 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054009 | 0.015003 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.32245 | 74.365727 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054009 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015003 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |             96.3 |       1.9 |               101.9 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           0.674482 |          1.9 |                 1.685952 |                      1.949426 |                         0.0 |                          0.0 |                      0.108789 |                           0.123526 |             1.623571 |                  1.535091 |     100.276429 |   1.225361 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  4.619722 |        1.148656 |                1.101438 | -0.010417 | 15.341501 |  0.011696 |              0.969265 |                      nan |                  0.180006 |                   20.87339 |                  3.287363 |                     0.0 |                      0.0 |     0.832292 |             0.0 |              0.0 |         17.586027 |             49.473684 |                 100.0 |               196.965142 |              193.259739 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.207904 |       0.0 | 0.078639 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052614 | 0.014615 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.300337 | 74.287088 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014615 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |             89.9 |       3.2 |               101.9 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           1.135969 |          3.2 |                 1.953031 |                       1.95097 |                         0.0 |                          0.0 |                      0.123726 |                           0.123612 |             1.755775 |                  1.538746 |     100.144225 |   1.215051 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 |  7.198198 |        1.917614 |                1.839976 | -0.010417 | 15.267791 | -0.039576 |              1.619179 |                      nan |                  0.300705 |                   20.87339 |                  3.279458 |                     0.0 |                      0.0 |     1.482535 |             0.0 |              0.0 |         17.593932 |                29.375 |                 100.0 |               189.864158 |              186.549347 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.442855 |       0.0 | 0.131304 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.051255 | 0.014238 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.329496 | 74.155785 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.051255 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014238 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |             79.2 |       4.0 |               101.9 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           1.419962 |          4.0 |                 2.256393 |                      1.957159 |                         0.0 |                          0.0 |                      0.140363 |                           0.123954 |             1.787063 |                  1.543872 |     100.112937 |   1.205358 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.604725 |  9.475573 |        2.356024 |                2.266269 | -0.010417 | 15.364988 | -0.087156 |              1.994317 |                      nan |                  0.370373 |                  20.772158 |                  3.250753 |                     0.0 |                      0.0 |     1.858869 |             0.0 |              0.0 |         17.521405 |                  23.5 |                 100.0 |               202.881364 |              198.836433 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.679013 |       0.0 | 0.175093 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049931 |  0.01387 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.406236 | 73.980691 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049931 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01387 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |             73.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |           1.632956 |          4.6 |                  2.43042 |                      1.962848 |                         0.0 |                          0.0 |                      0.149767 |                           0.124269 |             1.783928 |                  1.546397 |     100.116072 |   1.200448 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.719761 | 10.786309 |        2.873495 |                 2.75694 | -0.010417 | 15.620786 | -0.107551 |              2.426107 |                      nan |                  0.450563 |                   20.87339 |                  3.266966 |                     0.0 |                      0.0 |     2.289983 |             0.0 |              0.0 |         17.606424 |             20.434783 |                 100.0 |               224.441968 |              219.051923 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.901883 |       0.0 | 0.201952 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048642 | 0.013512 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  31.50337 |  73.77874 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048642 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013512 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |             67.3 |       4.1 |               101.9 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           1.455461 |          4.1 |                 2.552842 |                      1.968551 |                         0.0 |                          0.0 |                      0.156328 |                           0.124585 |             1.718063 |                  1.545886 |     100.181937 |   1.197482 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            23.949834 | 10.865527 |        2.894599 |                2.761725 | -0.010417 | 15.944568 | -0.117837 |              2.430318 |                      nan |                  0.451345 |                  21.075854 |                  3.309837 |                     0.0 |                      0.0 |     2.292408 |             0.0 |              0.0 |         17.766017 |             22.926829 |                 100.0 |               261.395836 |              253.311837 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.942976 |       0.0 | 0.209722 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047385 | 0.013163 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.610791 | 73.569018 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013163 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |             62.2 |       4.1 |               101.9 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           1.455461 |          4.1 |                 2.729939 |                      1.976351 |                         0.0 |                          0.0 |                      0.165743 |                           0.125016 |             1.698022 |                  1.546988 |     100.201978 |   1.193116 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |             24.06487 |  10.45226 |        2.784503 |                 2.65055 | -0.010417 | 16.302636 | -0.133863 |              2.332484 |                      nan |                  0.433176 |                  21.177086 |                  3.330376 |                     0.0 |                      0.0 |     2.193718 |             0.0 |              0.0 |         17.846709 |             22.926829 |                 100.0 |               315.484528 |              302.591605 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.983393 |       0.0 | 0.196078 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046161 | 0.012823 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.734237 | 73.372939 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046161 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012823 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |             59.6 |       4.4 |               101.9 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           1.561958 |          4.4 |                 2.830978 |                      1.983133 |                         0.0 |                          0.0 |                      0.171077 |                           0.125391 |             1.687263 |                  1.550892 |     100.212737 |   1.190747 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.294943 |  10.96314 |        2.920603 |                2.764822 | -0.010417 | 16.714124 | -0.137935 |              2.433044 |                      nan |                  0.451851 |                   21.37955 |                  3.363175 |                     0.0 |                      0.0 |     2.292911 |             0.0 |              0.0 |         18.016374 |             21.363636 |                 100.0 |               357.583565 |              340.253538 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.076015 |       0.0 | 0.188119 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044969 | 0.012491 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.861756 | 73.184821 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044969 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012491 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |             55.4 |       4.6 |               101.9 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |           1.632956 |          4.6 |                 2.882708 |                      1.990983 |                         0.0 |                          0.0 |                      0.173797 |                           0.125825 |              1.59702 |                  1.552718 |      100.30298 |   1.189943 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.409979 | 10.528069 |        2.804699 |                2.649346 | -0.010417 | 17.139185 | -0.135234 |              2.331424 |                      nan |                  0.432979 |                  21.480782 |                   3.38194 |                     0.0 |                      0.0 |      2.19051 |             0.0 |              0.0 |         18.098841 |             20.434783 |                 100.0 |               429.701201 |              403.402315 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.117473 |       0.0 | 0.165048 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043807 | 0.012169 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 31.986572 | 73.019773 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043807 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012169 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |             52.1 |       4.5 |               101.9 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           1.597457 |          4.5 |                 2.952961 |                      2.001493 |                         0.0 |                          0.0 |                      0.177481 |                           0.126406 |             1.538493 |                  1.555327 |     100.361507 |     1.1886 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.525016 |  8.882849 |        2.366409 |                 2.23335 | -0.010417 | 17.555242 | -0.134887 |              1.965348 |                      nan |                  0.364993 |                  21.582014 |                  3.400029 |                     0.0 |                      0.0 |      1.82368 |             0.0 |              0.0 |         18.181985 |             20.888889 |                 100.0 |               525.329939 |              484.576285 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.044793 |       0.0 | 0.135822 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042676 | 0.011854 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.111043 | 72.883951 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042676 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011854 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |             49.5 |       4.7 |               101.8 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           1.668455 |          4.7 |                 2.988644 |                      2.010993 |                         0.0 |                          0.0 |                      0.179348 |                            0.12693 |             1.479379 |                  1.556257 |     100.320621 |   1.186889 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  6.439028 |         1.71537 |                1.620343 | -0.010417 | 17.970142 | -0.130245 |              1.425902 |                      nan |                   0.26481 |                  21.683246 |                   3.42173 |                     0.0 |                      0.0 |      1.28333 |             0.0 |              0.0 |         18.261515 |                  20.0 |                 100.0 |               627.911114 |              568.561649 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.943159 |       0.0 | 0.103418 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041573 | 0.011548 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.230871 | 72.780532 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041573 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011548 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |             48.1 |       4.5 |               101.8 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           1.597457 |          4.5 |                 2.988644 |                      2.022657 |                         0.0 |                          0.0 |                      0.179348 |                           0.127573 |             1.437538 |                  1.557615 |     100.362462 |   1.187075 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  3.911896 |        1.042137 |                0.985011 | -0.010417 |  18.36957 | -0.121617 |               0.86681 |                      nan |                  0.160979 |                  21.683246 |                  3.423912 |                     0.0 |                      0.0 |     0.724147 |             0.0 |              0.0 |         18.259334 |             20.888889 |                 100.0 |               687.355785 |              615.833261 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.773152 |       0.0 | 0.081869 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040499 |  0.01125 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.342072 | 72.698663 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040499 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01125 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |             49.4 |       4.1 |               101.8 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           1.455461 |          4.1 |                  2.93526 |                      2.037589 |                         0.0 |                          0.0 |                      0.176554 |                           0.128396 |             1.450018 |                  1.560199 |     100.349982 |   1.188219 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  1.905864 |        0.507726 |                0.480125 | -0.010417 | 18.740238 | -0.107131 |               0.42251 |                      nan |                  0.078466 |                  21.683246 |                   3.42458 |                     0.0 |                      0.0 |     0.279819 |             0.0 |              0.0 |         18.258666 |             22.926829 |                 100.0 |               600.040442 |              546.050102 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.580608 |       0.0 | 0.073132 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039453 | 0.010959 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.438786 | 72.625531 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039453 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010959 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |             57.0 |       3.8 |               101.8 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           1.348964 |          3.8 |                 2.780058 |                      2.054233 |                         0.0 |                          0.0 |                      0.168392 |                           0.129312 |             1.584633 |                  1.565154 |     100.215367 |   1.191234 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      1.0 |                     15.438712 |            24.640052 |  0.576079 |        0.153469 |                0.145175 | -0.010417 | 19.062619 | -0.080727 |              0.127754 |                      nan |                  0.023726 |                  21.683246 |                  3.420162 |                     0.0 |                      0.0 |    -0.014752 |             0.0 |              0.0 |         18.263084 |             24.736842 |                 100.0 |               384.261969 |              363.812822 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.38361 |       0.0 | 0.070385 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.038434 | 0.010676 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.509096 | 72.555146 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.038434 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010676 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |             65.9 |       2.5 |               101.8 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           0.887476 |          2.5 |                 2.537251 |                      2.072623 |                         0.0 |                          0.0 |                      0.155495 |                           0.130323 |             1.672048 |                  1.574157 |     100.127952 |   1.196911 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                 0.072689 |                     15.438712 |            24.870125 |  0.002495 |        0.003269 |                0.003122 | -0.010417 | 19.296988 | -0.043265 |              0.002747 |                      nan |                   0.00051 |                   21.88571 |                  3.445356 |                     0.0 |                      0.0 |    -0.140809 |             0.0 |              0.0 |         18.440354 |                  37.6 |                 100.0 |               272.690342 |                73.37327 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.163075 |       0.0 | 0.099136 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.037441 |   0.0104 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.541945 | 72.456009 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037441 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0104 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |             73.0 |       1.9 |               101.8 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           0.674482 |          1.9 |                 2.242395 |                      2.091705 |                         0.0 |                          0.0 |                      0.139602 |                           0.131371 |             1.636948 |                  1.582026 |     100.163052 |   1.205255 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.406483 |    0.0023 |                   0.0 |                      nan |                       0.0 |                   21.88571 |                  3.434698 |                     0.0 |                      0.0 |    -0.143112 |             0.0 |              0.0 |         18.451012 |             49.473684 |                 100.0 |               228.581729 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.080802 |       0.0 | 0.053476 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.036474 | 0.010132 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.529228 | 72.402533 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036474 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010132 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |             76.7 |       2.2 |               101.8 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           0.780979 |          2.2 |                  2.10653 |                       2.11094 |                         0.0 |                          0.0 |                      0.132185 |                           0.132426 |             1.615709 |                  1.588483 |     100.184291 |   1.209486 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.364094 |  0.022984 |                   0.0 |                      nan |                       0.0 |                   21.88571 |                  3.427556 |                     0.0 |                      0.0 |    -0.142815 |             0.0 |              0.0 |         18.458154 |             42.727273 |                 100.0 |               220.437523 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.076279 |       0.0 | 0.047612 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035532 |  0.00987 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.495827 | 72.354921 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.035532 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00987 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |             81.8 |       2.7 |               101.8 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           0.958474 |          2.7 |                 1.990459 |                      2.129218 |                         0.0 |                          0.0 |                      0.125796 |                           0.133428 |             1.628195 |                  1.595583 |     100.171805 |   1.213176 |         0.288036 | 4.633978 | 20.072689 |          0.97278 |                 35.520271 |                      0.0 |                     15.438712 |            24.870125 |       0.0 |             0.0 |                     0.0 | -0.010417 | 19.252757 |   0.04002 |                   0.0 |                      nan |                       0.0 |                   21.88571 |                  3.418333 |                     0.0 |                      0.0 |    -0.142431 |             0.0 |              0.0 |         18.467377 |             34.814815 |                 100.0 |               211.424945 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.067295 |       0.0 | 0.038511 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.034614 | 0.009615 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 32.445391 |  72.31641 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009615 |

.. _lland_v4_acker_winter_hourly:

acre (winter)
-------------

>>> pub.timegrids = "2010-12-10", "2010-12-13", "1h"

>>> inputs.nied.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.3, 0.2, 0.4, 0.7, 0.5, 1.0, 2.0, 2.1, 1.8, 1.3,
...     0.5, 0.2, 0.5, 0.2, 0.1, 0.3, 0.1, 0.1, 0.0, 0.1, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.1, 0.1, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.teml.series = (
...     -3.7, -4.8, -5.6, -6.2, -7.3, -8.3, -8.6, -8.5, -9.5, -9.2, -6.9,
...     -4.0, -2.8, -1.4, -0.4, 0.5, 1.0, 1.1, 1.0, 1.4, 1.6, 2.0, 2.4, 2.5,
...     2.7, 3.2, 3.9, 4.4, 4.8, 4.8, 5.1, 5.1, 5.4, 5.3, 5.5, 6.1, 6.4, 6.4,
...     6.5, 6.3, 5.9, 5.5, 5.1, 4.8, 4.9, 4.9, 4.7, 3.8, 3.9, 3.7, 3.5, 3.3,
...     3.3, 3.0, 1.9, 1.2, 0.8, 0.8, 0.5, 0.3, -0.1, -0.4, -0.6, -0.8, -1.3,
...     -1.6, -2.0, -2.1, -2.2, -2.6, -3.1, -3.9)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.2, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0)
>>> inputs.windspeed.series = (
...     3.4, 2.9, 2.5, 3.0, 2.5, 2.3, 2.1, 1.4, 1.1, 1.0, 1.4, 1.3, 1.3, 1.4,
...     2.3, 2.8, 3.1, 3.5, 4.4, 3.8, 5.7, 5.6, 5.8, 6.2, 5.6, 5.2, 5.7, 6.6,
...     7.0, 7.3, 7.2, 7.6, 7.3, 7.1, 7.1, 8.9, 9.5, 7.9, 9.6, 9.1, 8.7, 8.9,
...     7.9, 7.1, 6.4, 7.2, 7.0, 6.8, 6.4, 6.3, 5.7, 5.2, 4.6, 4.9, 6.3, 5.8,
...     6.1, 5.9, 6.4, 6.8, 7.0, 6.3, 7.3, 7.5, 7.0, 6.0, 4.8, 4.2, 4.9, 4.0,
...     3.4, 3.4)
>>> inputs.relativehumidity.series = (
...     87.5, 90.7, 89.7, 91.3, 94.2, 93.2, 96.5, 95.9, 96.8, 94.3, 96.7,
...     97.4, 93.9, 93.4, 93.3, 94.1, 94.6, 94.4, 96.9, 98.3, 98.7, 99.6,
...     99.4, 99.2, 99.2, 99.6, 99.8, 99.8, 99.0, 97.6, 97.4, 96.3, 95.2,
...     96.3, 95.5, 91.6, 90.0, 85.7, 82.3, 82.5, 82.2, 76.5, 81.7, 83.4,
...     82.8, 85.8, 87.6, 85.6, 85.3, 86.2, 89.3, 91.4, 90.9, 89.2, 85.6,
...     86.3, 91.2, 87.5, 84.5, 76.7, 74.8, 69.7, 66.9, 64.4, 65.4, 67.6,
...     70.2, 70.4, 69.5, 71.6, 74.5, 80.2)
>>> inputs.atmosphericpressure.series = (
...     101.7, 101.8, 101.8, 101.9, 102.0, 102.0, 102.1, 102.1, 102.2, 102.2,
...     102.2, 102.2, 102.2, 102.1, 102.0, 102.0, 101.9, 101.8, 101.6, 101.5,
...     101.4, 101.3, 101.1, 101.0, 100.8, 100.7, 100.6, 100.5, 100.5, 100.4,
...     100.4, 100.4, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2, 100.2, 100.3,
...     100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.3, 100.2, 100.2, 100.2,
...     100.2, 100.3, 100.4, 100.5, 100.6, 100.8, 100.9, 101.0, 101.1, 101.2,
...     101.3, 101.4, 101.5, 101.5, 101.7, 101.7, 101.8, 101.9, 101.9, 102.0,
...     102.0, 102.0)

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.0, 0.04, 0.25, 0.59, 0.91, 0.97, 1.0,
...       0.65, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...       0.0, 0.1, 0.2, 0.4, 0.6, 0.5, 0.2, 0.0,
...       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [-2.8, -2.6, -2.2, -1.8, -1.8, -2.4, -1.5, -1.4,
...       -0.8, -0.8, -0.9, 0.9, -0.8, -1.0, -1.2, -1.8,
...       -2.8, -3.8, -4.3, -4, -3.9, -1.6, -1.6, 2.6]),
...     (logs.loggedrelativehumidity,
...      [99.4, 99.0, 99.6, 99.6, 99.3, 99.3, 99.8, 99.7,
...       100.0, 99.6, 99.3, 98.9, 96.6, 93.0, 91.4, 92.0,
...       94.5, 96.5, 98.2, 96.9, 98.0, 95.0, 91.8, 88.5]),
...     (logs.loggedwindspeed2m,
...      [2.2, 2.0, 2.5, 2.6, 2.3, 1.7, 2.8, 1.9,
...       2.6, 2.9, 3.5, 3.4, 3.5, 3.0, 3.9, 3.6,
...       2.2, 2.4, 2.5, 1.5, 2.5, 3.0, 3.2, 3.1]))

.. integration-test::

    >>> lnk(ACKER)
    >>> test("lland_v4_acker_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |      evi |      evb |       evs |    wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr | wlatinz |  wlatsnow | wsensinz | wsenssnow | wsurfinz |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                         0.0 |                     0.411372 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.337255 | -0.000889 |                   0.0 |                      0.0 |                       0.0 |                   0.705458 |                  4.014396 |                     0.0 |                 0.266912 |    -0.167266 |             0.0 |        -0.266912 |         -3.308938 |             62.648725 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.001835 |      0.0 |      0.0 |  0.000772 |      0.0 |      -5.2961 |     0.754325 |      0.0 |      0.0 | 0.698543 | 0.698543 |     0.0 |   0.00206 |      0.0 | -0.036548 |      0.0 |  0.232424 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.321403 | 20.349045 |       0.0 | 0.860328 | -1.569388 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                         0.0 |                     0.395912 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.231292 |  0.012212 |                   0.0 |                      0.0 |                       0.0 |                    0.70736 |                  4.024777 |                     0.0 |                 0.276546 |    -0.167699 |             0.0 |        -0.276546 |         -3.317416 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004543 |      0.0 |      0.0 |  0.001145 |      0.0 |    -5.796677 |     0.753662 |      0.0 |      0.0 | 0.739119 | 0.027642 |     0.0 |  0.003053 |      0.0 | -0.020521 |      0.0 |  0.259078 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |   20.3479 |   20.3479 | -0.237633 | 0.867636 |   -1.5507 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                         0.0 |                     0.323756 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 |    -5.169 |  0.069583 |                   0.0 |                      0.0 |                       0.0 |                   0.708045 |                  4.040774 |                     0.0 |                 0.253201 |    -0.168366 |             0.0 |        -0.253201 |         -3.332729 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004605 |      0.0 |      0.0 | -0.004547 |      0.0 |    -8.390436 |     0.753424 |      0.0 |      0.0 | 1.068326 |      0.0 |     0.0 | -0.012129 |      0.0 | -0.052301 |      0.0 |  0.188771 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.352447 | 20.352447 | -0.356821 | 0.870279 | -1.589383 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                         0.0 |                     0.292658 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.297943 |  0.094441 |                   0.0 |                      0.0 |                       0.0 |                   0.708596 |                  4.052814 |                     0.0 |                  0.24342 |    -0.168867 |             0.0 |         -0.24342 |         -3.344217 |             71.001889 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004915 |      0.0 |      0.0 | -0.008026 |      0.0 |    -9.670218 |     0.753232 |      0.0 |      0.0 | 1.231552 |      0.0 |     0.0 | -0.021408 |      0.0 | -0.073053 |      0.0 |  0.148958 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.360473 | 20.360473 | -0.411338 | 0.872414 | -1.652924 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                         0.0 |                     0.268567 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.509747 |  0.113135 |                   0.0 |                      0.0 |                       0.0 |                   0.709099 |                  4.064535 |                     0.0 |                 0.244706 |    -0.169356 |             0.0 |        -0.244706 |         -3.355436 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008645 |      0.0 |      0.0 | -0.007855 |      0.0 |    -10.74747 |     0.753057 |      0.0 |      0.0 | 1.369285 |      0.0 |     0.0 | -0.020953 |      0.0 | -0.064616 |      0.0 |  0.159138 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.368328 | 20.368328 | -0.457341 | 0.874366 | -1.735159 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                         0.0 |                      0.24802 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.783863 |  0.128575 |                   0.0 |                      0.0 |                       0.0 |                   0.709565 |                  4.080005 |                     0.0 |                 0.249609 |        -0.17 |             0.0 |        -0.249609 |          -3.37044 |             92.611159 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007726 |      0.0 |      0.0 | -0.006569 |      0.0 |   -11.736425 |     0.752895 |      0.0 |      0.0 |  1.49586 |      0.0 |     0.0 | -0.017522 |      0.0 | -0.061235 |      0.0 |  0.170851 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.374897 | 20.374897 | -0.499617 | 0.876177 | -1.832834 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                         0.0 |                     0.236559 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -6.109447 |  0.134151 |                   0.0 |                      0.0 |                       0.0 |                       0.71 |                  4.093695 |                     0.0 |                 0.243066 |    -0.170571 |             0.0 |        -0.243066 |         -3.383696 |             101.43127 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.010331 |      0.0 |      0.0 | -0.007889 |      0.0 |   -12.320119 |     0.752743 |      0.0 |      0.0 | 1.570761 |      0.0 |     0.0 | -0.021044 |      0.0 | -0.062855 |      0.0 |  0.159167 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.382786 | 20.382786 | -0.524634 | 0.877869 | -1.936085 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                         0.0 |                     0.229922 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 | -6.453616 |  0.134268 |                   0.0 |                      0.0 |                       0.0 |                   0.653538 |                  3.682358 |                     0.0 |                 0.225074 |    -0.153432 |             0.0 |        -0.225074 |          -3.02882 |            152.146905 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008377 |      0.0 |      0.0 | -0.007023 |      0.0 |   -12.669732 |     0.752598 |      0.0 |      0.0 | 1.615961 |      0.0 |     0.0 | -0.018734 |      0.0 | -0.056975 |      0.0 |  0.149365 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.389809 | 20.389809 | -0.539731 | 0.879495 | -2.039453 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                         0.0 |                     0.221928 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 | -6.798176 |  0.136184 |              0.000987 |                      0.0 |                  0.000987 |                   0.566356 |                  3.039527 |                     0.0 |                 0.214153 |     -0.12566 |             0.0 |        -0.213166 |         -2.473171 |            193.641515 |                 100.0 |               112.102145 |               99.616909 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005975 |      0.0 |      0.0 | -0.005283 |      0.0 |   -13.102969 |     0.752457 |      0.0 |      0.0 | 1.671794 |      0.0 |     0.0 | -0.014093 |      0.0 |  -0.04424 |      0.0 |  0.154832 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.395093 | 20.395093 | -0.558379 | 0.881082 | -2.144736 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                         0.0 |                     0.225852 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 | -7.149121 |  0.123972 |              0.016641 |                      0.0 |                  0.016641 |                   0.481694 |                  2.412046 |                     0.0 |                 0.187811 |    -0.083861 |             0.0 |         -0.17117 |         -1.930352 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001957 |      0.0 |      0.0 | -0.004725 |      0.0 |   -12.888585 |      0.75232 |      0.0 |      0.0 | 1.644867 |      0.0 |     0.0 | -0.012603 |      0.0 | -0.043589 |      0.0 |  0.114979 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.399817 | 20.399817 | -0.549386 | 0.882622 | -2.237809 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                         0.0 |                     0.260619 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -7.459363 |  0.079109 |              0.050199 |                      0.0 |                  0.050199 |                     0.4022 |                  1.815487 |                     0.0 |                 0.133683 |    -0.025446 |             0.0 |        -0.083484 |         -1.413287 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.002329 |      0.0 |      0.0 | -0.008206 |      0.0 |   -11.121795 |     0.752181 |      0.0 |      0.0 | 1.419714 |      0.0 |     0.0 |  -0.02189 |      0.0 | -0.057687 |      0.0 |  0.003908 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.408023 | 20.408023 | -0.474185 | 0.884185 | -2.286017 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                         0.0 |                     0.321537 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -7.620058 |  0.018533 |               0.09567 |                      0.0 |                   0.09567 |                   0.359477 |                  1.493002 |                     0.0 |                 0.088769 |     0.033462 |             0.0 |         0.006901 |         -1.133525 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008285 |      0.0 |      0.0 | -0.010503 |      0.0 |    -8.478066 |     0.752025 |      0.0 |      0.0 | 1.082674 |      0.0 |     0.0 | -0.028017 |      0.0 | -0.059121 |      0.0 | -0.094039 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.418526 | 20.418526 | -0.361613 |  0.88595 |  -2.27365 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                         0.0 |                     0.363399 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -7.578834 | -0.014477 |              0.081721 |                      0.0 |                  0.081721 |                   0.337806 |                  1.331032 |                     0.0 |                 0.078636 |     0.026261 |             0.0 |         0.003085 |         -0.993227 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010258 |      0.0 |      0.0 | -0.008974 |      0.0 |    -6.908605 |     0.751837 |      0.0 |      0.0 | 0.882703 |      0.0 |     0.0 | -0.023939 |      0.0 | -0.054243 |      0.0 | -0.081267 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.427501 | 20.427501 | -0.294823 | 0.888072 | -2.228273 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                         0.0 |                     0.401002 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -7.427578 | -0.038828 |              0.071957 |                      0.0 |                  0.071957 |                   0.334577 |                  1.306141 |                     0.0 |                 0.063829 |     0.017534 |             0.0 |         0.008128 |         -0.971563 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010713 |      0.0 |      0.0 | -0.010257 |      0.0 |    -5.629984 |     0.751627 |      0.0 |      0.0 | 0.719651 |      0.0 |     0.0 | -0.027361 |      0.0 | -0.057799 |      0.0 | -0.093287 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.437758 | 20.437758 | -0.240364 |  0.89045 | -2.158545 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                         0.0 |                     0.432355 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -7.195151 | -0.055133 |              0.038592 |                      0.0 |                  0.038592 |                    0.33489 |                  1.307301 |                     0.0 |                 0.054985 |    -0.015879 |             0.0 |        -0.016393 |         -0.972411 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01036 |      0.0 |      0.0 | -0.014237 |      0.0 |     -4.64268 |     0.751395 |      0.0 |      0.0 | 0.593747 |      0.0 |     0.0 | -0.037976 |      0.0 | -0.075602 |      0.0 | -0.097185 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.451995 | 20.451995 | -0.198312 | 0.893084 | -2.072512 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                         0.0 |                      0.45959 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -6.908373 |  -0.06638 |              0.009246 |                      0.0 |                  0.009246 |                   0.335231 |                  1.307629 |                     0.0 |                  0.04409 |    -0.045239 |             0.0 |        -0.034844 |         -0.972398 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007076 |      0.0 |      0.0 | -0.018122 |      0.0 |    -3.835233 |     0.751142 |      0.0 |      0.0 | 0.490826 |      0.0 |     0.0 |  -0.04834 |      0.0 |  -0.08726 |      0.0 | -0.100756 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.470117 | 20.470117 | -0.163936 | 0.895969 | -1.975232 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                         0.0 |                       0.4832 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 | -6.584107 |  -0.07377 |              0.000025 |                      0.0 |                  0.000025 |                   0.335355 |                   1.30686 |                     0.0 |                  0.04131 |    -0.054427 |             0.0 |        -0.041284 |         -0.971505 |             68.711505 |                 100.0 |               112.102145 |               98.953162 |    0.094 |  0.0235 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005889 |      0.0 |      0.0 | -0.019606 | 0.000344 |    -3.168837 |     0.751049 |      0.0 |      0.0 | 0.409167 |   0.0705 |     0.0 | -0.052299 |      0.0 | -0.089685 |      0.0 | -0.100699 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.583723 | 20.583723 | -0.113115 | 0.897026 | -1.870562 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                         0.0 |                     0.506673 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -6.235208 | -0.079941 |                   0.0 |                      0.0 |                       0.0 |                   0.335794 |                  1.305996 |                     0.0 |                  0.04931 |    -0.054417 |             0.0 |         -0.04931 |         -0.970202 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007388 |      0.0 |      0.0 | -0.018162 |      0.0 |    -2.534239 |     0.750724 |      0.0 |      0.0 | 0.326416 |      0.0 |     0.0 | -0.048447 |      0.0 | -0.084896 |      0.0 | -0.084033 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.601886 | 20.601886 | -0.109023 | 0.900755 | -1.759721 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                         0.0 |                     0.519787 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.865738 | -0.079376 |                   0.0 |                      0.0 |                       0.0 |                     0.3355 |                  1.303739 |                     0.0 |                 0.053522 |    -0.054322 |             0.0 |        -0.053522 |         -0.968239 |             48.410379 |                 100.0 |               112.102145 |               98.814229 |    0.282 |  0.0705 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.003845 |      0.0 |      0.0 | -0.021177 | 0.001033 |    -2.190931 |     0.750942 |      0.0 |      0.0 | 0.289221 |   0.2115 |     0.0 | -0.056488 |      0.0 |   -0.0878 |      0.0 | -0.090766 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.905062 | 20.905062 | -0.025959 |  0.89825 | -1.649446 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                         0.0 |                     0.555528 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.498152 | -0.090836 |                   0.0 |                      0.0 |                       0.0 |                   0.335912 |                  1.301296 |                     0.0 |                  0.05576 |    -0.054221 |             0.0 |         -0.05576 |         -0.965383 |             56.054123 |                 100.0 |               112.102145 |               98.814229 |    0.188 |  0.0282 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000004 |      0.0 |      0.0 | -0.017799 | 0.001019 |    -1.292768 |     0.750636 |      0.0 |      0.0 | 0.171929 |   0.1598 |     0.0 | -0.047478 |      0.0 | -0.066633 |      0.0 | -0.058352 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.110861 | 21.110861 | -0.004051 | 0.901761 | -1.527709 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                         0.0 |                     0.583215 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -5.092364 | -0.096356 |                   0.0 |                      0.0 |                       0.0 |                    0.33636 |                  1.298642 |                     0.0 |                 0.060864 |     -0.05411 |             0.0 |        -0.060864 |         -0.962282 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    0.376 |  0.0376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000346 |      0.0 |      0.0 | -0.020725 | 0.002393 |    -0.631424 |     0.750304 |      0.0 |      0.0 | 0.086239 | 0.086239 |     0.0 | -0.055282 |      0.0 | -0.074793 |      0.0 | -0.069211 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.255182 | 21.507586 |       0.0 | 0.905583 | -1.400453 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                         0.0 |                     0.600098 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -4.668177 | -0.095613 |                   0.0 |                      0.0 |                       0.0 |                   0.337214 |                  1.295068 |                     0.0 |                 0.055526 |    -0.053961 |             0.0 |        -0.055526 |         -0.957854 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |    0.658 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002532 |      0.0 |      0.0 | -0.022446 |  0.00551 |    -0.241637 |      0.74967 |      0.0 |      0.0 | 0.034896 | 0.034896 |     0.0 | -0.059874 |      0.0 |   -0.0741 |      0.0 | -0.078448 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.311637 | 22.188032 |       0.0 | 0.912919 |  -1.27394 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                         0.0 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -4.246466 | -0.091724 |                   0.0 |                      0.0 |                       0.0 |                   0.338061 |                  1.290674 |                     0.0 |                 0.049365 |    -0.053778 |             0.0 |        -0.049365 |         -0.952613 |             36.725115 |                 100.0 |               112.102145 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001838 |      0.0 |      0.0 | -0.024931 | 0.004723 |          0.0 |     0.749041 | 0.034996 | 0.034996 |      0.0 |      0.0 |     0.0 | -0.066503 |      0.0 | -0.081551 |      0.0 |  -0.09869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.300053 | 22.682964 |       0.0 |  0.92025 | -1.151316 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                         0.0 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -3.83772 | -0.082895 |                   0.0 |                      0.0 |                       0.0 |                   0.338903 |                   1.28612 |                     0.0 |                 0.047044 |    -0.053588 |             0.0 |        -0.047044 |         -0.947217 |             34.355753 |                 100.0 |               112.102145 |               98.814229 |     0.94 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000987 |      0.0 |      0.0 | -0.027164 | 0.009839 |          0.0 |     0.748415 | 0.125524 | 0.125524 |      0.0 |      0.0 |     0.0 | -0.072459 |      0.0 | -0.089566 |      0.0 | -0.114981 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.198878 | 23.650128 |       0.0 | 0.927602 | -1.037521 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                         0.0 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -3.458404 | -0.074702 |                   0.0 |                      0.0 |                       0.0 |                   0.337412 |                  1.277745 |                     0.0 |                 0.041776 |    -0.053239 |             0.0 |        -0.041776 |         -0.940333 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |     1.88 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001175 |      0.0 |      0.0 | -0.027338 | 0.021252 |          0.0 |     0.747791 |  0.20045 |  0.20045 |      0.0 |      0.0 |     0.0 | -0.072924 |      0.0 | -0.089252 |      0.0 |   -0.1204 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.020914 | 25.557466 |       0.0 | 0.934993 |  -0.93192 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                         0.0 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -3.106399 | -0.067098 |                   0.0 |                      0.0 |                       0.0 |                   0.338241 |                  1.268373 |                     0.0 |                 0.027852 |    -0.052849 |             0.0 |        -0.027852 |         -0.930132 |             40.962628 |                 100.0 |               112.102145 |               98.814229 |    1.974 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002517 |      0.0 |      0.0 | -0.031933 | 0.026447 |          0.0 |     0.747171 | 0.348941 | 0.348941 |      0.0 |      0.0 |     0.0 |  -0.08518 |      0.0 |  -0.09987 |      0.0 | -0.157197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.695949 | 27.563399 |       0.0 | 0.942391 | -0.833922 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                         0.0 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.779738 | -0.060042 |                   0.0 |                      0.0 |                       0.0 |                   0.339064 |                  1.257174 |                     0.0 |                 0.008709 |    -0.052382 |             0.0 |        -0.008709 |          -0.91811 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    1.692 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003094 |      0.0 |      0.0 | -0.043255 | 0.027628 |          0.0 |     0.746556 | 0.613704 | 0.613704 |      0.0 |      0.0 |     0.0 | -0.115381 |      0.0 |  -0.13072 |      0.0 | -0.237391 |      1.0 |      0.3 | 0.538583 | 0.183859 |  0.0 |  0.0 | 0.0 |  0.0 | 0.183859 | 0.002385 | 0.000662 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.111938 | 28.760072 |       0.0 | 0.949785 | -0.742979 | 71.800984 | 0.183859 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002385 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000662 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                         0.0 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.476597 | -0.053495 |                   0.0 |                      0.0 |                       0.0 |                   0.339871 |                  1.245134 |                     0.0 |                -0.004886 |    -0.051881 |             0.0 |         0.004886 |         -0.905263 |             32.273586 |                 100.0 |               111.594033 |               98.814229 |    1.222 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002985 |      0.0 |      0.0 | -0.055801 |      0.0 |          0.0 |     0.745953 | 0.796411 | 0.796411 |      0.0 |      0.0 |     0.0 | -0.148847 |      0.0 | -0.165762 |      0.0 | -0.319496 |      1.0 |      0.3 | 2.363078 | 0.808536 |  0.0 |  0.0 | 0.0 |  0.0 | 0.808536 | 0.026773 | 0.007437 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.353003 | 27.674795 |       0.0 | 0.957093 | -0.658585 | 73.355527 |      0.5 | 0.308536 |   0.0 |   0.0 |  0.0 | 0.011174 |   0.0156 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007437 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                         0.0 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.195282 | -0.047418 |                   0.0 |                      0.0 |                       0.0 |                   0.340672 |                  1.232336 |                     0.0 |                -0.014528 |    -0.051347 |             0.0 |         0.014528 |         -0.891664 |             30.429381 |                 100.0 |               109.379504 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000597 |      0.0 |      0.0 |  -0.06294 |      0.0 |          0.0 |     0.745354 | 0.972146 | 0.972146 |      0.0 |      0.0 |     0.0 |  -0.16789 |      0.0 | -0.189697 |      0.0 | -0.372115 |      1.0 |      0.3 |  1.86333 | 0.639128 |  0.0 |  0.0 | 0.0 |  0.0 | 0.639128 | 0.059949 | 0.016653 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 18.422661 | 26.344405 |       0.0 | 0.964401 | -0.580267 | 74.579729 |      0.5 | 0.139128 |   0.0 |   0.0 |  0.0 |   0.0238 | 0.036149 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016653 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                         0.0 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.934222 | -0.041779 |                   0.0 |                      0.0 |                       0.0 |                   0.341467 |                  1.218349 |                     0.0 |                -0.012189 |    -0.050765 |             0.0 |         0.012189 |         -0.876882 |             29.178858 |                 100.0 |               107.648893 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007215 |      0.0 |      0.0 | -0.061894 |      0.0 |          0.0 |      0.74476 | 0.993575 | 0.993575 |      0.0 |      0.0 |     0.0 | -0.165098 |      0.0 | -0.196345 |      0.0 | -0.373633 |      1.0 |      0.3 |  1.61215 | 0.554125 |  0.0 |  0.0 | 0.0 |  0.0 | 0.554125 | 0.078187 | 0.021719 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.470034 | 24.982149 |       0.0 | 0.971709 | -0.507587 | 75.637753 |      0.5 | 0.054125 |   0.0 |   0.0 |  0.0 | 0.036101 | 0.042086 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021719 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                         0.0 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.691958 | -0.036546 |                   0.0 |                      0.0 |                       0.0 |                   0.342256 |                  1.204312 |                     0.0 |                -0.020102 |     -0.05018 |             0.0 |         0.020102 |         -0.862056 |              29.58412 |                 100.0 |               106.162112 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008179 |      0.0 |      0.0 |  -0.06548 |      0.0 |          0.0 |      0.74417 | 1.091266 | 1.091266 |      0.0 |      0.0 |     0.0 | -0.174665 |      0.0 | -0.206262 |      0.0 | -0.401029 |      1.0 |      0.3 | 2.034601 | 0.700988 |  0.0 |  0.0 | 0.0 |  0.0 | 0.700988 | 0.098792 | 0.027442 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.421698 | 23.483028 |       0.0 | 0.979017 | -0.440141 | 76.971366 |      0.5 | 0.200988 |   0.0 |   0.0 |  0.0 | 0.048084 | 0.050708 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027442 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                         0.0 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.467137 |  -0.03169 |                   0.0 |                      0.0 |                       0.0 |                    0.34304 |                  1.190153 |                     0.0 |                -0.018227 |     -0.04959 |             0.0 |         0.018227 |         -0.847113 |             28.027061 |                 100.0 |               104.299198 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014145 |      0.0 |      0.0 | -0.065772 |      0.0 |          0.0 |     0.743584 | 1.130726 | 1.130726 |      0.0 |      0.0 |     0.0 | -0.175444 |      0.0 | -0.215681 |      0.0 | -0.409353 |      1.0 |      0.3 | 1.809467 |  0.62487 |  0.0 |  0.0 | 0.0 |  0.0 |  0.62487 | 0.121553 | 0.033765 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |   15.3338 | 21.927334 |       0.0 | 0.986325 | -0.377551 | 78.155963 |      0.5 |  0.12487 |   0.0 |   0.0 |  0.0 | 0.059757 | 0.061796 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033765 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                         0.0 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 | -1.258503 | -0.027184 |              0.000833 |                      0.0 |                  0.000833 |                   0.343817 |                  1.174779 |                     0.0 |                -0.024601 |    -0.048116 |             0.0 |         0.025434 |         -0.830962 |             29.178858 |                 100.0 |               102.654311 |               99.312271 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019008 |      0.0 |      0.0 | -0.065786 |      0.0 |          0.0 |     0.743003 | 1.181498 | 1.181498 |      0.0 |      0.0 |     0.0 | -0.175481 |      0.0 | -0.220889 |      0.0 | -0.421804 |      1.0 |      0.3 | 1.788611 | 0.619042 |  0.0 |  0.0 | 0.0 |  0.0 | 0.619042 | 0.138899 | 0.038583 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 14.194762 | 20.298509 |       0.0 | 0.993633 | -0.319467 | 79.325532 |      0.5 | 0.119042 |   0.0 |   0.0 |  0.0 | 0.071129 |  0.06777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038583 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                         0.0 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 | -1.064891 | -0.023002 |              0.016012 |                      0.0 |                  0.016012 |                   0.344589 |                   1.15845 |                     0.0 |                -0.023742 |    -0.032257 |             0.0 |         0.039754 |         -0.813861 |             30.000798 |                 100.0 |               101.038917 |              100.104451 |    0.282 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01638 |      0.0 |      0.0 | -0.065321 |      0.0 |          0.0 |     0.742426 | 1.206281 | 1.206281 |      0.0 |      0.0 |     0.0 | -0.174242 |      0.0 | -0.211904 |      0.0 | -0.425899 |      1.0 |      0.3 | 2.012533 | 0.698227 |  0.0 |  0.0 | 0.0 |  0.0 | 0.698227 | 0.159077 | 0.044188 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.030278 | 18.633298 |       0.0 | 1.000941 | -0.265566 | 80.639838 |      0.5 | 0.198227 |   0.0 |   0.0 |  0.0 | 0.082207 | 0.076869 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044188 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                         0.0 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 | -0.885219 | -0.019121 |              0.048899 |                      0.0 |                  0.048899 |                   0.345356 |                  1.144953 |                     0.0 |                 -0.02789 |     0.001193 |             0.0 |         0.076789 |         -0.799597 |             30.000798 |                 100.0 |                99.233307 |               99.921659 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026429 |      0.0 |      0.0 | -0.066592 |      0.0 |          0.0 |     0.741853 | 1.362873 | 1.362873 |      0.0 |      0.0 |     0.0 | -0.177631 |      0.0 |   -0.2199 |      0.0 |  -0.47432 |      1.0 |      0.3 | 2.049873 | 0.712982 |  0.0 |  0.0 | 0.0 |  0.0 | 0.712982 | 0.182676 | 0.050743 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 11.709103 | 16.744017 |       0.0 | 1.008249 | -0.215545 | 81.976729 |      0.5 | 0.212982 |   0.0 |   0.0 |  0.0 | 0.092999 | 0.089677 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050743 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                         0.0 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 | -0.718483 | -0.015519 |              0.074636 |                      0.0 |                  0.074636 |                   0.337009 |                  1.076886 |                     0.0 |                -0.043295 |     0.029766 |             0.0 |         0.117931 |         -0.739877 |             23.933221 |                 100.0 |                97.406638 |               99.730556 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060046 |      0.0 |      0.0 | -0.080429 |      0.0 |          0.0 |     0.741284 |  1.83095 |  1.83095 |      0.0 |      0.0 |     0.0 |  -0.21454 |      0.0 | -0.294586 |      0.0 | -0.627057 |      1.0 |      0.3 | 2.724835 |  0.95073 |  0.0 |  0.0 | 0.0 |  0.0 |  0.95073 | 0.217465 | 0.060407 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.925601 |  14.19361 |       0.0 | 1.015557 | -0.169126 | 83.750834 |      0.5 |  0.45073 |   0.0 |   0.0 |  0.0 | 0.103512 | 0.113953 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060407 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                         0.0 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 | -0.563752 | -0.012177 |               0.07493 |                      0.0 |                   0.07493 |                   0.337745 |                  1.067855 |                     0.0 |                -0.048721 |     0.030436 |             0.0 |         0.123651 |         -0.730111 |             22.421649 |                 100.0 |                94.997168 |               99.468435 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073917 |      0.0 |      0.0 | -0.085956 |      0.0 |          0.0 |      0.74072 | 1.998691 | 1.998691 |      0.0 |      0.0 |     0.0 | -0.229285 |      0.0 | -0.326804 |      0.0 |  -0.67974 |      1.0 |      0.3 | 2.875436 | 1.006812 |  0.0 |  0.0 | 0.0 |  0.0 | 1.006812 | 0.264014 | 0.073337 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  7.974916 |  11.40413 |       0.0 | 1.022865 | -0.126049 | 85.619458 |      0.5 | 0.506812 |   0.0 |   0.0 |  0.0 | 0.113754 |  0.15026 |   0.0 |   0.0 |  0.0 |   0.0 | 0.073337 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                         0.0 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 | -0.420162 | -0.009076 |              0.070938 |                      0.0 |                  0.070938 |                   0.338475 |                   1.06128 |                     0.0 |                -0.040462 |     0.026718 |             0.0 |         0.111401 |         -0.722805 |             26.962743 |                 100.0 |                92.476135 |               99.181147 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.085367 |      0.0 |      0.0 | -0.061599 |      0.0 |          0.0 |     0.740159 | 1.635212 | 1.635212 |      0.0 |      0.0 |     0.0 | -0.164312 |      0.0 | -0.279524 |      0.0 | -0.555236 |      1.0 |      0.3 | 2.444983 | 0.858829 |  0.0 |  0.0 | 0.0 |  0.0 | 0.858829 | 0.301946 | 0.083874 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.373948 |  9.114746 |       0.0 | 1.030173 | -0.086073 | 87.205612 |      0.5 | 0.358829 |   0.0 |   0.0 |  0.0 | 0.123731 | 0.178215 |   0.0 |   0.0 |  0.0 |   0.0 | 0.083874 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                         0.0 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 |  -0.28691 | -0.006197 |              0.039586 |                      0.0 |                  0.039586 |                     0.3392 |                  1.056546 |                     0.0 |                -0.036448 |    -0.004437 |             0.0 |         0.076034 |         -0.717346 |              22.18809 |                 100.0 |                90.348807 |               98.927637 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.116273 |      0.0 |      0.0 | -0.063414 |      0.0 |          0.0 |     0.739603 |  1.71827 |  1.71827 |      0.0 |      0.0 |     0.0 | -0.169154 |      0.0 | -0.334911 |      0.0 | -0.580099 |      1.0 |      0.3 | 2.474221 |  0.87177 |  0.0 |  0.0 | 0.0 |  0.0 |  0.87177 | 0.330271 | 0.091742 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  4.688069 |  6.703939 |       0.0 | 1.037481 | -0.048976 | 88.808063 |      0.5 |  0.37177 |   0.0 |   0.0 |  0.0 |  0.13345 | 0.196821 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091742 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                         0.0 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 | -0.163253 | -0.003526 |               0.01185 |                      0.0 |                   0.01185 |                   0.349106 |                  1.110113 |                     0.0 |                -0.025663 |    -0.034405 |             0.0 |         0.037513 |         -0.761008 |             23.407216 |                 100.0 |                88.210626 |               98.661911 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102953 |      0.0 |      0.0 | -0.057639 |      0.0 |          0.0 |      0.73905 | 1.490415 | 1.490415 |      0.0 |      0.0 |     0.0 |  -0.15375 |      0.0 | -0.310062 |      0.0 | -0.501325 |      1.0 |      0.3 | 2.243617 | 0.792795 |  0.0 |  0.0 | 0.0 |  0.0 | 0.792795 | 0.353134 | 0.098093 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.225148 |  4.611961 |       0.0 | 1.044789 |  -0.01455 | 90.258885 |      0.5 | 0.292795 |   0.0 |   0.0 |  0.0 | 0.142919 | 0.210215 |   0.0 |   0.0 |  0.0 |   0.0 | 0.098093 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                         0.0 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 | -0.048499 | -0.001048 |              0.000021 |                      0.0 |                  0.000021 |                    0.34984 |                  1.107879 |                     0.0 |                -0.014096 |    -0.046141 |             0.0 |         0.014117 |          -0.75804 |              24.48341 |                 100.0 |                86.283755 |               98.792568 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096543 |      0.0 |      0.0 | -0.047871 |      0.0 |          0.0 |     0.738502 | 1.258207 | 1.258207 |      0.0 |      0.0 |     0.0 | -0.127693 |      0.0 | -0.279479 |      0.0 | -0.421289 | 0.999349 | 0.299609 | 1.817912 | 0.643269 |  0.0 |  0.0 | 0.0 |  0.0 | 0.643269 | 0.363005 | 0.100835 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.987357 |   2.84192 |       0.0 | 1.052097 |  0.017398 | 91.433529 |      0.5 | 0.143269 |   0.0 |   0.0 |  0.0 | 0.152142 | 0.210863 |   0.0 |   0.0 |  0.0 |   0.0 | 0.100835 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                         0.0 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                      0.0 |                       0.0 |                   0.350568 |                   1.10789 |                     0.0 |                 0.008426 |    -0.046162 |             0.0 |        -0.008426 |         -0.757322 |             23.933221 |                 100.0 |                84.729616 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.129358 |      0.0 |      0.0 | -0.025712 |      0.0 |          0.0 |     0.737957 | 0.975359 | 0.975359 |      0.0 |      0.0 |     0.0 | -0.068585 |      0.0 |  -0.26561 |      0.0 |  -0.32577 | 0.998192 | 0.298916 | 1.407382 | 0.498053 |  0.0 |  0.0 | 0.0 |  0.0 | 0.498053 | 0.358003 | 0.099445 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.021154 |   1.46025 |       0.0 | 1.059405 |  0.048298 | 92.342857 | 0.498053 |      0.0 |   0.0 |   0.0 |  0.0 | 0.161103 | 0.196901 |   0.0 |   0.0 |  0.0 |   0.0 | 0.099445 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                         0.0 |                      0.61078 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                      0.0 |                       0.0 |                   0.351292 |                  1.107494 |                     0.0 |                 0.008574 |    -0.046146 |             0.0 |        -0.008574 |         -0.756202 |             26.962743 |                 100.0 |                83.529996 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.087608 |      0.0 |      0.0 | -0.030943 |      0.0 |          0.0 |     0.737416 | 0.888357 | 0.888357 |      0.0 |      0.0 |     0.0 |  -0.08254 |      0.0 | -0.222746 |      0.0 | -0.296711 | 0.997036 | 0.298224 | 1.297269 | 0.459025 |  0.0 |  0.0 | 0.0 |  0.0 | 0.459025 | 0.346631 | 0.096286 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.135611 |  0.193924 |       0.0 | 1.066713 |  0.079198 | 93.181101 | 0.459025 |      0.0 |   0.0 |   0.0 |  0.0 |   0.1693 | 0.177331 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096286 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                         0.0 |                      0.61078 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                      0.0 |                       0.0 |                    0.35201 |                  1.107885 |                     0.0 |                  0.01332 |    -0.046162 |             0.0 |         -0.01332 |         -0.755874 |             30.000798 |                 100.0 |                82.426755 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.070419 |     0.06 |  0.00372 | -0.028252 |      0.0 |          0.0 |     0.736879 | 0.760344 | 0.135611 |      0.0 |      0.0 |     0.0 | -0.075361 |      0.0 | -0.191913 |      0.0 | -0.253955 |  0.99588 | 0.297533 | 0.222177 | 0.078526 |  0.0 |  0.0 | 0.0 |  0.0 | 0.078526 | 0.331555 | 0.092098 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 1.074021 |  0.110098 | 93.321031 | 0.078526 |      0.0 |   0.0 |   0.0 |  0.0 | 0.171848 | 0.159706 |   0.0 |   0.0 |  0.0 |   0.0 | 0.092098 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                         0.0 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.108737 |                     0.0 |                      0.0 |    -0.046197 |             0.0 |              0.0 |          0.095308 |             33.282135 |                 100.0 |                82.242826 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.066191 |      0.0 | 0.025199 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.990762 | 0.294483 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.312252 | 0.086737 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.246838 | 93.295832 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.168419 | 0.143833 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086737 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                         0.0 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.109465 |                     0.0 |                      0.0 |    -0.046228 |             0.0 |              0.0 |          0.094579 |              29.58412 |                 100.0 |                82.275944 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060894 |      0.0 | 0.021594 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.985645 | 0.291449 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.293607 | 0.081557 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.383578 | 93.274239 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.164069 | 0.129538 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081557 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                         0.0 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10152 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                   1.11033 |                     0.0 |                      0.0 |    -0.046264 |             0.0 |              0.0 |          0.093714 |             30.429381 |                 100.0 |                82.304324 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.050931 | 0.050931 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.980689 | 0.288525 |    0.034 | 0.011735 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011735 | 0.276646 | 0.076846 | 0.009069 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.515998 | 93.296503 | 0.011735 |      0.0 |   0.0 |   0.0 |  0.0 | 0.159983 | 0.116663 |   0.0 |   0.0 |  0.0 |   0.0 | 0.076846 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                         0.0 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08208 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.112997 |                     0.0 |                      0.0 |    -0.046375 |             0.0 |              0.0 |          0.091047 |             31.324363 |                 100.0 |                82.275061 |               98.814229 | 0.043069 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.055782 | 0.055782 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 |  0.97646 | 0.286042 | 0.043069 | 0.014768 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014768 | 0.261261 | 0.072573 | 0.004218 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.628978 | 93.324804 | 0.014768 |      0.0 |   0.0 |   0.0 |  0.0 | 0.156193 | 0.105068 |   0.0 |   0.0 |  0.0 |   0.0 | 0.072573 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                         0.0 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08424 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.116023 |                     0.0 |                      0.0 |    -0.046501 |             0.0 |              0.0 |          0.080642 |             33.282135 |                 100.0 |                82.237867 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.053849 | 0.004218 | 0.018581 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.972151 | 0.283523 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.246974 | 0.068604 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.744118 | 93.306223 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.152348 | 0.094626 |   0.0 |   0.0 |  0.0 |   0.0 | 0.068604 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                         0.0 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07992 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.119656 |                     0.0 |                      0.0 |    -0.046652 |             0.0 |              0.0 |          0.077009 |             33.810423 |                 100.0 |                82.262287 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049163 |      0.0 | 0.018518 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.968004 | 0.281109 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.233634 | 0.064898 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.854938 | 93.287705 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.148413 | 0.085221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064898 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                         0.0 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0756 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.123475 |                     0.0 |                      0.0 |    -0.046811 |             0.0 |              0.0 |           0.07319 |             37.369415 |                 100.0 |                82.286625 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033548 | 0.033548 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.964018 | 0.278799 |    0.034 | 0.011431 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011431 | 0.221478 | 0.061522 | 0.026452 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  0.961438 | 93.310274 | 0.011431 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144727 | 0.076751 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061522 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                         0.0 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.127474 |                     0.0 |                      0.0 |    -0.046978 |             0.0 |              0.0 |           0.06919 |             40.962628 |                 100.0 |                82.256964 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023802 | 0.023802 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.960194 | 0.276592 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.210259 | 0.058405 |  0.00265 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.063618 | 93.310274 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.141136 | 0.069123 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058405 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                         0.0 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.131856 |                     0.0 |                      0.0 |    -0.047161 |             0.0 |              0.0 |          0.064808 |              46.30558 |                 100.0 |                82.256964 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.022296 |  0.00265 | 0.008765 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 |  0.95637 | 0.274393 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.199743 | 0.055484 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.165798 | 93.301508 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.13749 | 0.062253 |   0.0 |   0.0 |  0.0 |   0.0 | 0.055484 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                         0.0 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0648 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.136687 |                     0.0 |                      0.0 |    -0.047362 |             0.0 |              0.0 |          0.059978 |             43.470544 |                 100.0 |                82.268484 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.028418 |      0.0 |  0.01221 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.952788 | 0.272342 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.190004 | 0.052779 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.261498 | 93.289298 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.133939 | 0.056066 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052779 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                         0.0 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.04104 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.144252 |                     0.0 |                      0.0 |    -0.047677 |             0.0 |              0.0 |          0.052412 |             33.810423 |                 100.0 |                82.284531 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.047697 |      0.0 | 0.017452 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.950096 | 0.270805 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.180972 |  0.05027 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.333438 | 93.271846 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.130479 | 0.050493 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05027 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                         0.0 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.02592 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.152031 |                     0.0 |                      0.0 |    -0.048001 |             0.0 |              0.0 |          0.044634 |             36.725115 |                 100.0 |                82.307469 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.040381 |      0.0 | 0.015359 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.947969 | 0.269594 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.172584 |  0.04794 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.390258 | 93.256487 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.127109 | 0.045475 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04794 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                         0.0 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 |  -0.01728 |              0.002897 |                      0.0 |                  0.002897 |                   1.196665 |                  1.158654 |                     0.0 |                      0.0 |     -0.04538 |             0.0 |              0.0 |          0.038011 |             34.918962 |                 100.0 |                82.327657 |               98.490767 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02629 |      0.0 | 0.009663 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.946166 | 0.268569 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164781 | 0.045772 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.438438 | 93.246825 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.123825 | 0.040955 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045772 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                         0.0 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 |  -0.01728 |              0.057389 |                      0.0 |                  0.057389 |                   1.196665 |                  1.166643 |                     0.0 |                      0.0 |     0.008779 |             0.0 |              0.0 |          0.030022 |             36.102655 |                 100.0 |                82.340358 |               97.869977 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.045635 |      0.0 | 0.017181 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.944363 | 0.267546 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.157512 | 0.043753 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.486618 | 93.229643 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.120627 | 0.036885 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043753 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                         0.0 |                     0.555459 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |   0.02796 |              0.020003 |                      0.0 |                  0.020003 |                   0.136052 |                  1.175846 |                     0.0 |                 0.100918 |     -0.02899 |             0.0 |        -0.080915 |         -1.039794 |             33.282135 |                 100.0 |                82.362942 |               97.873214 |    0.034 | 0.01275 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052426 |      0.0 |      0.0 |  0.004903 | 0.000025 |    -1.294436 |     0.897677 |      0.0 |      0.0 | 0.000192 | 0.000192 |     0.0 |  0.013079 |      0.0 | -0.065945 |      0.0 |  0.028049 | 0.944253 | 0.267484 | 0.015768 | 0.005136 |  0.0 |  0.0 | 0.0 |  0.0 | 0.005136 | 0.150796 | 0.041888 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.009462 |  0.013329 |       0.0 |  0.00743 |  1.489558 | 93.240275 | 0.005136 |      0.0 |   0.0 |   0.0 |  0.0 | 0.117578 | 0.033219 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041888 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                         0.0 |                     0.535467 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 |  0.038672 |              0.030684 |                      0.0 |                  0.030684 |                   0.139046 |                   1.18712 |                     0.0 |                 0.111622 |    -0.018779 |             0.0 |        -0.080938 |         -1.048074 |             31.324363 |                 100.0 |                82.348967 |               97.871211 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.084311 |     0.06 | 0.008323 |  0.013329 |      0.0 |    -1.790392 |     0.895425 |      0.0 |      0.0 | 0.000193 | 0.000193 |     0.0 |  0.038481 |      0.0 | -0.080682 |      0.0 |  0.038737 | 0.944544 | 0.267649 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.144524 | 0.040145 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 0.014738 |  1.481786 | 93.231951 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.114606 | 0.029917 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040145 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                         0.0 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |   0.00216 |              0.264596 |                      0.0 |                  0.264596 |                   1.196665 |                  1.199167 |                     0.0 |                      0.0 |      0.21463 |             0.0 |              0.0 |         -0.002502 |             30.429381 |                 100.0 |                82.359908 |               97.872779 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.127998 |      0.0 | 0.042746 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.943468 |  0.26704 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.13859 | 0.038497 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.510526 | 93.189206 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.111646 | 0.026944 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038497 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                         0.0 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |   0.00864 |              0.249736 |                      0.0 |                  0.249736 |                   1.196665 |                   1.21168 |                     0.0 |                      0.0 |     0.199249 |             0.0 |              0.0 |         -0.015015 |             33.810423 |                 100.0 |                82.416101 |               97.880827 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.131541 |      0.0 |  0.04673 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.942635 | 0.266568 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.133028 | 0.036952 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.532786 | 93.142475 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.108762 | 0.024266 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036952 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                         0.0 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |   0.01296 |              0.195223 |                      0.0 |                  0.195223 |                   1.259762 |                  1.349977 |                     0.0 |                      0.0 |     0.138974 |             0.0 |              0.0 |         -0.090215 |             29.178858 |                 100.0 |                82.477539 |               97.889615 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146498 |      0.0 | 0.047227 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.941964 | 0.266189 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.127807 | 0.035502 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.550726 | 93.095248 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.105953 | 0.021854 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035502 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                         0.0 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |   0.01728 |              0.048073 |                      0.0 |                  0.048073 |                   1.291311 |                  1.428152 |                     0.0 |                      0.0 |    -0.011433 |             0.0 |              0.0 |         -0.136842 |             28.400756 |                 100.0 |                82.539637 |               97.898485 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135014 |      0.0 | 0.042617 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.941454 | 0.265901 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.122898 | 0.034138 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.564346 |  93.05263 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103216 | 0.019682 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034138 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                         0.0 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |   0.02808 |              0.000547 |                      0.0 |                  0.000547 |                   1.354408 |                  1.571379 |                     0.0 |                      0.0 |    -0.064927 |             0.0 |              0.0 |         -0.216971 |             30.429381 |                 100.0 |                82.595681 |               98.771733 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.11254 |      0.0 | 0.036657 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.941349 | 0.265841 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.118276 | 0.032854 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.567166 | 93.015973 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.10055 | 0.017726 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032854 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                         0.0 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.03456 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.583999 |                     0.0 |                      0.0 |       -0.066 |             0.0 |              0.0 |          -0.22959 |             35.500944 |                 100.0 |                82.643892 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.08847 |      0.0 |  0.03163 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.941486 | 0.265919 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.113917 | 0.031644 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.563506 | 92.984343 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.097952 | 0.015964 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031644 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                         0.0 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |    0.0432 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                    1.5975 |                     0.0 |                      0.0 |    -0.066562 |             0.0 |              0.0 |         -0.243092 |             44.376181 |                 100.0 |                82.685495 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.063209 |      0.0 | 0.025658 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.941946 | 0.266179 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |   0.1098 |   0.0305 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.551206 | 92.958685 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095422 | 0.014378 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0305 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                         0.0 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04536 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.611195 |                     0.0 |                      0.0 |    -0.067133 |             0.0 |              0.0 |         -0.256786 |             50.715635 |                 100.0 |                82.719245 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.054415 |      0.0 |  0.02375 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.942487 | 0.266485 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.105906 | 0.029418 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.536746 | 92.934935 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.092957 | 0.012949 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029418 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                         0.0 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04752 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.625096 |                     0.0 |                      0.0 |    -0.067712 |             0.0 |              0.0 |         -0.270688 |             43.470544 |                 100.0 |                82.750488 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.065452 |      0.0 |  0.02619 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.943109 | 0.266836 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.102218 | 0.028394 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.520126 | 92.908745 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.090556 | 0.011662 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028394 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                         0.0 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.05616 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.639665 |                     0.0 |                      0.0 |    -0.068319 |             0.0 |              0.0 |         -0.285257 |             53.251417 |                 100.0 |                82.784943 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.048111 |      0.0 | 0.021424 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.944054 | 0.267372 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.09872 | 0.027422 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.494866 |  92.88732 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088217 | 0.010503 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027422 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                         0.0 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.06696 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.653965 |                     0.0 |                      0.0 |    -0.068915 |             0.0 |              0.0 |         -0.299557 |             62.648725 |                 100.0 |                82.813131 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034966 |      0.0 | 0.016822 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.945404 | 0.268137 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095397 | 0.026499 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.458806 | 92.870498 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.085938 | 0.009459 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026499 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                         0.0 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.08424 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.664915 |                     0.0 |                      0.0 |    -0.069371 |             0.0 |              0.0 |         -0.310507 |             62.648725 |                 100.0 |                82.835264 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024921 |      0.0 |  0.01189 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 |   0.9474 |  0.26927 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.092237 | 0.025621 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.405466 | 92.858608 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.083719 | 0.008519 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025621 |

.. _lland_v4_nadelw_winter_hourly:

conifers (winter)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v4_nadelw_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_nadelw_winter)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |  nbes |   sbes | snowintmax | snowintrate |  nbesinz |  sbesinz | wniedinz | actualalbedoinz |  wadainz | schmpotinz |  schminz | gefrpotinz |  gefrinz |   evsinz |      evpo |      evi |       evb |       evs |    wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |  wlatinz |  wlatsnow |  wsensinz | wsenssnow |  wsurfinz |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp |    stinz |     sinz |  esnowinz |    asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |           1.206967 |          3.4 |                 0.464297 |                      0.526233 |                    0.464297 |                      0.48954 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.724943 |  0.027429 |                   0.0 |                      0.0 |                       0.0 |                    0.75989 |                  4.014396 |                0.174381 |                 0.091855 |    -0.167266 |       -0.174381 |        -0.091855 |         -3.254506 |             27.647059 |                 100.0 |              3030.581342 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439593 |      0.0 |      0.0 |      0.0 |        0.750856 |      0.0 |        0.0 |      0.0 |   0.356526 |      0.0 | 0.005419 |  0.023255 |      0.0 |   0.00866 |  0.007776 |      0.0 |    -2.994784 |     0.735369 |      0.0 |      0.0 |  0.28488 |  0.28488 | 0.014456 |  0.020743 | -0.071098 |   0.00998 |  0.117739 |  0.122578 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.049814 | 0.013837 |      2.2 | 6.525649 | 6.525649 |  -0.11908 | 0.899234 | 10.042278 | 12.470566 |       0.0 | 1.094846 | -0.514011 | 73.404346 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.049814 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013837 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           1.029472 |          2.9 |                 0.427219 |                      0.521048 |                    0.427219 |                     0.475444 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.713372 |  0.036097 |                   0.0 |                      0.0 |                       0.0 |                   0.761391 |                  4.024777 |                0.143217 |                 0.102631 |    -0.167699 |       -0.143217 |        -0.102631 |         -3.263386 |             32.413793 |                 100.0 |             19517.015931 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439577 |      0.0 |      0.0 |      0.0 |        0.750675 |      0.0 |        0.0 |      0.0 |   0.510268 |      0.0 | 0.003442 |   0.00778 |      0.0 |   0.00314 |   0.00762 |      0.0 |    -3.384514 |     0.734846 |      0.0 |      0.0 | 0.315703 | 0.315703 | 0.009181 |  0.020325 | -0.101048 |  0.018585 |   0.05135 |  0.141542 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.048527 |  0.01348 |      2.2 | 6.522207 | 6.522207 |  -0.17043 | 0.901318 | 10.351652 | 12.462946 |       0.0 | 1.102154 | -0.519208 | 73.401206 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.048527 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01348 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           0.887476 |          2.5 |                 0.401924 |                      0.514631 |                    0.401924 |                     0.463834 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.730694 |  0.042823 |                   0.0 |                      0.0 |                       0.0 |                   0.762882 |                  4.040774 |                0.131664 |                 0.110368 |    -0.168366 |       -0.131664 |        -0.110368 |         -3.277892 |                  37.6 |                 100.0 |             27328.680919 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439567 |      0.0 |      0.0 |      0.0 |        0.750535 |      0.0 |        0.0 |      0.0 |   0.597812 |      0.0 | 0.003363 |   0.00648 |      0.0 |  0.002817 |  0.008392 |      0.0 |    -3.713261 |     0.734327 |      0.0 |      0.0 | 0.338802 | 0.338802 | 0.008971 |  0.022386 | -0.111394 |  0.023229 |   0.02924 |  0.155983 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.047274 | 0.013132 |      2.2 | 6.518844 | 6.518844 | -0.199669 | 0.902922 | 10.683256 | 12.454554 |       0.0 | 1.109462 | -0.531132 | 73.398388 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.047274 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013132 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           1.064971 |          3.0 |                 0.383833 |                      0.510972 |                    0.383833 |                     0.453362 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.770439 |  0.048506 |                   0.0 |                      0.0 |                       0.0 |                   0.764362 |                  4.052814 |                0.129056 |                 0.114685 |    -0.168867 |       -0.129056 |        -0.114685 |         -3.288451 |             31.333333 |                 100.0 |             27329.679692 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439557 |      0.0 |      0.0 |      0.0 |        0.750415 |      0.0 |        0.0 |      0.0 |   0.634352 |      0.0 | 0.002938 |  0.006163 |      0.0 |  0.002398 |  0.009055 |      0.0 |    -4.016079 |     0.733812 |      0.0 |      0.0 | 0.357648 | 0.357648 | 0.007837 |  0.024154 | -0.124689 |  0.029121 |  0.012204 |   0.16796 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.046053 | 0.012792 |      2.2 | 6.515906 | 6.515906 | -0.211874 | 0.904306 | 11.032877 | 12.445498 |       0.0 |  1.11677 | -0.548737 | 73.395991 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.046053 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012792 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           0.887476 |          2.5 |                 0.352531 |                      0.505764 |                    0.352531 |                     0.439856 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.829125 |  0.055875 |                   0.0 |                      0.0 |                       0.0 |                   0.765832 |                  4.064535 |                0.129157 |                 0.124956 |    -0.169356 |       -0.129157 |        -0.124956 |         -3.298703 |                  37.6 |                 100.0 |             27330.529674 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439548 |      0.0 |      0.0 |      0.0 |        0.750302 |      0.0 |        0.0 |      0.0 |   0.684537 |      0.0 | 0.001661 | -0.003516 |      0.0 | -0.001501 |  0.008755 |      0.0 |    -4.415912 |       0.7333 |      0.0 |      0.0 | 0.383062 | 0.383062 | 0.004431 |  0.023353 | -0.116826 |  0.035508 |  0.016762 |  0.183817 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.044863 | 0.012462 |      2.2 | 6.514245 | 6.514245 | -0.228635 | 0.905605 | 11.407908 | 12.436744 |       0.0 | 1.124078 | -0.573712 | 73.397492 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.044863 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012462 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |           0.816478 |          2.3 |                 0.326061 |                      0.499513 |                    0.326061 |                     0.425471 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.912373 |  0.063537 |                   0.0 |                      0.0 |                       0.0 |                   0.767292 |                  4.080005 |                0.130209 |                 0.133686 |        -0.17 |       -0.130209 |        -0.133686 |         -3.312713 |             40.869565 |                 100.0 |             27329.997609 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439543 |      0.0 |      0.0 |      0.0 |        0.750198 |      0.0 |        0.0 |      0.0 |   0.738467 |      0.0 | 0.001741 | -0.002485 |      0.0 | -0.001099 |  0.009549 |      0.0 |    -4.853907 |     0.732791 |      0.0 |      0.0 | 0.409095 | 0.409095 | 0.004645 |   0.02547 | -0.116841 |  0.041019 |  0.018013 |  0.200175 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.043704 |  0.01214 |      2.2 | 6.512504 | 6.512504 | -0.246648 | 0.906799 |  11.80793 | 12.427195 |       0.0 | 1.131386 | -0.606349 | 73.398591 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.043704 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01214 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |            0.74548 |          2.1 |                 0.318472 |                      0.492103 |                    0.318472 |                      0.42117 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.021164 |  0.064071 |                   0.0 |                      0.0 |                       0.0 |                   0.768743 |                  4.093695 |                0.124263 |                 0.135531 |    -0.170571 |       -0.124263 |        -0.135531 |         -3.324953 |             44.761905 |                 100.0 |             27329.607878 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439538 |      0.0 |      0.0 |      0.0 |        0.750104 |      0.0 |        0.0 |      0.0 |   0.766198 |      0.0 | 0.000845 | -0.008706 |      0.0 | -0.004031 |  0.008634 |      0.0 |    -4.987414 |     0.732286 |      0.0 |      0.0 | 0.407226 | 0.407226 | 0.002255 |   0.02303 | -0.117256 |  0.041523 |  0.009262 |  0.200084 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.042575 | 0.011827 |      2.2 | 6.511659 | 6.511659 |  -0.25591 | 0.907887 | 12.206669 | 12.418561 |       0.0 | 1.138694 |  -0.63952 | 73.402622 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.042575 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011827 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           0.496987 |          1.4 |                 0.320984 |                      0.483429 |                    0.320984 |                     0.423593 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.131734 |  0.060055 |                   0.0 |                      0.0 |                       0.0 |                   0.708519 |                  3.682358 |                0.109805 |                 0.133121 |    -0.153432 |       -0.109805 |        -0.133121 |         -2.973839 |             67.142857 |                 100.0 |              27328.17901 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439535 |      0.0 |      0.0 |      0.0 |        0.750014 |      0.0 |        0.0 |      0.0 |   0.780744 |      0.0 | 0.000874 | -0.007626 |      0.0 | -0.004277 |  0.007686 |      0.0 |    -4.912051 |     0.731785 |      0.0 |      0.0 | 0.388244 | 0.211892 | 0.002331 |  0.020503 | -0.107278 |  0.036104 |  0.004858 |  0.189728 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.041476 | 0.011521 |      2.2 | 6.510785 | 6.510785 | -0.260769 | 0.908925 | 12.410875 | 12.410875 | -0.058902 | 1.146002 | -0.668675 | 73.406899 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.041476 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011521 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |           0.390489 |          1.1 |                 0.296635 |                      0.471927 |                    0.296635 |                     0.390224 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 | -2.228917 |  0.081136 |              0.001071 |                 0.000812 |                  0.000199 |                   0.614273 |                  3.039527 |                0.101025 |                 0.127327 |    -0.125576 |       -0.100213 |        -0.127129 |         -2.425254 |             85.454545 |                 100.0 |             27326.662803 |              106.750382 |   0.0 |    0.0 |       24.5 |    0.439532 |      0.0 |      0.0 |      0.0 |        0.749927 |      0.0 |        0.0 |      0.0 |    0.79926 |      0.0 | 0.000592 | -0.005971 |      0.0 |  -0.00303 |  0.006427 |      0.0 |     -5.98522 |     0.731513 |      0.0 |      0.0 | 0.464817 |      0.0 | 0.001579 |  0.017143 | -0.095607 |  0.033212 |  0.006184 |  0.177484 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.040404 | 0.011223 |      2.2 | 6.510193 | 6.510193 | -0.266953 | 0.909938 | 12.404448 | 12.404448 | -0.155249 | 1.149974 | -0.718911 | 73.409929 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.040404 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011223 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |            0.35499 |          1.0 |                 0.303761 |                      0.460234 |                    0.303761 |                     0.358644 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 | -2.396371 |  0.101135 |              0.018051 |                 0.013686 |                  0.003352 |                   0.522497 |                  2.412046 |                0.086779 |                 0.101393 |    -0.082451 |       -0.073093 |         -0.09804 |         -1.889549 |                  94.0 |                 100.0 |             27325.588601 |             4280.734103 |   0.0 |    0.0 |       24.5 |    0.439531 |      0.0 |      0.0 |      0.0 |        0.749842 |      0.0 |        0.0 |      0.0 |   0.762859 |      0.0 | 0.001056 |  -0.00035 |      0.0 |  -0.00001 |  0.004404 |      0.0 |    -7.078549 |     0.731339 |      0.0 |      0.0 | 0.549442 |      0.0 | 0.002817 |  0.011747 | -0.088068 |  0.019612 | -0.012158 |    0.1294 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.039361 | 0.010934 |      2.2 | 6.509137 | 6.509137 | -0.254795 | 0.910919 | 12.400044 | 12.400044 | -0.183513 |  1.15253 | -0.789146 | 73.409939 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.039361 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010934 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           0.496987 |          1.4 |                 0.363641 |                      0.451775 |                    0.363641 |                     0.376025 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -2.630487 |  0.082863 |              0.054453 |                 0.041277 |                  0.010113 |                   0.436284 |                  1.815487 |                0.065588 |                 0.068376 |    -0.021192 |       -0.024311 |        -0.058263 |         -1.379203 |             67.142857 |                 100.0 |             27325.585192 |             4280.734021 |   0.0 |    0.0 |       24.5 |    0.439527 |      0.0 |      0.0 |      0.0 |        0.749752 |      0.0 |        0.0 |      0.0 |   0.603496 |      0.0 | 0.000797 |  0.004983 |      0.0 |  0.000104 |  0.001619 |      0.0 |    -6.466727 |      0.73118 |      0.0 |      0.0 | 0.501773 |      0.0 | 0.002125 |  0.004318 | -0.079663 |   0.00436 | -0.053227 |  0.066942 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.038344 | 0.010651 |      2.2 |  6.50834 |  6.50834 | -0.201568 | 0.911963 | 12.398425 | 12.398425 | -0.167592 | 1.154877 | -0.841109 | 73.409835 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.038344 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010651 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           0.461488 |          1.3 |                 0.453913 |                      0.447245 |                    0.453913 |                      0.42179 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -2.803697 |  0.046751 |              0.103777 |                 0.078654 |                  0.019273 |                   0.389936 |                  1.493002 |                0.059322 |                 0.040237 |     0.041568 |        0.019331 |        -0.020964 |         -1.103066 |             72.307692 |                 100.0 |              3848.295347 |             2210.189944 |   0.0 |    0.0 |       24.5 |    0.439525 |      0.0 |      0.0 |      0.0 |        0.749634 |      0.0 |        0.0 |      0.0 |   0.388355 |      0.0 | 0.000768 |   0.01218 |      0.0 |  0.000543 | -0.001322 |      0.0 |    -4.968086 |     0.731013 |      0.0 |      0.0 | 0.385439 |      0.0 | 0.002048 | -0.003526 | -0.054573 | -0.009544 | -0.071857 |  0.007895 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.037354 | 0.010376 |      2.2 | 6.507572 | 6.507572 | -0.129711 | 0.913331 | 12.399747 | 12.399747 | -0.128737 | 1.157337 |  -0.85696 | 73.409292 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.037354 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010376 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           0.461488 |          1.3 |                 0.496723 |                      0.442058 |                    0.496723 |                     0.452204 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -2.856533 |  0.025777 |              0.088641 |                  0.06718 |                  0.016462 |                   0.366408 |                  1.331032 |                0.064954 |                 0.033247 |     0.033181 |        0.002226 |        -0.016785 |         -0.964624 |             72.307692 |                 100.0 |              1765.665368 |             1328.210612 |   0.0 |    0.0 |      26.95 |    0.439523 |      0.0 |      0.0 |      0.0 |        0.749465 |      0.0 |        0.0 |      0.0 |   0.278333 |      0.0 | 0.001971 |   0.01666 |      0.0 |  0.001226 | -0.000925 |      0.0 |    -4.049928 |     0.730824 |      0.0 |      0.0 | 0.314239 |      0.0 | 0.005257 | -0.002467 | -0.039779 | -0.012322 | -0.036748 |  0.001997 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.036389 | 0.010108 |      2.2 | 6.505602 | 6.505602 | -0.092963 |   0.9153 | 12.400672 | 12.400672 | -0.104956 | 1.160131 | -0.851837 | 73.408066 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.036389 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010108 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           0.496987 |          1.4 |                 0.551149 |                      0.441361 |                    0.551149 |                     0.482861 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -2.839457 |  0.007317 |              0.078043 |                 0.059152 |                  0.014494 |                   0.362879 |                  1.306141 |                0.061405 |                  0.02277 |     0.023621 |       -0.002253 |        -0.008276 |         -0.943262 |             67.142857 |                 100.0 |              1082.373603 |              906.322793 |   0.0 |    0.0 |       44.1 |    0.439517 |      0.0 |      0.0 |      0.0 |         0.74926 |      0.0 |        0.0 |      0.0 |   0.199155 |      0.0 | 0.002415 |  0.017988 |      0.0 |  0.001808 | -0.002119 |      0.0 |    -3.178189 |     0.730617 |      0.0 |      0.0 | 0.246618 |      0.0 | 0.006442 | -0.005652 | -0.035141 | -0.017893 | -0.026445 | -0.015269 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.035449 | 0.009847 |      2.2 | 6.503186 | 6.503186 | -0.066518 | 0.917691 | 12.402791 | 12.402791 |  -0.08237 | 1.163202 | -0.828254 | 73.406257 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.035449 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009847 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |           0.816478 |          2.3 |                 0.593188 |                      0.441918 |                    0.593188 |                     0.509295 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -2.760846 | -0.006391 |              0.041853 |                 0.031725 |                  0.007773 |                   0.363189 |                  1.307301 |                 0.05342 |                 0.016736 |    -0.012618 |       -0.021695 |        -0.008963 |         -0.944112 |             40.869565 |                 100.0 |               848.550168 |               739.14694 |   0.0 |    0.0 |       49.0 |     0.43951 |      0.0 |      0.0 |      0.0 |        0.749019 |      0.0 |        0.0 |      0.0 |   0.161691 |      0.0 | 0.003121 |  0.020649 |      0.0 |  0.001662 | -0.003467 |      0.0 |     -2.46497 |     0.730387 |      0.0 |      0.0 | 0.191307 |      0.0 | 0.008326 | -0.009249 | -0.042534 | -0.024579 | -0.012513 | -0.024865 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.034533 | 0.009593 |      2.2 | 6.500065 | 6.500065 | -0.054005 | 0.920504 | 12.406258 | 12.406258 | -0.063897 | 1.166629 | -0.790963 | 73.404595 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.034533 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009593 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           0.993973 |          2.8 |                 0.633417 |                      0.443734 |                    0.633417 |                     0.534559 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -2.636543 | -0.017782 |              0.010026 |                 0.007601 |                  0.001862 |                   0.363535 |                  1.307629 |                0.041609 |                 0.011087 |    -0.044458 |       -0.034008 |        -0.009225 |         -0.944094 |             33.571429 |                 100.0 |               708.854077 |              632.684921 |   0.0 |    0.0 |       49.0 |      0.4395 |      0.0 |      0.0 |      0.0 |        0.748755 |      0.0 |        0.0 |      0.0 |   0.138163 |      0.0 | 0.003187 |  0.017712 |      0.0 |   0.00141 | -0.005244 |      0.0 |     -1.81329 |      0.73013 |      0.0 |      0.0 | 0.140769 |      0.0 | 0.008501 | -0.013987 | -0.050368 |   -0.0299 | -0.007858 | -0.034662 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.033641 | 0.009345 |      2.2 | 6.496878 | 6.496878 | -0.046147 | 0.923601 | 12.411502 | 12.411502 | -0.047017 | 1.170465 | -0.742281 | 73.403185 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.033641 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009345 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |            1.10047 |          3.1 |                 0.656792 |                      0.447104 |                    0.656792 |                     0.554621 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 | -2.474269 | -0.025042 |              0.000027 |                 0.000021 |                  0.000005 |                   0.363683 |                   1.30686 |                0.037496 |                 0.010115 |    -0.054425 |       -0.037475 |         -0.01011 |         -0.943178 |             30.322581 |                 100.0 |               649.164736 |               73.451805 | 0.094 | 0.0235 |       49.0 |    0.439491 | 0.041312 | 0.010328 | 0.000151 |        0.748552 |      0.0 |        0.0 |      0.0 |   0.118161 | 0.030984 | 0.003168 |  0.016667 |      0.0 |   0.00677 | -0.005959 | 0.000344 |    -1.314909 |     0.730021 |      0.0 |      0.0 | 0.103563 | 0.080828 | 0.008451 | -0.015894 | -0.052455 | -0.031341 | -0.006529 | -0.037125 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.032772 | 0.009103 |      2.2 | 6.535022 | 6.535022 | -0.029117 | 0.925994 |  12.51146 |  12.51146 | -0.007593 | 1.172104 | -0.686338 | 73.396415 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.032772 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009103 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           1.242466 |          3.5 |                 0.661557 |                      0.450781 |                    0.661557 |                     0.577881 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.287795 | -0.033073 |                   0.0 |                      0.0 |                       0.0 |                   0.364275 |                  1.305996 |                0.042764 |                 0.017556 |    -0.054417 |       -0.042764 |        -0.017556 |         -0.941722 |             26.857143 |                 100.0 |               639.092907 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.439605 |      0.0 |      0.0 |      0.0 |        0.748206 |      0.0 |        0.0 |      0.0 |   0.095536 |      0.0 | 0.003509 |  0.020103 |      0.0 |  0.007884 | -0.004417 |      0.0 |    -0.756636 |     0.729581 |      0.0 |      0.0 | 0.059237 |      0.0 | 0.009361 | -0.011782 | -0.049333 | -0.026655 |  0.002792 | -0.020881 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.031925 | 0.008868 |      2.2 | 6.531513 | 6.531513 | -0.031909 | 0.930072 | 12.515877 | 12.515877 | -0.019785 | 1.178705 | -0.622365 | 73.388531 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.031925 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.008868 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           1.561958 |          4.4 |                 0.656792 |                      0.455629 |                    0.656792 |                     0.577842 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.074551 | -0.028447 |                   0.0 |                      0.0 |                       0.0 |                   0.364051 |                  1.303739 |                0.044112 |                 0.018618 |    -0.054322 |       -0.044112 |        -0.018618 |         -0.939688 |             21.363636 |                 100.0 |               644.333557 |               69.444444 | 0.282 | 0.0705 |       49.0 |    0.439595 | 0.123966 | 0.030991 | 0.000454 |        0.748112 |      0.0 |        0.0 |      0.0 |   0.090831 | 0.090831 | 0.002176 |  0.012597 |      0.0 |  0.004285 | -0.006261 | 0.001033 |    -0.757572 |     0.729747 |      0.0 |      0.0 | 0.061822 | 0.061822 | 0.005804 | -0.016702 | -0.051034 | -0.028467 | -0.001118 | -0.026551 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.031101 | 0.008639 |      2.2 |  6.65116 | 6.653302 |       0.0 | 0.931192 |  12.62338 | 12.804139 |       0.0 | 1.176208 | -0.563019 | 73.384246 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.031101 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.008639 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           1.348964 |          3.8 |                 0.676036 |                      0.460234 |                    0.676036 |                     0.597485 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.876729 | -0.034029 |                   0.0 |                      0.0 |                       0.0 |                   0.364409 |                  1.301296 |                0.048773 |                 0.018532 |    -0.054221 |       -0.048773 |        -0.018532 |         -0.936887 |             24.736842 |                 100.0 |               601.402258 |               69.444444 | 0.188 | 0.0282 |       49.0 |     0.43996 | 0.082712 | 0.012407 | 0.000448 |        0.747593 |      0.0 |        0.0 |      0.0 |   0.044348 | 0.044348 | 0.001135 |  0.004093 |      0.0 |  0.001535 | -0.006624 | 0.001019 |    -0.301321 |     0.729482 |      0.0 |      0.0 | 0.025165 | 0.025165 | 0.003028 | -0.017668 | -0.036541 | -0.025469 |  0.015261 | -0.024605 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.030297 | 0.008416 |      2.2 | 6.706784 |  6.73488 |       0.0 | 0.937352 | 12.670795 | 12.998762 |       0.0 | 1.180208 |  -0.49809 | 73.382711 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.030297 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.008416 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |           2.023445 |          5.7 |                 0.685844 |                      0.465172 |                    0.685844 |                     0.610278 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.660299 | -0.035619 |                   0.0 |                      0.0 |                       0.0 |                   0.364667 |                  1.298642 |                0.049546 |                 0.019604 |     -0.05411 |       -0.049546 |        -0.019604 |         -0.933976 |             16.491228 |                 100.0 |               582.434858 |               69.444444 | 0.376 | 0.0376 |       49.0 |    0.440205 | 0.165517 | 0.016552 | 0.001053 |         0.74711 |      0.0 |        0.0 |      0.0 |   0.028015 | 0.028015 | 0.001109 |  0.005159 |      0.0 |  0.001492 | -0.008292 | 0.002393 |    -0.011279 |      0.72929 |      0.0 |      0.0 | 0.000992 | 0.000992 | 0.002959 | -0.022118 | -0.042094 | -0.030381 |   0.01041 | -0.032895 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.029515 | 0.008199 |      2.2 | 6.750266 | 6.899287 |       0.0 | 0.943126 | 12.700704 | 13.383054 |       0.0 | 1.183098 | -0.431571 | 73.381219 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.029515 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.008199 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           1.987946 |          5.6 |                 0.705836 |                      0.471336 |                    0.705836 |                      0.61078 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -1.43857 | -0.031073 |                   0.0 |                      0.0 |                       0.0 |                   0.365319 |                  1.295068 |                0.045761 |                 0.012239 |    -0.053961 |       -0.045761 |        -0.012239 |         -0.929749 |             16.785714 |                 100.0 |               547.488202 |               69.444444 | 0.658 |    0.0 |       49.0 |    0.440698 | 0.289979 |      0.0 | 0.002428 |        0.746499 |      0.0 |        0.0 |      0.0 |   0.009603 | 0.009603 | 0.000347 | -0.001162 |      0.0 | -0.000343 |  -0.01135 |  0.00551 |          0.0 |     0.728806 | 0.089143 | 0.089143 |      0.0 |      0.0 | 0.000927 | -0.030275 | -0.041053 | -0.037301 |  0.005636 | -0.055337 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.028752 | 0.007987 |      2.2 | 6.759542 | 7.188919 |       0.0 | 0.950471 | 12.621756 | 14.052404 |       0.0 | 1.190448 | -0.369598 | 73.381562 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.028752 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007987 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           2.058944 |          5.8 |                  0.72634 |                      0.478769 |                     0.72634 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.231993 | -0.026611 |                   0.0 |                      0.0 |                       0.0 |                   0.365966 |                  1.290674 |                0.040197 |                 0.004964 |    -0.053778 |       -0.040197 |        -0.004964 |         -0.924708 |             16.206897 |                 100.0 |               518.290215 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.441567 | 0.207536 |      0.0 | 0.002085 |        0.745894 |      0.0 |   0.018471 | 0.018471 |        0.0 |      0.0 | 0.000548 |  0.000413 |      0.0 |   0.00012 | -0.013984 | 0.004723 |          0.0 |     0.728326 | 0.168242 | 0.168242 |      0.0 |      0.0 | 0.001462 | -0.037302 | -0.045743 | -0.045743 | -0.004084 | -0.078081 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.02801 |  0.00778 |      2.2 | 6.740571 | 7.395907 |       0.0 | 0.957809 | 12.465506 | 14.536388 |       0.0 | 1.197789 | -0.312087 | 73.381442 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.02801 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00778 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           2.200941 |          6.2 |                 0.731547 |                      0.486761 |                    0.731547 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.040289 |  -0.02247 |                   0.0 |                      0.0 |                       0.0 |                   0.366611 |                   1.28612 |                0.038308 |                 0.003183 |    -0.053588 |       -0.038308 |        -0.003183 |         -0.919509 |              15.16129 |                 100.0 |               511.763683 |               69.444444 |  0.94 |    0.0 |       49.0 |    0.442188 | 0.415656 |      0.0 | 0.004351 |        0.745291 |      0.0 |   0.040986 | 0.040986 |        0.0 |      0.0 | 0.000768 |  0.002303 |      0.0 |  0.000639 | -0.015071 | 0.009839 |          0.0 |     0.727847 |   0.2218 |   0.2218 |      0.0 |      0.0 | 0.002047 | -0.040202 | -0.049694 | -0.049694 | -0.009339 | -0.086712 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.027286 |  0.00758 |      2.2 | 6.698927 | 7.810796 |       0.0 | 0.965174 | 12.255629 |  15.49146 |       0.0 | 1.205159 | -0.258717 | 73.380804 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.027286 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00758 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           1.987946 |          5.6 |                 0.742061 |                      0.496568 |                    0.742061 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.862389 | -0.018628 |                   0.0 |                      0.0 |                       0.0 |                   0.364735 |                  1.277745 |                0.034017 |                  -0.0005 |    -0.053239 |       -0.034017 |           0.0005 |          -0.91301 |             16.785714 |                 100.0 |               498.496274 |               69.444444 |  1.88 |    0.0 |       49.0 |    0.443432 | 0.833653 |      0.0 | 0.009424 |        0.744689 |      0.0 |   0.071301 | 0.071301 |        0.0 |      0.0 | 0.000731 |  0.001791 |      0.0 |  0.000535 | -0.015425 | 0.021252 |          0.0 |     0.727368 | 0.283309 | 0.283309 |      0.0 |      0.0 | 0.001949 | -0.041144 | -0.050357 | -0.050357 | -0.014391 |    -0.092 |      1.0 |      0.3 | 0.251264 | 0.086066 |  0.0 |  0.0 | 0.0 |  0.0 | 0.086066 | 0.027698 | 0.007694 |      2.2 | 6.627066 | 8.643719 |       0.0 | 0.972586 | 11.982951 |  17.13562 |       0.0 | 1.212578 | -0.209189 | 73.545467 | 0.086066 |      0.0 |   0.0 |   0.0 |  0.0 | 0.027698 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007694 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |            1.84595 |          5.2 |                  0.76893 |                      0.509073 |                     0.76893 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.697297 | -0.015062 |                   0.0 |                      0.0 |                       0.0 |                   0.365362 |                  1.268373 |                0.022679 |                -0.009872 |    -0.052849 |       -0.022679 |         0.009872 |         -0.903012 |             18.076923 |                 100.0 |               466.431761 |               69.444444 | 1.974 |    0.0 |       49.0 |    0.445931 | 0.880268 |      0.0 |  0.01116 |        0.744091 | 0.238141 |   0.133473 | 0.133473 |        0.0 |      0.0 | 0.000362 | -0.001204 |      0.0 | -0.000382 | -0.018246 |      0.0 |          0.0 |     0.726899 | 0.301036 | 0.301036 |      0.0 |      0.0 | 0.000965 | -0.048671 | -0.057065 | -0.057065 |  -0.03342 | -0.115608 |      1.0 |      0.3 |  2.40494 | 0.825592 |  0.0 |  0.0 | 0.0 |  0.0 | 0.825592 | 0.051037 | 0.014177 |      2.2 | 6.493345 | 9.285484 |       0.0 |     0.98 | 11.694354 | 16.722926 |       0.0 | 1.219886 | -0.163227 | 75.125197 |      0.5 | 0.325592 |   0.0 |   0.0 |  0.0 | 0.034575 | 0.016462 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014177 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |           2.023445 |          5.7 |                 0.807986 |                      0.524283 |                    0.807986 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.544091 | -0.011752 |                   0.0 |                      0.0 |                       0.0 |                   0.365985 |                  1.257174 |                0.007092 |                -0.022991 |    -0.052382 |       -0.007092 |         0.022991 |          -0.89119 |             16.491228 |                 100.0 |               420.659902 |               69.444444 | 1.692 |    0.0 |       49.0 |    0.447856 | 0.757773 |      0.0 |      0.0 |        0.743506 | 1.039941 |   0.197325 | 0.197325 |        0.0 |      0.0 | 0.000201 | -0.002413 |      0.0 | -0.000729 | -0.024333 |      0.0 |          0.0 |     0.726434 | 0.448142 | 0.448142 |      0.0 |      0.0 | 0.000536 | -0.064906 | -0.073535 | -0.073535 | -0.065906 | -0.161432 |      1.0 |      0.3 | 2.333776 | 0.803499 |  0.0 |  0.0 | 0.0 |  0.0 | 0.803499 | 0.092667 | 0.025741 |      2.2 | 6.295884 | 9.003115 |       0.0 | 0.987308 | 11.262576 | 16.105483 |       0.0 | 1.227194 | -0.120575 | 76.656202 |      0.5 | 0.303499 |   0.0 |   0.0 |  0.0 | 0.046597 | 0.046069 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025741 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           2.342937 |          6.6 |                 0.836943 |                      0.541728 |                    0.836943 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.401917 | -0.008681 |                   0.0 |                      0.0 |                       0.0 |                   0.366603 |                  1.245134 |               -0.003979 |                -0.032391 |    -0.051881 |        0.003979 |         0.032391 |         -0.878531 |             14.242424 |                 100.0 |               390.309935 |               69.444444 | 1.222 |    0.0 |       49.0 |    0.447009 | 0.546245 |      0.0 |      0.0 |        0.742925 | 0.950524 |   0.282719 | 0.282719 |        0.0 |      0.0 | 0.000229 | -0.002135 |      0.0 | -0.000588 | -0.030654 |      0.0 |          0.0 |     0.725971 | 0.588433 | 0.588433 |      0.0 |      0.0 |  0.00061 | -0.081767 | -0.091059 | -0.091059 | -0.094428 | -0.205218 |      1.0 |      0.3 | 2.065061 | 0.712872 |  0.0 |  0.0 | 0.0 |  0.0 | 0.712872 | 0.125382 | 0.034828 |      2.2 | 6.013012 | 8.598608 |       0.0 | 0.994616 | 10.694459 | 15.293076 |       0.0 | 1.234502 | -0.080994 | 78.008979 |      0.5 | 0.212872 |   0.0 |   0.0 |  0.0 | 0.058309 | 0.067073 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034828 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           2.484933 |          7.0 |                 0.860763 |                      0.562266 |                    0.860763 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.269979 | -0.005832 |                   0.0 |                      0.0 |                       0.0 |                   0.367218 |                  1.232336 |                -0.01183 |                -0.039709 |    -0.051347 |         0.01183 |         0.039709 |         -0.865118 |             13.428571 |                 100.0 |               368.937554 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.445796 | 0.209524 |      0.0 |      0.0 |        0.742349 | 0.688269 |   0.334834 | 0.334834 |        0.0 |      0.0 | 0.001222 |  0.005737 |      0.0 |  0.001528 | -0.034262 |      0.0 |          0.0 |     0.725511 | 0.684233 | 0.684233 |      0.0 |      0.0 | 0.003259 | -0.091392 | -0.103263 | -0.103263 | -0.111835 | -0.234365 |      1.0 |      0.3 | 1.450645 | 0.501782 |  0.0 |  0.0 | 0.0 |  0.0 | 0.501782 | 0.140609 | 0.039058 |      2.2 | 5.677371 | 8.118641 |       0.0 | 1.001924 | 10.032652 | 14.346693 |       0.0 |  1.24181 | -0.044262 | 78.956315 |      0.5 | 0.001782 |   0.0 |   0.0 |  0.0 | 0.069718 |  0.07089 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039058 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |            2.59143 |          7.3 |                 0.860763 |                      0.585275 |                    0.860763 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -0.14754 | -0.003187 |                   0.0 |                      0.0 |                       0.0 |                   0.367829 |                  1.218349 |               -0.009926 |                -0.039275 |    -0.050765 |        0.009926 |         0.039275 |          -0.85052 |             12.876712 |                 100.0 |               366.456526 |               69.444444 | 0.188 |    0.0 |       49.0 |    0.444356 | 0.083539 |      0.0 |      0.0 |        0.741776 | 0.546141 |   0.323618 | 0.323618 |        0.0 |      0.0 | 0.003016 |  0.020252 |      0.0 |  0.005233 |  -0.03348 |      0.0 |          0.0 |     0.725055 | 0.693418 | 0.693418 |      0.0 |      0.0 | 0.008045 | -0.089306 | -0.106208 | -0.106208 | -0.108089 | -0.234788 |      1.0 |      0.3 | 1.181902 |  0.40946 |  0.0 |  0.0 | 0.0 |  0.0 |  0.40946 |  0.14359 | 0.039886 |      2.2 | 5.351764 | 7.653023 |       0.0 | 1.009232 |  9.361029 | 13.386271 |       0.0 | 1.249118 | -0.010175 | 79.723524 |  0.40946 |      0.0 |   0.0 |   0.0 |  0.0 | 0.079658 | 0.063932 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039886 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           2.555931 |          7.2 |                 0.879017 |                      0.610223 |                    0.879017 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.033917 | -0.000733 |                   0.0 |                      0.0 |                       0.0 |                   0.368436 |                  1.204312 |               -0.016369 |                -0.044906 |     -0.05018 |        0.016369 |         0.044906 |         -0.835877 |             13.055556 |                 100.0 |               352.261829 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.442959 | 0.208191 |      0.0 |      0.0 |        0.741208 | 0.718975 |   0.357346 | 0.357346 |        0.0 |      0.0 | 0.003306 |  0.022273 |      0.0 |   0.00585 | -0.035493 |      0.0 |          0.0 |     0.724601 | 0.750456 | 0.750456 |      0.0 |      0.0 | 0.008818 | -0.094676 | -0.111803 | -0.111803 | -0.119354 | -0.251385 | 0.999197 | 0.299518 | 1.545997 | 0.535856 |  0.0 |  0.0 | 0.0 |  0.0 | 0.535856 | 0.148742 | 0.041317 |      2.2 | 4.992261 | 7.138933 |       0.0 |  1.01654 |  8.633404 | 12.345767 |       0.0 | 1.256426 |  0.021457 | 80.727816 |      0.5 | 0.035856 |   0.0 |   0.0 |  0.0 | 0.089352 | 0.059391 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041317 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           2.697927 |          7.6 |                 0.879017 |                      0.635914 |                    0.879017 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                      0.0 |                       0.0 |                   0.369039 |                  1.190153 |               -0.014842 |                -0.044558 |     -0.04959 |        0.014842 |         0.044558 |         -0.821114 |             12.368421 |                 100.0 |               349.137797 |               69.444444 | 0.188 |    0.0 |       49.0 |    0.441417 | 0.082986 |      0.0 |      0.0 |        0.740644 | 0.586992 |   0.352693 | 0.352693 |        0.0 |      0.0 |  0.00488 |  0.035379 |      0.0 |  0.008934 | -0.035367 |      0.0 |          0.0 |     0.724151 | 0.763085 | 0.763085 |      0.0 |      0.0 | 0.013017 | -0.094338 | -0.115974 | -0.115974 | -0.117799 |  -0.25487 | 0.998041 | 0.298825 | 1.282337 | 0.444387 |  0.0 |  0.0 | 0.0 |  0.0 | 0.444387 | 0.154476 |  0.04291 |      2.2 | 4.636396 | 6.630047 |       0.0 | 1.023848 |  7.892865 | 11.286797 |       0.0 | 1.263734 |  0.052357 | 81.556832 | 0.444387 |      0.0 |   0.0 |   0.0 |  0.0 | 0.099238 | 0.055239 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04291 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |            2.59143 |          7.3 |                 0.897612 |                      0.665151 |                    0.897612 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 |       0.0 |       0.0 |              0.000896 |                 0.000686 |                  0.000166 |                   0.369638 |                  1.174779 |               -0.020032 |                -0.049916 |    -0.048053 |        0.020718 |         0.050082 |         -0.805141 |             12.876712 |                 100.0 |               336.997925 |               95.413772 | 0.094 |    0.0 |       49.0 |     0.43989 |  0.04135 |      0.0 |      0.0 |        0.740084 | 0.569278 |   0.369532 | 0.369532 |        0.0 |      0.0 |  0.00629 |  0.045774 |      0.0 |  0.009387 | -0.035585 |      0.0 |          0.0 |     0.723703 |  0.79188 |  0.79188 |      0.0 |      0.0 | 0.016779 | -0.094922 | -0.119484 | -0.119484 | -0.123424 | -0.264488 | 0.996884 | 0.298133 | 1.229959 | 0.426099 |  0.0 |  0.0 | 0.0 |  0.0 | 0.426099 | 0.157665 | 0.043796 |      2.2 | 4.262817 | 6.095828 |       0.0 | 1.031156 |  7.123373 | 10.186423 |       0.0 | 1.271042 |  0.083257 | 82.351305 | 0.426099 |      0.0 |   0.0 |   0.0 |  0.0 | 0.107916 | 0.049748 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043796 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           2.520432 |          7.1 |                 0.891375 |                      0.694736 |                    0.891375 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 |       0.0 |       0.0 |              0.017203 |                 0.013185 |                  0.003195 |                   0.370233 |                   1.15845 |               -0.019333 |                -0.048363 |    -0.031066 |        0.032518 |         0.051558 |         -0.788217 |             13.239437 |                 100.0 |               334.780842 |               319.92853 | 0.282 |    0.0 |       49.0 |    0.438287 | 0.123597 |      0.0 |      0.0 |        0.739528 | 0.701208 |   0.404237 | 0.404237 |        0.0 |      0.0 | 0.004726 |  0.036661 |      0.0 |  0.002679 | -0.035482 |      0.0 |          0.0 |     0.723258 | 0.782359 | 0.782359 |      0.0 |      0.0 | 0.012606 | -0.094646 | -0.115104 | -0.115104 | -0.135015 | -0.261308 | 0.995728 | 0.297442 | 1.404671 |  0.48654 |  0.0 |  0.0 | 0.0 |  0.0 |  0.48654 | 0.161723 | 0.044923 |      2.2 | 3.855588 | 5.513491 |       0.0 | 1.038464 |  6.363101 |  9.099234 |       0.0 |  1.27835 |  0.114157 | 83.266757 |  0.48654 |      0.0 |   0.0 |   0.0 |  0.0 | 0.116919 | 0.044804 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044923 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           2.520432 |          7.1 |                 0.903886 |                       0.72095 |                    0.903886 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 |       0.0 |       0.0 |              0.052505 |                 0.040262 |                  0.009751 |                   0.370824 |                  1.144953 |                -0.02271 |                -0.051921 |     0.004799 |        0.062972 |         0.061672 |         -0.774129 |             13.239437 |                 100.0 |               325.462462 |              311.572171 | 0.094 |    0.0 |       49.0 |     0.43654 | 0.041035 |      0.0 |      0.0 |        0.738976 | 0.754731 |   0.499616 | 0.499616 |        0.0 |      0.0 | 0.005828 |  0.051889 |      0.0 |  0.003906 | -0.036172 |      0.0 |          0.0 |     0.722816 | 0.831158 | 0.831158 |      0.0 |      0.0 | 0.015547 | -0.096487 | -0.119447 | -0.119447 | -0.166872 | -0.277607 | 0.994571 | 0.296752 |  1.28728 | 0.445793 |  0.0 |  0.0 | 0.0 |  0.0 | 0.445793 | 0.166289 | 0.046191 |      2.2 | 3.352424 | 4.793967 |       0.0 | 1.045772 |  5.553934 |  7.942126 |       0.0 | 1.285658 |  0.145057 | 84.104338 | 0.445793 |      0.0 |   0.0 |   0.0 |  0.0 | 0.125938 | 0.040351 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046191 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           3.159415 |          8.9 |                 0.942351 |                      0.742943 |                    0.942351 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 |       0.0 |       0.0 |              0.080091 |                 0.061446 |                  0.014874 |                   0.361639 |                  1.076886 |               -0.035254 |                -0.063181 |     0.035221 |        0.096701 |         0.078055 |         -0.715247 |             10.561798 |                 100.0 |               310.314611 |              297.925078 | 0.094 |    0.0 |       49.0 |    0.434382 | 0.040832 |      0.0 |      0.0 |        0.738428 | 0.964112 |   0.647442 | 0.647442 |        0.0 |      0.0 | 0.013265 |  0.120077 |      0.0 |  0.007797 | -0.042299 |      0.0 |          0.0 |     0.722377 | 1.035376 | 1.035376 |      0.0 |      0.0 | 0.035384 | -0.112831 | -0.154929 | -0.154929 | -0.216246 | -0.345815 | 0.993415 | 0.296062 | 1.582473 | 0.548042 |  0.0 |  0.0 | 0.0 |  0.0 | 0.548042 | 0.173673 | 0.048243 |      2.2 | 2.697498 | 3.857422 |       0.0 |  1.05308 |  4.542624 |  6.495952 |       0.0 | 1.292966 |  0.175957 | 85.130972 |      0.5 | 0.048042 |   0.0 |   0.0 |  0.0 | 0.134904 |  0.03877 |   0.0 |   0.0 |  0.0 |   0.0 | 0.048243 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           3.372409 |          9.5 |                 0.962118 |                      0.763489 |                    0.962118 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 |       0.0 |       0.0 |              0.080358 |                 0.061682 |                  0.014924 |                   0.362207 |                  1.067855 |               -0.039673 |                -0.068409 |     0.035864 |        0.101355 |         0.083332 |         -0.705648 |              9.894737 |                 100.0 |               300.386794 |              288.938191 |   0.0 |    0.0 |       49.0 |    0.431572 |      0.0 |      0.0 |      0.0 |        0.737884 | 0.966223 |   0.678654 | 0.678654 |        0.0 |      0.0 | 0.016902 |  0.150602 |      0.0 |  0.009557 | -0.044819 |      0.0 |          0.0 |      0.72194 | 1.117621 | 1.117621 |      0.0 |      0.0 | 0.045085 | -0.119553 |   -0.1704 |   -0.1704 |  -0.22667 | -0.373285 | 0.992258 | 0.295373 | 1.609224 | 0.557404 |  0.0 |  0.0 | 0.0 |  0.0 | 0.557404 | 0.184499 |  0.05125 |      2.2 | 2.009997 | 2.874296 |       0.0 | 1.060388 |  3.448634 |  4.931547 |       0.0 | 1.300274 |  0.206857 | 86.173235 |      0.5 | 0.057404 |   0.0 |   0.0 |  0.0 | 0.144334 | 0.040165 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05125 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           2.804424 |          7.9 |                 0.962118 |                      0.781298 |                    0.962118 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 |       0.0 |       0.0 |              0.076031 |                  0.05839 |                   0.01412 |                   0.362772 |                   1.06128 |               -0.032948 |                -0.066875 |     0.031811 |        0.091338 |         0.080995 |         -0.698508 |             11.898734 |                 100.0 |               302.438987 |              290.798665 | 0.094 |    0.0 |       49.0 |    0.428623 | 0.040291 |      0.0 |      0.0 |        0.737343 |  0.82333 |   0.551649 | 0.551649 |        0.0 |      0.0 |   0.0212 |  0.176302 |      0.0 |  0.013161 | -0.032937 |      0.0 |          0.0 |     0.721507 | 0.953039 | 0.953039 |      0.0 |      0.0 | 0.056549 | -0.087858 | -0.149462 | -0.149462 | -0.184251 | -0.318315 | 0.991102 | 0.294685 | 1.465949 | 0.507809 |  0.0 |  0.0 | 0.0 |  0.0 | 0.507809 | 0.192892 | 0.053581 |      2.2 | 1.447592 | 2.070057 |       0.0 | 1.067696 |  2.512262 |  3.592535 |       0.0 | 1.307582 |  0.237757 | 87.118213 |      0.5 | 0.007809 |   0.0 |   0.0 |  0.0 | 0.153521 |  0.03937 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053581 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           3.407908 |          9.6 |                 0.968788 |                      0.797357 |                    0.968788 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 |       0.0 |       0.0 |              0.042402 |                  0.03258 |                  0.007875 |                   0.363334 |                  1.056546 |               -0.029679 |                -0.067539 |    -0.001621 |         0.06226 |         0.075414 |         -0.693212 |              9.791667 |                 100.0 |               301.150454 |              289.630679 |   0.0 |    0.0 |       49.0 |     0.42621 |      0.0 |      0.0 |      0.0 |        0.736807 | 0.656764 |   0.466111 | 0.466111 |        0.0 |      0.0 | 0.030355 |  0.252179 |      0.0 |  0.015859 | -0.033021 |      0.0 |          0.0 |     0.721076 | 1.011636 | 1.011636 |      0.0 |      0.0 | 0.080971 | -0.088081 | -0.174392 | -0.174392 | -0.155681 | -0.337886 | 0.989945 | 0.293998 | 1.459936 | 0.505733 |  0.0 |  0.0 | 0.0 |  0.0 | 0.505733 |   0.1986 | 0.055167 |      2.2 | 0.967089 | 1.382938 |       0.0 | 1.075004 |  1.514419 |  2.165619 |       0.0 |  1.31489 |  0.268657 | 88.056557 |      0.5 | 0.005733 |   0.0 |   0.0 |  0.0 | 0.162471 | 0.036129 |   0.0 |   0.0 |  0.0 |   0.0 | 0.055167 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |           3.230413 |          9.1 |                 0.955489 |                       0.81108 |                    0.955489 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 |       0.0 |       0.0 |              0.012685 |                 0.009752 |                  0.002356 |                   0.373725 |                  1.110113 |               -0.020897 |                -0.062718 |    -0.033569 |        0.030649 |         0.065074 |         -0.736388 |              10.32967 |                 100.0 |               301.529977 |              289.974756 | 0.094 |    0.0 |       49.0 |    0.424149 |  0.03987 |      0.0 |      0.0 |        0.736274 | 0.531683 |   0.351155 | 0.351155 |        0.0 |      0.0 | 0.028472 |   0.22909 |      0.0 |  0.015032 | -0.030224 |      0.0 |          0.0 |     0.720647 | 0.922995 | 0.922995 |      0.0 |      0.0 | 0.075948 | -0.080621 | -0.162585 | -0.162585 | -0.117286 |  -0.30828 | 0.988789 | 0.293311 | 1.432304 | 0.496156 |  0.0 |  0.0 | 0.0 |  0.0 | 0.496156 | 0.203958 | 0.056655 |      2.2 | 0.603254 | 0.862653 |       0.0 | 1.082312 |  0.599678 |   0.85754 |       0.0 | 1.322198 |  0.299557 | 88.977673 | 0.496156 |      0.0 |   0.0 |   0.0 |  0.0 |  0.17114 | 0.032818 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056655 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           3.088417 |          8.7 |                 0.929372 |                      0.822835 |                    0.929372 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 |       0.0 |       0.0 |              0.000022 |                 0.000017 |                  0.000004 |                   0.374295 |                  1.107879 |               -0.011478 |                -0.054953 |     -0.04614 |        0.011495 |         0.054957 |         -0.733585 |             10.804598 |                 100.0 |               307.996057 |               72.420963 |   0.0 |    0.0 |       49.0 |    0.422588 |      0.0 |      0.0 |      0.0 |        0.735745 | 0.357286 |   0.258008 | 0.258008 |        0.0 |      0.0 | 0.027276 |  0.217285 |      0.0 |    0.0489 | -0.025254 |      0.0 |          0.0 |     0.720222 | 0.807656 | 0.599678 |      0.0 |      0.0 | 0.072757 | -0.067363 | -0.147437 | -0.147437 | -0.086175 | -0.269757 | 0.987633 | 0.292625 | 0.882793 | 0.305647 |  0.0 |  0.0 | 0.0 |  0.0 | 0.305647 |  0.20662 | 0.057395 |      2.2 | 0.334329 | 0.478091 |       0.0 |  1.08962 |       0.0 |       0.0 |       0.0 | 1.329506 |  0.330457 | 89.505919 | 0.305647 |      0.0 |   0.0 |   0.0 |  0.0 | 0.177064 | 0.029556 |   0.0 |   0.0 |  0.0 |   0.0 | 0.057395 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           3.159415 |          8.9 |                 0.903886 |                      0.833518 |                    0.903886 |                          0.0 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.1188 |                   0.0 |                      0.0 |                       0.0 |                   1.177288 |                   1.10789 |                0.006861 |                      0.0 |    -0.046162 |       -0.006861 |              0.0 |          0.069398 |             10.561798 |                 100.0 |               325.514283 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.421434 |      0.0 |      0.0 |      0.0 |         0.73522 | 0.150101 |   0.113409 | 0.113409 |        0.0 |      0.0 | 0.035596 |  0.290175 |      0.0 |  0.065707 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 0.094951 |       0.0 |  -0.13969 |       0.0 | -0.037879 |       0.0 |  0.98203 | 0.289315 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.203039 |   0.0564 |      2.2 | 0.204472 | 0.292394 |       0.0 | 1.096928 |       0.0 |       0.0 |       0.0 |      nan |  0.480157 | 89.440212 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.176421 | 0.026619 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0564 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           2.804424 |          7.9 |                 0.879017 |                      0.843583 |                    0.879017 |                          0.0 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.11016 |                   0.0 |                      0.0 |                       0.0 |                   1.177288 |                  1.107494 |                0.006982 |                      0.0 |    -0.046146 |       -0.006982 |              0.0 |          0.069794 |             11.898734 |                 100.0 |               329.248962 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.420877 |      0.0 |      0.0 |      0.0 |        0.734698 | 0.180266 |   0.137736 | 0.137736 |        0.0 |      0.0 | 0.024787 |  0.197149 |      0.0 |  0.048384 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 0.066117 |       0.0 | -0.119103 |       0.0 | -0.046004 |       0.0 | 0.976751 | 0.286213 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.195837 | 0.054399 |      2.2 | 0.061078 | 0.087341 |       0.0 | 1.104236 |       0.0 |       0.0 |       0.0 |      nan |  0.621217 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.171864 | 0.023973 |   0.0 |   0.0 |  0.0 |   0.0 | 0.054399 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           2.520432 |          7.1 |                 0.860763 |                       0.85201 |                    0.860763 |                          0.0 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10368 |                   0.0 |                      0.0 |                       0.0 |                   1.177288 |                  1.107885 |                0.010846 |                      0.0 |    -0.046162 |       -0.010846 |              0.0 |          0.069403 |             13.239437 |                 100.0 |               336.291164 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.420262 |      0.0 |      0.0 |      0.0 |         0.73418 | 0.066867 |   0.116118 | 0.061078 |        0.0 |      0.0 | 0.020475 |   0.15874 |  0.15874 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 0.054616 |       0.0 | -0.104245 |       0.0 | -0.038783 |       0.0 | 0.971714 | 0.283269 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.189015 | 0.052504 |  2.04126 |      0.0 |      0.0 |       0.0 | 1.111544 |       0.0 |       0.0 |       0.0 |      nan |  0.755797 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.167424 |  0.02159 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052504 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           2.271939 |          6.4 |                  0.86681 |                       0.86026 |                         0.0 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      nan |                       0.0 |                   1.177288 |                  1.108737 |                     0.0 |                      0.0 |    -0.046197 |             0.0 |              0.0 |          0.068551 |               14.6875 |                 100.0 |               334.010232 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.148995 | 0.148995 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.966597 | 0.280293 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.182544 | 0.050707 | 1.892265 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  0.892537 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.1631 | 0.019444 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050707 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           2.555931 |          7.2 |                  0.86681 |                      0.867569 |                         0.0 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      nan |                       0.0 |                   1.177288 |                  1.109465 |                     0.0 |                      0.0 |    -0.046228 |             0.0 |              0.0 |          0.067823 |             13.055556 |                 100.0 |               329.382578 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.137771 | 0.137771 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.961479 | 0.277333 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.176399 |    0.049 | 1.754494 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.029277 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.158887 | 0.017512 |   0.0 |   0.0 |  0.0 |   0.0 |    0.049 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           2.484933 |          7.0 |                 0.854752 |                      0.873404 |                         0.0 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10152 |                   0.0 |                      nan |                       0.0 |                   1.177288 |                   1.11033 |                     0.0 |                      0.0 |    -0.046264 |             0.0 |              0.0 |          0.066958 |             13.428571 |                 100.0 |               333.139242 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.115706 | 0.115706 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.956523 | 0.274481 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.170554 | 0.047376 | 1.732788 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.161697 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.154783 | 0.015771 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047376 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |           2.413935 |          6.8 |                 0.802302 |                      0.876717 |                         0.0 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08208 |                   0.0 |                      nan |                       0.0 |                   1.177288 |                  1.112997 |                     0.0 |                      0.0 |    -0.046375 |             0.0 |              0.0 |          0.064291 |             13.823529 |                 100.0 |               369.069697 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |   0.12638 |  0.12638 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.952295 |  0.27206 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164988 |  0.04583 | 1.700408 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.274677 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.150784 | 0.014204 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04583 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           2.271939 |          6.4 |                 0.807986 |                      0.879785 |                         0.0 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08424 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.116023 |                     0.0 |                      0.0 |    -0.046501 |             0.0 |              0.0 |          0.054049 |               14.6875 |                 100.0 |                365.55183 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.121939 | 0.121939 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.947986 | 0.269603 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.159682 | 0.044356 |  1.57847 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.389817 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.146889 | 0.012792 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044356 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |            2.23644 |          6.3 |                 0.796653 |                      0.881066 |                         0.0 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07992 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.119656 |                     0.0 |                      0.0 |    -0.046652 |             0.0 |              0.0 |          0.050416 |             14.920635 |                 100.0 |               372.207112 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.111599 | 0.111599 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.943838 | 0.267249 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.154616 | 0.042949 |  1.46687 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.500637 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.143095 | 0.011521 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042949 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |           2.023445 |          5.7 |                  0.78546 |                      0.880041 |                         0.0 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0756 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.123475 |                     0.0 |                      0.0 |    -0.046811 |             0.0 |              0.0 |          0.046597 |             16.491228 |                 100.0 |               375.816236 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.076984 | 0.076984 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.939853 | 0.264997 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.149775 | 0.041604 | 1.483887 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.607137 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.139399 | 0.010376 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041604 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |            1.84595 |          5.2 |                 0.774405 |                      0.877228 |                         0.0 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.127474 |                     0.0 |                      0.0 |    -0.046978 |             0.0 |              0.0 |          0.042598 |             18.076923 |                 100.0 |               381.399302 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.055406 | 0.055406 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.936029 | 0.262845 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.145143 | 0.040317 | 1.428481 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.709317 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.135798 | 0.009345 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040317 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |           1.632956 |          4.6 |                 0.774405 |                      0.873404 |                         0.0 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.131856 |                     0.0 |                      0.0 |    -0.047161 |             0.0 |              0.0 |          0.038216 |             20.434783 |                 100.0 |               382.164664 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.051903 | 0.051903 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.932204 | 0.260702 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.140706 | 0.039085 | 1.376578 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.811497 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.13229 | 0.008416 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039085 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           1.739453 |          4.9 |                 0.758081 |                      0.868834 |                         0.0 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0648 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.136687 |                     0.0 |                      0.0 |    -0.047362 |             0.0 |              0.0 |          0.033385 |             19.183673 |                 100.0 |               398.852717 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.065481 | 0.065481 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.928623 | 0.258702 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.136453 | 0.037903 | 1.311096 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.907197 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.128873 | 0.007579 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037903 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |            2.23644 |          6.3 |                  0.70079 |                      0.860763 |                         0.0 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.04104 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.144252 |                     0.0 |                      0.0 |    -0.047677 |             0.0 |              0.0 |           0.02582 |             14.920635 |                 100.0 |               467.074523 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.108633 | 0.108633 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 |  0.92593 | 0.257204 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.13237 |  0.03677 | 1.202463 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  1.979137 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.125544 | 0.006826 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03677 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           2.058944 |          5.8 |                 0.666353 |                      0.851015 |                         0.0 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.02592 |                   0.0 |                      nan |                       0.0 |                   1.170072 |                  1.152031 |                     0.0 |                      0.0 |    -0.048001 |             0.0 |              0.0 |          0.018042 |             16.206897 |                 100.0 |               517.187941 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.092335 | 0.092335 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.923804 | 0.256024 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.128449 |  0.03568 | 1.110129 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.035957 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.122301 | 0.006148 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03568 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           2.165441 |          6.1 |                 0.647352 |                      0.839643 |                         0.0 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 |  -0.01728 |              0.002833 |                      nan |                  0.000526 |                   1.170072 |                  1.158654 |                     0.0 |                      0.0 |    -0.045445 |             0.0 |              0.0 |          0.011418 |             15.409836 |                 100.0 |               543.089971 |               98.153323 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.060806 | 0.060806 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.922001 | 0.255026 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.124679 | 0.034633 | 1.049322 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.084137 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.119142 | 0.005537 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034633 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           2.094443 |          5.9 |                 0.647352 |                      0.828647 |                         0.0 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 |  -0.01728 |              0.056113 |                      nan |                  0.010421 |                   1.170072 |                  1.166643 |                     0.0 |                      0.0 |     0.007503 |             0.0 |              0.0 |          0.003429 |             15.932203 |                 100.0 |               549.868272 |              504.951032 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.093283 | 0.093283 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.920198 | 0.254029 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.121051 | 0.033625 | 0.956039 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.132317 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.116065 | 0.004986 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033625 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           2.271939 |          6.4 |                 0.633417 |                      0.816579 |                         0.0 |                          0.0 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |   -0.0108 |              0.172031 |                      nan |                  0.031949 |                   1.170072 |                  1.175846 |                     0.0 |                      0.0 |     0.123037 |             0.0 |              0.0 |         -0.005774 |               14.6875 |                 100.0 |               585.123822 |              533.908719 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.140468 | 0.140468 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.918637 | 0.253168 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.117557 | 0.032655 | 0.909571 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.174017 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.113067 | 0.004491 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032655 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |           2.413935 |          6.8 |                 0.624275 |                      0.802774 |                         0.0 |                          0.0 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 |  -0.00648 |              0.258205 |                      nan |                  0.047952 |                   1.170072 |                   1.18712 |                     0.0 |                      0.0 |     0.208742 |             0.0 |              0.0 |         -0.017048 |             13.823529 |                 100.0 |                622.68293 |              564.355905 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.221518 | 0.221518 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.917238 | 0.252398 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.11419 |  0.03172 | 0.688053 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.211397 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.110146 | 0.004044 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03172 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           2.484933 |          7.0 |                 0.606339 |                      0.787548 |                         0.0 |                          0.0 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |   0.00216 |              0.258716 |                      nan |                  0.048047 |                   1.170072 |                  1.199167 |                     0.0 |                      0.0 |      0.20875 |             0.0 |              0.0 |         -0.029095 |             13.428571 |                 100.0 |               676.190967 |               607.03086 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.238232 | 0.238232 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.916163 | 0.251806 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.110943 | 0.030818 | 0.449821 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.240137 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.107301 | 0.003642 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030818 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |            2.23644 |          6.3 |                 0.593188 |                      0.771892 |                         0.0 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |   0.00864 |              0.244186 |                      nan |                  0.045349 |                   1.170072 |                   1.21168 |                     0.0 |                      0.0 |     0.193699 |             0.0 |              0.0 |         -0.041608 |             14.920635 |                 100.0 |               730.950059 |              649.872347 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.249511 | 0.249511 |       0.0 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.915329 | 0.251348 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.10781 | 0.029947 | 0.200311 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.262397 | 89.391828 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.104529 |  0.00328 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029947 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |            2.59143 |          7.3 |                  0.58456 |                      0.755838 |                         0.0 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |   0.01296 |              0.190885 |                      nan |                   0.03545 |                   1.231767 |                  1.349977 |                     0.0 |                      0.0 |     0.134636 |             0.0 |              0.0 |          -0.11821 |             12.876712 |                 100.0 |               769.993016 |              679.917834 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.294252 | 0.200311 |  0.002809 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.914658 |  0.25098 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.104783 | 0.029107 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.280337 |  89.38902 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.101829 | 0.002954 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029107 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           2.662428 |          7.5 |                 0.576044 |                       0.74008 |                         0.0 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |   0.01728 |              0.047005 |                      nan |                  0.008729 |                   1.262615 |                  1.428152 |                     0.0 |                      0.0 |    -0.012502 |             0.0 |              0.0 |         -0.165538 |             12.533333 |                 100.0 |               811.899181 |              711.714211 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.296611 |      0.0 |  0.008227 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.914148 |   0.2507 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.10186 | 0.028294 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.293957 | 89.380793 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.099199 | 0.002661 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028294 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           2.484933 |          7.0 |                 0.555231 |                      0.724396 |                         0.0 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |   0.02808 |              0.000535 |                      nan |                  0.000099 |                    1.32431 |                  1.571379 |                     0.0 |                      0.0 |    -0.064939 |             0.0 |              0.0 |         -0.247069 |             13.428571 |                 100.0 |               915.122139 |               72.512894 |   0.0 |    0.0 |     45.325 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.256087 |      0.0 |  0.058496 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.914043 | 0.250642 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.099033 | 0.027509 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.296777 | 89.322297 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.096636 | 0.002396 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027509 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           2.129942 |          6.0 |                 0.543065 |                      0.709217 |                         0.0 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.03456 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.583999 |                     0.0 |                      0.0 |       -0.066 |             0.0 |              0.0 |         -0.259688 |             15.666667 |                 100.0 |                987.92442 |               69.444444 |   0.0 |    0.0 |      41.65 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.202105 |      0.0 |  0.053184 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 |  0.91418 | 0.250717 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.096298 |  0.02675 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.293117 | 89.269113 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.09414 | 0.002158 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02675 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           1.703954 |          4.8 |                  0.52721 |                       0.69432 |                         0.0 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |    0.0432 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                    1.5975 |                     0.0 |                      0.0 |    -0.066562 |             0.0 |              0.0 |          -0.27319 |             19.583333 |                 100.0 |              1109.093677 |               69.444444 |   0.0 |    0.0 |      36.75 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |   0.14534 |      0.0 |  0.044429 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 |  0.91464 |  0.25097 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.093652 | 0.026014 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.280817 | 89.224684 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091708 | 0.001944 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026014 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |            1.49096 |          4.2 |                 0.523311 |                      0.680107 |                         0.0 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04536 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.611195 |                     0.0 |                      0.0 |    -0.067133 |             0.0 |              0.0 |         -0.286884 |             22.380952 |                 100.0 |              1146.737517 |               69.444444 |   0.0 |    0.0 |     35.525 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.125512 |      0.0 |  0.041865 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.915181 | 0.251267 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.09109 | 0.025303 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.266357 | 89.182819 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.08934 |  0.00175 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025303 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           1.739453 |          4.9 |                 0.519437 |                      0.665752 |                         0.0 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04752 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.625096 |                     0.0 |                      0.0 |    -0.067712 |             0.0 |              0.0 |         -0.300786 |             19.183673 |                 100.0 |              1190.723506 |               69.444444 |   0.0 |    0.0 |       34.3 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.150562 |      0.0 |  0.045237 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.915803 | 0.251609 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.088608 | 0.024613 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.249737 | 89.137582 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.087032 | 0.001577 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024613 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           1.419962 |          4.0 |                 0.504195 |                      0.650878 |                         0.0 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.05616 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.639665 |                     0.0 |                      0.0 |    -0.068319 |             0.0 |              0.0 |         -0.315355 |                  23.5 |                 100.0 |              1378.722537 |               69.444444 |   0.0 |    0.0 |       29.4 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.111614 |      0.0 |  0.038133 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.916749 | 0.252128 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.086204 | 0.023945 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.224477 | 89.099449 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084784 |  0.00142 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023945 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |           1.206967 |          3.4 |                 0.485698 |                      0.635721 |                         0.0 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.06696 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.653965 |                     0.0 |                      0.0 |    -0.068915 |             0.0 |              0.0 |         -0.329655 |             27.647059 |                 100.0 |              1724.589241 |               69.444444 |   0.0 |    0.0 |       24.5 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.082224 |      0.0 |  0.030873 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.918098 | 0.252871 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.083872 | 0.023298 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.188417 | 89.068576 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.082594 | 0.001279 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023298 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |           1.206967 |          3.4 |                 0.457351 |                      0.621065 |                         0.0 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.08424 |                   0.0 |                      nan |                       0.0 |                    1.32431 |                  1.664915 |                     0.0 |                      0.0 |    -0.069371 |             0.0 |              0.0 |         -0.340605 |             27.647059 |                 100.0 |              2928.258272 |               69.444444 |   0.0 |    0.0 |       24.5 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.060145 |      0.0 |  0.022339 |       0.0 |      0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |       0.0 |       0.0 |       0.0 |       0.0 | 0.920094 | 0.253972 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.081612 |  0.02267 |      0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan |  2.135077 | 89.046237 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.08046 | 0.001152 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02267 |

.. _lland_v4_acker_winter_ktschnee_hourly:

acre (snow surface temperature)
-------------------------------

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(0.018)
    >>> test("lland_v4_acker_winter_ktschnee_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter_ktschnee)
    |                date | nied | teml | sunshineduration | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair | solardeclination |      tsa |       tsu | earthsundistance | extraterrestrialradiation | possiblesunshineduration | dailypossiblesunshineduration | dailyglobalradiation |        sp | globalradiation | adjustedglobalradiation |      g |        tz |        wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |      evi |      evb |       evs |    wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr | wlatinz |  wlatsnow | wsensinz | wsenssnow | wsurfinz |     wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |      ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |             87.5 |       3.4 |               101.7 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                 0.464297 |                      0.526233 |                         0.0 |                     0.335938 |                      0.034962 |                           0.039056 |              0.40626 |                  0.509766 |      101.29374 |   1.312855 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.470836 | -0.004029 |                   0.0 |                      0.0 |                       0.0 |                   0.708429 |                  4.014396 |                     0.0 |                 0.226362 |    -0.167266 |             0.0 |        -0.226362 |         -3.305967 |             62.648725 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.001835 |      0.0 |      0.0 | -0.010625 |      0.0 |    -7.918721 |      0.75329 |      0.0 |      0.0 | 0.315715 | 0.315715 |     0.0 | -0.028341 |      0.0 | -0.096602 |      0.0 |   0.10142 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.067836 |  20.42268 |       0.0 | 0.871765 | -0.706321 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |             90.7 |       2.9 |               101.8 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 0.427219 |                      0.521048 |                         0.0 |                      0.31051 |                      0.032479 |                           0.038716 |             0.387487 |                  0.504505 |     101.412513 |   1.319627 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.354405 |  0.002798 |                   0.0 |                      0.0 |                       0.0 |                   0.710308 |                  4.024777 |                     0.0 |                 0.228637 |    -0.167699 |             0.0 |        -0.228637 |         -3.314469 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004543 |      0.0 |      0.0 | -0.010458 |      0.0 |    -8.921626 |     0.752636 |      0.0 |      0.0 | 0.338563 | 0.338563 |     0.0 | -0.027895 |      0.0 | -0.084863 |      0.0 |  0.115878 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.416336 | 20.433137 |       0.0 | 0.879073 |  -0.67822 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |             89.7 |       2.5 |               101.8 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 0.401924 |                      0.514631 |                         0.0 |                     0.289024 |                      0.030771 |                           0.038294 |             0.360526 |                  0.497155 |     101.439474 |   1.323706 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.260732 |  0.008912 |                   0.0 |                      0.0 |                       0.0 |                   0.712173 |                  4.040774 |                     0.0 |                   0.2316 |    -0.168366 |             0.0 |          -0.2316 |         -3.328601 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004605 |      0.0 |      0.0 | -0.008842 |      0.0 |    -9.827574 |     0.751986 |      0.0 |      0.0 | 0.358876 | 0.358876 |     0.0 | -0.023587 |      0.0 | -0.079237 |      0.0 |  0.128777 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.783769 |  20.44198 |       0.0 | 0.886381 | -0.656232 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |             91.3 |       3.0 |               101.9 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 0.383833 |                      0.510972 |                         0.0 |                     0.283861 |                       0.02954 |                           0.038053 |             0.350439 |                  0.492194 |     101.549561 |   1.328036 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.187439 |  0.011645 |                   0.0 |                      0.0 |                       0.0 |                   0.714023 |                  4.052814 |                     0.0 |                 0.237699 |    -0.168867 |             0.0 |        -0.237699 |          -3.33879 |             71.001889 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004915 |      0.0 |      0.0 | -0.009248 |      0.0 |   -10.054129 |     0.751342 |      0.0 |      0.0 | 0.360034 | 0.360034 |     0.0 | -0.024668 |      0.0 | -0.081135 |      0.0 |  0.131896 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.152917 | 20.451227 |       0.0 | 0.893689 | -0.636976 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |             94.2 |       2.5 |               102.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                 0.352531 |                      0.505764 |                         0.0 |                     0.261706 |                      0.027395 |                            0.03771 |             0.332084 |                  0.486608 |     101.667916 |   1.334932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.123255 |  0.017953 |                   0.0 |                      0.0 |                       0.0 |                    0.71586 |                  4.064535 |                     0.0 |                 0.239957 |    -0.169356 |             0.0 |        -0.239957 |         -3.348675 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008645 |      0.0 |      0.0 | -0.008703 |      0.0 |   -11.069985 |     0.750703 |      0.0 |      0.0 | 0.383617 | 0.298311 |     0.0 | -0.023216 |      0.0 | -0.070661 |      0.0 |  0.146081 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.459931 | 20.459931 | -0.028492 | 0.900997 | -0.624029 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |             93.2 |       2.3 |               102.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                 0.326061 |                      0.499513 |                         0.0 |                     0.239833 |                      0.025563 |                           0.037298 |             0.303889 |                  0.479553 |     101.696111 |   1.340112 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.080097 |  0.031686 |                   0.0 |                      0.0 |                       0.0 |                    0.71732 |                  4.080005 |                     0.0 |                 0.243578 |        -0.17 |             0.0 |        -0.243578 |         -3.362686 |             92.611159 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007726 |      0.0 |      0.0 | -0.007531 |      0.0 |   -12.150848 |     0.750194 |      0.0 |      0.0 | 0.454118 |      0.0 |     0.0 | -0.020089 |      0.0 |  -0.06862 |      0.0 |  0.154869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.467462 | 20.467462 | -0.151675 | 0.906848 | -0.624815 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |             96.5 |       2.1 |               102.1 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 0.318472 |                      0.492103 |                         0.0 |                     0.225911 |                      0.025034 |                           0.036808 |             0.307325 |                  0.472439 |     101.792675 |   1.342932 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.871508 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.082716 |  0.067396 |                   0.0 |                      0.0 |                       0.0 |                    0.71813 |                  4.093695 |                     0.0 |                 0.234903 |    -0.170571 |             0.0 |        -0.234903 |         -3.375566 |             101.43127 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.010331 |      0.0 |      0.0 | -0.009076 |      0.0 |   -12.885431 |     0.749912 |      0.0 |      0.0 | 0.666361 |      0.0 |     0.0 |  -0.02421 |      0.0 | -0.072407 |      0.0 |  0.138285 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.476538 | 20.476538 | -0.222565 |  0.91011 | -0.661311 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |             95.9 |       1.4 |               102.1 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 0.320984 |                      0.483429 |                         0.0 |                     0.218053 |                       0.02521 |                           0.036234 |             0.307823 |                  0.464394 |     101.792177 |   1.342422 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             2.641605 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -2.20437 |  0.086056 |                   0.0 |                      0.0 |                       0.0 |                   0.661257 |                  3.682358 |                     0.0 |                 0.215754 |    -0.153432 |             0.0 |        -0.215754 |         -3.021102 |            152.146905 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008377 |      0.0 |      0.0 | -0.008093 |      0.0 |   -13.318015 |     0.749676 |      0.0 |      0.0 | 0.792936 |      0.0 |     0.0 | -0.021589 |      0.0 | -0.065834 |      0.0 |  0.128332 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.484632 | 20.484632 | -0.264841 | 0.912845 | -0.716467 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |             96.8 |       1.1 |               102.2 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 0.296635 |                      0.471927 |                         0.0 |                     0.206259 |                      0.023505 |                            0.03547 |             0.287143 |                  0.454289 |     101.912857 |   1.348938 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.355232 |                       7.41713 |             2.287909 |  0.266196 |        0.003492 |                0.003987 | 0.0309 | -2.388223 |  0.097368 |              0.000999 |                      0.0 |                  0.000999 |                   0.573213 |                  3.039527 |                     0.0 |                 0.201424 |    -0.125648 |             0.0 |        -0.200425 |         -2.466314 |            193.641515 |                 100.0 |               112.102145 |               99.616909 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005974 |      0.0 |      0.0 | -0.006553 |      0.0 |   -13.994011 |      0.74946 |      0.0 |      0.0 | 0.883944 |      0.0 |     0.0 |  -0.01748 |      0.0 | -0.055181 |      0.0 |  0.127764 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.491185 | 20.491185 | -0.295237 |  0.91536 | -0.782935 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |             94.3 |       1.0 |               102.2 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 0.303761 |                      0.460234 |                         0.0 |                     0.221608 |                      0.024006 |                           0.034691 |             0.286447 |                  0.443589 |     101.913553 |   1.347408 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.944823 |  4.919884 |        0.064534 |                0.067188 | 0.0309 | -2.609783 |  0.097657 |              0.016847 |                      0.0 |                  0.016847 |                   0.487653 |                  2.412046 |                     0.0 |                 0.184476 |    -0.083655 |             0.0 |        -0.167629 |         -1.924393 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001942 |      0.0 |      0.0 | -0.005056 |      0.0 |    -13.12059 |     0.749256 |      0.0 |      0.0 | 0.914353 |      0.0 |     0.0 | -0.013485 |      0.0 |  -0.04633 |      0.0 |  0.107814 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.49624 |  20.49624 | -0.305394 | 0.917741 | -0.849692 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |             96.7 |       1.4 |               102.2 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 0.363641 |                      0.451775 |                         0.0 |                     0.294444 |                      0.028159 |                           0.034126 |             0.351641 |                  0.436132 |     101.848359 |   1.335446 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.622959 | 14.873457 |        0.195096 |                0.202564 | 0.0309 | -2.832306 |  0.080117 |              0.050832 |                      0.0 |                  0.050832 |                   0.407272 |                  1.815487 |                     0.0 |                 0.156329 |    -0.024813 |             0.0 |        -0.105497 |         -1.408215 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.002388 |      0.0 |      0.0 | -0.005157 |      0.0 |    -9.593506 |     0.749056 |      0.0 |      0.0 | 0.838967 |      0.0 |     0.0 | -0.013755 |      0.0 | -0.036804 |      0.0 |  0.054938 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.501397 | 20.501397 | -0.280215 | 0.920079 | -0.898908 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |             97.4 |       1.3 |               102.2 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 0.453913 |                      0.447245 |                         0.0 |                     0.400568 |                      0.034269 |                           0.033823 |             0.442111 |                  0.431908 |     101.757889 |   1.320615 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.449648 | 22.189872 |        0.349279 |                0.385806 | 0.0309 | -2.996361 |  0.052262 |              0.096896 |                      0.0 |                  0.096896 |                   0.364083 |                  1.493002 |                     0.0 |                 0.132363 |     0.034688 |             0.0 |        -0.035467 |         -1.128918 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008412 |      0.0 |      0.0 | -0.003619 |      0.0 |    -5.644128 |     0.748847 |      0.0 |      0.0 | 0.694792 |      0.0 |     0.0 | -0.009653 |      0.0 | -0.021706 |      0.0 |  0.004108 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.505015 | 20.505015 | -0.232061 | 0.922526 | -0.920271 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |             93.9 |       1.3 |               102.2 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 0.496723 |                      0.442058 |                         0.0 |                      0.43315 |                      0.037113 |                           0.033476 |             0.466423 |                  0.425978 |     101.733577 |   1.314635 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.361224 | 22.197307 |        0.291164 |                0.329304 | 0.0309 | -3.067569 |  0.033666 |              0.082781 |                      0.0 |                  0.082781 |                   0.342186 |                  1.331032 |                     0.0 |                 0.114382 |     0.027321 |             0.0 |        -0.031601 |         -0.988846 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010374 |      0.0 |      0.0 | -0.002898 |      0.0 |    -4.618483 |     0.748619 |      0.0 |      0.0 | 0.593583 |      0.0 |     0.0 | -0.007731 |      0.0 | -0.024008 |      0.0 | -0.000138 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.507914 | 20.507914 | -0.198257 | 0.925208 | -0.923036 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |             93.4 |       1.4 |               102.1 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 0.551149 |                      0.441361 |                         0.0 |                     0.475403 |                      0.040686 |                           0.033429 |             0.514774 |                  0.424221 |     101.585226 |   1.306346 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 20.962863 |        0.274971 |                0.289712 | 0.0309 | -3.076788 |  0.018844 |              0.072899 |                      0.0 |                  0.072899 |                    0.33896 |                  1.306141 |                     0.0 |                 0.099356 |     0.018477 |             0.0 |        -0.026457 |         -0.967181 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010825 |      0.0 |      0.0 |  -0.00355 |      0.0 |    -3.385661 |     0.748374 |      0.0 |      0.0 | 0.506793 |      0.0 |     0.0 | -0.009468 |      0.0 | -0.027132 |      0.0 | -0.010144 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.511463 | 20.511463 | -0.169269 | 0.928097 | -0.910981 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |             93.3 |       2.3 |               102.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 0.593188 |                      0.441918 |                         0.0 |                     0.498831 |                      0.043416 |                           0.033467 |             0.553444 |                  0.423597 |     101.446556 |   1.300092 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 | 11.702396 |        0.153501 |                0.155234 | 0.0309 | -3.036602 |  0.008856 |              0.039102 |                      0.0 |                  0.039102 |                   0.339314 |                  1.307301 |                     0.0 |                 0.085326 |    -0.015369 |             0.0 |        -0.046224 |         -0.967987 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010429 |      0.0 |      0.0 | -0.006421 |      0.0 |    -2.743297 |     0.748111 |      0.0 |      0.0 | 0.442374 |      0.0 |     0.0 | -0.017128 |      0.0 | -0.041756 |      0.0 |  -0.01266 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.517884 | 20.517884 | -0.147753 | 0.931203 | -0.888937 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |             94.1 |       2.8 |               102.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 0.633417 |                      0.443734 |                         0.0 |                      0.52257 |                      0.046005 |                           0.033588 |             0.596046 |                  0.424247 |     101.403954 |   1.295611 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      1.0 |                       7.41713 |             1.347076 |  2.880181 |         0.03778 |                0.037152 | 0.0309 | -2.963123 |  0.001439 |              0.009369 |                      0.0 |                  0.009369 |                   0.339689 |                  1.307629 |                     0.0 |                 0.071723 |    -0.045116 |             0.0 |        -0.062354 |          -0.96794 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007094 |      0.0 |      0.0 | -0.009758 |      0.0 |    -2.119079 |     0.747832 |      0.0 |      0.0 | 0.388989 |      0.0 |     0.0 | -0.026029 |      0.0 | -0.052717 |      0.0 | -0.016392 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.527642 | 20.527642 | -0.129922 | 0.934507 | -0.859476 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |             94.6 |       3.1 |               101.9 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 0.656792 |                      0.447104 |                         0.0 |                     0.542945 |                        0.0475 |                           0.033814 |             0.621325 |                  0.426519 |     101.278675 |   1.291856 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                 0.061898 |                       7.41713 |             1.347076 |  0.007844 |        0.000103 |                0.000101 | 0.0309 | -2.864919 | -0.004442 |              0.000026 |                      0.0 |                  0.000026 |                   0.339839 |                   1.30686 |                     0.0 |                 0.066688 |    -0.054427 |             0.0 |        -0.066663 |         -0.967021 |             68.711505 |                 100.0 |               112.102145 |               98.953162 |    0.094 |  0.0235 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005889 |      0.0 |      0.0 | -0.011126 | 0.000344 |    -1.602996 |     0.747721 |      0.0 |      0.0 |  0.34433 |   0.0705 |     0.0 | -0.029677 |      0.0 | -0.055999 |      0.0 | -0.019013 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.632768 | 20.632768 | -0.091459 |  0.93583 | -0.824134 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |             94.4 |       3.5 |               101.8 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 0.661557 |                      0.450781 |                         0.0 |                     0.555593 |                      0.047804 |                            0.03406 |              0.62451 |                  0.429012 |      101.17549 |     1.2901 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.747113 | -0.013894 |                   0.0 |                      0.0 |                       0.0 |                   0.340301 |                  1.305996 |                     0.0 |                 0.069563 |    -0.054417 |             0.0 |        -0.069563 |         -0.965695 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007388 |      0.0 |      0.0 | -0.010622 |      0.0 |     -1.29117 |     0.747378 |      0.0 |      0.0 |  0.27163 |      0.0 |     0.0 | -0.028334 |      0.0 | -0.055858 |      0.0 | -0.014629 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.64339 |  20.64339 | -0.090724 | 0.939917 |  -0.77934 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |             96.9 |       4.4 |               101.6 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 0.656792 |                      0.455629 |                         0.0 |                     0.566657 |                        0.0475 |                           0.034384 |             0.636431 |                   0.43317 |     100.963569 |   1.287972 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |   -2.5978 |  -0.01377 |                   0.0 |                      0.0 |                       0.0 |                   0.340023 |                  1.303739 |                     0.0 |                 0.072606 |    -0.054322 |             0.0 |        -0.072606 |         -0.963716 |             48.410379 |                 100.0 |               112.102145 |               98.814229 |    0.282 |  0.0705 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.003845 |      0.0 |      0.0 | -0.012668 | 0.001033 |    -1.023516 |     0.747585 |      0.0 |      0.0 | 0.259282 |   0.2115 |     0.0 |  -0.03379 |      0.0 | -0.055678 |      0.0 | -0.016862 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.938058 | 20.938058 | -0.015959 | 0.937452 |  -0.73467 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |             98.3 |       3.8 |               101.5 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 0.676036 |                      0.460234 |                         0.0 |                     0.591017 |                      0.048726 |                           0.034691 |             0.664544 |                  0.437356 |     100.835456 |   1.284691 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.448899 |  -0.03339 |                   0.0 |                      0.0 |                       0.0 |                   0.340458 |                  1.301296 |                     0.0 |                 0.069648 |    -0.054221 |             0.0 |        -0.069648 |         -0.960838 |             56.054123 |                 100.0 |               112.102145 |               98.814229 |    0.188 |  0.0282 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000004 |      0.0 |      0.0 | -0.012005 | 0.001019 |    -0.450069 |     0.747262 |      0.0 |      0.0 | 0.120288 | 0.120288 |     0.0 | -0.032022 |      0.0 |  -0.04578 |      0.0 | -0.008154 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.098528 | 21.138063 |       0.0 |  0.94131 |  -0.67038 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |             98.7 |       5.7 |               101.4 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 0.685844 |                      0.465172 |                         0.0 |                      0.61078 |                      0.049349 |                            0.03502 |             0.676928 |                  0.441874 |     100.723072 |   1.282429 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -2.234599 | -0.040392 |                   0.0 |                      0.0 |                       0.0 |                   0.340883 |                  1.298642 |                     0.0 |                 0.071334 |     -0.05411 |             0.0 |        -0.071334 |         -0.957759 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    0.376 |  0.0376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000346 |      0.0 |      0.0 | -0.014629 | 0.002393 |          0.0 |     0.746946 |      0.0 |      0.0 | 0.049951 | 0.049951 |     0.0 | -0.039021 |      0.0 | -0.053629 |      0.0 | -0.021316 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.200484 | 21.528691 |       0.0 | 0.945091 | -0.599088 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |             99.6 |       5.6 |               101.3 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 0.705836 |                      0.471336 |                         0.0 |                      0.61078 |                      0.050616 |                           0.035431 |             0.703012 |                   0.44773 |     100.596988 |   1.279174 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -1.99696 | -0.043134 |                   0.0 |                      0.0 |                       0.0 |                   0.341702 |                  1.295068 |                     0.0 |                 0.059541 |    -0.053961 |             0.0 |        -0.059541 |         -0.953365 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |    0.658 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002532 |      0.0 |      0.0 | -0.020116 |  0.00551 |          0.0 |     0.746338 | 0.067683 | 0.067683 |      0.0 |      0.0 |     0.0 |  -0.05366 |      0.0 | -0.066112 |      0.0 | -0.060231 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.151962 | 22.206808 |       0.0 | 0.952427 | -0.525054 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |             99.4 |       5.8 |               101.1 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                  0.72634 |                      0.478769 |                         0.0 |                      0.61078 |                       0.05191 |                           0.035924 |             0.721982 |                   0.45487 |     100.378018 |   1.274698 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.750179 | -0.037804 |                   0.0 |                      0.0 |                       0.0 |                   0.342515 |                  1.290674 |                     0.0 |                 0.049365 |    -0.053778 |             0.0 |        -0.049365 |         -0.948158 |             36.725115 |                 100.0 |               112.102145 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001838 |      0.0 |      0.0 | -0.024931 | 0.004723 |          0.0 |     0.745734 | 0.196432 | 0.196432 |      0.0 |      0.0 |     0.0 | -0.066503 |      0.0 | -0.081551 |      0.0 |  -0.09869 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.978569 | 22.701739 |       0.0 | 0.959757 |  -0.45635 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |             99.2 |       6.2 |               101.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 0.731547 |                      0.486761 |                         0.0 |                      0.61078 |                      0.052238 |                           0.036454 |             0.725695 |                  0.462423 |     100.274305 |   1.272954 |         -0.40453 | 8.644768 | 16.061898 |         1.030879 |                  5.246826 |                      0.0 |                       7.41713 |             1.347076 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.521166 | -0.032857 |                   0.0 |                      0.0 |                       0.0 |                   0.343325 |                   1.28612 |                     0.0 |                 0.047044 |    -0.053588 |             0.0 |        -0.047044 |         -0.942796 |             34.355753 |                 100.0 |               112.102145 |               98.814229 |     0.94 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000987 |      0.0 |      0.0 | -0.027164 | 0.009839 |          0.0 |     0.745134 | 0.275337 | 0.275337 |      0.0 |      0.0 |     0.0 | -0.072459 |      0.0 | -0.089566 |      0.0 | -0.114981 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.72702 | 23.668903 |       0.0 | 0.967109 | -0.392593 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |             99.2 |       5.6 |               100.8 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 0.742061 |                      0.496568 |                         0.0 |                      0.61078 |                      0.052899 |                           0.037103 |             0.736124 |                   0.47416 |     100.063876 |   1.269455 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -1.308642 | -0.028267 |                   0.0 |                      0.0 |                       0.0 |                   0.341769 |                  1.277745 |                     0.0 |                 0.041776 |    -0.053239 |             0.0 |        -0.041776 |         -0.935976 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |     1.88 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001175 |      0.0 |      0.0 | -0.027338 | 0.021252 |          0.0 |     0.744534 | 0.339477 | 0.339477 |      0.0 |      0.0 |     0.0 | -0.072924 |      0.0 | -0.089252 |      0.0 |   -0.1204 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.409359 | 25.576242 |       0.0 | 0.974498 | -0.333426 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |             99.6 |       5.2 |               100.7 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                  0.76893 |                      0.509073 |                         0.0 |                      0.61078 |                      0.054584 |                           0.037928 |             0.765854 |                  0.487989 |      99.934146 |   1.265756 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -1.11142 | -0.024007 |                   0.0 |                      0.0 |                       0.0 |                   0.342566 |                  1.268373 |                     0.0 |                 0.027852 |    -0.052849 |             0.0 |        -0.027852 |         -0.925808 |             40.962628 |                 100.0 |               112.102145 |               98.814229 |    1.974 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002517 |      0.0 |      0.0 | -0.031933 | 0.026447 |          0.0 |     0.743939 | 0.477958 | 0.477958 |      0.0 |      0.0 |     0.0 |  -0.08518 |      0.0 |  -0.09987 |      0.0 | -0.157197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.954503 | 27.582175 |       0.0 | 0.981895 | -0.278519 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |             99.8 |       5.7 |               100.6 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                 0.807986 |                      0.524283 |                         0.0 |                      0.61078 |                       0.05702 |                           0.038928 |              0.80637 |                  0.504776 |       99.79363 |   1.261108 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.928398 | -0.020053 |                   0.0 |                      0.0 |                       0.0 |                   0.343352 |                  1.257174 |                     0.0 |                 0.008709 |    -0.052382 |             0.0 |        -0.008709 |         -0.913822 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    1.692 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003094 |      0.0 |      0.0 | -0.043255 | 0.015557 |          0.0 |     0.743351 | 0.697291 | 0.697291 |      0.0 |      0.0 |     0.0 | -0.115381 |      0.0 |  -0.13072 |      0.0 | -0.237391 |      1.0 |      0.3 | 1.737874 | 0.593896 |  0.0 |  0.0 | 0.0 |  0.0 | 0.593896 | 0.011233 |  0.00312 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.286403 | 27.579556 |       0.0 | 0.989252 | -0.227566 | 72.590238 |      0.5 | 0.093896 |   0.0 |   0.0 |  0.0 | 0.006486 | 0.004747 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00312 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |             99.8 |       6.6 |               100.5 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                 0.836943 |                      0.541728 |                         0.0 |                      0.61078 |                      0.058816 |                           0.040071 |             0.835269 |                   0.52349 |      99.664731 |   1.257444 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.758553 | -0.016385 |                   0.0 |                      0.0 |                       0.0 |                   0.344127 |                  1.245134 |                     0.0 |                -0.004886 |    -0.051881 |             0.0 |         0.004886 |         -0.901007 |             32.273586 |                 100.0 |               110.467257 |               98.814229 |    1.222 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002985 |      0.0 |      0.0 | -0.055801 |      0.0 |          0.0 |     0.742771 | 0.907519 | 0.907519 |      0.0 |      0.0 |     0.0 | -0.148847 |      0.0 | -0.165762 |      0.0 | -0.319496 |      1.0 |      0.3 | 2.522377 | 0.864443 |  0.0 |  0.0 | 0.0 |  0.0 | 0.864443 | 0.046521 | 0.012922 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  18.41607 | 26.334981 |       0.0 |  0.99656 | -0.180281 | 74.248173 |      0.5 | 0.364443 |   0.0 |   0.0 |  0.0 | 0.019234 | 0.027287 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012922 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |             99.0 |       7.0 |               100.5 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 0.860763 |                      0.562266 |                         0.0 |                      0.61078 |                      0.060288 |                            0.04141 |             0.852155 |                  0.544461 |      99.647845 |   1.255555 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.600937 |  -0.01298 |                   0.0 |                      0.0 |                       0.0 |                   0.344897 |                  1.232336 |                     0.0 |                -0.014528 |    -0.051347 |             0.0 |         0.014528 |         -0.887439 |             30.429381 |                 100.0 |               108.116487 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000597 |      0.0 |      0.0 |  -0.06294 |      0.0 |          0.0 |     0.742196 | 1.075253 | 1.075253 |      0.0 |      0.0 |     0.0 |  -0.16789 |      0.0 | -0.189697 |      0.0 | -0.372115 |      1.0 |      0.3 | 2.011287 | 0.691128 |  0.0 |  0.0 | 0.0 |  0.0 | 0.691128 | 0.083686 | 0.023246 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.382262 | 24.856634 |       0.0 | 1.003868 | -0.136401 | 75.568331 |      0.5 | 0.191128 |   0.0 |   0.0 |  0.0 | 0.031652 | 0.052033 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023246 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |             97.6 |       7.3 |               100.4 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 0.860763 |                      0.585275 |                         0.0 |                      0.61078 |                      0.060288 |                           0.042904 |             0.840104 |                  0.567814 |      99.559896 |   1.254358 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -0.45467 | -0.009821 |                   0.0 |                      0.0 |                       0.0 |                   0.345661 |                  1.218349 |                     0.0 |                -0.012189 |    -0.050765 |             0.0 |         0.012189 |         -0.872687 |             29.178858 |                 100.0 |               106.259422 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007215 |      0.0 |      0.0 | -0.061894 |      0.0 |          0.0 |     0.741625 | 1.089258 | 1.089258 |      0.0 |      0.0 |     0.0 | -0.165098 |      0.0 | -0.196345 |      0.0 | -0.373633 |      1.0 |      0.3 | 1.749518 | 0.602536 |  0.0 |  0.0 | 0.0 |  0.0 | 0.602536 | 0.105129 | 0.029202 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.333573 |  23.35701 |       0.0 | 1.011176 |  -0.09568 | 76.715313 |      0.5 | 0.102536 |   0.0 |   0.0 |  0.0 |  0.04375 | 0.061379 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029202 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |             97.4 |       7.2 |               100.4 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 0.879017 |                      0.610223 |                         0.0 |                      0.61078 |                      0.061412 |                           0.044515 |             0.856163 |                  0.592247 |      99.543837 |    1.25293 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 | -0.318933 | -0.006889 |                   0.0 |                      0.0 |                       0.0 |                    0.34642 |                  1.204312 |                     0.0 |                -0.020102 |     -0.05018 |             0.0 |         0.020102 |         -0.857893 |              29.58412 |                 100.0 |               104.655945 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008179 |      0.0 |      0.0 |  -0.06548 |      0.0 |          0.0 |     0.741058 | 1.180061 | 1.180061 |      0.0 |      0.0 |     0.0 | -0.174665 |      0.0 | -0.206262 |      0.0 | -0.401029 |      1.0 |      0.3 | 2.162217 | 0.746558 |  0.0 |  0.0 | 0.0 |  0.0 | 0.746558 | 0.128287 | 0.035635 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 15.195995 | 21.730273 |       0.0 | 1.018484 | -0.057891 | 78.130972 |      0.5 | 0.246558 |   0.0 |   0.0 |  0.0 | 0.055536 | 0.072751 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035635 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |             96.3 |       7.6 |               100.4 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 0.879017 |                      0.635914 |                         0.0 |                      0.61078 |                      0.061412 |                           0.046165 |             0.846494 |                  0.617287 |      99.553506 |   1.252976 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |  -0.19297 | -0.004168 |                   0.0 |                      0.0 |                       0.0 |                   0.347173 |                  1.190153 |                     0.0 |                -0.018227 |     -0.04959 |             0.0 |         0.018227 |          -0.84298 |             28.027061 |                 100.0 |               102.688921 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014145 |      0.0 |      0.0 | -0.065772 |      0.0 |          0.0 |     0.740495 | 1.213127 | 1.213127 |      0.0 |      0.0 |     0.0 | -0.175444 |      0.0 | -0.215681 |      0.0 | -0.409353 |      1.0 |      0.3 | 1.928023 | 0.667344 |  0.0 |  0.0 | 0.0 |  0.0 | 0.667344 | 0.153037 |  0.04251 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  14.02519 | 20.056022 |       0.0 | 1.025792 | -0.022823 | 79.391651 |      0.5 | 0.167344 |   0.0 |   0.0 |  0.0 | 0.067016 |  0.08602 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04251 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |             95.2 |       7.3 |               100.3 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 0.897612 |                      0.665151 |                         0.0 |                      0.61078 |                      0.062554 |                           0.048033 |             0.854526 |                  0.645224 |      99.445474 |   1.250338 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.347041 |                      7.400749 |             1.337827 |  0.254871 |         0.00332 |                0.003243 | 0.0309 | -0.076076 | -0.001643 |              0.000843 |                      0.0 |                  0.000843 |                   0.347921 |                  1.174779 |                     0.0 |                -0.024601 |    -0.048106 |             0.0 |         0.025444 |         -0.826858 |             29.178858 |                 100.0 |               100.947841 |               99.255106 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01901 |      0.0 |      0.0 | -0.065786 |      0.0 |          0.0 |     0.739936 | 1.257996 | 1.257996 |      0.0 |      0.0 |     0.0 | -0.175481 |      0.0 | -0.220889 |      0.0 | -0.421814 | 0.999636 | 0.299782 | 1.898835 |  0.65841 |  0.0 |  0.0 | 0.0 |  0.0 |  0.65841 | 0.171852 | 0.047737 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.809073 | 18.316974 |       0.0 |   1.0331 |   0.00972 | 80.632075 |      0.5 |  0.15841 |   0.0 |   0.0 |  0.0 | 0.078201 | 0.093651 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047737 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |             96.3 |       7.1 |               100.3 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 0.891375 |                      0.694736 |                         0.0 |                      0.61078 |                      0.062171 |                           0.049913 |             0.858394 |                  0.674502 |      99.441606 |   1.250768 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  4.882681 |        0.063603 |                0.062163 | 0.0309 |       0.0 |       0.0 |              0.016201 |                      0.0 |                  0.016201 |                   0.348663 |                   1.15845 |                     0.0 |                -0.023742 |    -0.032068 |             0.0 |         0.039943 |         -0.809787 |             30.000798 |                 100.0 |                99.243943 |               99.922753 |    0.282 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.016415 |      0.0 |      0.0 | -0.065321 |      0.0 |          0.0 |     0.739381 | 1.275715 | 1.275715 |      0.0 |      0.0 |     0.0 | -0.174242 |      0.0 | -0.211904 |      0.0 | -0.426089 |  0.99848 | 0.299089 | 2.112778 | 0.733098 |  0.0 |  0.0 | 0.0 |  0.0 | 0.733098 |  0.19296 |   0.0536 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 11.574488 | 16.551517 |       0.0 | 1.040408 |   0.04062 | 82.011755 |      0.5 | 0.233098 |   0.0 |   0.0 |  0.0 | 0.089096 | 0.103864 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0536 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |             95.5 |       7.1 |               100.3 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 0.903886 |                       0.72095 |                         0.0 |                      0.61078 |                      0.062939 |                            0.05157 |             0.863211 |                  0.699592 |      99.436789 |   1.249848 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 | 14.862249 |        0.193599 |                0.189423 | 0.0309 |       0.0 |       0.0 |              0.049472 |                      0.0 |                  0.049472 |                     0.3494 |                  1.144953 |                     0.0 |                 -0.02789 |     0.001765 |             0.0 |         0.077361 |         -0.795553 |             30.000798 |                 100.0 |                 97.35891 |               99.725477 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026537 |      0.0 |      0.0 | -0.066592 |      0.0 |          0.0 |      0.73883 | 1.421835 | 1.421835 |      0.0 |      0.0 |     0.0 | -0.177631 |      0.0 |   -0.2199 |      0.0 | -0.474893 | 0.997323 | 0.298396 | 2.135404 | 0.741544 |  0.0 |  0.0 | 0.0 |  0.0 | 0.741544 | 0.216845 | 0.060235 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |   10.1935 | 14.576705 |       0.0 | 1.047716 |   0.07152 | 83.405616 |      0.5 | 0.241544 |   0.0 |   0.0 |  0.0 |  0.09971 | 0.117135 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060235 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |             91.6 |       8.9 |               100.2 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 0.942351 |                      0.742943 |                         0.0 |                      0.61078 |                       0.06529 |                           0.052955 |             0.863194 |                  0.719138 |      99.336806 |   1.245915 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 |  22.23196 |        0.289599 |                0.288488 | 0.0309 |       0.0 |       0.0 |              0.075502 |                      0.0 |                  0.075502 |                   0.340919 |                  1.076886 |                     0.0 |                -0.043295 |     0.030632 |             0.0 |         0.118797 |         -0.735967 |             23.933221 |                 100.0 |                95.464765 |               99.520229 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060214 |      0.0 |      0.0 | -0.080429 |      0.0 |          0.0 |     0.738283 | 1.880007 | 1.880007 |      0.0 |      0.0 |     0.0 |  -0.21454 |      0.0 | -0.294586 |      0.0 | -0.627922 | 0.996167 | 0.297705 | 2.797244 | 0.972757 |  0.0 |  0.0 | 0.0 |  0.0 | 0.972757 |  0.25124 | 0.069789 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  8.359363 |  11.95389 |       0.0 | 1.055024 |   0.10242 | 85.230103 |      0.5 | 0.472757 |   0.0 |   0.0 |  0.0 |  0.11005 |  0.14119 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069789 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |             90.0 |       9.5 |               100.2 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 0.962118 |                      0.763489 |                         0.0 |                      0.61078 |                      0.066493 |                           0.054244 |             0.865906 |                  0.737784 |      99.334094 |   1.244565 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 22.247803 |        0.289805 |                0.288994 | 0.0309 |       0.0 |       0.0 |              0.075792 |                      0.0 |                  0.075792 |                   0.341627 |                  1.067855 |                     0.0 |                -0.048721 |     0.031298 |             0.0 |         0.124513 |         -0.726229 |             22.421649 |                 100.0 |                93.000068 |               99.241996 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.074086 |      0.0 |      0.0 | -0.085956 |      0.0 |          0.0 |      0.73774 | 2.037727 | 2.037727 |      0.0 |      0.0 |     0.0 | -0.229285 |      0.0 | -0.326804 |      0.0 | -0.680601 |  0.99501 | 0.297014 | 2.934903 | 1.022476 |  0.0 |  0.0 | 0.0 |  0.0 | 1.022476 | 0.296781 | 0.082439 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.367093 |  9.104943 |       0.0 | 1.062332 |   0.13332 |  87.14253 |      0.5 | 0.522476 |   0.0 |   0.0 |  0.0 | 0.120123 | 0.176658 |   0.0 |   0.0 |  0.0 |   0.0 | 0.082439 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |             85.7 |       7.9 |               100.2 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 0.962118 |                      0.781298 |                         0.0 |                      0.61078 |                      0.066493 |                           0.055357 |             0.824535 |                  0.752488 |      99.375465 |    1.24476 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 20.992726 |        0.273456 |                0.273007 | 0.0309 |       0.0 |       0.0 |              0.071746 |                      0.0 |                  0.071746 |                   0.342329 |                   1.06128 |                     0.0 |                -0.040462 |     0.027526 |             0.0 |         0.112208 |         -0.718951 |             26.962743 |                 100.0 |                  90.4332 |               98.937896 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.085524 |      0.0 |      0.0 | -0.061599 |      0.0 |          0.0 |       0.7372 | 1.664802 | 1.664802 |      0.0 |      0.0 |     0.0 | -0.164312 |      0.0 | -0.279524 |      0.0 | -0.556044 | 0.993854 | 0.296324 | 2.490773 | 0.869036 |  0.0 |  0.0 | 0.0 |  0.0 | 0.869036 | 0.333206 | 0.092557 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  4.734104 |  6.769768 |       0.0 |  1.06964 |   0.16422 | 88.764268 |      0.5 | 0.369036 |   0.0 |   0.0 |  0.0 | 0.129935 |  0.20327 |   0.0 |   0.0 |  0.0 |   0.0 | 0.092557 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |             82.3 |       9.6 |               100.2 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 0.968788 |                      0.797357 |                         0.0 |                      0.61078 |                      0.066898 |                           0.056358 |             0.797312 |                    0.7643 |      99.402688 |   1.244443 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.302624 | 11.677032 |        0.152108 |                 0.15202 | 0.0309 |       0.0 |       0.0 |              0.040032 |                      0.0 |                  0.040032 |                   0.343027 |                  1.056546 |                     0.0 |                -0.036448 |     -0.00399 |             0.0 |         0.076481 |         -0.713519 |              22.18809 |                 100.0 |                88.268922 |               98.669307 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.116361 |      0.0 |      0.0 | -0.063414 |      0.0 |          0.0 |     0.736665 | 1.738162 | 1.738162 |      0.0 |      0.0 |     0.0 | -0.169154 |      0.0 | -0.334911 |      0.0 | -0.580546 | 0.992698 | 0.295635 | 2.508855 | 0.876536 |  0.0 |  0.0 | 0.0 |  0.0 | 0.876536 | 0.359619 | 0.099894 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.024005 |  4.324328 |       0.0 | 1.076948 |   0.19512 | 90.396586 |      0.5 | 0.376536 |   0.0 |   0.0 |  0.0 | 0.139494 | 0.220125 |   0.0 |   0.0 |  0.0 |   0.0 | 0.099894 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |             82.5 |       9.1 |               100.3 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 0.955489 |                       0.81108 |                         0.0 |                      0.61078 |                       0.06609 |                           0.057212 |             0.788278 |                  0.773534 |      99.511722 |   1.246623 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      1.0 |                      7.400749 |             1.337827 |  2.844748 |        0.044468 |                 0.04541 | 0.0309 |       0.0 |       0.0 |              0.011982 |                      0.0 |                  0.011982 |                   0.353008 |                  1.110113 |                     0.0 |                -0.025663 |    -0.034272 |             0.0 |         0.037645 |         -0.757105 |             23.407216 |                 100.0 |                86.101298 |               98.388338 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102979 |      0.0 |      0.0 | -0.057639 |      0.0 |          0.0 |     0.736133 | 1.501369 | 1.501369 |      0.0 |      0.0 |     0.0 |  -0.15375 |      0.0 | -0.310062 |      0.0 | -0.501457 | 0.991541 | 0.294946 | 2.269574 | 0.793867 |  0.0 |  0.0 | 0.0 |  0.0 | 0.793867 | 0.380297 | 0.105638 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.542932 |  2.206393 |       0.0 | 1.084256 |   0.22602 | 91.872294 |      0.5 | 0.293867 |   0.0 |   0.0 |  0.0 | 0.148807 |  0.23149 |   0.0 |   0.0 |  0.0 |   0.0 | 0.105638 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |             82.2 |       8.7 |               100.3 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 0.929372 |                      0.822835 |                         0.0 |                      0.61078 |                      0.064498 |                           0.057942 |             0.763944 |                  0.780493 |      99.536056 |   1.248525 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                 0.053708 |                      7.400749 |             1.337827 |   0.00593 |        0.000077 |                0.000079 | 0.0309 |       0.0 |       0.0 |              0.000021 |                      0.0 |                  0.000021 |                   0.353715 |                  1.107879 |                     0.0 |                -0.014096 |    -0.046141 |             0.0 |         0.014117 |         -0.754165 |              24.48341 |                 100.0 |                84.150408 |               98.776942 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096543 |      0.0 |      0.0 | -0.047871 |      0.0 |          0.0 |     0.735605 | 1.261344 | 1.261344 |      0.0 |      0.0 |     0.0 | -0.127693 |      0.0 | -0.279479 |      0.0 | -0.421289 | 0.990385 | 0.294259 | 1.842857 | 0.645067 |  0.0 |  0.0 | 0.0 |  0.0 | 0.645067 | 0.388045 |  0.10779 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.287697 |  0.411407 |       0.0 | 1.091564 |   0.25692 | 93.070083 |      0.5 | 0.145067 |   0.0 |   0.0 |  0.0 | 0.157878 | 0.230166 |   0.0 |   0.0 |  0.0 |   0.0 |  0.10779 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |             76.5 |       8.9 |               100.3 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 0.903886 |                      0.833518 |                         0.0 |                      0.61078 |                      0.062939 |                           0.058604 |             0.691473 |                   0.78441 |      99.608527 |   1.250659 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |       0.0 |                   0.0 |                      0.0 |                       0.0 |                   0.354416 |                   1.10789 |                     0.0 |                 0.008426 |    -0.046162 |             0.0 |        -0.008426 |         -0.753474 |             23.933221 |                 100.0 |                82.572729 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.129358 |     0.06 | 0.021721 | -0.025712 |      0.0 |          0.0 |     0.735081 | 0.975359 | 0.287697 |      0.0 |      0.0 |     0.0 | -0.068585 |      0.0 |  -0.26561 |      0.0 |  -0.32577 | 0.989228 | 0.293572 | 0.437119 | 0.152898 |  0.0 |  0.0 | 0.0 |  0.0 | 0.152898 | 0.376587 | 0.104607 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 1.098872 |   0.28782 | 93.332584 | 0.152898 |      0.0 |   0.0 |   0.0 |  0.0 | 0.162213 | 0.214374 |   0.0 |   0.0 |  0.0 |   0.0 | 0.104607 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |             81.7 |       7.9 |               100.3 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 0.879017 |                      0.843583 |                         0.0 |                          0.0 |                      0.061412 |                           0.059227 |             0.718157 |                  0.788539 |      99.581843 |   1.252331 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.11016 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.107494 |                     0.0 |                      0.0 |    -0.046146 |             0.0 |              0.0 |           0.09655 |             26.962743 |                 100.0 |                82.227644 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.087608 |      0.0 | 0.029418 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.983949 | 0.290447 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.353057 | 0.098071 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   0.42888 | 93.303165 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.159989 | 0.193068 |   0.0 |   0.0 |  0.0 |   0.0 | 0.098071 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |             83.4 |       7.1 |               100.3 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 0.860763 |                       0.85201 |                         0.0 |                          0.0 |                      0.060288 |                           0.059747 |             0.717876 |                  0.791127 |      99.582124 |   1.253684 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10368 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.107885 |                     0.0 |                      0.0 |    -0.046162 |             0.0 |              0.0 |          0.096159 |             30.000798 |                 100.0 |                82.266306 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.070419 |      0.0 | 0.025143 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.978912 | 0.287481 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.329735 | 0.091593 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   0.56346 | 93.278022 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.155856 | 0.173879 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091593 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |             82.8 |       6.4 |               100.3 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                  0.86681 |                       0.86026 |                         0.0 |                          0.0 |                       0.06066 |                           0.060257 |             0.717719 |                  0.793088 |      99.582281 |   1.253234 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.108737 |                     0.0 |                      0.0 |    -0.046197 |             0.0 |              0.0 |          0.095308 |             33.282135 |                 100.0 |                82.299352 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.066191 |      0.0 | 0.025199 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.973795 | 0.284483 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.308428 | 0.085674 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |    0.7002 | 93.252823 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.15183 | 0.156597 |   0.0 |   0.0 |  0.0 |   0.0 | 0.085674 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |             85.8 |       7.2 |               100.3 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                  0.86681 |                      0.867569 |                         0.0 |                          0.0 |                       0.06066 |                           0.060707 |             0.743723 |                  0.794837 |      99.556277 |   1.253111 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10584 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.109465 |                     0.0 |                      0.0 |    -0.046228 |             0.0 |              0.0 |          0.094579 |              29.58412 |                 100.0 |                82.332473 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060894 |      0.0 | 0.021594 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.968677 | 0.281501 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.288942 | 0.080262 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   0.83694 | 93.231229 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.147909 | 0.141033 |   0.0 |   0.0 |  0.0 |   0.0 | 0.080262 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |             87.6 |       7.0 |               100.3 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 0.854752 |                      0.873404 |                         0.0 |                          0.0 |                      0.059917 |                           0.061066 |             0.748763 |                  0.795889 |      99.551237 |   1.253989 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.10152 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                   1.11033 |                     0.0 |                      0.0 |    -0.046264 |             0.0 |              0.0 |          0.093714 |             30.429381 |                 100.0 |                82.360857 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.050931 | 0.050931 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.963722 | 0.278628 |    0.034 | 0.011424 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011424 | 0.271252 | 0.075348 | 0.009069 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   0.96936 | 93.253805 | 0.011424 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144236 | 0.127016 |   0.0 |   0.0 |  0.0 |   0.0 | 0.075348 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |             85.6 |       6.8 |               100.2 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 0.802302 |                      0.876717 |                         0.0 |                          0.0 |                      0.056666 |                            0.06127 |              0.68677 |                   0.79394 |       99.51323 |   1.257101 |        -0.405619 | 8.652959 | 16.053708 |         1.031074 |                  5.210498 |                      0.0 |                      7.400749 |             1.337827 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08208 |                   0.0 |                      0.0 |                       0.0 |                   1.204044 |                  1.112997 |                     0.0 |                      0.0 |    -0.046375 |             0.0 |              0.0 |          0.091047 |             31.324363 |                 100.0 |                82.331182 |               98.814229 | 0.043069 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.055782 | 0.055782 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.959493 | 0.276188 | 0.043069 | 0.014376 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014376 | 0.255236 | 0.070899 | 0.004218 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.08234 | 93.282498 | 0.014376 |      0.0 |   0.0 |   0.0 |  0.0 | 0.140844 | 0.114392 |   0.0 |   0.0 |  0.0 |   0.0 | 0.070899 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |             85.3 |       6.4 |               100.2 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                 0.807986 |                      0.879785 |                         0.0 |                          0.0 |                       0.05702 |                           0.061459 |             0.689212 |                  0.791623 |      99.510788 |   1.256635 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.08424 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.116023 |                     0.0 |                      0.0 |    -0.046501 |             0.0 |              0.0 |          0.080642 |             33.282135 |                 100.0 |                82.293469 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.053849 | 0.004218 | 0.018581 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.955184 | 0.273713 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.240413 | 0.066782 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.19748 | 93.263916 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.137391 | 0.103023 |   0.0 |   0.0 |  0.0 |   0.0 | 0.066782 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |             86.2 |       6.3 |               100.2 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 0.796653 |                      0.881066 |                         0.0 |                          0.0 |                      0.056315 |                           0.061538 |             0.686715 |                  0.787857 |      99.513285 |   1.257555 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07992 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.119656 |                     0.0 |                      0.0 |    -0.046652 |             0.0 |              0.0 |          0.077009 |             33.810423 |                 100.0 |                82.317892 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049163 |      0.0 | 0.018518 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.951037 | 0.271341 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.226625 | 0.062951 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |    1.3083 | 93.245398 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.133842 | 0.092784 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062951 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |             89.3 |       5.7 |               100.2 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                  0.78546 |                      0.880041 |                         0.0 |                          0.0 |                      0.055617 |                           0.061475 |             0.701415 |                   0.78309 |      99.498585 |   1.258394 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0756 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.123475 |                     0.0 |                      0.0 |    -0.046811 |             0.0 |              0.0 |           0.07319 |             37.369415 |                 100.0 |                82.342232 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033548 | 0.033548 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.947051 | 0.269072 |    0.034 | 0.011126 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011126 | 0.214091 |  0.05947 | 0.026452 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |    1.4148 | 93.268273 | 0.011126 |      0.0 |   0.0 |   0.0 |  0.0 | 0.130529 | 0.083562 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05947 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |             91.4 |       5.2 |               100.3 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 0.774405 |                      0.877228 |                         0.0 |                          0.0 |                      0.054927 |                           0.061302 |             0.707807 |                  0.777516 |      99.592193 |   1.260534 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.127474 |                     0.0 |                      0.0 |    -0.046978 |             0.0 |              0.0 |           0.06919 |             40.962628 |                 100.0 |                82.312166 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023802 | 0.023802 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.943227 | 0.266903 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.202557 | 0.056266 |  0.00265 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.51698 | 93.268273 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.1273 | 0.075257 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056266 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |             90.9 |       4.6 |               100.4 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 0.774405 |                      0.873404 |                         0.0 |                          0.0 |                      0.054927 |                           0.061066 |             0.703935 |                  0.771179 |      99.696065 |   1.261813 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.07128 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.131856 |                     0.0 |                      0.0 |    -0.047161 |             0.0 |              0.0 |          0.064808 |              46.30558 |                 100.0 |                82.312166 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.022296 |  0.00265 | 0.008765 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.939403 | 0.264743 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.191789 | 0.053275 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.61916 | 93.259507 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124012 | 0.067777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053275 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |             89.2 |       4.9 |               100.5 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 0.758081 |                      0.868834 |                         0.0 |                          0.0 |                      0.053905 |                           0.060785 |             0.676208 |                  0.764103 |      99.823792 |   1.264577 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   -0.0648 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.136687 |                     0.0 |                      0.0 |    -0.047362 |             0.0 |              0.0 |          0.059978 |             43.470544 |                 100.0 |                82.323687 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.028418 |      0.0 |  0.01221 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.935821 | 0.262728 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.181849 | 0.050514 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.71486 | 93.247297 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.120808 | 0.061041 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050514 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |             85.6 |       6.3 |               100.6 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                  0.70079 |                      0.860763 |                         0.0 |                          0.0 |                      0.050296 |                           0.060288 |             0.599876 |                  0.752773 |     100.000124 |   1.271267 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.04104 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.144252 |                     0.0 |                      0.0 |    -0.047677 |             0.0 |              0.0 |          0.052412 |             33.810423 |                 100.0 |                82.339736 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.047697 |      0.0 | 0.017452 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.933129 | 0.261219 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.172662 | 0.047962 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |    1.7868 | 93.229845 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.117688 | 0.054974 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047962 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |             86.3 |       5.8 |               100.8 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 0.666353 |                      0.851015 |                         0.0 |                          0.0 |                       0.04811 |                           0.059686 |             0.575062 |                  0.740702 |     100.224938 |   1.277169 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.329627 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |  -0.02592 |                   0.0 |                      0.0 |                       0.0 |                   1.196665 |                  1.152031 |                     0.0 |                      0.0 |    -0.048001 |             0.0 |              0.0 |          0.044634 |             36.725115 |                 100.0 |                82.362677 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.040381 |      0.0 | 0.015359 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.931002 |  0.26003 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164158 |   0.0456 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.84362 | 93.214486 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.114648 |  0.04951 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0456 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |             91.2 |       6.1 |               100.9 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 0.647352 |                      0.839643 |                         0.0 |                          0.0 |                      0.046897 |                           0.058983 |             0.590385 |                  0.729404 |     100.309615 |   1.280232 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.339738 |                      7.386143 |             1.329627 |  0.244956 |        0.003171 |                0.003219 | 0.0309 |       0.0 |  -0.01728 |              0.002897 |                      0.0 |                  0.002897 |                   1.196665 |                  1.158654 |                     0.0 |                      0.0 |     -0.04538 |             0.0 |              0.0 |          0.038011 |             34.918962 |                 100.0 |                82.382867 |               98.493489 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02629 |      0.0 | 0.009662 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.929199 | 0.259023 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.156276 |  0.04341 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |    1.8918 | 93.204824 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.111686 |  0.04459 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04341 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |             87.5 |       5.9 |               101.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 0.647352 |                      0.828647 |                         0.0 |                          0.0 |                      0.046897 |                           0.058302 |             0.566433 |                  0.716814 |     100.433567 |   1.281619 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 |  4.849391 |        0.062779 |                0.063765 | 0.0309 |       0.0 |  -0.01728 |              0.057389 |                      0.0 |                  0.057389 |                   1.196665 |                  1.166643 |                     0.0 |                      0.0 |     0.008779 |             0.0 |              0.0 |          0.030022 |             36.102655 |                 100.0 |                82.395569 |               97.877887 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.045635 |      0.0 |  0.01718 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.927396 | 0.258019 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.148959 | 0.041378 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |   1.93998 | 93.187643 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.108802 | 0.040158 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041378 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |             84.5 |       6.4 |               101.1 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 0.633417 |                      0.816579 |                         0.0 |                     0.548474 |                      0.046005 |                           0.057554 |             0.535238 |                  0.702632 |     100.564762 |   1.284447 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 14.852029 |         0.19227 |                0.195489 | 0.0309 |       0.0 |  0.014361 |              0.020003 |                      0.0 |                  0.020003 |                   0.136052 |                  1.175846 |                     0.0 |                 0.098108 |     -0.02899 |             0.0 |        -0.078105 |         -1.039794 |             33.282135 |                 100.0 |                82.418155 |               97.881121 |    0.034 | 0.01275 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052426 |      0.0 |      0.0 |  0.003209 | 0.000025 |    -1.465907 |     0.897677 |      0.0 |      0.0 | 0.000099 | 0.000099 |     0.0 |  0.008561 |      0.0 | -0.072247 |      0.0 |  0.014419 | 0.926777 | 0.257675 | 0.015768 | 0.004993 |  0.0 |  0.0 | 0.0 |  0.0 | 0.004993 | 0.142223 | 0.039506 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.010587 |  0.015023 |       0.0 |  0.00743 |  1.956519 | 93.198417 | 0.004993 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106056 | 0.036167 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039506 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |             76.7 |       6.8 |               101.2 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                 0.624275 |                      0.802774 |                         0.0 |                     0.526367 |                      0.045418 |                           0.056696 |             0.478819 |                   0.68577 |     100.721181 |   1.286932 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.269583 |        0.288296 |                0.293415 | 0.0309 |       0.0 |  0.019828 |              0.030684 |                      0.0 |                  0.030684 |                   0.139046 |                   1.18712 |                     0.0 |                 0.107855 |    -0.018779 |             0.0 |        -0.077171 |         -1.048074 |             31.324363 |                 100.0 |                82.403991 |               97.879093 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.084311 |      0.0 |      0.0 |  0.012109 |      0.0 |    -2.021578 |     0.895425 |      0.0 |      0.0 | 0.000112 | 0.000112 |     0.0 |  0.032299 |      0.0 | -0.089605 |      0.0 |  0.019865 | 0.926363 | 0.257444 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.135953 | 0.037765 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.002076 |  0.002914 |       0.0 | 0.014738 |  1.967591 | 93.198417 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103381 | 0.032572 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037765 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |             74.8 |       7.0 |               101.3 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 0.606339 |                      0.787548 |                         0.0 |                     0.507806 |                      0.044264 |                           0.055747 |             0.453542 |                  0.667775 |     100.846458 |   1.290215 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 22.293018 |        0.288599 |                0.293995 | 0.0309 |       0.0 |  0.024583 |              0.031397 |                      0.0 |                  0.031397 |                   0.141997 |                  1.199167 |                     0.0 |                 0.113262 |    -0.018568 |             0.0 |        -0.081865 |          -1.05717 |             30.429381 |                 100.0 |                82.403991 |               97.879093 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092017 |     0.06 | 0.010692 |  0.002914 |      0.0 |     -2.50427 |     0.893205 |      0.0 |      0.0 | 0.000027 | 0.000027 |     0.0 |  0.037744 |      0.0 | -0.095017 |      0.0 |  0.024591 | 0.926126 | 0.257313 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.130045 | 0.036124 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 0.022046 |  1.973909 | 93.187725 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.10071 | 0.029335 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036124 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |             69.7 |       6.3 |               101.4 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 0.593188 |                      0.771892 |                         0.0 |                          0.0 |                      0.043416 |                           0.054769 |             0.413452 |                  0.649354 |     100.986548 |   1.293105 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.329627 | 21.019348 |        0.272111 |                0.277484 | 0.0309 |       0.0 |   0.00864 |              0.249736 |                      0.0 |                  0.249736 |                   1.196665 |                   1.21168 |                     0.0 |                      0.0 |     0.199249 |             0.0 |              0.0 |         -0.015015 |             33.810423 |                 100.0 |                82.418047 |               97.881105 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.131541 |      0.0 |  0.04673 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.925293 |  0.25685 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.124528 | 0.034591 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.996169 | 93.140995 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098109 | 0.026419 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034591 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |             66.9 |       7.3 |               101.5 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                  0.58456 |                      0.755838 |                         0.0 |                          0.0 |                      0.042857 |                           0.053764 |             0.391071 |                  0.630999 |     101.108929 |    1.29544 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |             1.399736 | 11.654171 |         0.21122 |                0.216915 | 0.0309 |       0.0 |   0.01296 |              0.195223 |                      0.0 |                  0.195223 |                   1.259762 |                  1.349977 |                     0.0 |                      0.0 |     0.138974 |             0.0 |              0.0 |         -0.090215 |             29.178858 |                 100.0 |                82.479485 |               97.889893 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146498 |      0.0 | 0.047227 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.924622 | 0.256478 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.119368 | 0.033158 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.014109 | 93.093768 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095574 | 0.023794 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033158 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |             64.4 |       7.5 |               101.5 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 0.576044 |                       0.74008 |                         0.0 |                          0.0 |                      0.042305 |                           0.052775 |             0.370972 |                  0.612262 |     101.129028 |   1.296488 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      1.0 |                      7.386143 |              1.43479 |  2.813062 |        0.050984 |                0.053414 | 0.0309 |       0.0 |   0.01728 |              0.048073 |                      0.0 |                  0.048073 |                   1.291311 |                  1.428152 |                     0.0 |                      0.0 |    -0.011433 |             0.0 |              0.0 |         -0.136842 |             28.400756 |                 100.0 |                82.541584 |               97.898763 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135014 |      0.0 | 0.042617 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.924112 | 0.256195 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.114534 | 0.031815 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.027729 | 93.051151 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093106 | 0.021429 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031815 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |             65.4 |       7.0 |               101.7 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 0.555231 |                      0.724396 |                         0.0 |                          0.0 |                      0.040952 |                           0.051788 |             0.363121 |                  0.594216 |     101.336879 |   1.301474 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                 0.046405 |                      7.386143 |             1.504898 |  0.004444 |        0.000553 |                0.000608 | 0.0309 |       0.0 |   0.02808 |              0.000547 |                      0.0 |                  0.000547 |                   1.354408 |                  1.571379 |                     0.0 |                      0.0 |    -0.064927 |             0.0 |              0.0 |         -0.216971 |             30.429381 |                 100.0 |                82.597627 |               98.771746 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.11254 |      0.0 | 0.036657 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.924006 | 0.256136 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |     0.11 | 0.030555 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.030549 | 93.014493 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.090701 | 0.019299 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030555 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |             67.6 |       6.0 |               101.7 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 0.543065 |                      0.709217 |                         0.0 |                          0.0 |                      0.040158 |                           0.050829 |             0.367112 |                  0.579135 |     101.332888 |   1.302892 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.03456 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.583999 |                     0.0 |                      0.0 |       -0.066 |             0.0 |              0.0 |          -0.22959 |             35.500944 |                 100.0 |                82.645838 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.08847 |      0.0 |  0.03163 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.924143 | 0.256212 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.105739 | 0.029372 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.026889 | 92.982864 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088358 | 0.017381 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029372 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |             70.2 |       4.8 |               101.8 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                  0.52721 |                       0.69432 |                         0.0 |                          0.0 |                       0.03912 |                           0.049887 |             0.370102 |                  0.563643 |     101.429898 |   1.306085 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |    0.0432 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                    1.5975 |                     0.0 |                      0.0 |    -0.066562 |             0.0 |              0.0 |         -0.243092 |             44.376181 |                 100.0 |                82.687441 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.063209 |      0.0 | 0.025658 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.924604 | 0.256468 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.101729 | 0.028258 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.014589 | 92.957205 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086075 | 0.015653 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028258 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |             70.4 |       4.2 |               101.9 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 0.523311 |                      0.680107 |                         0.0 |                          0.0 |                      0.038864 |                           0.048985 |             0.368411 |                  0.548422 |     101.531589 |    1.30786 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04536 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.611195 |                     0.0 |                      0.0 |    -0.067133 |             0.0 |              0.0 |         -0.256786 |             50.715635 |                 100.0 |                82.721192 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.054415 |      0.0 |  0.02375 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.925145 | 0.256768 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.09795 | 0.027208 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  2.000129 | 92.933455 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.083852 | 0.014098 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027208 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |             69.5 |       4.9 |               101.9 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 0.519437 |                      0.665752 |                         0.0 |                          0.0 |                       0.03861 |                           0.048071 |             0.361009 |                  0.533156 |     101.538991 |   1.308379 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.04752 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.625096 |                     0.0 |                      0.0 |    -0.067712 |             0.0 |              0.0 |         -0.270688 |             43.470544 |                 100.0 |                82.752435 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.065452 |      0.0 |  0.02619 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.925767 | 0.257113 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.094383 | 0.026217 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.983509 | 92.907265 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.081686 | 0.012696 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026217 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |             71.6 |       4.0 |               102.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 0.504195 |                      0.650878 |                         0.0 |                          0.0 |                      0.037607 |                           0.047123 |             0.361003 |                  0.517394 |     101.638997 |   1.311601 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.05616 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.639665 |                     0.0 |                      0.0 |    -0.068319 |             0.0 |              0.0 |         -0.285257 |             53.251417 |                 100.0 |                 82.78689 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.048111 |      0.0 | 0.021424 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.926712 | 0.257639 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.091011 | 0.025281 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.958249 | 92.885841 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.079576 | 0.011435 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025281 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |             74.5 |       3.4 |               102.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 0.485698 |                      0.635721 |                         0.0 |                          0.0 |                      0.036384 |                           0.046153 |             0.361845 |                  0.501875 |     101.638155 |   1.314025 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.06696 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.653965 |                     0.0 |                      0.0 |    -0.068915 |             0.0 |              0.0 |         -0.299557 |             62.648725 |                 100.0 |                82.815078 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034966 |      0.0 | 0.016822 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.928062 |  0.25839 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.087819 | 0.024394 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.922189 | 92.869018 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.077521 | 0.010298 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024394 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |             80.2 |       3.4 |               102.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                 0.457351 |                      0.621065 |                         0.0 |                          0.0 |                      0.034499 |                           0.045212 |             0.366795 |                  0.488908 |     101.633205 |   1.317905 |        -0.406588 | 8.660262 | 16.046405 |          1.03126 |                  5.178293 |                      0.0 |                      7.386143 |             1.504898 |       0.0 |             0.0 |                     0.0 | 0.0309 |       0.0 |   0.08424 |                   0.0 |                      0.0 |                       0.0 |                   1.354408 |                  1.664915 |                     0.0 |                      0.0 |    -0.069371 |             0.0 |              0.0 |         -0.310507 |             62.648725 |                 100.0 |                82.837211 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024921 |      0.0 |  0.01189 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |       0.0 |      0.0 |       0.0 |      0.0 |       0.0 | 0.930058 | 0.259502 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.084793 | 0.023554 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan |  1.868849 | 92.857128 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.075518 | 0.009275 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023554 |
"""
# import...
# ...from HydPy
from hydpy.exe.modelimports import *
from hydpy.core import masktools
from hydpy.core import modeltools

# ...from lland
from hydpy.models.lland import lland_model
from hydpy.models.lland import lland_masks
from hydpy.models.lland.lland_constants import *


class Model(modeltools.AdHocModel):
    """Penman-Monteith and Knauf with snow interception version of HydPy-L-Land
    (|lland_v4|)."""

    INLET_METHODS = (lland_model.Pick_QZ_V1,)
    RECEIVER_METHODS = ()
    ADD_METHODS = (
        lland_model.Return_AdjustedWindSpeed_V1,
        lland_model.Return_ActualVapourPressure_V1,
        lland_model.Return_NetLongwaveRadiationInz_V1,
        lland_model.Return_NetLongwaveRadiationSnow_V1,
        lland_model.Return_DailyGlobalRadiation_V1,
        lland_model.Return_Penman_V1,
        lland_model.Return_PenmanMonteith_V1,
        lland_model.Return_EnergyGainSnowSurface_V1,
        lland_model.Return_SaturationVapourPressure_V1,
        lland_model.Return_SaturationVapourPressureSlope_V1,
        lland_model.Return_NetRadiation_V1,
        lland_model.Return_WSensInz_V1,
        lland_model.Return_WSensSnow_V1,
        lland_model.Return_WLatInz_V1,
        lland_model.Return_WLatSnow_V1,
        lland_model.Return_WSurfInz_V1,
        lland_model.Return_WSurf_V1,
        lland_model.Return_TempS_V1,
        lland_model.Return_TempSInz_V1,
        lland_model.Return_WG_V1,
        lland_model.Return_BackwardEulerErrorInz_V1,
        lland_model.Return_BackwardEulerError_V1,
        lland_model.Return_ESnowInz_V1,
        lland_model.Return_ESnow_V1,
        lland_model.Return_TempSSurface_V1,
    )
    RUN_METHODS = (
        lland_model.Calc_QZH_V1,
        lland_model.Update_LoggedTemL_V1,
        lland_model.Calc_TemLTag_V1,
        lland_model.Update_LoggedRelativeHumidity_V1,
        lland_model.Calc_DailyRelativeHumidity_V1,
        lland_model.Update_LoggedSunshineDuration_V1,
        lland_model.Calc_NKor_V1,
        lland_model.Calc_TKor_V1,
        lland_model.Calc_TKorTag_V1,
        lland_model.Calc_WindSpeed2m_V1,
        lland_model.Calc_ReducedWindSpeed2m_V1,
        lland_model.Update_LoggedWindSpeed2m_V1,
        lland_model.Calc_DailyWindSpeed2m_V1,
        lland_model.Calc_WindSpeed10m_V1,
        lland_model.Calc_SolarDeclination_V1,
        lland_model.Calc_TSA_TSU_V1,
        lland_model.Calc_DailySunshineDuration_V1,
        lland_model.Calc_EarthSunDistance_V1,
        lland_model.Calc_ExtraterrestrialRadiation_V1,
        lland_model.Calc_PossibleSunshineDuration_V1,
        lland_model.Calc_DailyPossibleSunshineDuration_V1,
        lland_model.Calc_SP_V1,
        lland_model.Calc_GlobalRadiation_V1,
        lland_model.Update_LoggedGlobalRadiation_V1,
        lland_model.Calc_DailyGlobalRadiation_V1,
        lland_model.Calc_AdjustedGlobalRadiation_V1,
        lland_model.Calc_SaturationVapourPressure_V1,
        lland_model.Calc_DailySaturationVapourPressure_V1,
        lland_model.Calc_SaturationVapourPressureSlope_V1,
        lland_model.Calc_DailySaturationVapourPressureSlope_V1,
        lland_model.Calc_ActualVapourPressure_V1,
        lland_model.Calc_DailyActualVapourPressure_V1,
        lland_model.Calc_NBes_Inzp_V1,
        lland_model.Calc_SNRatio_V1,
        lland_model.Calc_SBes_V1,
        lland_model.Calc_SnowIntMax_V1,
        lland_model.Calc_SnowIntRate_V1,
        lland_model.Calc_NBesInz_V1,
        lland_model.Calc_SBesInz_V1,
        lland_model.Calc_STInz_V1,
        lland_model.Calc_WaDaInz_SInz_V1,
        lland_model.Calc_WNiedInz_ESnowInz_V1,
        lland_model.Calc_TempSInz_V1,
        lland_model.Update_ASInz_V1,
        lland_model.Calc_ActualAlbedoInz_V1,
        lland_model.Calc_NetShortwaveRadiationInz_V1,
        lland_model.Calc_RLAtm_V1,
        lland_model.Update_ESnowInz_V1,
        lland_model.Calc_SchmPotInz_V1,
        lland_model.Calc_SchmInz_STInz_V1,
        lland_model.Calc_GefrPotInz_V1,
        lland_model.Calc_GefrInz_STInz_V1,
        lland_model.Calc_EvSInz_SInz_STInz_V1,
        lland_model.Update_WaDaInz_SInz_V1,
        lland_model.Update_ESnowInz_V2,
        lland_model.Calc_WATS_V2,
        lland_model.Calc_WaDa_WAeS_V2,
        lland_model.Calc_WNied_ESnow_V1,
        lland_model.Calc_TempS_V1,
        lland_model.Update_TauS_V1,
        lland_model.Calc_ActualAlbedo_V1,
        lland_model.Calc_NetShortwaveRadiationSnow_V1,
        lland_model.Calc_TZ_V1,
        lland_model.Calc_WG_V1,
        lland_model.Update_ESnow_V1,
        lland_model.Calc_SchmPot_V2,
        lland_model.Calc_Schm_WATS_V1,
        lland_model.Calc_GefrPot_V1,
        lland_model.Calc_Gefr_WATS_V1,
        lland_model.Calc_EvS_WAeS_WATS_V1,
        lland_model.Update_WaDa_WAeS_V1,
        lland_model.Update_ESnow_V2,
        lland_model.Calc_NetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetLongwaveRadiation_V1,
        lland_model.Calc_NetRadiation_V1,
        lland_model.Calc_DailyNetRadiation_V1,
        lland_model.Calc_DryAirPressure_V1,
        lland_model.Calc_DensityAir_V1,
        lland_model.Calc_AerodynamicResistance_V1,
        lland_model.Calc_SoilSurfaceResistance_V1,
        lland_model.Calc_LanduseSurfaceResistance_V1,
        lland_model.Calc_ActualSurfaceResistance_V1,
        lland_model.Calc_G_V2,
        lland_model.Calc_EvPo_V2,
        lland_model.Calc_EvI_Inzp_V3,
        lland_model.Calc_EvB_V2,
        lland_model.Update_EBdn_V1,
        lland_model.Calc_SFF_V1,
        lland_model.Calc_FVG_V1,
        lland_model.Calc_QKap_V1,
        lland_model.Calc_QBB_V1,
        lland_model.Calc_QIB1_V1,
        lland_model.Calc_QIB2_V1,
        lland_model.Calc_QDB_V1,
        lland_model.Update_QDB_V1,
        lland_model.Calc_BoWa_V1,
        lland_model.Calc_QBGZ_V1,
        lland_model.Calc_QIGZ1_V1,
        lland_model.Calc_QIGZ2_V1,
        lland_model.Calc_QDGZ_V1,
        lland_model.Calc_QBGA_V1,
        lland_model.Update_QDGZ_QBGZ_QBGA_V1,
        lland_model.Update_QDGZ_QBGZ_QBGA_V2,
        lland_model.Calc_QIGA1_V1,
        lland_model.Calc_QIGA2_V1,
        lland_model.Calc_QDGZ1_QDGZ2_V1,
        lland_model.Calc_QDGA1_V1,
        lland_model.Calc_QDGA2_V1,
        lland_model.Calc_QAH_V1,
        lland_model.Calc_QA_V1,
    )
    OUTLET_METHODS = (lland_model.Pass_QA_V1,)
    SENDER_METHODS = ()
    SUBMODELS = (
        lland_model.PegasusESnowInz,
        lland_model.PegasusESnow,
        lland_model.PegasusTempSSurface,
    )
    idx_hru = modeltools.Idx_HRU()


class Masks(masktools.Masks):
    """Masks applicable to |lland_v4|."""

    # pylint: disable=no-member
    # bug of pylint 2.4?
    CLASSES = lland_masks.Masks.CLASSES


tester = Tester()
cythonizer = Cythonizer()
cythonizer.finalise()
